#
# $Id: HostAdapterRcsp.tcl,v 1.8 1996-05-29 14:34:22-07 bmah Exp $
#   based on:
# $ID: HostAdapterFifo.tcl,v 1.4 1995-09-28 11:44:26-07 bmah Exp $
#
# HostAdapterRcsp.tcl
#
# Copyright (c) 1995, 1996 The Regents of the University of California
# All Rights Reserved
# Created by Bruce A. Mah
# Department of Electrical Engineering and Computer Sciences,
#   Computer Science Division
#

#
# HostAdapterRcsp
#
# Input:  host adapter name (name), max number of VCIDs (v),
#         default link bandwidth (bw), number of priority levels (np),
#         host ATM address (addr)
#
# Output:  none
#
# Create an ATM host adapter for RCSP queueing.  RCSP parameters are
# set as for the HostAdapterFifo.
#
proc HostAdapterRcsp {name v bw np addr} {

    # We need to use a non-work-conserving scheduler at the entrance
    # to the network to smooth out bursts.  If we don't, the combination
    # of large buffers (to absorb packets) and the non-work-conserving
    # scheduler will very likely overwhelm the next downstream switch.
    global ConfHostRcspMode
    if {! [info exists ConfHostRcspMode] } {
	set confRcspMode non-work-conserving
    } else {
	set confRcspMode $ConfHostRcspMode
    }

    # Set up some parameters
    set minCellTimes 16
    set tickCellTimes [ expr {$minCellTimes / 2} ]
    set subqueues [ expr 1 << ($np + 1) ]

    set cellSize [ expr { 53 * 8 } ]
    set cellTime [ expr { $cellSize / double($bw) } ]

    # We need more subqueues because the host might be sending data
    # in huge bursts.  Our current heuristic is to have enough subqueues
    # to support up to two seconds' worth of buffering.
    while {$subqueues * $tickCellTimes * $cellTime < 2.0} {
	set subqueues [ expr $subqueues * 2 ]
#	puts "increased subqueues to $subqueues now can buffer for [expr {$subqueues * $tickCellTimes * $cellTime}] "
    }

    # Compute vector of delay times
    set cells $minCellTimes
    set delayVector [ list ]
    set i 0
    while {$i < $np} {
	set delay [ expr { $cells * $cellTime } ]
#	puts "delay bound for $cells cells is [ expr { $delay*1000 } ] ms"
	lappend delayVector $delay
	set cells [ expr { $cells * 2 } ]
	incr i
    }
    set tickTime [ expr { $tickCellTimes * $cellTime } ]

    # Create switch module (3 ports, 0 == sighost, 1 == host, 2 == net)
    SwitchModule $name.switch 3 $v

    # Create signalling entity (100 concurrent signalling 
    # calls max)
    SigHostRcsp $name.sighost $v 100 $addr
    $name.sighost mode $confRcspMode
    $name.sighost switch $name.gobackn
    $name.sighost aal $name.aal
    $name.sighost net $name.output2 $bw $tickTime $delayVector

    global ConfHostRcspExcess
    if { [info exists ConfHostRcspExcess] } {
	$name.sighost excess $ConfHostRcspExcess
    }

    # Create input and output ports for signalling entity on port 0
    # with no delay and arbitrarily large bandwidth
    CellInputPort $name.input0 $name.switch 0 0
    CellQueueFifo $name.output0
    $name.output0 output $name.gobackn
    $name.output0 latency 0
    $name.output0 bitrate 1000000000
    $name.switch output 0 $name.output0

    # Create input and output ports for host (100 vcids, 1 us gap, port 1)
    Aal $name.aal $v 0.000001 1
    $name.aal down $name.switch
    $name.switch output 1 $name.aal
    
    # Create input and output ports for network
    CellInputPort $name.input2 $name.switch 2 0
    CellQueueRcsp $name.output2 $v $np $subqueues $tickTime
    $name.output2 mode $confRcspMode
    $name.output2 latency 0
    $name.output2 bitrate $bw
    $name.switch output 2 $name.output2

    # Create GoBackN entity for signalling messages
    CellGoBackN $name.gobackn 3 512 0.50 0.10
    $name.gobackn up $name.sighost
    $name.gobackn down $name.input0

    # Create initial translation tables for control channels in
    # switch module.  We want a passthrough for all data channels
    # (both inbound and outbound), and to divert control messages
    # to the SigHost object.
    for {set i 1} {$i < $v} {incr i} {
	$name.switch set 1 $i 2 $i
	$name.switch set 2 $i 1 $i
    }
    $name.switch set 1 0 0 1
    $name.switch set 0 1 1 0
    $name.switch set 2 0 0 2
    $name.switch set 0 2 2 0

    proc $name {} [ list puts -nonewline "$name (HostAdapterRcsp)" ]

}

