#
# $Id: LanHost.tcl,v 1.8 1996-04-22 08:51:50-07 bmah Exp $
#
# LanHost.tcl
# Bruce A. Mah (bmah@CS.Berkeley.EDU)
#
# Copyright (c) 1996 The Regents of the University of California
# All Rights Reserved
# Created by Bruce A. Mah
# Department of Electrical Engineering and Computer Sciences,
#   Computer Science Division
#

#
# LanHost
#
# Input:  host name, Lan object, IP address, netmask, 
#         default gateway address (optional)
#
# Output:  None
#
# Create an IP host single-attached to a Lan object, with the specified
# addresses.
#
proc LanHost {name lan ip netmask {gateway none}} {

    # First create the IP module for the host.  Specify loopback (lo0)
    # interface and some appropriate routes.
    Ip $name.ip
    $name.ip ifconfig $name.lo0 127.0.0.1 255.0.0.0 127.255.255.255

    # Create loopback interface, 0.1 ms latency
    LoopbackDeviceDriver $name.lo0 .0001
    $name.lo0 up $name.ip
    
    # Create and attach primary LAN interface
    lanAdd $name $lan 0 $ip $netmask

    # Specify routes.  We do this last so that the routes can get
    # get bound to the appropriate interfaces.
    $name.ip route add 127.0.0.1 255.0.0.0 127.0.0.1
    if {$gateway != "none"} {
	$name.ip route add 0.0.0.0 0.0.0.0 $gateway G
    }

    # Create and attach transport protocol objects
    Tcp $name.tcp
    $name.tcp down $name.ip
    $name.ip proto attach tcp $name.tcp

    Udp $name.udp
    $name.udp down $name.ip
    $name.ip proto attach udp $name.udp

    # Create a placeholder object in Tcl-land.  This is mostly
    # used if we need to abstract objects in some way...
    proc $name {} [ list puts -nonewline "$name (LanHost)" ]

}

#
# lanAdd
#
# Input:  name (of host), Lan object, interface number, ip address,
#         netmask
#
# Output:  None
#
# Attach a LAN interface to an existing host's IP module.
#
proc lanAdd {name lan num ip netmask} {

    # Get binary representations for the dotted quads making up
    # the IP address and netmask
    set ipAddr [ inet_addr $ip ]
    set netmaskAddr [ inet_addr $netmask ]

    # Compute network address and broadcast address
    set netAddr [ expr $ipAddr & $netmaskAddr ]
    set broadcastAddr [ expr $netAddr | ~$netmaskAddr]
    set net [ inet_ntoa $netAddr ]
    set broadcast [ inet_ntoa $broadcastAddr ]

    # Create LAN driver, 64 packet queue, 0.1 ms latency
    LanDeviceDriver $name.lan$num 64 .0001
    $name.lan$num down $lan
    $name.lan$num up $name.ip
    $lan arp $ip $name.lan$num

    # Specify LAN (lan) interface to the IP module
    $name.ip ifconfig $name.lan$num $ip $netmask $broadcast
    $name.ip route add $net $netmask $ip

}


#
# inet_addr
#
# Input:  Dotted-quad representation of an IP address (quad)
#
# Output:  Integer containing the binary representation of the IP address
#
# This function is the Tcl equivalent of inet_addr(3), from the C library.
#
proc inet_addr {quad} {

    scan $quad "%d.%d.%d.%d" b1 b2 b3 b4
    expr ($b1 << 24) + ($b2 << 16) + ($b3 << 8) + $b4

}

#
# inet_ntoa
#
# Input:  Integer containing the binary representation of an IP address
#
# Output:  String containing the dotted-quad representation of the
# IP address
#
# Tcl equivalent of inet_ntoa(3) from the C library.
#
proc inet_ntoa {addr} {

    set result ""
    append result [ expr "($addr >> 24) & 255" ]
    append result "."
    append result [ expr "($addr >> 16) & 255" ]
    append result "."
    append result [ expr "($addr >> 8) & 255" ]
    append result "."
    append result [ expr "($addr >> 0) & 255" ]

}
