#
# $Id: Site.tcl,v 1.4 1996-05-29 14:34:25-07 bmah Exp $
#
# Site.tcl
# Bruce A. Mah (bmah@CS.Berkeley.EDU)
#
# Copyright (c) 1996 The Regents of the University of California
# All Rights Reserved
# Created by Bruce A. Mah
# Department of Electrical Engineering and Computer Sciences,
#   Computer Science Division
#

#
# Site
#
# Input:  Site prefix (name), LAN bandwidth (bw), LAN latency (latency),
#         number of workstation clients (clients),
#         number of servers (servers), IP prefix (prefix).
#
# Output:  Names and IP addresses of servers, clients, and routers,
#          contained in global variables as shown.
#
# Create a LAN and populate it with some number of hosts.  There is
# one router off the LAN, and a number of clients and servers as
# specified by caller.
#
# Restrictions:
#
# o The IP prefix must be a 24-bit prefix, expressed exactly in the
#   form "128.32.33" (three octets, no trailing dot).
#
proc Site {name bw latency clients servers prefix} {

    # Initialize the various lists that we're going to create
    # The calling procedure can use these names to refer to the Internet
    # hosts on the newly-created LAN.
    global $name.serverNames $name.serverAddresses $name.numberServers
    global $name.clientNames $name.clientAddresses $name.numberClients
    global $name.routerNames $name.routerAddresses $name.numberRouters
    list $name.serverNames {}
    list $name.serverAddresses {}
    set $name.numberServers 0
    list $name.clientNames {}
    list $name.clientAddresses {}
    set $name.numberClients 0
    list $name.routerNames {}
    list $name.routerAddresses {}

    # Create network and hosts
    set routerAddress $prefix.1
    SiteHelper $name $bw $latency $clients $servers $prefix $routerAddress

    # Create router.  Note that the router looks just like a normal
    # host (for now).  We assume that the caller building up the
    # internetwork will provide the appropriate entries in its routing
    # table.  Note also that it is configured without a default route.
    LanHost $name.router0 $name.lan.$prefix $routerAddress 255.255.255.0
    lappend $name.routerNames $name.router0
    lappend $name.routerAddresses $routerAddress
    
    proc $name {} [ list puts -nonewline "$name (Site)" ]

}

#
# subnetAdd
#
# Input:  Site prefix (name), LAN bandwidth (bw), LAN latency (latency),
#         number of workstation clients (clients),
#         number of servers (servers), IP prefix (prefix), router name
#         (router), interface number to use on router (num).
#
# Output:  Names and IP addresses of servers, clients, and routers,
#          contained in global variables as shown.
#
# Create a LAN and populate it with some number of hosts.  An interface
# is added to an existing router.  See Site(3) for additional details.
#
proc subnetAdd {name bw latency clients servers prefix router num} {

    # Create network and hosts
    set routerAddress $prefix.1
    SiteHelper $name $bw $latency $clients $servers $prefix $routerAddress

    # Create router interface and add it to an existing router.
    set routerAddress $prefix.$num
    lanAdd $router $name.lan.$prefix $num $prefix.1 255.255.255.0
}

#
# SiteHelper
#
# Input:  Site prefix (name), LAN bandwidth (bw), LAN latency (latency),
#         number of workstation clients (clients),
#         number of servers (servers), IP prefix (prefix), default
#         router IP address (routerAddress).
#
# Output:  None (side effects only)
#
# Helper procedure for Site and addSubnetToSite.  Creates LAN and hosts,
# but does not create a router or a router interface
#
proc SiteHelper {name bw latency clients servers prefix routerAddress} {

    global $name.serverNames $name.serverAddresses $name.numberServers
    global $name.clientNames $name.clientAddresses $name.numberClients
    global $name.routerNames $name.routerAddresses $name.numberRouters

    # Set epsilon to some small value (here set to 1 us).  This is the
    # starting time for all the daemons.
    set epsilon 0.000001

    # Initialize host number on the subnet.  Remember to skip over
    # IP broadcast address (well, actually, INSANE's architecture
    # doesn't mandate this, but it might someday).  Also skip over
    # some hosts to leave room for router(s).
    set hostNumber 16

    # We need a Lan object, so create it.  Make the LAN
    # handle at least the number of clients and servers, plus some
    # fudge-factor.
    Lan $name.lan.$prefix [expr $clients + $servers + 8] $latency $bw

    # Create servers.
    for {set i 0} {$i < $servers} {incr i} {

	# Get the server number for the purposes of Tcl naming
	set serverNum [set ${name}.numberServers] 
	incr $name.numberServers

	# Create the machine
#	puts $name.server$serverNum
	LanHost $name.server$serverNum $name.lan.$prefix \
	        $prefix.$hostNumber 255.255.255.0 \
		$routerAddress
	lappend $name.serverNames $name.server$serverNum
	lappend $name.serverAddresses $prefix.$hostNumber
	incr hostNumber

	# Install daemons on server
	# FTP server uses port 21 for control and port 20 for data
	FtpServer $name.server$serverNum.ftpd $epsilon $name.server$serverNum.tcp 21 20

	# telnet server uses port 23
	TelnetServer $name.server$serverNum.telnetd $epsilon $name.server$serverNum.tcp 23

	# httpd uses port 80
	HttpServer $name.server$serverNum.httpd $epsilon $name.server$serverNum.tcp 80

	# install a sink server to simulate SMTP, port 25
	SampleTcpServer $name.server$serverNum.smtpd $epsilon $name.server$serverNum.tcp 25

	# install a sink server to simulate NNTP, port 119
	SampleTcpServer $name.server$serverNum.nntpd $epsilon $name.server$serverNum.tcp 119

    }

    # Create workstations.
    for {set i 0} {$i < $clients} {incr i} {

	# Get the client number for the purposes of Tcl naming
	set clientNum [set ${name}.numberClients]
	incr $name.numberClients

#	puts $name.client$clientNum
	LanHost $name.client$clientNum $name.lan.$prefix \
		$prefix.$hostNumber \
	        255.255.255.0 $routerAddress
	lappend $name.clientNames $name.client$clientNum
	lappend $name.clientAddresses $prefix.$hostNumber
	incr hostNumber
	
	# Start daemons on workstations.
	# telnet server uses port 23
	TelnetServer $name.client$clientNum.telnetd $epsilon $name.client$clientNum.tcp 23
    }

}
