#
# $Id: SwitchRcspN.tcl,v 1.7 1996-05-29 14:34:28-07 bmah Exp $
#
# SwitchRcspN.tcl
# Bruce A. Mah (bmah@CS.Berkeley.EDU)
#
# Copyright (c) 1995, 1996 The Regents of the University of California
# All Rights Reserved
# Created by Bruce A. Mah
# Department of Electrical Engineering and Computer Sciences,
#   Computer Science Division
#

#
# SwitchRcspN
#
# Input:  switch name (name), number of data ports (n),
#         max number of VCIDs (v), default link bandwidth (bw),
#	  number of priority levels (np).
#
# Output:  none
#
# Create an nxn ATM switch with RCSP output queues.  Delay bounds
# for the RCSP queues are set as follows:  The minimum delay corresponds
# to a cell service rate of 16 cell times at the default link bitrate.
# The Tick parameter (granularity of delay bounds) is set to half that
# value.
#
proc SwitchRcspN {name n v bw np} {
  
    global ConfRcspMode
    if {! [info exists ConfRcspMode] } {
	set confRcspMode non-work-conserving
    } else {
	set confRcspMode $ConfRcspMode
    }

    # Set up some parameters
    set minCellTimes 16
    set tickCellTimes [ expr {$minCellTimes / 2} ]
    set subqueues [ expr 1 << ($np + 1) ]

    set cellSize [ expr { 53 * 8 } ]
    set cellTime [ expr { $cellSize / double($bw) } ]

    # We need more subqueues to handle large eligibility offsets for
    # for bursts.  Our current heuristic is to have enough subqueues
    # to support up to one second's worth of buffering.
    while {$subqueues * $tickCellTimes * $cellTime < 1.0} {
	set subqueues [ expr $subqueues * 2 ]
#	puts "increased subqueues to $subqueues now can buffer for [expr {$subqueues * $tickCellTimes * $cellTime}] "
    }

    # Compute vector of delay times
    set cells $minCellTimes
    set delayVector [ list ]
    set i 0
    while {$i < $np} {
	set delay [ expr { $cells * $cellTime } ]
#	puts "delay bound for $cells cells is [ expr { $delay*1000 } ] ms"
	lappend delayVector $delay
	set cells [ expr { $cells * 2 } ]
	incr i
    }
    set tickTime [ expr { $tickCellTimes * $cellTime } ]

    # Create switch module (use n+1 ports...port 0 goes to the
    # switch controller)
    SwitchModule $name.switch [expr {$n+1}] $v

    # Create signalling entity (100 VCIs max, 256 routes max)
    SigRcsp $name.sig [expr {$n+1}] $v 256
    $name.sig mode $confRcspMode
    $name.sig switch $name.switch
    $name.sig gobackn $name.gobackn

    global ConfRcspExcess
    if { [info exists ConfRcspExcess] } {
	$name.sig excess $ConfRcspExcess
    }

    # Create input port and output port for switch controller
    # on port 0, with no delay and arbitrarily large bandwidth
    CellInputPort $name.input0 $name.switch 0 0
    CellQueueFifo $name.output0
    $name.output0 output $name.gobackn
    $name.output0 latency 0
    $name.output0 bitrate 1000000000
    $name.switch output 0 $name.output0

    # Create CellGoBackN entity for signalling messages.
    CellGoBackN $name.gobackn [expr {$n+1}] 512 0.50 0.10
    $name.gobackn up $name.sig
    $name.gobackn down $name.input0

    # Loop over all data ports and create both input ports and output
    # ports.
    for {set i 1} {$i < [expr $n+1]} {incr i} {

	CellInputPort $name.input$i $name.switch $i 0
	CellQueueRcsp $name.output$i $v $np $subqueues $tickTime
	$name.output$i mode $confRcspMode
	$name.output$i latency 0
	$name.output$i bitrate $bw
	$name.switch output $i $name.output$i
	$name.sig output $i $name.output$i $bw $tickTime $delayVector

    }

    # Create initial translation tables for control channels.  We
    # follow a scheme based on XUNET, where VCID 0 along each link is
    # used for signalling.  Cells arriving on port p are translated
    # to port 0 (to controller), VCID p to get send to the switch
    # controller.  Similarly, outgoing cells from port 0 (VCID v) are
    # translated to port v, VCID 0 to go to the next hop switch.
    for {set i 1} {$i < [expr $n+1]} {incr i} {
	$name.switch set $i 0 0 $i
	$name.switch set 0 $i $i 0
    }

    proc $name {} [ list puts -nonewline "$name (SwitchRcspN)" ]

}



