static char rcsid[] = "$Id: Aal.cc,v 1.12 1996-02-02 15:56:06-08 bmah Exp $";
//
// $Id: Aal.cc,v 1.12 1996-02-02 15:56:06-08 bmah Exp $
//
// Aal.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Simple simulated ATM adaptation layer (SSAAL?).
//
#include "Aal.h"
#include "AalFrame.h"
#include "AtmCell.h"
#include "util.h"
#include "inline.h"

//
// Aal::Handler
//
// (Event handler)
//
void Aal::Handler(Event *e) {

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    register int vcid;

    switch (type) {

	//
	// Cell arrival (presumably from below)
	//
	case (EventCellArrival):
	{

	    AtmCell *cell = (AtmCell *) data;
	    register bool accept;

	    vcid = cell->vcid;
	    
	    if (vcid >= maxVcids) {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("Aal");
		    *(sim->debugStream) <<
		      "received cell for out-of-bounds VCID " <<
		      vcid << "\n";
		});
		return;
	    }

	    IF_DEBUG(3, {
		DEBUG_PREFIX("Aal");
		*(sim->debugStream) << "inbound vcid " << vcid << 
		  " packet " << cell->packet << " cell " << cell->seq << "\n";
	    });

	    // Common case path:  Cell matches what was expected.
	    if ((cell->seq == inCellNumber[vcid]) &&
		(cell->packet == inPacketNumber[vcid])) {
		accept = true;
		inCellNumber[vcid]++;

		IF_DEBUG(4, {
		    DEBUG_PREFIX("Aal");
		    *(sim->debugStream) << "accept in-sequence cell\n";
		});
		
	    }
	    else {

		// If we got the first cell for the next packet, then that
		// means we missed the end of the current packet.  Reset,
		// throw away all cells for current packet, and start
		// accepting cells for the next packet.
		if ((cell->seq == 0) &&
		    (cell->packet > inPacketNumber[vcid])) {
		    
		    accept = true;
		    inCellNumber[vcid] = 1;
		    inPacketNumber[vcid] = cell->packet;

		    IF_DEBUG(3, {
			DEBUG_PREFIX("Aal");
			*(sim->debugStream) << "reset for next packet\n";
		    });
		}

		// Missed some cells, but still in the current packet, or
		// else we got cells from another packet (and missed the start).
		// Throw all these cells away, prepare to receive the next
		// packet.
		else {
		    accept = false;
		    inCellNumber[vcid] = 0;
		    inPacketNumber[vcid] = cell->packet + 1; // assumes that
		    // cells stay in
		    // order.
		    IF_DEBUG(3, {
			DEBUG_PREFIX("Aal");
			*(sim->debugStream) << "reject cell\n";
		    });
		}
	    }

	    // If the cell was good, now check for end of packet indicator.
	    // Pass up the packet if this was the last cell, prepare to receive
	    // first cell of next packet.
	    //
	    // Note:  cell->payload is a pointer to the AAL payload, not the
	    // payload of an individual cell.  This allows us to simulate
	    // the fragmentation and reassembly efficiently, without all of
	    // the buffering and copying that would be necessary in a real
	    // implementation.
	    if (accept && (cell->type == AtmCellTypeEndOfPacket)) {

		Lookup(up, upName);
		Post(time + 0, up, EventFrameArrival,
		     (void *) cell->payload); // XXX

		inCellNumber[vcid] = 0;
		inPacketNumber[vcid]++;

		cell->Detatch();

		IF_DEBUG(2, {
		    DEBUG_PREFIX("Aal");
		    *(sim->debugStream) << "passed up vcid " <<
		      vcid << " packet " << cell->packet << "\n";
		});

	    }

	    // In any case, we can get rid of the cell.  In real life, we would
	    // buffer up accepted cells, but in this simulation we can get away
	    // with not doing so and merely manipulating pointers to the
	    // higher-layer payloads (see AtmCell.h for more details).
	    delete cell;

	    break;
	}    

	//
	// Packet arrival from upper-layer protocol
	//
	case (EventFrameSend):
	{
	    AalFrame *aalFrame = (AalFrame *) data;

	    int numCells = (aalFrame->length + (AtmPayloadSize - 1))
	      / AtmPayloadSize;
	    
	    vcid = aalFrame->vcid;

	    if (vcid >= maxVcids) {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("Aal");
		    *(sim->debugStream) <<
		      "received cell for out-of-bounds VCID " <<
		      vcid << "\n";
		});
		return;
	    }

	    IF_DEBUG(2, {
		DEBUG_PREFIX("Aal");
		*(sim->debugStream) << "outbound vcid " <<
		  vcid << " packet " << outPacketNumber[vcid] <<
		    " cells " << numCells << "\n";
	    });

	    // When we send the packet out as cells, we need to preserve
	    // the inter-cell gap across cells for different packets.
	    // We maintain an expected transmission time to ensure that
	    // cells don't leave the AAL object separated by less than
	    // the gap.
	    if (time > expectedTime) {
		expectedTime = time;
	    }

	    // Send all cells up to (but not including) the last.
	    for (int i = 0; i < numCells-1; i++) {
		AtmCell *cell = new AtmCell(AtmCellTypeData,
					    aalFrame,
					    i,
					    outPacketNumber[vcid]);
		cell->vcid = aalFrame->vcid;
		cell->port = port;
		Lookup(down, downName);
		Post(expectedTime, down, EventCellArrival, cell);
		expectedTime += gap;
	    }

	    // Send last cell.
	    AtmCell *last = new AtmCell(AtmCellTypeEndOfPacket,
					aalFrame,
					numCells-1,
					outPacketNumber[vcid]);
	    last->vcid = aalFrame->vcid;
	    last->port = port;
	    Lookup(down, downName);
	    Post(expectedTime, down, EventCellArrival, last);
	    expectedTime += gap;

	    // Get ready for next packet.
	    outPacketNumber[vcid]++;

	    break;
	}
	
	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("Aal");
		*(sim->debugStream) << "bad EventType " << type << "\n";
	    });
	}
	break;

    }

}

//
// Aal::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String Aal::Output()
{
    String out;
    char buf[30];

    sprintf(buf, "%d", maxVcids);
    out = name + " [Aal] maxVcids " + buf;

    sprintf(buf, "%f", gap);
    out += (String) " gap " + buf;

    sprintf(buf, "%d", port);
    out += (String) " port " + buf;

    out += (String) " up " + upName + " down " + downName;

    return out;
}

//
// Aal::ResetIn, Aal::ResetOut
//
// Input:  VCID to reset (vcid), initial packet sequence number (seqno)
//
// Output:  None.
//
// Reset either the input or output sides of the AAL, for a given
// virtual circuit.  Note that since virtual circuits are simplex only,
// the space of input and output VCIDs are completely separate from
// each other.
//
void Aal::ResetIn(int vcid, int seqno)
{

    if (vcid < maxVcids) {
	inPacketNumber[vcid] = seqno;
	inCellNumber[vcid] = 0;
    }
    else {
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("Aal");
	    *(sim->debugStream) <<
	      "ResetIn for out-of-bounds VCID " <<
		vcid << "\n";
	});
	return;
    }
}

void Aal::ResetOut(int vcid, int seqno)
{
    if (vcid < maxVcids) {
	outPacketNumber[vcid] = seqno;
    }
    else {
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("Aal");
	    *(sim->debugStream) <<
	      "ResetOut for out-of-bounds VCID " <<
		vcid << "\n";
	});
	return;
    }
}

//
// AalCmd
//
// (Tcl command handler procedure)
//
// Aal name vcids gap port
//
int AalCmd(ClientData clientData, Tcl_Interp *interp,
	   int argc, char *argv[])
{
    if (argc != 5) {
	Tcl_SetResult(interp, "usage:  Aal name maxvcid gap port",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of new CellGoBackN
	int mv;			// maximum number of VCIs
	double g;		// inter-cell gap
	int p;			// port number on host "switch"
	int rc;			// return code from Tcl library
	Aal *aal;		// new object
	extern Simulation *sim;	// global simulation object

	// get name
	nm = argv[1];

	// get numeric arguments
	rc = Tcl_GetInt(interp, argv[2], &mv);
	if (rc == TCL_OK) {

	    rc = Tcl_GetDouble(interp, argv[3], &g);
	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[4], &p);
		if (rc == TCL_OK) {

		    aal = new Aal(nm, sim, mv, g, p);
		    Tcl_CreateCommand(interp, nm,
				      AalObjectCmd,
				      (ClientData) aal,
				      AalDeleteProc);
		}
	    }
	}

	return rc;
	
    }
}

//
// AalObjectCmd
//
// (Tcl object handler procedure)
//
int AalObjectCmd(ClientData clientData, Tcl_Interp *interp,
		 int argc, char *argv[])
{
    
    Aal *aal = (Aal *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }
    else {

	//
	// aal up object
	//
	// Connect adaptation layer output to upper-layer protocol
	//
	if (!strcmp(argv[1], "up")) {

	    if (argc == 3) {

		String outName;	// object name

		outName = argv[2];

		// set new object...zero out cached object pointer
		aal->upName = outName;
		aal->up = NULL;

		return (TCL_OK);
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}
	//
	// aal down object
	//
	// Connect adaptation layer output to lower-layer object
	// (usually SwitchModule type object)
	//
	else if (!strcmp(argv[1], "down")) {

	    if (argc == 3) {

		String outName;	// object name

		outName = argv[2];

		// set new object...zero out cached object pointer
		aal->downName = outName;
		aal->down = NULL;

		return (TCL_OK);
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}
	else {
	    return(ScheduledObjectCmd(clientData, interp, argc, argv));
	}
	
    }
}

//
// AalDeleteProc
//
// (Tcl DeleteProc)
//
void AalDeleteProc(ClientData clientData)
{
    delete ((Aal *) clientData);
}
