// -*- c++ -*-
// $Id: AtmCell.h,v 1.8 1996-02-09 10:03:31-08 bmah Exp $
//
// AtmCell.h
//
// Copyright (c) 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//

#ifndef ATMCELL_H
#define ATMCELL_H

#include <stdlib.h>
#include <iostream.h>
#include "Pdu.h"

enum AtmCellType {
    AtmCellTypeData,
    AtmCellTypeEndOfPacket,
    AtmCellTypeAck
};

class AtmCell : public Pdu {

    // Public members
    public:

    int vcid;			// VCI
    int vpid;			// VPI (not currently supported)
    const AtmCellType type;	// cell type

    // These public members are designed for efficient simulation
    // of an AAL5- or AALX-like protocol.  Instead of actually
    // calculating the length and checksum of the cells making up
    // an AAL5 frame, we watch the cells for the current packet,
    // using the sequence and packet numbers to detect losses.
    // The payload can then point to the AAL5 (or whatever) frame;
    // it just "magically" appears out of the ATM network.
    Pdu *payload; 		// payload pointer
    int seq;			// sequence number
    const int packet;		// packet number

    // This member is designed to more efficiently simulate an
    // ATM switch backplane.  When a cell appears on an input port
    // at some switch, the address translation mechanism needs to
    // know the input port of the cell.  Rather than encapsulate
    // the cell in some kind of "inside the switch" data structure,
    // we just tag the cell with its input port before it gets
    // to the address translation table.
    // 
    // This member only has a useful meaning during forwarding.
    int port; 			// input port

    // This member is used to make queueing more efficient for those
    // queueing disciplines that maintain linked lists of cells.
    // The philosophy here is similar to that of the port member above.
    //
    // This member only has a useful meaning when a cell is enqueued
    // somewhere.
    AtmCell *next; 		// next cell in SLL of cells

    // This member acts as a per-cell timestamp for various jitter-control
    // schemes.
    double time;		// timestamp

    AtmCell(AtmCellType t, Pdu *p, int s, int pn) :
    type(t), payload(p), seq(s), packet(pn) {
	if (payload) {
	    (payload->refcount)++;
	}
    }

    AtmCell(const AtmCell &c) :
    vcid(c.vcid), type(c.type), seq(c.seq),
    packet(c.packet), port(c.port) {
	if (c.payload) {
	    payload = c.payload->Copy();
	    (payload->refcount)++;
	}
    }

    virtual ~AtmCell() {
	if (payload) {
	    (payload->refcount)--;
	    if (payload->refcount == 0) {
		delete payload;
	    }
	}
    }

    void Detatch() {
	if (payload) {
	    (payload->refcount)--;
	}
	payload = NULL;
    }

    virtual ostream& Print(ostream &s);

};

typedef AtmCell *AtmCellPtr;

const int AtmPayloadSize = 48;	// payload size in bytes
const int AtmCellSize = (5+48)*8; // transmission time on wire in bits

extern ostream& operator<<(ostream &s, AtmCellType t);

#endif // ATMCELL_H
