static char rcsid[] = "$Id: AtmDeviceDriver.cc,v 1.44 1996-07-31 21:36:07-07 bmah Exp $";
//
// $Id: AtmDeviceDriver.cc,v 1.44 1996-07-31 21:36:07-07 bmah Exp $
//
// AtmDeviceDriver.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// ATM device driver base class.  Functionally equivalent to the XUNET
// IP device driver plus some parts of the XUNET SigHost signalling
// software.
//

#include <iostream.h>
#include <iomanip.h>

#include "AtmDeviceDriver.h"
#include "DeviceDriver.h"
#include "AalFrame.h"
#include "SigMessage.h"
#include "Event.h"
#include "inline.h"

Uniform *AtmDeviceDriver::skewGenerator = NULL;

//
// Data for enum printing routines
//
char *AtmDeviceDriverQosTypeName[] = {
    "null",
    "noqos",
    "qos1",
    "fallback",
    "telnet",
    "ftp",
    "http",
    "audio",
    "video",
    "av",
    "isp",
};

char *AtmDeviceDriverMuxTypeName[] = {
    "null",
    "router",
    "conv",
    "app",
};

char *AtmDeviceDriverVcTypeName[] = {
    "null",
    "pvc",
    "svc",
    "svccache",
};

char *AtmDeviceDriverConvTypeName[] = {
    "default",
    "telnet+",
    "telnet-",
    "ftpdata+",
    "ftpdata-",
    "ftpctl+",
    "ftpctl-",
    "http+",
    "http-",
    "audio",
    "video",
};

char *AtmRouteStateName[] = {
    "?",			// unused
    "^",			// establishing
    "+",			// established
    ".",			// idle
    "z",			// zombie
};


//
// AtmDeviceDriver::AtmDeviceDriver
//
// (object constructor)
//
AtmDeviceDriver::AtmDeviceDriver(const String &name, Simulation *si, int ma,
				 int mh, int mr) :
DeviceDriver(name, si), maxAddresses(ma), hashSize(mh), maxRequests(mr) {

    int i;			// straw loop counter

    // Check skew random number generator, initialize if not already
    // done.
    if (skewGenerator == NULL) {
        skewGenerator = new Uniform(0, AtmDeviceDriverPollInterval,
				    sim->rng);
    }

    // Allocate and zero out ATM address map
    atmArpMap = new AtmArpEntry[maxAddresses];
    for (i = 0; i < maxAddresses; i++) {
	atmArpMap[i].ip.s_addr = 0;
    }

    // Allocate and zero out VCID map
    atmVcidMap = new (AtmRoute *) [hashSize];
    for (i = 0; i < hashSize; i++) {
	atmVcidMap[i] = NULL;
    }

    // Allocate and initialize array of pending signalling requests
    pendingRoute = new (AtmRoute *) [maxRequests];
    for (i = 0; i < maxRequests; i++) {
        pendingRoute[i] = NULL;
    }

    // Set initial default behavior
    qosType = AtmDeviceDriverQosTypeNull;
    muxType = AtmDeviceDriverMuxTypeNull;
    vcType = AtmDeviceDriverVcTypeNull;
    maxIdle = 60.0;
    maxCache = 300.0;

    // Zero out statistics
    numConnCached = 0;
    numConnRequests = 0;
    numConnFirstSucceeded = 0;
    numConnFallbacks = 0;
    numConnFallbackCached = 0;
    numConnFallbackRequests = 0;
    numConnFallbackSucceeded = 0;
    numConnSucceeded = 0;
    numConnFailed = 0;
    for (i = 0; i < AtmDeviceDriverConvTypeMAX; i++) {
        cacheCount[i] = 0;
        connCount[i] = 0;
	failCount[i] = 0;
	byteCount[i] = 0;
	packetCount[i] = 0;
    }
    nowEstablished = 0;

    // Post initial timeout event.
    Post((*skewGenerator)(), this, EventTimer, NULL);

}

//
// AtmDeviceDriver::~AtmDeviceDriver
//
// (object destructor)
//
AtmDeviceDriver::~AtmDeviceDriver() {

    int i;
    
    *(sim->logStream) << name <<
      " [AtmDeviceDriver] connections cached " << numConnCached <<
	" requested " << numConnRequests <<
	" first-succeeded " << numConnFirstSucceeded <<
	" fallbacks " << numConnFallbacks <<
	" fallbacks-cached " << numConnFallbackCached <<
	" fallbacks-requested " << numConnFallbackRequests <<
	" fallbacks-succeeded " << numConnFallbackSucceeded <<
	" total-succeeded " << numConnSucceeded <<
	" total-failed " << numConnFailed << "\n";
    *(sim->logStream) << name <<
      " [AtmDeviceDriver] packets in " << if_ipackets << " out " <<
	if_opackets << " bytes in " <<
	if_ibytes << " out " << if_obytes << "\n";

    *(sim->logStream) << name << " [AtmDeviceDriver] cached connections";
    for (i = 0; i < AtmDeviceDriverConvTypeMAX; i++) {
	*(sim->logStream) << " " << (AtmDeviceDriverConvType) i << " " << cacheCount[i];
    }
    *(sim->logStream) << "\n";

    *(sim->logStream) << name << " [AtmDeviceDriver] successful establishments";
    for (i = 0; i < AtmDeviceDriverConvTypeMAX; i++) {
	*(sim->logStream) << " " << (AtmDeviceDriverConvType) i << " " << connCount[i];
    }
    *(sim->logStream) << "\n";

    *(sim->logStream) << name << " [AtmDeviceDriver] failed establishments";
    for (i = 0; i < AtmDeviceDriverConvTypeMAX; i++) {
	*(sim->logStream) << " " << (AtmDeviceDriverConvType) i << " " << failCount[i];
    }
    *(sim->logStream) << "\n";

    *(sim->logStream) << name << " [AtmDeviceDriver] success rate%";
    for (i = 0; i < AtmDeviceDriverConvTypeMAX; i++) {
	*(sim->logStream) << " " << (AtmDeviceDriverConvType) i << " ";
	if (connCount[i] + failCount[i] != 0) {
	    *(sim->logStream) << (double) connCount[i] /
	                         (double) (connCount[i] + failCount[i]);
	}
	else {
	    *(sim->logStream) << "0";
	}
    }
    *(sim->logStream) << "\n";

    *(sim->logStream) << name << " [AtmDeviceDriver] classified bytes";
    for (i = 0; i < AtmDeviceDriverConvTypeMAX; i++) {
	*(sim->logStream) << " " << (AtmDeviceDriverConvType) i << " " << byteCount[i];
    }
    *(sim->logStream) << "\n";

    *(sim->logStream) << name << " [AtmDeviceDriver] classified packets";
    for (i = 0; i < AtmDeviceDriverConvTypeMAX; i++) {
	*(sim->logStream) << " " << (AtmDeviceDriverConvType) i << " " << packetCount[i];
    }
    *(sim->logStream) << "\n";
    
    delete [] atmArpMap;
    delete atmVcidMap;
}

//
// AtmDeviceDriver::Handler
//
// (Event handler)
//
void AtmDeviceDriver::Handler(Event *e)
{
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    switch (type) {

	//
	// EventPacketSend:  Packet outbound to the ATM network.
	//
	case EventPacketSend:
	{

	    IpPacket *ipPacket = (IpPacket *) data;
	    AtmRoute *route;	// route to use for this packet

	    IF_DEBUG(2, {
		DEBUG_PREFIX("AtmDeviceDriver");
		*(sim->debugStream) << "EventPacketSend " << ipPacket << "\n";
	    });

	    // Determine an existing VC (if any) for this packet
	    route = GetAtmRoute(ipPacket);

	    // If no existing VC can be used for this packet,
	    // then kick off a virtual circuit establishment and
	    // queue up the packet for later transmission.
	    if (!route) {

		IF_DEBUG(1, {
		    DEBUG_PREFIX("AtmDeviceDriver");
		    *(sim->debugStream) << "needs new route " <<
		      ipPacket << "\n";
		});

		AtmRoute *newRoute;
		int hash;
		
		// Get an AtmRoute structure with the correct
		// parameters, etc. for establishment.  Store it in
		// the hash chain.
		newRoute = MakeAtmRoute(ipPacket);

		if (newRoute == NULL) {

		    IF_DEBUG(0, {
			DEBUG_PREFIX("AtmDeviceDriver");
			*(sim->debugStream) << "couldn't get new AtmRoute\n";
		    });

		    if_oerrors++;
		    delete ipPacket;
		    break;
		}

		hash = AtmRouteHash(newRoute);
		newRoute->lastUse = -1.0; // route never used
		newRoute->state = AtmRouteStateEstablishing;
		newRoute->next = atmVcidMap[hash];
		newRoute->prev = NULL;
		newRoute->hash = hash;
		atmVcidMap[hash] = newRoute;
		if (newRoute->next) {
		    newRoute->next->prev = newRoute;
		}

		// Enqueue packet on pending queue.  The queue
		// is implemented as an SLL with both head and tail
		// pointers.
		QueueElem *newQueueElem = new QueueElem;

		newQueueElem->payload = ipPacket;
		newQueueElem->next = NULL;
		newRoute->queueHead = newQueueElem;
		newRoute->queueTail = newQueueElem;
		newRoute->queueCount = 1;

		// Do the call for this route
		DoCallForRoute(time, newRoute);

	    }

	    // Otherwise, we have an AtmRoute we can use for this packet.
	    else {

		IF_DEBUG(2, {
		    DEBUG_PREFIX("AtmDeviceDriver");
		    *(sim->debugStream) << "uses existing route\n";
		});

		// If route already has an active VC, send the packet
		// immediately.
		if (route->state == AtmRouteStateEstablished) {

		    AalFrame *aalFrame = new AalFrame(ipPacket);

		    aalFrame->vcid = route->vcid;
		    aalFrame->length = ipPacket->ip_len + AalHeaderSize; 
		    
		    Lookup(down, downName);
		    Post(time + 0, down, EventFrameSend,
			 (void *) aalFrame);

		    // Update statistics
		    route->lastUse = time;
		    if_opackets++;
		    if_obytes += ipPacket->ip_len;
		    packetCount[route->type]++;
		    byteCount[route->type] += ipPacket->ip_len;
		    
		    IF_DEBUG(2, {
			DEBUG_PREFIX("AtmDeviceDriver");
			*(sim->debugStream) << "send packet vcid " <<
			  route->vcid << "\n";
		    });

		}

		// If route has a VC establishment in progress, queue
		// the packet.
		else if (route->state == AtmRouteStateEstablishing) {

		    IF_DEBUG(2, {
			DEBUG_PREFIX("AtmDeviceDriver");
			*(sim->debugStream) << "queue packet\n";
		    });

		    QueueElem *newQueueElem = new QueueElem;

		    newQueueElem->payload = ipPacket;
		    newQueueElem->next = NULL;
		    
		    if (route->queueHead == NULL) {
			route->queueHead = newQueueElem;
			route->queueTail = newQueueElem;
		    }
		    else {
			route->queueTail->next = newQueueElem;
			route->queueTail = newQueueElem;
		    }

		    route->queueCount++;
		    if (route->queueCount > 10) {
			IF_DEBUG(1, {
			    DEBUG_PREFIX("AtmDeviceDriver");
			    *(sim->debugStream) << "warning:  >= 10 packets queued...last was" << ipPacket << "\n";
			});
		    }
		    
		}

	    }

	}
        break;

	//
	// EventFrameArrival:  AAL frame showed up.
	//
	case EventFrameArrival:
	{

	    AalFrame *aalFrame = (AalFrame *) data;
	    IpPacket *ipPacket = (IpPacket *) aalFrame->payload;

	    IF_DEBUG(2, {
		DEBUG_PREFIX("AtmDeviceDriver");
		*(sim->debugStream) << "input packet " << ipPacket << "\n";
	    });

	    if (ipPacket) {

		// Detatching the AalFrame will properly decrement
		// the reference count.
		aalFrame->Detatch();

		// Pass the packet up to the IP layer
		Lookup(up, upName);
		Post(time + 0, up, EventPacketArrival, (void *) ipPacket);

		// Update statistics
		if_ipackets++;
		if_ibytes += ipPacket->ip_len;

	    }
	    else {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("AtmDeviceDriver");
		    *(sim->debugStream) << "got null packet\n";
		});
	    }
	    
	    // Throw away what's left of the AAL frame
	    delete aalFrame;
	    
	}
	break;

	//
	// EventTimer:  Timer went off.  We may need to idle some
	// VCs.
	//
	case (EventTimer):
	{

	    IF_DEBUG(4, {
		DEBUG_PREFIX("AtmDeviceDriver");
		*(sim->debugStream) << "timer\n";
	    });

	    // Walk the hash table...get rid of anything that should be
	    // timed out.
	    for (int hash = 0; hash < hashSize; hash++) {
		AtmRoute *currentRoute = atmVcidMap[hash]; // route we're
							   // looking at
		
		while (currentRoute) {
		    if (TimeoutAtmRoute(currentRoute, time) == true) {

			IF_DEBUG(1, {
			    DEBUG_PREFIX("AtmDeviceDriver");
			    *(sim->debugStream) << "ready to timeout route\n";
			});

			// Try to get a request ID for the signalling
			// downcall.  If we can, then proceed with doing
			// the timeout actions.  If not, the driver can't
			// handle any more concurrent signalling requests.
			// In this particular case, it's not really an
			// error; we can just keep the route around for
			// for another polling interval; eventually it'll
			// go away (we hope).
			int call = GetRequestId(currentRoute);
			if (call > -1) {

			    // Rejoin the DLL around the deleted route
			    // entry.  Don't get rid of it yet, because
			    // we need to hold some of the state around
			    // for a little while...at least until the
			    // VC assigned to this route actually gets
			    // torn down.
			    if (currentRoute->next) {
				currentRoute->next->prev = currentRoute->prev;
			    }
			    if (currentRoute->prev) {
				currentRoute->prev->next = currentRoute->next;
			    }
			    else {
				atmVcidMap[hash] = currentRoute->next;
			    }

			    // Now free everything related to currentRoute.
			    // First start a teardown for the VC it was using.
			    currentRoute->state = AtmRouteStateZombie;
			    ApiMessage *teardown = new ApiMessage;
			    teardown->client = this;
			    teardown->type = ApiTypeDelete;
			    teardown->vcid = currentRoute->vcid;
			    teardown->requestId = call;

			    Lookup(sigHost, sigHostName);
			    Post(time + 0, sigHost, EventCall,
				 (void *) teardown); // XXX

			    // Free the traffic descriptor and ancillary
			    // data structures.
			    if (currentRoute->traffic) {
				delete currentRoute->traffic;
			    }

			    // Free all packets that were queued up for
			    // this AtmRoute, by walking the packet queue
			    // for this route
			    while (currentRoute->queueHead != NULL) {

				QueueElem *deadPacket = currentRoute->queueHead;
				delete deadPacket->payload;
				
				currentRoute->queueHead =
				  currentRoute->queueHead->next;

				if_oerrors++;

				delete deadPacket;

			    } // while (currentRoute->queueHead
			} // if call
		    } // if (TimeoutAtmRoute
		    
		    currentRoute = currentRoute->next;

		} // while (currentRoute
	    } // for (int hash

	    // Set timer to go off again
	    Post(time + AtmDeviceDriverPollInterval, this, EventTimer, NULL);

	}
	break;

	//
	// EventCall:  Call to the signalling system returned.
	//
	case (EventCall):
	{

	    ApiMessage *apiMessage = (ApiMessage *) data;

	    IF_DEBUG(1, {
		DEBUG_PREFIX("AtmDeviceDriver");
		*(sim->debugStream) << "EventCall for call " <<
		  apiMessage->requestId << "\n";
	    });

	    switch (apiMessage->type) {

		case (ApiTypeReturn):
		{

		    IF_DEBUG(1, {
			DEBUG_PREFIX("AtmDeviceDriver");
			*(sim->debugStream) << "ApiTypeReturn received\n";
		    });

		    // Find call ID and do a little sanity checking.
		    if (apiMessage->requestId >= maxRequests) {
			IF_DEBUG(0, {
			    DEBUG_PREFIX("AtmDeviceDriver");
			    *(sim->debugStream) << "received bad requestId in return " <<
			      apiMessage->requestId << "\n";
			});
			break;
		    }

		    AtmRoute *callRoute = pendingRoute[apiMessage->requestId];
		    ReleaseRequestId(apiMessage->requestId);

		    if (callRoute == NULL) {
			IF_DEBUG(0, {
			    DEBUG_PREFIX("AtmDeviceDriver");
			    *(sim->debugStream) << "has null route for returned requestID" <<
			      apiMessage->type << "\n";
			});
			break;
		    }
		    
		    // Use the route's state to determine how to handle the
		    // return.
		    switch (callRoute->state) {

			// Channel being established.
			case (AtmRouteStateEstablishing):
			{

			    // If establishment was successful, then
			    // update route state and send all queued
			    // packets.
			    if (apiMessage->code == SigOK) {

				IF_DEBUG(1, {
				    DEBUG_PREFIX("AtmDeviceDriver");
				    *(sim->debugStream) << "route VC successful request " << apiMessage->requestId << " vcid " << apiMessage->vcid << "\n";
				});
				
				callRoute->vcid = apiMessage->vcid;
				callRoute->state = AtmRouteStateEstablished;

				TransmitRoute(time, callRoute);

				if (callRoute->fallback) {
				    numConnFallbackSucceeded++;
				}
				else {
				    numConnFirstSucceeded++;
				}
				numConnSucceeded++;
				nowEstablished++;
				connCount[callRoute->type]++;

			    }

			    // If establishment failed, then try the
			    // fallback virtual circuit for this route.
			    // If that doesn't work (or there isn't
			    // one, then delete the route and flush
			    // all queued packets.
			    else {

				failCount[callRoute->type]++;
				numConnFailed++;

				IF_DEBUG(1, {
				    DEBUG_PREFIX("AtmDeviceDriver");
				    *(sim->debugStream) <<
				    "setup failed request " <<
				    apiMessage->requestId <<
				    " " << apiMessage->code << "\n";
				});

				if (FallbackRoute(callRoute)) {
				    numConnFallbacks++;

				    AtmRoute *cachedRoute = NULL;

				    // If doing VC caching, try to get
				    // a cached fallback route and VC.
				    if (vcType ==
					AtmDeviceDriverVcTypeSvccache) {
					 cachedRoute =
					   GetCachedFallback(callRoute);
				    }

				    // If we have a cached route, use it.
				    if (cachedRoute) {

					// Temporarily give the original
					// route the VCID of the cached
					// route, long enough to forward
					// the queued packets.
					callRoute->vcid = cachedRoute->vcid;
					TransmitRoute(time, callRoute);

					// Now get rid of the requested
					// route.
					DeleteRoute(callRoute);

					IF_DEBUG(1, {
					    DEBUG_PREFIX("AtmDeviceDriver");
					    *(sim->debugStream) <<
					      "using cached fallback connection\n";
					});

					numConnFallbackCached++;

				    }
				    else {
					
					// No cached route, so go request
					// one.
					DoCallForRoute(time, callRoute);

					numConnFallbackRequests++;

					IF_DEBUG(1, {
					    DEBUG_PREFIX("AtmDeviceDriver");
					    *(sim->debugStream) <<
					      "initiate fallback connection\n";
					});

				    }

				}
				else {
				    DeleteRoute(callRoute);
				}
			    }

			}
			break;

			// Channel being torn down.  We can finally
			// get rid of the associated route information.
			case (AtmRouteStateZombie):
			{

			    IF_DEBUG(1, {
				DEBUG_PREFIX("AtmDeviceDriver");
				*(sim->debugStream) << "route VC teardown completed vcid " << apiMessage->vcid << "\n";
			    });

			    nowEstablished--;
			    
			    delete callRoute;
			}  
			break;

			default:
			{
			    IF_DEBUG(0, {
				DEBUG_PREFIX("AtmDeviceDriver");
				*(sim->debugStream) << "unexpected route state " <<
				  callRoute->state << "\n";
			    });
			}

		    }

		}
		break;
		
		default:
		{
		    IF_DEBUG(0, {
			DEBUG_PREFIX("AtmDeviceDriver");
			*(sim->debugStream) << "received bad ApiMessage type " <<
			  apiMessage->type << "\n";
		    });
		}
		break;
	    }

	    delete apiMessage;

	}
	break;

	case (EventTimer2):
	{

	    // Print out statistics over this interval
	    IF_DEBUG(0, {
		DEBUG_PREFIX("AtmDeviceDriver");
		*(sim->debugStream) << "interval " << interval << " current-routes " << nowEstablished << "\n";
	    });

	    // Wake us up again later
	    Post(time + interval, this, EventTimer2, NULL);

	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("AtmDeviceDriver");
		*(sim->debugStream) << "bad EventType " << type << "\n";
	    });
	}
	break;

    }
}

//
// The methods GetAtmRoute, PacketClassifier, BindRoute, MakeAtmRoute,
// FallbackRoute, and TimeoutAtmRoute define much of the behavior of
// the AtmDeviceDriver.  Together, they specify different QOS,
// multiplexing, and timeout policies.
//

//
// AtmDeviceDriver::GetAtmRoute
//
// Input;  IP packet
//
// Output:  Pointer to the ATM route structure corresponding to the IP packet
//          (NULL if no match found)
//
// Return the ATM route structure appropriate to an IP packet.
// If there is no such structure, return NULL.  This function
// implements most of the multiplexing policy for the system.
//
AtmRoute *AtmDeviceDriver::GetAtmRoute(IpPacket *ipPacket)
{
    int hash = AtmIpHash(ipPacket); // hash bucket to search
    AtmRoute *best = NULL;	    // best route found so far
    AtmRoute *cached = NULL;	    // cached route if we can find one
    
    // Walk the hash bucket and try to find a key that fits the
    // packet.
    for (AtmRoute *current = atmVcidMap[hash];
	 current != NULL;
	 current = current->next) {

	//
	// Use multiplexing type to determine the type of tests that
	// need to be run on the packet.
	//
	switch (muxType) {
	    
	    // Multiplexing per router pair...make sure the ATM subnet
	    // egress router matches.
	    case (AtmDeviceDriverMuxTypeRouter):
	    {
		if (current->rk.ip_next.s_addr != ipPacket->ip_next.s_addr) {
		    continue;
		}
	    }
	    break;

	    // Multiplexing policy is per-conversation:
	    // All TCP connections and UDP streams map to
	    // unique VCs, other transport protocols
	    // (if any) are multiplexed per-host pair.
	    case (AtmDeviceDriverMuxTypeConv):
	    {

		// If IP addresses and protocol number aren't equal,
		// this can't possibly be a match.
		if ((current->rk.ip_src.s_addr != ipPacket->ip_src.s_addr) ||
		    (current->rk.ip_dst.s_addr != ipPacket->ip_dst.s_addr) ||
		    (current->rk.ip_p != ipPacket->ip_p)) {
		    continue;
		}
		
		// If it's TCP, make sure port numbers match up.
		if (ipPacket->ip_p == IPPROTO_TCP) {
		    TcpSegment *tcpHeader = (TcpSegment *) ipPacket->payload;

		    if ((current->rk.tp.tcp.th_sport != tcpHeader->th_sport) ||
			(current->rk.tp.tcp.th_dport != tcpHeader->th_dport)) {
			continue;
		    }
		}

		// If it's UDP, then do the same check.  Otherwise we let it
		// through.
		else if (ipPacket->ip_p == IPPROTO_UDP) {
		    UdpPacket *udpHeader = (UdpPacket *) ipPacket->payload;

		    if ((current->rk.tp.udp.uh_sport != udpHeader->uh_sport) ||
			(current->rk.tp.udp.uh_dport != udpHeader->uh_dport)) {
			continue;
		    }
		}

	    }
	    break;

	    // Multiplexing policy is per-application per host pair.
	    // Server ports define unique applications for UDP and TCP;
	    // all other transport protocols each define unique applications.
	    case (AtmDeviceDriverMuxTypeApp):
	    {
		//
		// If IP addresses and protocol number aren't equal, this can't
		// possibly be a match.
		if ((current->rk.ip_src.s_addr != ipPacket->ip_src.s_addr) ||
		    (current->rk.ip_dst.s_addr != ipPacket->ip_dst.s_addr) ||
		    (current->rk.ip_p != ipPacket->ip_p)) {
		    continue;
		}
		
		// If it's TCP, make sure port numbers match up.
		if (ipPacket->ip_p == IPPROTO_TCP) {
		    TcpSegment *tcpHeader = (TcpSegment *) ipPacket->payload;

		    //
		    // If either of the ports is ephemeral, we assume that that
		    // port is the client side of the connection and set it to
		    // zero as a wildcard.
		    //
		    u_short sport = tcpHeader->th_sport; // copy of source port
		    u_short dport = tcpHeader->th_dport; // copy of dest port
		    if (sport >= TcpMinEphemeralPort) {
			sport = 0;
		    }
		    if (dport >= TcpMinEphemeralPort) {
			dport = 0;
		    }

		    if ((current->rk.tp.tcp.th_sport != sport) ||
			(current->rk.tp.tcp.th_dport != dport)) {
			continue;
		    }
		}

		// If it's UDP, then do the same check.  Otherwise we let it
		// through.
		else if (ipPacket->ip_p == IPPROTO_UDP) {
		    UdpPacket *udpHeader = (UdpPacket *) ipPacket->payload;

		    if ((current->rk.tp.udp.uh_sport != udpHeader->uh_sport) ||
			(current->rk.tp.udp.uh_dport != udpHeader->uh_dport)) {
			continue;
		    }
		}
	    }
	    break;

	    default:
	    IF_DEBUG(0, {
		DEBUG_PREFIX2("AtmDeviceDriver");
		*(sim->debugStream) << "bad muxType\n";
	    });
	    break;
	}
	
	// Check state of VC
	if ((current->state == AtmRouteStateEstablished) ||
	     (current->state == AtmRouteStateEstablishing)) {
	    best = current;
	}

    }

    //
    // If we haven't yet found a suitable virtual circuit and we're doing
    // VC caching, then classify the packet and try to find a suitable
    // idled VC appropriate to this classification.
    //
    if ((best == NULL) &&
	(vcType == AtmDeviceDriverVcTypeSvccache)) {

	// Classify the packet, then run through the hash table looking
	// for a VC of the same type going to the same destination.
        AtmDeviceDriverConvType convType;

	switch (qosType) {
	    case (AtmDeviceDriverQosTypeNoqos):
	    {
	        convType = AtmDeviceDriverConvTypeDefault;
	    }
	    break;

	    case (AtmDeviceDriverQosTypeQos1):
	    case (AtmDeviceDriverQosTypeFallback):
	    case (AtmDeviceDriverQosTypeTelnet):
	    case (AtmDeviceDriverQosTypeFtp):
	    case (AtmDeviceDriverQosTypeHttp):
	    case (AtmDeviceDriverQosTypeAudio):
	    case (AtmDeviceDriverQosTypeVideo):
	    case (AtmDeviceDriverQosTypeAv):
	    case (AtmDeviceDriverQosTypeIsp):
	    {
	        convType = PacketClassifier(ipPacket);
	    }
	    break;

	    default:
	      IF_DEBUG(0, {
		  DEBUG_PREFIX2("AtmDeviceDriver");
		  *(sim->debugStream) << "bad qosType\n";
		  abort();
	      });
	    break;

	}

	for (AtmRoute *current = atmVcidMap[hash];
	     current != NULL;
	     current = current->next) {

	    // Weed out unsuitable idled VCs
	    if (current->state != AtmRouteStateIdle) {
		continue;
	    }
	    if (current->type != convType) {
		continue;
	    }
	    if (current->rk.ip_next.s_addr != ipPacket->ip_next.s_addr) {
		continue;
	    }
	    best = current;
	    BindRoute(current, ipPacket);
	    current->state = AtmRouteStateEstablished;
	    numConnCached++;
	    cacheCount[convType]++;

	    IF_DEBUG(1, {
		DEBUG_PREFIX2("AtmDeviceDriver");
		*(sim->debugStream) << "used cached VCID\n";
	    });

	    break;

	}

    }

    return best;

}

//
// AtmDeviceDriver::GetCachedFallback
//
// Input:  Existing route (callRoute)
//
// Output:  Cached route
//
// Find a cached fallback connection that can be used for packets
// queued for the input route.  The cached route gets bound to the
// conversation from the original route.  Note that the caller is
// responsible for deallocating the original route, and that this
// needs to be done quasi-atomically before another packet shows up.
//
// If NULL returned, no fallback is available.
//
AtmRoute *AtmDeviceDriver::GetCachedFallback(AtmRoute *callRoute) {
    
    AtmRoute *best = NULL;	// no route found

    // We loop through the hash bucket for the called route.
    // A lot of this code was originally lifted from the cache code
    // in AtmDeviceDriver::GetAtmRoute().
    for (AtmRoute *current = atmVcidMap[callRoute->hash];
	 current != NULL;
	 current = current->next) {

	// Weed out VCs that aren't idle, aren't default (best-effort)
	// type, or aren't headed in the right direction.
	if (current->state != AtmRouteStateIdle) {
	    continue;
	}
	if (current->type != AtmDeviceDriverConvTypeDefault) {
	    continue;
	}
	if (current->rk.ip_next.s_addr != callRoute->rk.ip_next.s_addr) {
	    continue;
	}

	// Found a route.  Note that we need to bind the route to a
	// conversation.  To do this we must get the first packet from
	// the requested route.  Note that by logically, there *must* be
	// a packet enqueued, since the original route was motivated by
	// a packet arriving at the outbound interface.  If there isn't
	// anything in queue, something is very very wrong.
	best = current;
	if (callRoute->queueHead) {
	    BindRoute(current, callRoute->queueHead->payload);
	}
	else {
	    IF_DEBUG(0, {
		DEBUG_PREFIX2("AtmDeviceDriver");
		*(sim->debugStream) <<
		  "Can't find packet to compute binding for cached fallback\n";
		abort();	// give up and die.
	    });
	}
	current->state = AtmRouteStateEstablished;
	numConnCached++;
	break;

    }

    return best;

}

//
// AtmDeviceDriver::PacketClassifier
//
// Input:  IP packet
//
// Output:  One-byte packet type
//
// Attempt to find the conversation type of an IP packet for the
// purpose of computing QOS parameters.
//
AtmDeviceDriverConvType AtmDeviceDriver::PacketClassifier(IpPacket *ipPacket)
{
    AtmDeviceDriverConvType convType = AtmDeviceDriverConvTypeDefault;

    if (ipPacket->ip_p == IPPROTO_TCP) {
	TcpSegment *tcpHeader = (TcpSegment *) ipPacket->payload;

	// Try to use TCP header fields to determine the conversation
	// type.  Note there are a bunch of hardwired voodoo constants
	// in here...this is kind of bad but there isn't any really 
	// easier way to deal with this.
	if (tcpHeader->th_sport == 23) {
	    convType = AtmDeviceDriverConvTypeTelnetDown;
	}
	else if (tcpHeader->th_dport == 23) {
	    convType = AtmDeviceDriverConvTypeTelnetUp;
	}
	else if (tcpHeader->th_sport == 21) {
	    convType = AtmDeviceDriverConvTypeFtpControlDown;
	}
	else if (tcpHeader->th_dport == 21) {
	    convType = AtmDeviceDriverConvTypeFtpControlUp;
	}
	else if (tcpHeader->th_sport == 20) {
	    convType = AtmDeviceDriverConvTypeFtpDataDown;
	}
	else if (tcpHeader->th_dport == 20) {
	    convType = AtmDeviceDriverConvTypeFtpDataUp;
	}
	else if (tcpHeader->th_sport == 80) {
	    convType = AtmDeviceDriverConvTypeHttpDown;
	}
	else if (tcpHeader->th_dport == 80) {
	    convType = AtmDeviceDriverConvTypeHttpUp;
	}
	else if (tcpHeader->th_dport == 9) {
	    convType = AtmDeviceDriverConvTypeDiscard;
	}

    }
    else if (ipPacket->ip_p == IPPROTO_UDP) {
	UdpPacket *udpHeader = (UdpPacket *) ipPacket->payload;

	// The UDP applications we're simulating don't have any
	// well-known
	// port numbers.  A "real" device driver here would
	// probably look
	// in the RTP headers to find the type of the data
	// being carried
	// in the packet.  We simulate that using a fake type
	// field in the UdpMessage structure.
	if (udpHeader->type == UdpPayloadTypeAudio) {
	    convType = AtmDeviceDriverConvTypeAudio;
	}
	else if (udpHeader->type == UdpPayloadTypeVideo) {
	    convType = AtmDeviceDriverConvTypeVideo;
	}
    }

    return convType;
}

//
// AtmDeviceDriver::BindRoute
//
// Input:  route key (route), IP packet (ipPacket)
//
// Output:  None
//
// Bind a route to an IP packet.  Basically this consists of filling
// in fields in the route key.
//
void AtmDeviceDriver::BindRoute(AtmRoute *route, IpPacket *ipPacket)
{

    //
    // Based on multiplexing type, figure out what packet header fields
    // to use in the route key.
    //
    switch (muxType) {

	// Multiplexing per-router
	case (AtmDeviceDriverMuxTypeRouter):
	{
	    // Fill in the RouteKey structure, for this policy only using
	    // the ip_next field.
	    route->rk.ip_next.s_addr = ipPacket->ip_next.s_addr;
	    route->rk.ip_src.s_addr = 0;
	    route->rk.ip_dst.s_addr = 0;
	    route->rk.ip_p = IPPROTO_IP;
	    
	}
	break;
	
	// Multiplexing per-conversation
	case (AtmDeviceDriverMuxTypeConv):
	{
	    // Fill in the RouteKey structure with the source and destination
	    // IP addresses and protocol number.  If it's TCP or UDP, then
	    // fill in the port numbers.
	    route->rk.ip_src.s_addr = ipPacket->ip_src.s_addr;
	    route->rk.ip_dst.s_addr = ipPacket->ip_dst.s_addr;
	    route->rk.ip_next.s_addr = ipPacket->ip_next.s_addr;
	    route->rk.ip_p = ipPacket->ip_p;

	    if (ipPacket->ip_p == IPPROTO_TCP) {
		TcpSegment *tcpHeader = (TcpSegment *) ipPacket->payload;
		route->rk.tp.tcp.th_sport = tcpHeader->th_sport;
		route->rk.tp.tcp.th_dport = tcpHeader->th_dport;
	    }
	    else if (ipPacket->ip_p == IPPROTO_UDP) {
		UdpPacket *udpHeader = (UdpPacket *) ipPacket->payload;
		route->rk.tp.udp.uh_sport = udpHeader->uh_sport;
		route->rk.tp.udp.uh_dport = udpHeader->uh_dport;
	    }
	}
	break;

	// Multiplexing per application
	case (AtmDeviceDriverMuxTypeApp):
	{
	    // Fill in the RouteKey structure with the source and destination
	    // IP addresses and protocol number.  If it's TCP or UDP, then
	    // fill in the port numbers.
	    route->rk.ip_src.s_addr = ipPacket->ip_src.s_addr;
	    route->rk.ip_dst.s_addr = ipPacket->ip_dst.s_addr;
	    route->rk.ip_next.s_addr = ipPacket->ip_next.s_addr;
	    route->rk.ip_p = ipPacket->ip_p;

	    if (ipPacket->ip_p == IPPROTO_TCP) {
		TcpSegment *tcpHeader = (TcpSegment *) ipPacket->payload;
		if (tcpHeader->th_sport >= TcpMinEphemeralPort) {
		    route->rk.tp.tcp.th_sport = 0;
		}
		else {
		    route->rk.tp.tcp.th_sport = tcpHeader->th_sport;
		}
		if (tcpHeader->th_dport >= TcpMinEphemeralPort) {
		    route->rk.tp.tcp.th_dport = 0;
		}
		else {
		    route->rk.tp.tcp.th_dport = tcpHeader->th_dport;
		}
	    }
	    else if (ipPacket->ip_p == IPPROTO_UDP) {
		UdpPacket *udpHeader = (UdpPacket *) ipPacket->payload;
		route->rk.tp.udp.uh_sport = udpHeader->uh_sport;
		route->rk.tp.udp.uh_dport = udpHeader->uh_dport;
	    }

	}
	break;

	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("AtmDeviceDriver");
	    *(sim->debugStream) << "bad muxType\n";
	});
	break;
    }
}

//
// AtmDeviceDriver::MakeAtmRoute
//
// Input:  IP packet
//
// Output:  A new ATM route structure, appropriate to the packet
//          (if a new route could not be created, NULL is returned).
//
// Generate a new ATM route structure.  The RouteKey field is
// correctly filled in with the criteria for the packet and the
// ATM layer parameters (including the endpoint and
// traffic fields) are also computed.  The new AtmRoute structure
// is *not* inserted into the hash table; it is the caller's
// responsibility to do this.
//
// This function effectively controls the multiplexing and QOS
// policies of the IP-over-ATM service.
//
AtmRoute *AtmDeviceDriver::MakeAtmRoute(IpPacket *ipPacket)
{

    SigAddress atmDest = 0;	// ATM destination address

    // Get ATM address for IP gateway.
    for (int i = 0; i < maxAddresses; i++) {
	if (atmArpMap[i].ip.s_addr == ipPacket->ip_next.s_addr) {
	    atmDest = atmArpMap[i].atm;
	    break;
	}
    }
    if (atmDest == 0) {
	return(NULL);
    }

    // Make the AtmRoute structure and bind it to the packet.
    AtmRoute *route = new AtmRoute;

    route->state = AtmRouteStateUnused;
    route->fallback = false;
    route->endpoint = atmDest;
    route->refCount = 0;
    route->queueHead = NULL;
    route->queueTail = NULL;
    route->type = AtmDeviceDriverConvTypeDefault;

    BindRoute(route, ipPacket);

    //
    // Determine conversation type for QOS policies that need it.
    //
    switch (qosType) {
	case (AtmDeviceDriverQosTypeNoqos):
	{
	    route->type = AtmDeviceDriverConvTypeDefault;
	}
	break;

	case (AtmDeviceDriverQosTypeQos1):
	case (AtmDeviceDriverQosTypeFallback):
	case (AtmDeviceDriverQosTypeTelnet):
	case (AtmDeviceDriverQosTypeFtp):
	case (AtmDeviceDriverQosTypeHttp):
	case (AtmDeviceDriverQosTypeAudio):
	case (AtmDeviceDriverQosTypeVideo):
	case (AtmDeviceDriverQosTypeAv):
	case (AtmDeviceDriverQosTypeIsp):
	{
	    route->type = PacketClassifier(ipPacket);
	}
	break;

	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("AtmDeviceDriver");
	    *(sim->debugStream) << "bad qosType\n";
	    abort();
	});
	break;

    }
    IF_DEBUG(3, {
	DEBUG_PREFIX2("AtmDeviceDriver");
	*(sim->debugStream) << "classified conversation as " << route->type << "\n";
    });
    
    //
    // Create and fill in the traffic spec
    //
    QosRcsp *traffic = new QosRcsp;
    traffic->type = QosTypeRcsp;

    //
    // Set QOS parameters for virtual circuit setup
    //
    switch (qosType) {

	// No QOS policy (best effort)
	case (AtmDeviceDriverQosTypeNoqos):
	{
	    traffic->d = -1.0;		// d < 0 => best effort
	    traffic->Xmin = 0.0;	// these parameters don't mean anything
	    traffic->Xave = 0.0;	// for best-effort
	    traffic->I = 1.0;
	}
	break;

	// QOS policy 1
	case (AtmDeviceDriverQosTypeQos1):
	{
	    // Application-dependent processing
	    switch (route->type) {

		case (AtmDeviceDriverConvTypeTelnetUp):
		traffic->d = 0.01;	// 10 ms bound per switch
		traffic->Xmin = 0.02;	// peak 50 packets per second
		traffic->Xave = 0.05;	// average 20 packets per second
		traffic->I = 0.5;	// averaged over 0.5 seconds
		break;

		case (AtmDeviceDriverConvTypeTelnetDown):
		traffic->d = 0.01;	// 10 ms bound per switch
		traffic->Xmin = 0.01;	// peak 100 packets per second
		traffic->Xave = 0.02;	// average 50 packets per second
		traffic->I = 1.0;	// averaged over 1.0 seconds
		break;

		case (AtmDeviceDriverConvTypeFtpControlUp):
		case (AtmDeviceDriverConvTypeFtpControlDown):
		traffic->d = 0.02;	// 20 ms bound per switch
		traffic->Xmin = 0.1;	// peak 10 packets per second
		traffic->Xave = 0.5;	// average 2 packets per second
		traffic->I = 5.0;	// averaged over 5.0 seconds
		break;

		case (AtmDeviceDriverConvTypeFtpDataUp):
		traffic->d = 0.08;	// 80 ms bound per switch
		traffic->Xmin = 0.1;	// peak 10 packets per second
		traffic->Xave = 0.1;	// average 10 packets per second
		traffic->I = 2.0;	// averaged over 2.0 seconds
		break;
		  
		case (AtmDeviceDriverConvTypeFtpDataDown):
		traffic->d = 0.08;	// 80 ms bound per switch
		traffic->Xmin = 0.02;	// peak 50 packets per second
		traffic->Xave = 0.02;	// average 50 packets per second
		traffic->I = 2.0;	// averaged over 2.0 seconds
		break;
		  
		case (AtmDeviceDriverConvTypeHttpUp):
		traffic->d = 0.04;	// 40 ms bound per switch
		traffic->Xmin = 0.05;	// peak 20 packets per second
		traffic->Xave = 0.1;	// average 10 packets per second
		traffic->I = 10.0;	// averaged over 10.0 seconds
		break;
		  
		case (AtmDeviceDriverConvTypeHttpDown):
		traffic->d = 0.04;	// 40 ms bound per switch
		traffic->Xmin = 0.004;	// peak 250 packets per second
		traffic->Xave = 0.008;	// average 125 packets per second
		traffic->I = 1.0;	// averaged over 1.0 seconds
		break;
		  
		case (AtmDeviceDriverConvTypeAudio):
		traffic->d = 0.02;	// 20 ms bound per switch
		traffic->Xmin = 0.004;	
		traffic->Xave = 0.005;	// 328-byte packets + 8-byte UDP +
					// 20-byte IP header +
					// 8-byte AAL header = 364 bytes
					// == 7.58 -> 8 cells
					// 21 times per second = 168 cells per
					// second = .005 seconds between cells
		traffic->I = 0.1;	// averaged over 0.1 seconds
		break;

		case (AtmDeviceDriverConvTypeVideo):
		traffic->d = 0.05;	// 50 ms bound per switch
		traffic->Xmin = 0.0026;	// 512-byte packets +  8-byte UDP +
					// 20-byte IP header +
					// 8-byte AAL header = 548 bytes
					// = 11.4 -> 12 cells
					// 32 times per second = 384 cells
					// per second = 0.0026 seconds between
					// cells
					// (put another way, 147692 bps)
		traffic->Xave = 0.0035;	// 286 cells per second (109714 bps)
		traffic->I = 2.0;	// not clear if this is the right
					// averaging interval
		break;
		
		default:
		traffic->d = -1.0;	// d < 0 => best effort
		traffic->Xmin = 0.0;	// these parameters don't mean anything
		traffic->Xave = 0.0;	// for best-effort
		traffic->I = 1.0;
		break;
	    }
	}
	break;

	// Fallback mode...give the signalling system an impossible
	// set of parameters so it will be forced to try fallback
	// connections.
	case (AtmDeviceDriverQosTypeFallback):
	{
	    traffic->d = 0.0;	// no one could possibly meet this!
	}
	break;

	// telnet policy:  send all telnet with guaranteed performance
	case (AtmDeviceDriverQosTypeTelnet):
	{
	    // Application-dependent processing
	    switch (route->type) {

		case (AtmDeviceDriverConvTypeTelnetUp):
		traffic->d = 0.01;
		traffic->Xmin = 0.02;
		traffic->Xave = 0.05;
		traffic->I = 0.5;
		break;

		case (AtmDeviceDriverConvTypeTelnetDown):
		traffic->d = 0.01;
		traffic->Xmin = 0.01;
		traffic->Xave = 0.02;
		traffic->I = 1.0;
		break;

		default:
		traffic->d = -1.0;
		traffic->Xmin = 0.0;
		traffic->Xave = 0.0;
		traffic->I = 1.0;
		break;
	    }
	}
	break;

	// ftp policy:  send all ftp traffic at priority/guaranteed
	case (AtmDeviceDriverQosTypeFtp):
	{
	    switch (route->type) {

		case (AtmDeviceDriverConvTypeFtpControlUp):
		case (AtmDeviceDriverConvTypeFtpControlDown):
		traffic->d = 0.02;
		traffic->Xmin = 0.1;
		traffic->Xave = 0.5;
		traffic->I = 5.0;
		break;

		case (AtmDeviceDriverConvTypeFtpDataUp):
		traffic->d = 0.08;
		traffic->Xmin = 0.1;
		traffic->Xave = 0.1;
		traffic->I = 2.0;
		break;
		  
		case (AtmDeviceDriverConvTypeFtpDataDown):
		traffic->d = 0.08;
		traffic->Xmin = 0.02;
		traffic->Xave = 0.02;
		traffic->I = 2.0;
		break;
		  
		default:
		traffic->d = -1.0;
		traffic->Xmin = 0.0;
		traffic->Xave = 0.0;
		traffic->I = 1.0;
		break;
	    }

	}
	break;

	// HTTP:  Send all HTTP traffic with guaranteed performance.
	case (AtmDeviceDriverQosTypeHttp):
	{
	    switch (route->type) {
		case (AtmDeviceDriverConvTypeHttpUp):
		traffic->d = 0.04;
		traffic->Xmin = 0.05;
		traffic->Xave = 0.1;
		traffic->I = 10.0;
		break;
		  
		case (AtmDeviceDriverConvTypeHttpDown):
		traffic->d = 0.04;
		traffic->Xmin = 0.004;
		traffic->Xave = 0.008;
		traffic->I = 1.0;
		break;
		  
		default:
		traffic->d = -1.0;
		traffic->Xmin = 0.0;
		traffic->Xave = 0.0;
		traffic->I = 1.0;
		break;
	    }
	}
	break;

	// Audio QOS policy:  All audio traffic sent guaranteed
	case (AtmDeviceDriverQosTypeAudio):
	{
	    switch (route->type) {
		case (AtmDeviceDriverConvTypeAudio):
		traffic->d = 0.02;
		traffic->Xmin = 0.004;
		traffic->Xave = 0.005;
		traffic->I = 0.1;
		break;

		default:
		traffic->d = -1.0;
		traffic->Xmin = 0.0;
		traffic->Xave = 0.0;
		traffic->I = 1.0;
		break;
	    }
	}
	break;

	// Video QOS policy:  All video traffic sent guaranteed
	case (AtmDeviceDriverQosTypeVideo):
	{
	    switch (route->type) {
		case (AtmDeviceDriverConvTypeVideo):
		traffic->d = 0.05;
		traffic->Xmin = 0.0026;
		traffic->Xave = 0.0035;
		traffic->I = 2.0;
		break;

		default:
		traffic->d = -1.0;
		traffic->Xmin = 0.0;
		traffic->Xave = 0.0;
		traffic->I = 1.0;
		break;
	    }
	}
	break;

	// AV QOS policy:  All audio and video traffic sent guaranteed
	case (AtmDeviceDriverQosTypeAv):
	{
	    switch (route->type) {
		case (AtmDeviceDriverConvTypeAudio):
		traffic->d = 0.02;
		traffic->Xmin = 0.004;	
		traffic->Xave = 0.005;
		traffic->I = 0.1;
		break;

		case (AtmDeviceDriverConvTypeVideo):
		traffic->d = 0.05;
		traffic->Xmin = 0.0026;
		traffic->Xave = 0.0035;
		traffic->I = 2.0;
		break;
		
		default:
		traffic->d = -1.0;
		traffic->Xmin = 0.0;
		traffic->Xave = 0.0;
		traffic->I = 1.0;
		break;
	    }
	}
	break;

	// ISP policy:  Telnet and HTTP are sent guaranteed.
	case (AtmDeviceDriverQosTypeIsp):
	{
	    switch (route->type) {

		case (AtmDeviceDriverConvTypeTelnetUp):
		traffic->d = 0.01;
		traffic->Xmin = 0.02;
		traffic->Xave = 0.05;
		traffic->I = 0.5;
		break;

		case (AtmDeviceDriverConvTypeTelnetDown):
		traffic->d = 0.01;
		traffic->Xmin = 0.01;
		traffic->Xave = 0.02;
		traffic->I = 1.0;
		break;

		case (AtmDeviceDriverConvTypeHttpUp):
		traffic->d = 0.04;
		traffic->Xmin = 0.05;
		traffic->Xave = 0.1;
		traffic->I = 10.0;
		break;
		  
		case (AtmDeviceDriverConvTypeHttpDown):
		traffic->d = 0.04;
		traffic->Xmin = 0.004;
		traffic->Xave = 0.008;
		traffic->I = 1.0;
		break;
		  
		default:
		traffic->d = -1.0;
		traffic->Xmin = 0.0;
		traffic->Xave = 0.0;
		traffic->I = 1.0;
		break;
	    }
	}
	break;

	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("AtmDeviceDriver");
	    *(sim->debugStream) << "bad qosType\n";
	    abort();
	});
	break;
	
    }
    route->traffic = traffic;
    route->trafficLength = sizeof(QosRcsp);
    
    // Set VC timeout value (if any)
    switch (vcType) {
	
	case (AtmDeviceDriverVcTypePvc):
	{
	}
	break;

	case (AtmDeviceDriverVcTypeSvc):
	{
	    // Set a 60 second fixed timeout.
	    route->maxIdle = maxIdle;
	}
	break;

	case (AtmDeviceDriverVcTypeSvccache):
	{
	    // Set a 60 second fixed timeout, 300 second cache.
	    route->maxIdle = maxIdle;
	    route->maxCache = maxCache;
	}
	break;

	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("AtmDeviceDriver");
	    *(sim->debugStream) << "bad vcType\n";
	});
	break;
    }
    return (route);
    
}

//
// AtmDeviceDriver::FallbackRoute
//
// Input:  AtmRoute entry (atmRoute)
//
// Output:  Boolean (true if an alternate fallback route exists, false
//          otherwise)
//
// This method gets called whenever a call establishment fails.
// The intended use is to provide an alternate set of virtual circuit
// parameters in case a call with the original parameters fails
// (for example, a virtual circuit policy could use this facility to
// set up a best-effort virtual circuit if a real-time virtual circuit
// cannot be established).
//
// If a policy uses fallbacks, this method must modify the
// atmRoute parameters accordingly and return true.  Otherwise, it must
// return false.
//
bool AtmDeviceDriver::FallbackRoute(AtmRoute *atmRoute)
{
    switch (qosType) {

	case (AtmDeviceDriverQosTypeNoqos):
	{
	    return false;
	}
	break;

	case (AtmDeviceDriverQosTypeQos1):
	case (AtmDeviceDriverQosTypeFallback):
	case (AtmDeviceDriverQosTypeTelnet):
	case (AtmDeviceDriverQosTypeFtp):
	case (AtmDeviceDriverQosTypeHttp):
	case (AtmDeviceDriverQosTypeAudio):
	case (AtmDeviceDriverQosTypeVideo):
	case (AtmDeviceDriverQosTypeAv):
	case (AtmDeviceDriverQosTypeIsp):
	{
	    if (atmRoute->traffic) {

		if (((QosRcsp *) atmRoute->traffic)->d >= 0) {

		    // real-time vc failed, switch to best-effort
		    // XXX assumption here is that traffic descriptor is for
		    // RCSP
		    ((QosRcsp *) atmRoute->traffic)->d = -1.0;
		    atmRoute->type = AtmDeviceDriverConvTypeDefault;
		    atmRoute->fallback = true;
		    return true;

		}
		else {

		    IF_DEBUG(0, {
			DEBUG_PREFIX2("AtmDeviceDriver");
			*(sim->debugStream) << "fallback connection failed!\n";
		    });
		    return false;

		}
	    }
	    else {
		return false;
	    }
	}
	break;

	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("AtmDeviceDriver");
	    *(sim->debugStream) << "bad qosType\n";
	    abort();
	});
	break;

    }
}

//
// AtmDeviceDriver::TimeoutAtmRoute
//
// Input:  AtmRoute entry under consideration (atmRoute),
//         current simulation time (time)
//
// Output:  Boolean return value (true == route should be deleted and VC
//          closed)
//
// Determine whether or not to timeout a given AtmRoute entry.  We note
// that this routine may include (in the future?) some adaptive algorithms
// that take into account more than just these two fields in the atmRoute
// structure.
//
bool AtmDeviceDriver::TimeoutAtmRoute(AtmRoute *atmRoute, double time) {
    
    IF_DEBUG(4, {
	DEBUG_PREFIX("AtmDeviceDriver");
	*(sim->debugStream) << "TimeoutAtmRoute entry\n";
    });

    switch (vcType) {
	case (AtmDeviceDriverVcTypePvc):
	{
	    return false;
	}
	break;

	case (AtmDeviceDriverVcTypeSvc):
	{
	    // Delete only idle or established VCs (i.e. not those
	    // being established).
	    if ((atmRoute->state == AtmRouteStateEstablished) ||
		(atmRoute->state == AtmRouteStateIdle)) {
		return (atmRoute->lastUse + atmRoute->maxIdle < time);
	    }
	    else {
		return false;
	    }
	}
	break;

	case (AtmDeviceDriverVcTypeSvccache):
	{
	    // If VC is idle longer than cache timeout, then tear it down.
	    // Note:  Don't try to tear down channels pending establishment
	    // because we don't have a VCID for them.
	    if (atmRoute->lastUse + atmRoute->maxCache < time) {

		if ((atmRoute->state == AtmRouteStateEstablished) ||
		    (atmRoute->state == AtmRouteStateIdle)) {

		    IF_DEBUG(1, {
			DEBUG_PREFIX("AtmDeviceDriver");
			*(sim->debugStream) << "delete route " <<
			  atmRoute << "\n";
		    });

		    return true;
		}
		else {
		    return false;
		}
	    }

	    // Otherwise, if VC is idle too long (but still younger than
	    // the cachable threshhold), mark it as cacheable and keep
	    // it around.
	    else if (atmRoute->lastUse + atmRoute->maxIdle < time) {
		if (atmRoute->state == AtmRouteStateEstablished) {

		    atmRoute->state = AtmRouteStateIdle;

		    IF_DEBUG(1, {
			DEBUG_PREFIX("AtmDeviceDriver");
			*(sim->debugStream) << "idle route " <<
			  atmRoute->vcid << "\n";
		    });

		}
		return false;
	    }
	}
	break;
	
	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("AtmDeviceDriver");
	    *(sim->debugStream) << "bad vcType\n";
	});
	break;
    }

}

//
// AtmDeviceDriver::DoCallForRoute
//
// Input:  Current time (time), AtmRoute (atmRoute)
//
// Output:  None
//
// Start a call for the virtual circuit necessary for an AtmRoute.
// If there are any immediate problems, the route is deleted.
// In the normal case, the signalling system handles the call
// and this method returns (while the call is still in progress).
//
void AtmDeviceDriver::DoCallForRoute(double time, AtmRoute *atmRoute)
{

    // Start VC call...first get a request ID.
    int call = GetRequestId(atmRoute);
    if (call == -1) {

	// Couldn't get a request ID!  This is a pretty big problem...
	// throw away the route and all its queued up packets.
	// Want to send an ICMP message?  XXX
	IF_DEBUG(0, {
	    DEBUG_PREFIX2("AtmDeviceDriver");
	    *(sim->debugStream) << "couldn't get request ID\n";
	});
	
	if_oerrors++;
	DeleteRoute(atmRoute);
    }
    else {

	// Fill in a downcall message to the signalling
	// entity and send it.
	ApiMessage *apiMessage = new ApiMessage;
	apiMessage->type = ApiTypeEstablish;
	apiMessage->client = this;
	apiMessage->vcid = 0;
	apiMessage->requestId = call;
	apiMessage->code = SigOK;
	apiMessage->sourceAddress = 0; // SigHost will fill this in
	apiMessage->sourceService = SigServiceIP;
	apiMessage->destinationAddress = atmRoute->endpoint;
	apiMessage->destinationService = SigServiceIP;
	apiMessage->traffic = atmRoute->traffic;
	apiMessage->trafficLength = atmRoute->trafficLength;
	
	Lookup(sigHost, sigHostName);
	Post(time + 0, sigHost, EventCall, (void *) apiMessage);

	numConnRequests++;

	IF_DEBUG(1, {
	    DEBUG_PREFIX("AtmDeviceDriver");
	    *(sim->debugStream) << "sent call requestId " <<
	      apiMessage->requestId << "\n";
	});

    }
}

//
// AtmDeviceDriver::TransmitRoute
//
// Input: AtmRoute (atmRoute)
//
// Output:  None
//
// Transmit the packets queued up for an AtmRoute.
//
void AtmDeviceDriver::TransmitRoute(double time, AtmRoute *atmRoute)
{
    double delta = AtmDeviceDriverSendGap;

    while (atmRoute->queueHead) {

	// Find the packet
	QueueElem *sendPacket = atmRoute->queueHead;
	IpPacket *packet = sendPacket->payload;

	// Construct the AAL frame
	AalFrame *aalFrame = new AalFrame (packet);
	aalFrame->vcid = atmRoute->vcid;
	aalFrame->length = packet->ip_len + AalHeaderSize;

	// Send the frame
	Lookup(down, downName);
	Post(time + delta, down, EventFrameSend, (void *) aalFrame);

	// Update statistics
        atmRoute->lastUse = time;
	if_opackets++;
	if_obytes += packet->ip_len;
	packetCount[atmRoute->type]++;
	byteCount[atmRoute->type] += packet->ip_len;

	// Go to the next queue element
	atmRoute->queueHead = atmRoute->queueHead->next;
	delta += AtmDeviceDriverSendGap;
	delete sendPacket;

    }

    // Nothing left in the queue now!
    atmRoute->queueHead = NULL;
    atmRoute->queueTail = NULL;
    atmRoute->queueCount = 0;

}


//
// AtmDeviceDriver::DeleteRoute
//
// Input:  AtmRoute (atmRoute)
//
// Output:  None
//
// Delete an AtmRoute and make sure all data structures remain
// consistent.  This includes flushing packets queued up waiting
// for a virtual circuit establishment.
//
void AtmDeviceDriver::DeleteRoute(AtmRoute *atmRoute)
{

    // Rebuild the DLL around the doomed route
    if (atmRoute->next) {
	atmRoute->next->prev = atmRoute->prev;
    }
    if (atmRoute->prev) {
	atmRoute->prev->next = atmRoute->next;
    }
    else {
	atmVcidMap[atmRoute->hash] =
	  atmRoute->next;
    }

    // Free auxiliary data structures.
    if (atmRoute->traffic) {
	delete atmRoute->traffic;
    }
    
    // Flush queued packets
    while (atmRoute->queueHead) {
	
	QueueElem *deadPacket
	  = atmRoute->queueHead;
	delete deadPacket->payload;
	
	atmRoute->queueHead =
	  atmRoute->queueHead->next;
	
	if_oerrors++;
	
	delete deadPacket;
	
    }
    delete atmRoute;
}


//
// AtmDeviceDriver::GetRequestId
//
// Input:  AtmRoute (atmRoute)
//
// Output:  RequestId to use for signalling request (in return value)
//
// Get a new request ID to use for a signalling message.  If no request
// IDs are available, returns -1.
//
int AtmDeviceDriver::GetRequestId(AtmRoute *atmRoute)
{
    int call = -1;
    for (int i = 0; i < maxRequests; i++) {
	if (pendingRoute[i] == NULL) {
	    call = i;
	    pendingRoute[i] = atmRoute;
	    break;
	}
    }
    return call;
}

//
// AtmDeviceDriver::ReleaseRequestId
//
// Input:  Request ID to free (r)
//
// Output:  None
//
// Release a given Request ID.  To be used when a signalling call
// returns.
//
void AtmDeviceDriver::ReleaseRequestId(int r)
{
    pendingRoute[r] = NULL;
}

//
// AtmDeviceDriver::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String AtmDeviceDriver::Output()
{
    String out;
    char buf[128];

    out = name +
      " [AtmDeviceDriver] qostype " + AtmDeviceDriverQosTypeName[qosType] +
      " muxtype " + AtmDeviceDriverMuxTypeName[muxType] +
      " vctype " + AtmDeviceDriverVcTypeName[vcType];

    sprintf(buf,
	    " maxIdle %f maxCache %f if_ipackets %d if_ierrors %d if_opackets %d if_oerrors %d",
	    maxIdle, maxCache, if_ipackets, if_ierrors, if_opackets,
	    if_oerrors);

    out += (String) buf;

    return out;
    
}

//
// AtmDeviceDriverCmd
//
// (Tcl command procedure)
//
int
AtmDeviceDriverCmd(ClientData clientData, Tcl_Interp *interp,
		   int argc, char *argv[])
{
    if (argc != 5) {
	Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of new AtmDeviceDriver object
	int ma;			// maximum address
	int mh;			// maximum hash table size
	int mr;			// maximum number of pending requests
	AtmDeviceDriver *add;	// new AtmDeviceDriverSvcAppNoqos object

	int rc;			// return code from Tcl library
	extern Simulation *sim;	// global simulation object

	// get name
	nm = argv[1];

	// get numeric argument into ma
	rc = Tcl_GetInt(interp, argv[2], &ma);
	if (rc == TCL_OK) {

	    rc = Tcl_GetInt(interp, argv[3], &mh);
	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[4], &mr);
		if (rc == TCL_OK) {

		    add = new AtmDeviceDriver(nm, sim, ma, mh, mr);
		    Tcl_CreateCommand(interp, nm,
				      AtmDeviceDriverObjectCmd,
				      (ClientData) add,
				      AtmDeviceDriverDeleteProc);
		}
	    }
	}
	
	return rc;

    }
}

//
// AtmDeviceDriverObjectCmd
//
// (Tcl object command procedure)
//
int
AtmDeviceDriverObjectCmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{

    AtmDeviceDriver *add = (AtmDeviceDriver *) clientData;

    if (argc == 1) {
	return(DeviceDriverObjectCmd(clientData, interp, argc, argv));
    }
    else {

	//
	// add maxidle idle
	//
	if (!strcmp(argv[1], "maxidle")) {
	    int rc = TCL_OK;
	    double mi;

	    if (argc == 3) {
		rc = Tcl_GetDouble(interp, argv[2], &mi);
		if (rc == TCL_OK) {
		    add->maxIdle = mi;
		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// add maxcache cache
	//
	else if (!strcmp(argv[1], "maxcache")) {
	    int rc = TCL_OK;
	    double mc;

	    if (argc == 3) {
		rc = Tcl_GetDouble(interp, argv[2], &mc);
		if (rc == TCL_OK) {
		    add->maxCache = mc;
		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// add vctype vc-type
	//
	else if (!strcmp(argv[1], "vctype")) {

	    int rc = TCL_OK;
	    
	    if (argc == 3) {
		if (!strcmp(argv[2], "pvc")) {
		    add->vcType = AtmDeviceDriverVcTypePvc;
		}
		else if (!strcmp(argv[2], "svc")) {
		    add->vcType = AtmDeviceDriverVcTypeSvc;
		}
		else if (!strcmp(argv[2], "svccache")) {
		    add->vcType = AtmDeviceDriverVcTypeSvccache;
		}
		else {
		    Tcl_SetResult(interp, "add vctype pvc/svc/svccache",
				  TCL_STATIC);
		    rc = TCL_ERROR;
		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// add muxtype mux-type
	//
	else if (!strcmp(argv[1], "muxtype")) {

	    int rc = TCL_OK;
	    
	    if (argc == 3) {
		if (!strcmp(argv[2], "router")) {
		    add->muxType = AtmDeviceDriverMuxTypeRouter;
		}
		else if (!strcmp(argv[2], "conv")) {
		    add->muxType = AtmDeviceDriverMuxTypeConv;
		}
		else if (!strcmp(argv[2], "app")) {
		    add->muxType = AtmDeviceDriverMuxTypeApp;
		}
		else {
		    Tcl_SetResult(interp, "add muxtype router/conv/app",
				  TCL_STATIC);
		    rc = TCL_ERROR;
		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// add qostype qos-type
	//
	else if (!strcmp(argv[1], "qostype")) {

	    int rc = TCL_OK;
	    
	    if (argc == 3) {
		if (!strcmp(argv[2], "noqos")) {
		    add->qosType = AtmDeviceDriverQosTypeNoqos;
		}
		else if (!strcmp(argv[2], "qos1")) {
		    add->qosType = AtmDeviceDriverQosTypeQos1;
		}
		else if (!strcmp(argv[2], "fallback")) {
		    add->qosType = AtmDeviceDriverQosTypeFallback;
		}
		else if (!strcmp(argv[2], "telnet")) {
		    add->qosType = AtmDeviceDriverQosTypeTelnet;
		}
		else if (!strcmp(argv[2], "ftp")) {
		    add->qosType = AtmDeviceDriverQosTypeFtp;
		}
		else if (!strcmp(argv[2], "http")) {
		    add->qosType = AtmDeviceDriverQosTypeHttp;
		}
		else if (!strcmp(argv[2], "audio")) {
		    add->qosType = AtmDeviceDriverQosTypeAudio;
		}
		else if (!strcmp(argv[2], "video")) {
		    add->qosType = AtmDeviceDriverQosTypeVideo;
		}
		else if (!strcmp(argv[2], "av")) {
		    add->qosType = AtmDeviceDriverQosTypeAv;
		}
		else if (!strcmp(argv[2], "isp")) {
		    add->qosType = AtmDeviceDriverQosTypeIsp;
		}
		else {
		    Tcl_SetResult(interp, "add qostype noqos/qos1/fallback/telnet/ftp/http/audio/video/av/isp",
				  TCL_STATIC);
		    rc = TCL_ERROR;
		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// add atmarp ip-addr atm-addr
	//
	else if (!strcmp(argv[1], "atmarp")) {

	    if (argc == 4) {

		InAddr ipAddr; // ip address
		SigAddress atmAddr;    // atm address

		int rc = TCL_OK;       // return code
		
		// Get IP address
		ipAddr.s_addr = InetAddr(argv[2]);
		
		// Get ATM address
		rc = Tcl_GetInt(interp, argv[3], &atmAddr);

		if (rc == TCL_OK) {

		    int unusedEntry = -1;
		    bool done = false;
		    
		    // Insert entry into address mapping table.
		    // If the ATM address is 0, we should delete
		    // the entry.  If it's >0, replace an existing
		    // entry and reset the IP address.  Keep track
		    // of an empty entry in case this is a new
		    // atmarp entry.
		    for (int i = 0; i < add->maxAddresses; i++) {
			if (add->atmArpMap[i].ip.s_addr == ipAddr.s_addr) {
			    if (atmAddr == 0) {
				add->atmArpMap[i].ip.s_addr = 0;
			    }
			    else {
				add->atmArpMap[i].atm = atmAddr;
			    }
			    done = true;
			    break;
			}
			else if ((add->atmArpMap[i].ip.s_addr == 0) &&
				 (unusedEntry < 0)) {
			    unusedEntry = i;
			}
		    }

		    // If we couldn't find a matching entry, but we did
		    // find an unused one, make up an entry.  Otherwise,
		    // return an error.
		    if (!done) {
			if (unusedEntry >= 0) {
			    add->atmArpMap[unusedEntry].ip.s_addr =
			      ipAddr.s_addr;
			    add->atmArpMap[unusedEntry].atm = atmAddr;
			}
			else {
			    Tcl_SetResult(interp,
					  "couldn't find atmarp entry",
					  TCL_STATIC);
			    rc = TCL_ERROR;
			}
		    }
		    
		}
		return(rc);
		
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	//
	// add sighost object
	//
	else if (!strcmp (argv[1], "sighost")) {
	    if (argc == 3) {
		String on;	// object name

		on = argv[2];
		add->sigHostName = on;
		add->sigHost = NULL;
		return TCL_OK;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// add netstat
	//
	else if (!strcmp(argv[1], "netstat")) {
	    int count = 0;

	    if (argc > 2) {

		if (!strcmp(argv[2], "-d")) {
	    
		    int i;
		    
		    cout << "Downcall state\n";
		    cout << "ID  Ds Pk Key\n";
		    
		    for (i = 0; i < add->maxRequests; i++) {
			
			if (add->pendingRoute[i]) {
			    
			    AtmRoute *r = add->pendingRoute[i];
			    
			    cout << setw(3) << i << " " <<
			      setw(2) << r->endpoint << " " <<
			      setw(2) << r->queueCount << " " <<
			      r->rk << "\n";
			    
			    count++;
			}
		    }
		    cout << "(Total " << count <<
		      " of " << add->maxRequests << " in use)\n";
		    return(TCL_OK);

		}

		else if (!strcmp(argv[2], "-a")) {

		    int i;
		    AtmRoute *r;
		    count = 0;
		    
		    cout << "Hsh S Ds Type     Pk Used      Key\n";

		    for (i = 0; i < add->hashSize; i++) {

			for (r = add->atmVcidMap[i]; r != NULL; r = r->next) {

			    cout << setw(3) << i << " " <<
			      setw(1) << r->state << " " <<
			      setw(2) << r->endpoint << " " <<
			      setw(8) << r->type << " " <<
			      setw(2) << r->queueCount << " " <<
			      setw(7) << r->lastUse << " " <<
			      r->rk << "\n";

			    count++;

			}
		    }

		    cout << "(Total " << count << " routes in use)\n";
		    return(TCL_OK);

		}

		else {
		    Tcl_SetResult(interp,
				  "bad netstat subcommand", TCL_STATIC);
		    return(TCL_ERROR);
		}
		
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// add statinterval i
	//
	if (!strcmp(argv[1], "statinterval")) {

	    if (argc == 3) {
		double inter;      // statistics printing interval
		int rc;		   // Tcl return code
		
		rc = Tcl_GetDouble(interp, argv[2], &inter);
		if (rc == TCL_OK) {
		    
		    add->interval = inter;
		    add->Post(0.0 + inter, add, EventTimer2, NULL);

		}
		return rc;
	    }
	}

	else {
	    return(DeviceDriverObjectCmd(clientData, interp, argc, argv));
	}

    }

}

//
// AtmDeviceDriverDeleteProc
//
// (Tcl delete procedure)
//
void
AtmDeviceDriverDeleteProc(ClientData clientData)
{
    delete (AtmDeviceDriver *) clientData;
}


//
// operator<<
//
// Input:  Output stream (s), enumerated value (varies)
//
// Output:  Output stream
//
// Print the name of the enumerated value on the output stream s.
//
ostream& operator<<(ostream &s, AtmDeviceDriverConvType ct)
{
    char *output;
    
    output = AtmDeviceDriverConvTypeName[ct];
    return s << output;
}

ostream& operator<<(ostream &s, AtmDeviceDriverQosType qt)
{
    char *output;
    
    output = AtmDeviceDriverQosTypeName[qt];
    return s << output;
}

ostream& operator<<(ostream &s, AtmDeviceDriverMuxType mt)
{
    char *output;
    
    output = AtmDeviceDriverMuxTypeName[mt];
    return s << output;
}

ostream& operator<<(ostream &s, AtmDeviceDriverVcType vt)
{
    char *output;
    
    output = AtmDeviceDriverVcTypeName[vt];
    return s << output;
}

ostream& operator<<(ostream &s, AtmRouteState st)
{
    char *output;

    output = AtmRouteStateName[st];
    return s << output;
}


ostream& operator<<(ostream &s, RouteKey &rk)
{

    s << "(" << rk.ip_src << "->" <<
      rk.ip_dst << " via " <<
      rk.ip_next;

    switch (rk.ip_p) {

	case (IPPROTO_UDP):
	{
	    s << " udp " << rk.tp.udp.uh_sport << "->" << rk.tp.udp.uh_dport;
	}
	break;

	case (IPPROTO_TCP):
	{
	    s << " tcp " << rk.tp.tcp.th_sport << "->" << rk.tp.tcp.th_dport;
	}
	break;

	default:
	s << " proto" << (int) rk.ip_p;
	break;
    }

    s << ")";

    return s;
}
