// -*- c++ -*-
//
// $Id: AtmDeviceDriver.h,v 1.31 1996-07-31 21:36:10-07 bmah Exp $
//
// AtmDeviceDriver.h
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//

#ifndef ATMDEVICEDRIVER_H
#define ATMDEVICEDRIVER_H

#include <iostream.h>
#include <Uniform.h>

#include "ScheduledObject.h"
#include "Simulation.h"
#include "DeviceDriver.h"
#include "AtmCell.h"
#include "SigHost.h"
#include "Tcl.h"
#include "SigMessage.h"
#include "apiproto.h"
#include "Event.h"

// Configuration constants
const double AtmDeviceDriverPollInterval = 1.0; // idle timeout poll at 1 sec
const double AtmDeviceDriverSendGap = 0.000001;	// back to back sending gap
						// when clearing queue

// QOS policy types
enum AtmDeviceDriverQosType {
    AtmDeviceDriverQosTypeNull,
    AtmDeviceDriverQosTypeNoqos,
    AtmDeviceDriverQosTypeQos1,
    AtmDeviceDriverQosTypeFallback,
    AtmDeviceDriverQosTypeTelnet,
    AtmDeviceDriverQosTypeFtp,
    AtmDeviceDriverQosTypeHttp,
    AtmDeviceDriverQosTypeAudio,
    AtmDeviceDriverQosTypeVideo,
    AtmDeviceDriverQosTypeAv,
    AtmDeviceDriverQosTypeIsp,
};

// Multiplexing types
enum AtmDeviceDriverMuxType {
    AtmDeviceDriverMuxTypeNull,
    AtmDeviceDriverMuxTypeRouter,
    AtmDeviceDriverMuxTypeConv,
    AtmDeviceDriverMuxTypeApp
};

// Virtual circuit use types
enum AtmDeviceDriverVcType {
    AtmDeviceDriverVcTypeNull,
    AtmDeviceDriverVcTypePvc,
    AtmDeviceDriverVcTypeSvc,
    AtmDeviceDriverVcTypeSvccache
};

// Conversation types, necessary for setting ATM QOS parameters
enum AtmDeviceDriverConvType {
    AtmDeviceDriverConvTypeDefault,
    AtmDeviceDriverConvTypeTelnetUp,
    AtmDeviceDriverConvTypeTelnetDown,
    AtmDeviceDriverConvTypeFtpDataUp,
    AtmDeviceDriverConvTypeFtpDataDown,
    AtmDeviceDriverConvTypeFtpControlUp,
    AtmDeviceDriverConvTypeFtpControlDown,
    AtmDeviceDriverConvTypeHttpUp,
    AtmDeviceDriverConvTypeHttpDown,
    AtmDeviceDriverConvTypeAudio,
    AtmDeviceDriverConvTypeVideo,
    AtmDeviceDriverConvTypeDiscard,
    AtmDeviceDriverConvTypeMAX
};

// VCI mapping structure, performs the ARP-like function, but static,
// table-driven (for now)
typedef struct {
    InAddr ip;			// IP address
    SigAddress atm;		// ATM address
} AtmArpEntry;

// RouteKey structure captures the features of a class of IP packets that
// are necessary to properly route it onto an ATM virtual circuit.
typedef struct {

    // Internetwork-layer
    InAddr ip_src;		// source IP address
    InAddr ip_dst;		// destination IP address
    InAddr ip_next;		// next-hop gateway IP address
    u_char ip_tos;		// IP type of service

    // Transport-layer
    u_char ip_p;		// transport protocol
    union {

	// TCP
	struct {
	    u_short th_sport;	// source port
	    u_short th_dport;	// destination port
	} tcp;

	// UDP
	struct {
	    u_short uh_sport;	// source port
	    u_short uh_dport;	// destination port
	} udp;
	
    } tp;
} RouteKey;

// State of a given AtmRoute structure (see below).
//     AtmRouteStateUnused        Entry free (with current hash table
//                                this shouldn't show up)
//     AtmRouteStateEstablishing  VC being established for this route
//                                (packets should be queued)
//     AtmRouteStateEstablished   VC active (send packets)
//     AtmRouteStateIdle          VC still up but idle (can be reused
//                                by packets for another RouteKey)
//     AtmRouteStateZombie	  VC being deleted
enum AtmRouteState {
    AtmRouteStateUnused,
    AtmRouteStateEstablishing,
    AtmRouteStateEstablished,
    AtmRouteStateIdle,
    AtmRouteStateZombie
};

// QueueElems form a SLL of packets queued up while waiting for the
// VCID for a particular route to become active.  Yes, we really ought
// to use a template class for this, but we don't quite trust g++
// templates yet...
struct QueueElem {
    IpPacket *payload;
    struct QueueElem *next;
};

// ATM routing entry.  A hash table of DLLs of these entries performs the
// function of mapping individual IP packets to ATM virtual circuits.
struct AtmRoute {

    RouteKey rk;		// IP routing key
    AtmRouteState state;	// state of this routing entry
    bool fallback;		// set to true if a fallback connection

    int hash;			// position in hash table

    SigAddress endpoint;	// ATM endpoint
    int vcid;			// VCID to use for these packets
    void *traffic;		// traffic descriptor
    int trafficLength;		// length of traffic descriptor

    AtmDeviceDriverConvType type; // IP conversation type

    int refCount;		// How many times this route was referenced
    double lastUse;		// Last reference time
    double maxIdle;		// Maximum amount of idle time before entry
				// gets idled (AtmRouteStateIdle) or deleted
    double maxCache;		// Maximum amount of idle time to hold onto
				// a cached virtual circuit before it gets
				// deleted.

    QueueElem *queueHead,	// packet queue for pending connections
      *queueTail;
    int queueCount;		// count of queued packets

    AtmRoute *next, *prev; 	// Link pointers for DLL

};

class AtmDeviceDriver : public DeviceDriver {

    friend int
    AtmDeviceDriverCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]);
    friend int
    AtmDeviceDriverObjectCmd(ClientData clientData,
			     Tcl_Interp *interp,
			     int argc, char *argv[]);

    public:

    const int maxAddresses;	// maximum number of ATM addresses we
				// need to allocate map entries for

    const int hashSize;		// AtmRoute hash table size

    const int maxRequests;	// maximum number of pending signalling
				// requests

    AtmDeviceDriver(const String &name, Simulation *si, int ma,
		    int mh, int mr);
    virtual ~AtmDeviceDriver();

    virtual String Output(void);
    virtual void Handler (Event *e);

    protected:				  

    AtmArpEntry *atmArpMap;	// IP->ATM address mapping table
    AtmRoute **atmVcidMap;	// IP->VCID hash table
    AtmRoute **pendingRoute;	// table of pending route requests

    String sigHostName;		// name of signalling entity
    ScheduledObject *sigHost;	// signalling entity

    static Uniform *skewGenerator; // offset to skew timeouts
				   // so the network doesn't
				   // get hit with timeouts
				   // all at once.

    int numConnCached;		// number of connection requests satisfied
				// by cached VCs		
    int numConnRequests;	// number of connections requested
    int numConnFirstSucceeded;	// number of first-try connections succeeded
    int numConnFallbacks;	// number of fallback connections needed
    int numConnFallbackCached;	// number of times fallback connection cached
    int numConnFallbackRequests; // number of times fallback connection tried
    int numConnFallbackSucceeded; // number of times fallback succeeded
    int numConnSucceeded;	// total number of connections successful
    int numConnFailed;		// total number of failed connection attempts
    int cacheCount[AtmDeviceDriverConvTypeMAX];  // Connections cached...
    int connCount[AtmDeviceDriverConvTypeMAX];   // ...established...
    int failCount[AtmDeviceDriverConvTypeMAX];	 // ...and failed,
    int byteCount[AtmDeviceDriverConvTypeMAX];   // IP bytes sent and ...
    int packetCount[AtmDeviceDriverConvTypeMAX]; // ... output packets,
						 // indexed by VC traffic type

    int nowEstablished;		// number of connections currently in use

    AtmDeviceDriverQosType qosType; // QOS policy in use
    AtmDeviceDriverMuxType muxType; // multiplexing policy in use
    AtmDeviceDriverVcType vcType;   // virtual circuit use policy

    double maxIdle;		// default timeout
    double maxCache;		// cached VC timeout

    double interval;		// statistics-printing interval

    // Hash function for IP->VCID table.  There aren't a lot of choices
    // for this function, it it needs to be able to determine the hash
    // bucket for a packet regardless of which criteria might be used
    // to match against the RouteKeys in the bucket.  For example, with
    // certain multiplexing policies, hashing on the source or destination
    // IP addresses won't work.
    int AtmRouteHash (AtmRoute *route) {
	return (route->rk.ip_next.s_addr % hashSize);
    }
    int AtmIpHash (IpPacket *ipPacket) {
	return (ipPacket->ip_next.s_addr % hashSize);
    }
				  
    // Routing management functions
    void DoCallForRoute(double time, AtmRoute *atmRoute);
    void TransmitRoute(double time, AtmRoute *cllRoute);
    void DeleteRoute(AtmRoute *atmRoute);

    // Request ID management functions for downcalls.
    int GetRequestId(AtmRoute *atmRoute);
    void ReleaseRequestId(int r);

    //
    // Policy methods.  These functions are specified by device drivers
    // implementing IP-over-ATM policies.
    //
    virtual AtmRoute *GetAtmRoute(IpPacket *ipPacket);
    virtual AtmDeviceDriverConvType PacketClassifier(IpPacket *ipPacket);
    virtual AtmRoute *MakeAtmRoute(IpPacket *ipPacket);
    virtual void BindRoute(AtmRoute *atmRoute, IpPacket *ipPacket);
    virtual bool FallbackRoute(AtmRoute *atmRoute);
    virtual AtmRoute *GetCachedFallback(AtmRoute *atmRoute);
    virtual bool TimeoutAtmRoute(AtmRoute *atmRoute, double time);

};

extern int AtmDeviceDriverCmd(ClientData clientData, Tcl_Interp *interp,
			      int argc, char *argv[]);
extern int AtmDeviceDriverObjectCmd(ClientData clientData, Tcl_Interp *interp,
				    int argc, char *argv[]);
extern void AtmDeviceDriverDeleteProc(ClientData clientData);

extern ostream& operator<<(ostream &s, AtmDeviceDriverConvType ct);
extern ostream& operator<<(ostream &s, AtmDeviceDriverQosType qt);
extern ostream& operator<<(ostream &s, AtmDeviceDriverMuxType mt);
extern ostream& operator<<(ostream &s, AtmDeviceDriverVcType vt);
extern ostream& operator<<(ostream &s, AtmRouteState s);
extern ostream& operator<<(ostream &s, RouteKey &rk);

#endif // ATMDEVICEDRIVER_H
