static char rcsid[] = "$Id: AudioSink.cc,v 1.6 1996-04-17 11:00:54-07 bmah Exp $";
//
// $Id: AudioSink.cc,v 1.6 1996-04-17 11:00:54-07 bmah Exp $
//
// AudioSink.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Audio server (destination) application.  Receive time-stamped UDP
// packets and record statistics on delays (and loss?).
//

#include "AudioSink.h"
#include "Event.h"
#include "inline.h"

//
// AudioSink::AudioSink
//
// Input:  name, simulation object, start time, UDP object, port
//
// Output:  None
//
// Object constructor
//
AudioSink::AudioSink(const String &name, Simulation *si, double initTime,
		     String &u, short p) :
		     AppUdpServer(name, si, initTime, u, p)
{

    InAddr local;
    InAddr remote;

    startTime = initTime;

    local.s_addr = 0;		// XXX INADDR_ANY
    remote.s_addr = 0;		// XXX INADDR_ANY
    up = udpModule->UserBind(initTime, remote, 0, local, p, this);

    // Initialize accounting variables
    lastPacket = -1;
    packetsInSequence = 0;
    packetsOutOfOrder = 0;
    packetsGapped = 0;
    packets = 0;
    delays = new SampleHistogram(0.0, 1.0, 0.05);

}

//
// AudioSink::~AudioSink
//
// C++ destructor
//
AudioSink::~AudioSink()
{
    int i;

    *(sim->logStream) << name << " [AudioSink] lifetime " << lifeTime <<
      " in-sequence " << packetsInSequence <<
      " out-of-order " << packetsOutOfOrder <<
      " gaps " << packetsGapped <<
      " lost " << lastPacket + 1 - packetsInSequence <<
      " mean-delay " << delays->mean() <<
      " var-delay " << delays->var() <<
      " min-delay " << delays->min() <<
      " max-delay " << delays->max() << "\n";
    *(sim->logStream) << name << " [AudioSink] delays ";
    for (i = 0; i < delays->buckets(); i++) {
	*(sim->logStream) << delays->bucketThreshold(i) << " " << delays->inBucket(i) << " ";
    }
    *(sim->logStream) << "\n";
}

//
// AudioSink::Handler
//
// (Event handler)
//
void
AudioSink::Handler(Event *e)
{
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    IF_DEBUG(3, {
	DEBUG_PREFIX("AudioSink");
	*(sim->debugStream) << "got event " << type << "\n";
    });

    switch (type) {

	//
	// EventTimer:  The initial "startup" timer can cause
	// an EventTimer event, which we just ignore.  We'll also get
	// a timer event when we're supposed to quit.
	//
	case (EventTimer):
	{
	    if (time == startTime) {
		// Ignore the startup timer
	    }
	    else {
		// Ending timer fired...we're done.
		udpModule->UserClose(time, up);
		delete this;
	    }
	}
	break;

	case (EventUdpArrival):
	{

	    // Get datagram data
	    UdpCall *call = (UdpCall *) data;
	    UdpData *data = call->data.Dequeue();

	    if (data) {

		IF_DEBUG(2, {
		    DEBUG_PREFIX("AudioSink");
		    *(sim->debugStream) << "received packet " <<
		      data->seq << " ";
		});

		// Check packet status
		if (data->seq == lastPacket + 1) {
		    // in-sequence
		    packetsInSequence++;

		    IF_DEBUG(2, {
			*(sim->debugStream) << "in-sequence";
		    });
		}
		else if (data->seq < lastPacket + 1) {
		    // out of order or duplicate
		    packetsOutOfOrder++;

		    IF_DEBUG(2, {
			*(sim->debugStream) << "out-of-order";
		    });

		}
		else {
		    // gap (packet dropped or misordered)
		    packetsGapped++;

		    IF_DEBUG(2, {
			*(sim->debugStream) << "gap";
		    });
		}
		packets++;
		if (data->seq > lastPacket) {
		    lastPacket = data->seq;
		}

		// Compute end-to-end delay of packet
		double delay = time - data->data_time;
		*(delays) += delay;

		delete call;
		delete data;

		IF_DEBUG(2, {
		    *(sim->debugStream) << " delay " << delay << "\n";
		});

	    }
	    else {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("AudioSink");
		    *(sim->debugStream) << "got packet without payload\n";
		});
	    }
	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("AudioSink");
		*(sim->debugStream) << "got bad EventType " << type << "\n";
	    });
	}
	break;

    }

}

//
// AudioSink::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
AudioSink::Output()
{
    return name + " [AudioSink]";
}

//
// AudioSinkCmd
//
// (Tcl command procedure)
//
int
AudioSinkCmd(ClientData clientData, Tcl_Interp *interp,
	     int argc, char *argv[])
{
    if (argc != 5) {
	Tcl_SetResult(interp, "usage:  AudioSink name start udp port", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	AudioSink *as;	// new client object
	char *nm;		// name of new client object
	double startTime;	// starting time to run
	String udpObjectName;	// name of Udp object
	int port;		// port number
	extern Simulation *sim;	// simulation object
	int i;			// yet another loop counter

	// Get name
	nm = argv[1];

	// Get start time
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    // Get UDP object name
	    udpObjectName = argv[3];

	    // Get port number
	    rc = Tcl_GetInt(interp, argv[4], &port);
	    if (rc == TCL_OK) {
		
		as = new AudioSink(nm, sim, startTime,
				   udpObjectName,
				   port);
		Tcl_CreateCommand(interp, nm,
				  AudioSinkObjectCmd,
				  (ClientData) as,
				  AudioSinkDeleteProc);

	    }
	}
    return rc;
    }
}


//
// AudioSinkObjectCmd
//
// (Tcl object command procedure)
//
int
AudioSinkObjectCmd(ClientData clientData, Tcl_Interp *interp,
		   int argc, char *argv[])
{
    
    AudioSink *as = (AudioSink *) clientData;

    if (argc == 1) {
	return ScheduledObjectCmd(clientData, interp, argc, argv);
    }
    else {

	//
	// as lifetime ...
	//
	if (!strcmp(argv[1], "lifetime")) {

	    //
	    // as lifetime
	    //
	    if (argc == 2) {
		char ch[80];
		sprintf(ch, "%f", as->lifeTime);
		Tcl_SetResult(interp, ch, TCL_VOLATILE);
		return TCL_OK;
	    }
	    
	    //
	    // as lifetime lt
	    //
	    else if (argc == 3) {
		int rc = TCL_OK;
		double lt;

		rc = Tcl_GetDouble(interp, argv[2], &lt);
		if (rc == TCL_OK) {
		    as->lifeTime = lt;
		    as->endTime = as->lifeTime + as->startTime;

		    // Post a timer event when we're supposed to go away.
		    as->Post(as->endTime, as, EventTimer, NULL);

		}
		return rc;
	    }		

	}
	else {
	    return ScheduledObjectCmd(clientData, interp, argc, argv);
	}
    }
}

//
// AudioSinkDeleteProc
//
// (Tcl delete procedure)
//
void
AudioSinkDeleteProc(ClientData clientData)
{
    delete (AudioSink *) clientData;
}
