static char rcsid[] = "$Id: AudioSource.cc,v 1.6 1996-04-16 11:03:14-07 bmah Exp $";
//
// $Id: AudioSource.cc,v 1.6 1996-04-16 11:03:14-07 bmah Exp $
//
// AudioSource.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Audio client (sender) application.  Sends UDP packets containing
// audio data, with on/off times determined by tcplib distribution.
//

#include "AudioSource.h"
#include "EmpiricalContinuous.h"
#include "Event.h"
#include "inline.h"

//
// Imported probability distributions
//
extern RNG *rng;		// global random number generator

// Talk spurt length distribution
extern const DoubleDistElement *PhoneTalkspurtDistTable;
extern int PhoneTalkspurtDistLength;

// Silence length distribution
extern const DoubleDistElement *PhonePauseDistTable;
extern int PhonePauseDistLength;

//
// AudioSource::AudioSource
//
// Input:  name, simulation object, start time, UDP object, server
//         IP address, server port
//
// Output:  None
//
// Object constructor
//
AudioSource::AudioSource(const String &name, Simulation *si, double initTime,
			 String &u, InAddr s, short p, int pay,
			 double g) :
			 AppUdpClient(name, si, initTime, u, s, p),
			 payloadBytes(pay), gap(g)
{

    InAddr local;

    lifeTime = 100.0;		// XXX
    startTime = initTime;
    endTime = lifeTime + initTime;

    local.s_addr = 0;		// XXX INADDR_ANY
    up = udpModule->UserBind(initTime, s, p, local, 0, this);

    packetsSent = 0;

    // Configure the AudioSource to come up in the off state, but
    // start up after gap seconds.  The parent class constructor
    // posts the initial timer event to get the ball rolling.
    state = AudioSourceStateOff;
    stateChangeTime = initTime + gap;

}

//
// AudioSource::Handler
//
// (Event handler)
//
void
AudioSource::Handler(Event *e)
{
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalContinuous PhoneTalkspurtDist(PhoneTalkspurtDistLength,
						  PhoneTalkspurtDistTable,
						  rng);
    static EmpiricalContinuous PhonePauseDist(PhonePauseDistLength,
					      PhonePauseDistTable,
					      rng);

    IF_DEBUG(3, {
	DEBUG_PREFIX("AudioSource");
	*(sim->debugStream) << "got event " << type << "\n";
    });

    switch (state) {

	//
	// AudioSourceStateOn:  Currently transmitting audio data.
	//
	case (AudioSourceStateOn):
	{

	    // See if we're done.
	    if (time >= endTime) {

		// Delete object
		udpModule->UserClose(time, up);
		delete this;

	    }

	    // See if we've come to the end of a talk spurt.
	    // Our implementation has the effect of rounding the length
	    // of talk spurts up to the next higher multiple of the
	    // packet time.
	    else if (time >= stateChangeTime) {

		// Change state
		state = AudioSourceStateOff;
		stateChangeTime = time + PhonePauseDist();
		Post(stateChangeTime, this, EventTimer, NULL);

		IF_DEBUG(2, {
		    DEBUG_PREFIX("AudioSource");
		    *(sim->debugStream) << "change to off state, ends at " << stateChangeTime<< "\n";
		});
	    }

	    // Send packet, set a timer for us to check again
	    else {

		UdpData *ud = new UdpData;
		ud->data_time = time;
		ud->seq = packetsSent;

		udpModule->UserSend(time, up, payloadBytes, ud,
				    UdpPayloadTypeAudio);
		Post(time + gap, this, EventTimer, NULL);
		
		IF_DEBUG(2, {
		    DEBUG_PREFIX("AudioSource");
		    *(sim->debugStream) << "sent packet " << packetsSent << " " << payloadBytes << " bytes\n";
		});

		packetsSent++;

	    }

	}
	break;

	//
	// AudioSourceStateOff:  Pausing
	//
	case (AudioSourceStateOff):
	{

	    switch (type) {

		// Got timer, signifying end of pause
		case (EventTimer):
		{

		    state = AudioSourceStateOn;
		    stateChangeTime = time + PhoneTalkspurtDist();
		    
		    // Cause a timeout to have AudioSourceStateOn
		    // part of handler to send packet.
		    Post(time + gap, this, EventTimer, NULL);
		    
		    IF_DEBUG(2, {
			DEBUG_PREFIX("AudioSource");
			*(sim->debugStream) << "change to on state, ends at " << stateChangeTime<< "\n";
		    });

		}
		break;

		default:
		{
		    IF_DEBUG(0, {
			DEBUG_PREFIX("AudioSource");
			*(sim->debugStream) << "bad event type " <<
			  " received in state " << state << "\n";
		    });
		}
		break;

	    }

	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("AudioSource");
		*(sim->debugStream) << "Unknown state\n";
	    });
	}
	break;
	
    }

}

//
// AudioSource::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
AudioSource::Output()
{
    return name + " [AudioSource]";
}

//
// AudioSourceCmd
//
// (Tcl command procedure)
//
int
AudioSourceCmd(ClientData clientData, Tcl_Interp *interp,
	       int argc, char *argv[])
{
    if (argc != 8) {
	Tcl_SetResult(interp, "usage:  AudioSource name start udp server port payload gap", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	AudioSource *as;	// new client object
	char *nm;		// name of new client object
	double startTime;	// starting time to run
	String udpObjectName;	// name of Udp object
	InAddr destaddr;	// destination address
	int port;		// port number
	int pay;		// payload size
	double g;		// gap
	extern Simulation *sim;	// simulation object
	int i;			// yet another loop counter

	// Get name
	nm = argv[1];

	// Get start time
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    // Get UDP object name
	    udpObjectName = argv[3];

	    // Get server address
	    destaddr.s_addr = InetAddr(argv[4]);

	    // Get port number
	    rc = Tcl_GetInt(interp, argv[5], &port);
	    if (rc == TCL_OK) {
		
		// Get payload
		rc = Tcl_GetInt(interp, argv[6], &pay);
		if (rc == TCL_OK) {

		    // Get gap
		    rc = Tcl_GetDouble(interp, argv[7], &g);
		    if (rc == TCL_OK) {

			as = new AudioSource(nm, sim, startTime,
					     udpObjectName,
					     destaddr, port,
					     pay, g);
			Tcl_CreateCommand(interp, nm,
					  AudioSourceObjectCmd,
					  (ClientData) as,
					  AudioSourceDeleteProc);

		    }
		}
	    }
	}
    return rc;
    }
}


//
// AudioSourceObjectCmd
//
// (Tcl object command procedure)
//
int
AudioSourceObjectCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
    
    AudioSource *as = (AudioSource *) clientData;

    if (argc == 1) {
	return ScheduledObjectCmd(clientData, interp, argc, argv);
    }
    else {

	//
	// as lifetime ...
	//
	if (!strcmp(argv[1], "lifetime")) {

	    //
	    // as lifetime
	    //
	    if (argc == 2) {
		char ch[80];
		sprintf(ch, "%f", as->lifeTime);
		Tcl_SetResult(interp, ch, TCL_VOLATILE);
		return TCL_OK;
	    }
	    
	    //
	    // as lifetime lt
	    //
	    else if (argc == 3) {
		int rc = TCL_OK;
		double lt;

		rc = Tcl_GetDouble(interp, argv[2], &lt);
		if (rc == TCL_OK) {
		    as->lifeTime = lt;
		    as->endTime = as->lifeTime + as->startTime;
		}
		return rc;
	    }		

	}
	else {
	    return ScheduledObjectCmd(clientData, interp, argc, argv);
	}
    }
}

//
// AudioSourceDeleteProc
//
// (Tcl delete procedure)
//
void
AudioSourceDeleteProc(ClientData clientData)
{
    delete (AudioSource *) clientData;
}
