static char rcsid[] = "$Id: Calendar.cc,v 1.9 1996-02-05 22:14:54-08 bmah Exp $";
//
// $Id: Calendar.cc,v 1.9 1996-02-05 22:14:54-08 bmah Exp $
//   based on
// $HEADER: /tmp_mnt/n/truth/da/users/bmah/c268/src/RCS/calendar.c,v 1.4 1992/05/01 04:15:32 bmah Exp $
//
// Calendar.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Based on original C code by:
// Hui Zhang (hzhang@tenet.berkeley.edu)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Hui Zhang and Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Calendar queue scheduling package.
//

#include <stdio.h>
#include "Calendar.h"
#include "util.h"

/*
 * Compile-time configuration
 */
#define QUEUE_SIZE     1024*6  /* maximum bucket is 4096                */
#define HALF_MAX_DAYS  1024*2  

//
// Calendar::Calendar
//
// Input:  None.
//
// Output:  None.
//
// Initialization for the calendar queue.
//
Calendar::Calendar()
{
    clock = 0.0;
    LocalInit(0, 2, 1.0, 0.0);	// plenty of voodoo constants here
    cq_resizeEnabled = 1;
}

//
// Calendar::Insert
//
// Input:  event to insert into calendar queue (event)
//
// Output:  None
//
// Insert an event into the calendar queue for later scheduling.
//
void Calendar::Insert(Event *event)
{

    extern int events;
    unsigned int i;
    
    // Zap pointers
    event->ev_next = NULL;
    event->ev_prev = NULL;

    //
    // A small amount of sanity checking:  Make sure we don't try to
    // insert an event for a time that's already occurred.
    //
    if (event->ev_time < clock) {
	char errormsg[256];
	sprintf(errormsg, "Calendar::Insert:  tried to insert an event in the past (%f) at time %f\n", event->ev_time, clock);
	panic(errormsg);
    }

    //
    // Also make sure the event has a valid recipient (non-NULL)
    //
    if (event->ev_object == NULL) {
	panic("Calendar::Insert:  tried to insert an event for a NULL object");
    }

    i = (unsigned int) (event->ev_time/cq_interval); // find virtual bucket 
    i = i % cq_bucketNum;		    // find actual bucket

    InsertEvent(&cq_bucket[i], event);
    if (cq_eventNum > cq_topThreshold && cq_bucketNum < HALF_MAX_DAYS)
	Resize(2 * cq_bucketNum);
}

//
// Calendar::Get
//
// Input:  None
//
// Output:  Next event on the calendar queue (in return value)
//
// Dequeue the next event from the calendar queue and update the
// current time.
//
Event *Calendar::Get()
{
    Event *e = NextEvent();
    if (e) {
	clock = e->ev_time;
    }
    return e;
}

//
// Calendar::NextEvent
//
// Input:  None
//
// Output:  Next event on the calendar queue (in return value)
//
// Dequeue the next event from the calendar queue (no side effects).
// This is a protected routine...programs using this class should
// use Calendar::Get() to get objects.
//
Event *Calendar::NextEvent()
{
    register unsigned int i, virtualBucket;
    int year;
    struct Event *result;
    
    if (cq_eventNum == 0) return(NULL);
    again:
    for (i = cq_lastBucket;;) {       /* search buckets */
	if (cq_bucket[i] && cq_bucket[i]->ev_time < cq_bucketTop) {
	    result = cq_bucket[i];
	    cq_lastTime = result->ev_time;	
	    cq_bucket[i] = cq_bucket[i]->ev_next;
	    if (cq_bucket[i]) 
	      cq_bucket[i]->ev_prev = NULL;
	    cq_lastBucket = i;
	    if (--cq_eventNum < cq_bottomThreshold)
	      Resize(cq_bucketNum/2);

	    if (result == cq_bucket[i]) {
		cout << "!!! Calendar::NextEvent die !!! " << result << "\n";
	    }

	    return(result);
	} else {
	    if ( ++i == cq_bucketNum) 
	      i = 0;
	    cq_bucketTop += cq_interval;
	    if (i == cq_lastBucket)
	      break; /* go to direct search */
	}
    }
    
    /* 
     * find lowest priority by examining first event of each bucket 
     * set cq_lastBucket, cq_lastTime, cq_bucketTop                 
     */
    result = NULL;
    for (i=0; i<cq_bucketNum; i++) 
      if (cq_bucket[i]) {
	  result = cq_bucket[i];
	  cq_lastBucket = i;
	  break;
      }
    if (result == NULL) 
      panic("Calendar::NextEvent:  panic");
    for (i=cq_lastBucket+1; i<cq_bucketNum; i++) {
	if (cq_bucket[i] && cq_bucket[i]->ev_time < result->ev_time)  {
	    result = cq_bucket[i];
	    cq_lastBucket = i;
	}
    }
    
    cq_lastTime = result->ev_time;
    year = (int) (cq_lastTime/(cq_interval*cq_bucketNum));
    cq_bucketTop = year*cq_interval*cq_bucketNum + 
      (cq_lastBucket+1.5)*cq_interval;
    goto again;
}

//
// Calendar::LocalInit
//
// Input:  Starting position of new bucket array (qbase)
//         Number of buckets (nbuck).
//         Bucket width (startInterval).
//         Time to begin dequeueing this bucket (lastTime).
//
// Output:  None.
//
// Initialize a bucket array within the array CalendarQ.
//
void Calendar::LocalInit(int qbase, int nbuck,
			 double startInterval, double lastTime)
{
    int i;
    unsigned int n;

/* set position and size of new queue */
    
    cq_firstSub = qbase;
    cq_bucket = CalendarQ + qbase;
    cq_interval = startInterval;
    cq_bucketNum = nbuck;
    
/* initialize as empty */
    
    cq_eventNum = 0;
    for (i=0; i< cq_bucketNum; i++) cq_bucket[i] = NULL;
    
/* set up initial position in queue */

    cq_lastTime = lastTime;
    n = (unsigned int) (lastTime/cq_interval);
    cq_lastBucket = n % cq_bucketNum;
    cq_bucketTop = (n+1.5)*cq_interval;

/* set up queue size change thresholds */
    
    cq_bottomThreshold = cq_bucketNum/2 - 2;
    cq_topThreshold = 2*cq_bucketNum;
}

//
// Calendar::Resize
//
// Input:  New size of calendar (newSize)
//
// Output:  None.
//
// Copy the queue onto a calendar with newSize buckets.  The new
// bucket array is now the opposite end of the array.
//
void Calendar::Resize(int newSize)
{
    double newInterval;
    int i;
    int oldBucketNum;
    struct Event *currentEvent, **oldBucket;

    if (!cq_resizeEnabled) return;

    newInterval = NewInterval();

/* save location and size of old calendar for use when copying calendar */

    oldBucket = cq_bucket;
    oldBucketNum = cq_bucketNum;
    

/* initialize new calendar */

    if (cq_firstSub) 
	LocalInit(0, newSize, newInterval, cq_lastTime);
    else
	LocalInit(QUEUE_SIZE - newSize, newSize, newInterval, cq_lastTime);

    for (i = oldBucketNum - 1; i>=0; i--) {
	currentEvent = oldBucket[i];
	while (currentEvent) {
	    struct Event *nextEvent = currentEvent->ev_next;
	    unsigned int virtualBucket;

	    virtualBucket = (unsigned int) (currentEvent->ev_time/cq_interval);
	    virtualBucket = virtualBucket % cq_bucketNum; 
	    InsertEvent(&cq_bucket[virtualBucket], currentEvent);
	    currentEvent = nextEvent;
	}
    }
	    
}

//
// Calendar::NewInterval
//
double Calendar::NewInterval()
{
    int sampleNum;
    int i;
    int validSeparationNum;
    double eventSeparation[25], totalSeparation, twiceAverageSeparation;
    struct Event *sampledEvent[25];
    double save_lastTime, save_bucketTop;
    int save_lastBucket;
    

/* decide how many queue elements to sample */

    if (cq_eventNum < 2) 
	return(1.0);
    if (cq_eventNum <= 5)
	sampleNum = cq_eventNum;
    else
	sampleNum = 5 + cq_eventNum/10;
    if (sampleNum > 25)
	sampleNum = 25;

/* save important dynamic values */    

    save_lastTime = cq_lastTime;
    save_lastBucket = cq_lastBucket;
    save_bucketTop = cq_bucketTop;

    cq_resizeEnabled = 0;
    
    sampledEvent[0] = NextEvent();
    eventSeparation[0] = sampledEvent[0]->ev_time;

    totalSeparation = 0;
    for (i=1; i<sampleNum; i++) {
	sampledEvent[i] = NextEvent();
	eventSeparation[i] = sampledEvent[i]->ev_time;
	eventSeparation[i-1] = 
	    eventSeparation[i] - eventSeparation[i-1];
	totalSeparation += eventSeparation[i-1];
    }
    
    twiceAverageSeparation = 2*totalSeparation/(sampleNum - 1);
    
    totalSeparation = 0;
    validSeparationNum = 0;
    for (i=0; i<sampleNum-1; i++) 
	if (eventSeparation[i] < twiceAverageSeparation) {
	    totalSeparation += eventSeparation[i];
	    validSeparationNum++;
	}

/* insert sampled events back */

    for (i=0; i<sampleNum; i++) {
	unsigned int temp;
	temp = (unsigned int)
	  (sampledEvent[i]->ev_time/cq_interval); /* find virtual bucket */
	temp = temp % cq_bucketNum;		  /* find actual bucket  */
	InsertEvent(&cq_bucket[temp], sampledEvent[i]);
    }	

/* save important dynamic values */    

    cq_lastTime = save_lastTime;
    cq_lastBucket = save_lastBucket;
    cq_bucketTop = save_bucketTop;
    cq_resizeEnabled = 1;
    if (validSeparationNum && totalSeparation)
	return((double)(3*totalSeparation/validSeparationNum));
    else 
	return((double)cq_interval);
}
    
//
// Calendar::InsertEvent
//
void Calendar::InsertEvent(Event **bucket, Event *event)
{
    register unsigned int i, virtualBucket;
    register int save_resizeEnabled;
    struct Event *current, *least_event = NULL;
    double year;
    
    if (cq_eventNum) {
	save_resizeEnabled = cq_resizeEnabled;
	cq_resizeEnabled = 0;
	least_event = NextEvent();
	i = (unsigned int) (least_event->ev_time/cq_interval);
	i = i % cq_bucketNum;
	if (cq_bucket[i]) {
	    least_event->ev_next = cq_bucket[i];
	    if (least_event->ev_next == least_event) {
		cout << "!!! Calendar::InsertEvent reinsert non-empty die !!! " << least_event << "\n";
		abort();
	    }
	    least_event->ev_prev = NULL;
	    cq_bucket[i]->ev_prev = least_event;
	} else {
	    least_event->ev_next = least_event->ev_prev = NULL;
	    if (least_event->ev_next == least_event) {
		cout << "!!! Calendar::InsertEvent reinsert empty die !!! " << least_event << "\n";
		abort();
	    }
	}
	cq_bucket[i]= least_event;
	cq_resizeEnabled = save_resizeEnabled;
	cq_eventNum++;

    }
    if (least_event && event->ev_time < least_event->ev_time) {
	cq_lastTime = event->ev_time;
	virtualBucket = (unsigned int) (cq_lastTime/cq_interval);
	cq_lastBucket = virtualBucket % cq_bucketNum;
	cq_bucketTop = (virtualBucket+1.5)*cq_interval;
    }
    if (*bucket) {
	current = *bucket;
	while (current->ev_time < event->ev_time) {
	    if (current == event) {
		cout << "!!! Calendar::InsertEvent traversal current == event die !!! " << current << "\n";
		cout << "bucket = " << bucket << "\n";
		cout << "*bucket = " << *bucket << "\n";
		abort();
	    }
	    if (current == current->ev_next) {
		cout << "!!! Calendar::InsertEvent traversal ev_next die !!! " << current << "\n";
		cout << "bucket = " << bucket << "\n";
		cout << "*bucket = " << *bucket << "\n";
		abort();
	    }
	    if (current->ev_next)
		current = current->ev_next;
	    else {
		if (event == event->ev_next) {
		    cout << "!!! Calendar::InsertEvent end insert die !!! " << event << "\n";
		}
		current->ev_next = event;
		event->ev_prev = current;
		event->ev_next = NULL;
		cq_eventNum++;
		return;
	    }
	}
	if (current->ev_prev) {
	    event->ev_prev = current->ev_prev;
	    event->ev_next = current;
	    current->ev_prev->ev_next = event;
	    current->ev_prev = event;
	    if (event == event->ev_next) {
		cout << "!!! Calendar::InsertEvent middle insert die !!! " << event << "\n";
	    }
	    if (current->ev_prev == current->ev_prev->ev_next) {
		cout << "!!! Calendar::InsertEvent middle insert ev_pref die !!! " << current->ev_prev << "\n";
	    }
	}  else {
	    if (event == current) {
		cout << "!!! Calendar::InsertEvent start insert = die !!! " << event << "\n";
	    }
	    event->ev_prev = NULL;
	    event->ev_next = current;
	    current->ev_prev = event;
	    *bucket = event;
	    if (event == event->ev_next) {
		cout << "!!! Calendar::InsertEvent start insert die !!! " << event << "\n";
	    }
	}
    } else { /* empty bucket */
	event->ev_prev = event->ev_next = NULL;
	*bucket = event;
    }

    cq_eventNum++;
    return;
    
}

//
// Calendar::Sanity
//
// Input:  None
//
// Output:  None
//
// Do a sanity check on the calendar queue data structure.  Originally
// invented to catch a bug where a bad pointer was freeing up data
// structures inside the calendar queue.
//
void Calendar::Sanity() {

    int i;
    int count = 0;
    Event *e, *enext;

    for (i = 0; i < cq_bucketNum; i++) {

	for (e = cq_bucket[i]; e != NULL; e = enext) {
	    enext = e->ev_next;
	    if (enext == e) {
		cout << "!!! Calendar::Sanity die !!! Sanity detected loop " << e << " bucket " << i << " time " << clock << "\n";
		abort();
	    }
	    count++;
	}
	
    }

    if (count != cq_eventNum) {
	cout << "!!! Calendar::Sanity die !!! event count mismatch count " << count << " cq_eventNum " << cq_eventNum << " time " << clock << "\n";
	abort();
    }

}
