static char rcsid[] = "$Id: CellGoBackN.cc,v 1.12 1996-05-29 14:34:46-07 bmah Exp $";
//
// $Id: CellGoBackN.cc,v 1.12 1996-05-29 14:34:46-07 bmah Exp $
//
// CellGoBackN.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Reliable datagram protocol using a go-back-N scheme.  Used to
// provide reliability for signalling messages.
//

#include <Uniform.h>
#include "CellGoBackN.h"

//
// Inter-cell gap
//
static double epsilon = 0.00001; // 1 us

//
// CellGoBackN::CellGoBackN
//
// Input:
//
// Output:  none
//
// Object constructor
//
CellGoBackN::CellGoBackN(const String &name, Simulation *si, int p,
			 int mc, double t, double po) :
ScheduledObject(name, si),
ports(p),
maxCells(mc),
timeout(t),
poll(po) {

    Uniform skewGenerator(0, poll, si->rng); // skew random variable generator
    int i;

    // Allocate memory for translation tables and sequence number
    // tables.
    retransmitQueue = new AtmCellList[ports];
    incomingSeq = new (int)[ports];
    outgoingSeq = new (int)[ports];
    
    // Initialize arrays
    for (i = 0; i < ports; i++) {
	incomingSeq[i] = 0;
	outgoingSeq[i] = 0;
    }
    
    // Post retransmit timer event.
    nextTimeout = skewGenerator();
    Post(nextTimeout, this, EventTimer, NULL);
    
}

//
// CellGoBackN::Handler
//
// (Event handler)
//
void
CellGoBackN::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    switch (type) {

	// Cell arrived at module.
	// We need to figure out what type it is (ACK or DATA)
	case (EventCellArrival):
	{
	    AtmCell *cell = (AtmCell *) data;

	    switch (cell->type) {

		// Data cell.  Check the sequence number to see what to
		// do with the cell.
		case (AtmCellTypeData):
		case (AtmCellTypeEndOfPacket):
	        {

		    // If cell sequence number matched expected number, we
		    // successfully got the next data cell in the sequence.
		    if ((cell->seq) == incomingSeq[cell->vcid]) {

			// Pass the cell up.
			Lookup(up, upName);
			Post(time, up, EventCellArrival, data);

			// Make up an acknowledgement and send it.
			AtmCell *ack = new AtmCell(AtmCellTypeAck,
						   NULL,
						   cell->seq,
						   0);
			ack->vcid = cell->vcid;
			Lookup(down, downName);
			Post(time, down, EventCellArrival, (void *) ack);

			// Prepare to get the next cell in sequence.
			incomingSeq[cell->vcid]++;

			IF_DEBUG(2, {
			    DEBUG_PREFIX("CellGoBackN");
			    *(sim->debugStream) <<
			      "received in-sequence cell " << cell->seq <<
			      " on vcid " << cell->vcid << " " << cell <<
			      "\n";
			});

		    }
		    
		    // If the cell sequence number was smaller than expected,
		    // this is a retransmitted cell that we've seen before.
		    else if ((cell->seq) < incomingSeq[cell->vcid]) {

			// Send acknowledgement, but take no other action.
			AtmCell* ack = new AtmCell(AtmCellTypeAck,
						   NULL,
						   cell->seq,
						   0);
			ack->vcid = cell->vcid;
			Lookup(down, downName);
			Post(time, down, EventCellArrival, (void *) ack);
			
			IF_DEBUG(2, {
			    DEBUG_PREFIX("CellGoBackN");
			    cout << "received duplicate cell " <<
			      cell->seq <<
			      " (expecting " << incomingSeq[cell->vcid] <<
			      ") " <<
			      " on vcid " << cell->vcid << " " << cell <<
			      "\n";
			});

		    }

		    // Final alternative is that the sequence number was too
		    // big; in this case some intervening cells must have been
		    // dropped.  Drop the current cell and hope the sender will
		    // retransmit from the last cell we ACKed + 1.
		    else {

			IF_DEBUG(2, {
			    DEBUG_PREFIX("CellGoBackN");
			    *(sim->debugStream) <<
			      "received out of sequence cell " << cell->seq <<
			      " (expecting " << incomingSeq[cell->vcid] <<
			      ") " <<
			      " on vcid " << cell->vcid << " " << cell << "\n";
			});

			delete cell;

		    }
		}
		break;

		// Cell was an acknowledgement cell (presumably for a cell
		// we transmitted).
		case (AtmCellTypeAck):
		{

		    int i;
		    
		    // If the ACK was for the cell at the head of our
		    // list, then dequeue it...we're done with that cell
		    if ((retransmitQueue[cell->vcid][0]) &&
			(cell->seq == retransmitQueue[cell->vcid][0]->seq)) {

			delete (retransmitQueue[cell->vcid].Dequeue());

			IF_DEBUG(2, {
			    DEBUG_PREFIX("CellGoBackN");
			    *(sim->debugStream) <<
			      "accepted ack cell " << cell->seq <<
			      " on vcid " << cell->vcid << " " << cell << "\n";
			});
		    
		    }
		    else {

			IF_DEBUG(2, {
			    DEBUG_PREFIX("CellGoBackN");
			    *(sim->debugStream) <<
			      "rejected ack cell " << cell->seq <<
			      " on vcid " << cell->vcid << " " << cell << "\n";
			});

		    }
		    
		    // Done with the ACK cell now.
		    delete cell;
		}
		break;
	    }
	}
	break;
	
	// EventTimer:  Retransmit timer went off.
	case (EventTimer):
	{
	    
	    int i, j;
	    double delta = 0.0;

	    // Walk the retransmit list, resend cells and reschedule
	    // retransmission.
	    for (i = 0; i < ports; i++) {

		AtmCell *queuedCell, *retransmittedCell;
		int retransmittedCells = 0;

		for (j = 0; j < retransmitQueue[i].length; j++) {

		    queuedCell = retransmitQueue[i][j];
		    if ((queuedCell) && (retransmittedCells < maxCells / 2)) {
			retransmittedCell = new AtmCell(*queuedCell);

			if (retransmittedCell->vcid != i) {
			    IF_DEBUG(2, {
				DEBUG_PREFIX("CellGoBackN");
				*(sim->debugStream) <<
				  "warning:  queue number != vcid\n";
			    });
			}
			
			Lookup(down, downName);
			Post(time + delta, down, EventCellArrival,
			     (void *) retransmittedCell);
			delta += epsilon;
			queuedCell->time = time + timeout;
			retransmittedCells++;

			IF_DEBUG(2, {
			    DEBUG_PREFIX("CellGoBackN");
			    *(sim->debugStream) <<
			      "retransmit cell " << retransmittedCell->seq <<
			      " on vcid " << retransmittedCell->vcid <<
			      " " <<
			      retransmittedCell << "\n";
			});
		    }
		}
	    }

	    // Post another timeout event to wake us up.
	    nextTimeout = time + poll;
	    Post(nextTimeout, this, EventTimer, NULL);
	}
	break;

	// EventCellSend:  Upper-layer object requested
	// to send a cell.
	case (EventCellSend):
	{
	    AtmCell *cell = (AtmCell *) data;
	    int i;

	    // Plug a new sequence number into the cell
	    cell->seq = outgoingSeq[cell->vcid];
	    outgoingSeq[cell->vcid]++;

	    // Put a copy of the cell on the retransmit queue, setting
	    // the retransmission time.  time field in the AtmCell header
	    // now denotes the retransmission time of the cell, and will
	    // be updated during retransmissions.
	    if (retransmitQueue[cell->vcid].length < maxCells) {

		AtmCell *copyCell = new AtmCell(*cell);
		copyCell->time = time + timeout;
		retransmitQueue[cell->vcid].EnqueueTail(copyCell);

		// Transmission window is half the max number of cells
		// we can queue.
		if (retransmitQueue[cell->vcid].length < (maxCells / 2)) {
		    Lookup(down, downName);
		    Post(time, down, EventCellArrival, data);
		
		    IF_DEBUG(2, {
		      DEBUG_PREFIX("CellGoBackN");
		      *(sim->debugStream) <<
			"sent cell " << cell->seq <<
			" on vcid " << cell->vcid << " " << cell << "\n";
		    });
		}
		else {
		    delete cell;
		}
		
	    }
	    else {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("CellGoBackN");
		    *(sim->debugStream) << "ran out of slots on retransmit queue\n";
		});
	    }

	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("CellGoBackN");
		*(sim->debugStream) << "bad EventType " << type << "\n";
	    });
	}
	break;

    }
    
}

//
// CellGoBackN::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String CellGoBackN::Output()
{

    String out;
    char buf[30];

    sprintf(buf, "%d", ports);
    out = name + " [CellGoBackN] ports " + buf;

    sprintf(buf, "%d", maxCells);
    out += (String) " maxCells " + buf;

    sprintf(buf, "%f", timeout);
    out += (String) " timeout " + buf;

    sprintf(buf, "%f", poll);
    out += (String) " poll " + buf;

    return out;
    
}
//
// CellGoBackNCmd
//
// (Tcl command handler procedure)
//
// CellGoBackN name numports maxcells timeout poll
//
int CellGoBackNCmd(ClientData clientData, Tcl_Interp *interp,
		    int argc, char *argv[])
{

    if (argc != 6) {
	Tcl_SetResult(interp,
		      "Usage:  CellGoBackN name ports maxvci timeout poll",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of new CellGoBackN
	int np;			// num ports
	int mc;			// maximum number of VCIs
	double t;		// timeout
	double p;		// polling interval
	int rc;			// return code from Tcl library
	CellGoBackN *gbn;	// new object
	extern Simulation *sim;	// global simulation object

	// get name
	nm = argv[1];

	// get numeric arguments
	rc = Tcl_GetInt(interp, argv[2], &np);
	if (rc == TCL_OK) {

	    rc = Tcl_GetInt(interp, argv[3], &mc);
	    if (rc == TCL_OK) {

		rc = Tcl_GetDouble(interp, argv[4], &t);
		if (rc == TCL_OK) {

		    rc = Tcl_GetDouble(interp, argv[5], &p);
		    if (rc == TCL_OK) {

			gbn = new CellGoBackN(nm, sim, np, mc, t, p);
			Tcl_CreateCommand(interp, nm,
					  CellGoBackNObjectCmd,
					  (ClientData) gbn,
					  CellGoBackNDeleteProc);
		    }
		}
	    }
	}

	return rc;
	
    }

}

//
// CellGoBackNObjectCmd
//
// (Tcl object handler procedure)
//
int CellGoBackNObjectCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[])
{
    
    CellGoBackN *gbn = (CellGoBackN *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }
    else {

	//
	// gbn up object
	//
	// Connect switch module output to upper-layer protocol
	// (usually Sig type object)
	//
	if (!strcmp(argv[1], "up")) {

	    if (argc == 3) {

		String outName;	// object name

		outName = argv[2];

		// set new object...zero out cached object pointer
		gbn->upName = outName;
		gbn->up = NULL;

		return (TCL_OK);
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}
	//
	// gbn down object
	//
	// Connect switch module output to lower-layer object
	// (usually SwitchModule type object)
	//
	else if (!strcmp(argv[1], "down")) {

	    if (argc == 3) {

		String outName;	// object name

		outName = argv[2];

		// set new object...zero out cached object pointer
		gbn->downName = outName;
		gbn->down = NULL;

		return (TCL_OK);
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}
	
	else if (!strcmp(argv[1], "netstat")) {

	    int i;
	    int count = 0;
	    extern Simulation *sim;
	    bool verbose = false;

	    if ((argc > 2) && (!strcmp(argv[2], "-v"))) {
		verbose = true;
	    }

	    cout << "Sequence number state\n";
	    cout << "port  ";
	    for (i = 0; i < gbn->ports; i++) {
		cout.width(6);
		cout << i;
	    }
	    cout << "\n";
	    
	    cout << "in    ";
	    for (i = 0; i < gbn->ports; i++) {
		cout.width(6);
		cout << gbn->incomingSeq[i];
	    }
	    cout << "\n";

	    cout << "out   ";
	    for (i = 0; i < gbn->ports; i++) {
		cout.width(6);
		cout << gbn->outgoingSeq[i];
	    }
	    cout << "\n";

	    cout << "queued";
	    for (i = 0; i < gbn->ports; i++) {
		cout.width(6);
		cout << gbn->retransmitQueue[i].length;
	    }
	    cout << "\n";

	    cout << "head  ";
	    for (i = 0; i < gbn->ports; i++) {

		if (gbn->retransmitQueue[i].length > 0) {
		    AtmCell *headCell = gbn->retransmitQueue[i][0];
		    cout.width(6);
		    cout << headCell->seq;
		}
		else {
		    cout << "     *";
		}
	    }
	    cout << "\n";

	    cout << "tail  ";
	    for (i = 0; i < gbn->ports; i++) {

		if (gbn->retransmitQueue[i].length > 0) {
		    AtmCell *headCell = gbn->retransmitQueue[i]
		      [gbn->retransmitQueue[i].length - 1];

		    cout.width(6);
		    cout << headCell->seq;
		}
		else {
		    cout << "     *";
		}
	    }
	    cout << "\n";

	    return(TCL_OK);

	}

	else {
	    return(ScheduledObjectCmd(clientData, interp, argc, argv));
	}
	
    }
}

//
// CellGoBackNDeleteProc
//
// (Tcl DeleteProc)
//
void CellGoBackNDeleteProc(ClientData clientData)
{
    delete ((CellGoBackN *) clientData);
}
