static char rcsid[] = "$Id: CellQueueRcsp.cc,v 1.28 1996-06-12 20:15:19-07 bmah Exp $";
//
// $Id: CellQueueRcsp.cc,v 1.28 1996-06-12 20:15:19-07 bmah Exp $
//
// CellQueueRcsp.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Rate-Controlled Static Priority scheduler for ATM cells
//

#include <Uniform.h>
#include "CellQueueRcsp.h"
#include "Event.h"
#include "inline.h"

char *CellQueueRcspModeName[] = {
    "work-conserving",
    "non-work-conserving",
    "no-rate-control",
};

// Configuration
// 
const int CellQueueRcspReservedBuffers = 96;
				// Number of buffers to allocate for every
				// reserved channel.  This is a voodoo
				// constant that needs to be configurable.

//
// CellQueueRcsp::CellQueueRcsp
//
// Input:
//
// Output:  None
//
// Object constructor
//
CellQueueRcsp::CellQueueRcsp(const String &name, Simulation *si, int mv,
			     int l, int s, double t) :
CellQueue(name, si),
maxVcid(mv),
numLevels(l),
numSubqueues(s),
tick(t)
{

    int i;			// straw loop counter
    Uniform skewGenerator(0, tick, si->rng); // skew random variable generator

    // Initialize global state
    mode = CellQueueRcspModeNonWorkConserving;
    busy = false;
    buffersAllocated = 0;
    buffersUsed = 0;
    offset = 0;
    cellsIn = 0;
    cellsBestEffort = 0;
    cellsRealTime = 0;
    cellsOut = 0;
    cellsNoMem = 0;
    cellsScheduler = 0;
    cellsEpd = 0;
    cellsNoVc = 0;
    intervalCells = 0;
    regulatorTop = -1;
    regulatorBottom = -1;

    // Allocate memory for per-VCID stuff
    level = new int[maxVcid];
    cellsBuffered = new int[maxVcid];
    cellsMax = new int[maxVcid];
    discarding = new bool[maxVcid];

    Xmin = new double[maxVcid];
    Xave = new double[maxVcid];
    I = new double[maxVcid];
    
    eligibleTime = new (double *)[maxVcid];
    for (i = 0; i < maxVcid; i++) {
	eligibleTime[i] = NULL;
    }
    eligibleTimeOffset = new int[maxVcid];
    eligibleTimeSize = new int[maxVcid];

    // Allocate queue pointers
    eligible = new AtmCellList[numLevels];
    queues = new (AtmCellList *)[numLevels];
    for (i = 0; i < numLevels; i++) {
	queues[i] = new AtmCellList[numSubqueues];
    }

    // Set up best-effort queues for signalling (perhaps these should
    // really be real-time?)
    for (i = 0; i < SigProtoReservedChannels; i++) {
	level[i] = CellQueueRcspLevelSig;
	cellsMax[i] = CellQueueRcspReservedBuffers;
	buffersAllocated += CellQueueRcspReservedBuffers;
	cellsBuffered[i] = 0;
	discarding[i] = false;
    }

    // Set up all "normal" data queues as best-effort.  On the off chance
    // that a cell might show up for a non-established channel, we don't
    // want it messing up real-time guarantees.
    for (i = SigProtoReservedChannels; i < maxVcid; i++) {
	level[i] = CellQueueRcspLevelBestEffort;
    }

    // Start up tick timer
    Post(skewGenerator(), this, EventTimer, NULL);

}


//
// CellQueueRcsp::~CellQueueRcsp
//
// Object destructor
//
CellQueueRcsp::~CellQueueRcsp()
{

    int i;
    
    delete [] level;
    delete [] cellsBuffered;
    delete [] cellsMax;
    delete [] discarding;
    delete [] Xmin;
    delete [] Xave;
    delete [] I;
    for (i = 0; i < maxVcid; i++) {
	if (eligibleTime[i]) {
	    delete [] eligibleTime[i];
	}
    }
    delete [] eligibleTime;
    delete [] eligibleTimeOffset;
    delete [] eligibleTimeSize;
    delete [] eligible;
    for (i = 0; i < numLevels; i++) {
	delete [] queues[i];
    }
    delete [] queues;

    *(sim->logStream) << name << " [CellQueueRcsp] cellsIn " << cellsIn << " cellsBestEffort " << cellsBestEffort << " cellsRealTime " << cellsRealTime << " cellsOut " << cellsOut << " cellsNoMem " << cellsNoMem << " cellsScheduler " << cellsScheduler << " cellsEpd " << cellsEpd << " cellsNoVc " << cellsNoVc << "\n";

}

//
// CellQueueRcsp::Handler
//
// (Event handler)
//
void
CellQueueRcsp::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    switch (type) {

	// Cell arrived
	case (EventCellArrival):
	{

	    AtmCell *cell = (AtmCell *) data;
	    int vcid = cell->vcid;

	    IF_DEBUG(2, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "got event " << type << "\n";
	    });
    
	    cellsIn++;

	    // First: if we're doing early packet discard, drop cells
	    // unless we've come to the end of a packet.
	    if (discarding[vcid]) {
		if (cell->type == AtmCellTypeEndOfPacket) {
		    discarding[vcid] = false;

		    IF_DEBUG(3, {
			DEBUG_PREFIX("CellQueueRcsp");
			*(sim->debugStream) << "end of packet and early packet discard\n";
		    });
		}
		else {
		    cellsEpd++;
		    goto dropcell;
		}
	    }
	    
	    // If there's no buffer space left, drop packet.  For
	    // non-EOP cells, we want to leave space for at least one
	    // cell left in the buffer, to hold an EOP cell.  This
	    // ensures that if we have to drop cells from a packet,
	    // we can always store the EOP cell.  We want to do this
	    // to mark the end of the current packet and reset
	    // the Aal entity for the next packet.
	    if (((cell->type == AtmCellTypeEndOfPacket) &&
		 (cellsBuffered[vcid] >= cellsMax[vcid])) ||
		((cell->type != AtmCellTypeEndOfPacket) &&
		 (cellsBuffered[vcid] >= cellsMax[vcid] - 1))) {
		cellsNoMem++;

		IF_DEBUG(1, {
		    DEBUG_PREFIX("CellQueueRcsp");
		    *(sim->debugStream) << "out of memory, dropping (level " << level[vcid] << ") " << cell << "\n";
		});

		goto droppacket;
	    }

	    // Check to see if it was a signalling cell.  If so, put it in
	    // the (higher-than-best-effort-priority) signalling queue.
	    if (level[vcid] == CellQueueRcspLevelSig) {

		cellsSig++;
		sigQueue.EnqueueTail(cell);
		cellsBuffered[vcid]++;
		buffersUsed++;

		IF_DEBUG(4, {
		    DEBUG_PREFIX("CellQueueRcsp");
		    *(sim->debugStream) << "sig cell " << vcid << "\n";
		});

		// Wake up the scheduler if it was idle.
		if (!busy) {
		    SelectAndTransmit(time);
		}

	    }

	    // Next see if this is a best-effort cell.  If so,
	    // bypass the regulator and enqueue it on the best-effort queue.
	    else if (level[vcid] == CellQueueRcspLevelBestEffort) {

		cellsBestEffort++;
		bestEffortQueue.EnqueueTail(cell);
		cellsBuffered[vcid]++;
		buffersUsed++;

		IF_DEBUG(4, {
		    DEBUG_PREFIX("CellQueueRcsp");
		    *(sim->debugStream) << "best-effort cell " << vcid << "\n";
		});

		// If the scheduler wasn't busy, it can be now...
		// because we just gave it work to do.
		if (!busy) {
		    SelectAndTransmit(time);
		}

	    }

	    // This is a real-time cell.

	    else {

		double et; 		// eligible time of this cell
		int subqueue;		// subqueue to queue cell at
		double etXave, etXmin;

		cellsRealTime++;

		// If not doing rate control, then put the cell directly
		// into the eligible queue.  Activate the scheduler if
		// it wasn't already doing work.
		if (mode == CellQueueRcspModeNoRateControl) {

		    eligible[level[vcid]].EnqueueTail(cell);
		    cellsBuffered[vcid]++;
		    buffersUsed++;

		    IF_DEBUG(4, {
			DEBUG_PREFIX("CellQueueRcsp");
			*(sim->debugStream) << "real-time cell eligible for transmission\n";
		    });

		    if (!busy) {
			SelectAndTransmit(time);
		    }
		    
		}
		else {

		    //
		    // Make sure we actually have an eligibleTime array
		    // for this VC.  The only times we wouldn't are when
		    // we've closed a VC but there are still cells for
		    // the VC in question floating around the network.
		    //
		    if (eligibleTime[vcid] == NULL) {

			IF_DEBUG(1, {
			    DEBUG_PREFIX("CellQueueRcsp");
			    *(sim->debugStream) << "dropped cell for non-existent vcid " << vcid << "\n";
			});

			cellsNoVc++;
			goto dropcell;
		    }

		    //
		    // Compute eligible time.
		    // First constraint: a cell can't become eligible before
		    // it has arrived.
		    //
		    et = time;

		    // Determine Xmin constrant.  A packet can't become
		    // eligible at less than Xmin spacing from the last
		    // packet become eligible.  The rather odd coding of
		    // the secondary index is an artifact of the fact that
		    // the result of the mod operation for a negative argument
		    // is implementation dependent.  So instead of backstepping
		    // (potentially off the low end of the array), we forward
		    // step (around the high end).
		    etXmin = eligibleTime[vcid][(eligibleTimeOffset[vcid] +
						 eligibleTimeSize[vcid] - 1)
						% (eligibleTimeSize[vcid])] +
						  Xmin[vcid];

		    // Determine Xave constraint.  A packet can't become eligible
		    // at less than I spacing from the last ceil(I/Xave) packet.
		    // The secondary index into eligibleTime here isn't a bug:
		    // since the array is sized to ceil(I/Xave),
		    // and ceil(I/Xave) % ceil(I/Xave) == 0, we can just use
		    // the current offset into the array.
		    etXave = eligibleTime[vcid][eligibleTimeOffset[vcid]] +
		      I[vcid];

		    // Now take the maximum of all these constraints.
		    if (et < etXmin) {
			et = etXmin;
		    }
		    if (et < etXave) {
			et = etXave;
		    }

		    // Store the eligibility time for later reference
		    eligibleTime[vcid][eligibleTimeOffset[vcid]] = et;
		    eligibleTimeOffset[vcid] = (eligibleTimeOffset[vcid] + 1) %
		      eligibleTimeSize[vcid];

		    IF_DEBUG(4, {
			DEBUG_PREFIX("CellQueueRcsp");
			*(sim->debugStream) << "real-time cell vcid " << vcid << " eligible at " << et << " nextTick " << nextTick << "\n";
		    });

		    // Queue the cell in the eligible queue if it's already
		    // eligible, start transmitting if scheduler idle.
		    if (et < nextTick) {
			eligible[level[vcid]].EnqueueTail(cell);
			cellsBuffered[vcid]++;
			buffersUsed++;

			IF_DEBUG(4, {
			    DEBUG_PREFIX("CellQueueRcsp");
			    *(sim->debugStream) << "real-time cell eligible for transmission\n";
			});

			if (!busy) {
			    SelectAndTransmit(time);
			}
		    }
		    else {
			subqueue = (int) ((et - nextTick) / tick);

			IF_DEBUG(4, {
			    DEBUG_PREFIX("CellQueueRcsp");
			    *(sim->debugStream) << "real-time cell queued in subqueue +" << subqueue << ", level " << level[vcid] << "\n";
			});

			// It's possible that the cell arrived so far ahead of
			// time, we can't even hold it in the regulator because
			// it doesn't enough subqueues.  This really shouldn't
			// happen, but we can just drop the cell and let the
			// simulation continue.
			if (subqueue >= numSubqueues) {
			    IF_DEBUG(1, {
				DEBUG_PREFIX("CellQueueRcsp");
				*(sim->debugStream) << "need to hold cell too long, dropping (level " << level[vcid] << " subqueue " << subqueue << ") " << cell << "\n";
			    });
			    cellsScheduler++;
			    goto droppacket;
			    
			}

			// Update bounds in regulator.
			if ((regulatorTop < 0) || (regulatorTop > subqueue)) {
			    regulatorTop = subqueue;
			}
			if ((regulatorBottom < 0) || (regulatorBottom < subqueue)) {
			    regulatorBottom = subqueue;
			}
			
			IF_DEBUG(4, {
			    DEBUG_PREFIX("CellQueueRcsp");
			    *(sim->debugStream) << "top " << regulatorTop << " bottom " << regulatorBottom << "\n";
			});

			subqueue = (subqueue + offset) % numSubqueues;

			// Having picked the right subqueue and all, put the
			// cell in its place.  If we're running a
			// work-conserving scheduler, then we might have
			// a cell to transmit.
			(queues[level[vcid]][subqueue]).EnqueueTail(cell);
			cellsBuffered[vcid]++;
			buffersUsed++;
			if ((mode == CellQueueRcspModeWorkConserving) &&
			    (!busy)) {
			    SelectAndTransmit(time);
			}
		    }
		}		    
	    }

	    return;

droppacket:	    
            // Turn on EPD iff the cell didn't come from one of the
            // "Reserved" VCs.
            if (vcid >= SigProtoReservedChannels) {
		discarding[vcid] = true;
	    }
dropcell:	    
	    delete cell;

	    IF_DEBUG(3, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "drop cell\n";
	    });

	    return;

	}
	break;

	// Transmit complete
	case (EventCellTransmitComplete):
	{

	    IF_DEBUG(2, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "got event " << type << "\n";
	    });

	    busy = false;
	    SelectAndTransmit(time);
	}
	break;

	// Timer event.
	case (EventTimer):
	{

	    int i;		// Acme Loop Counters, Inc.

	    IF_DEBUG(5, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "got event " << type << "\n";
	    });

	    // Make new subqueues eligible for scheduling
	    for (i = 0; i < numLevels; i++) {
		(queues[i][offset]).Append(eligible[i]);
	    }

	    // Update regulator subqueue bounds
	    if (regulatorTop > 0) {
		regulatorTop--;
	    }
	    if (regulatorBottom > 0) {
		regulatorBottom--;
	    }

	    IF_DEBUG(5, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "tick top " << regulatorTop << " bottom " << regulatorBottom << "\n";
	    });

	    // Increment offset
	    offset = (offset + 1) % numSubqueues;

	    // If we weren't busy, and there are now cells eligible
	    // for transmission, wake up the scheduler.
	    if (!busy) {
		SelectAndTransmit(time);
	    }
	    
	    // Wake us up after tickTime time.
	    nextTick = time + tick;
	    Post(nextTick, this, EventTimer, NULL);

	}
	break;

	case (EventTimer2):
	{

	    // Print out statistics over this interval
	    IF_DEBUG(0, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "interval " << interval << " cells " << intervalCells << " userbytes " << intervalCells * AtmPayloadSize << " utilization " << (intervalCells * AtmCellSize) / (interval * bitrate) << " buffers-allocated " << buffersAllocated << " buffers-used " << buffersUsed << "\n";
	    });

	    // Reset counters
	    intervalCells = 0;

	    // Wake us up again later
	    Post(time + interval, this, EventTimer2, NULL);

	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "bad EventType " << type << "\n";
	    });
	}
	break;

    }

}

//
// CellQueueRcsp::SelectAndTransmit
//
// Input:  Current time (time)
//
// Output:  None
//
// If the scheduler isn't busy (busy == false), then
// select a cell to transmit from the eligible queue and transmit
// it.  Set the busy flag appropriately.
//
void
CellQueueRcsp::SelectAndTransmit(double time) {

    int i;
    AtmCell *cell;		// pointer to cell
    int level;			// priority level to transmit

    IF_DEBUG(5, {
	DEBUG_PREFIX("CellQueueRcsp");
	*(sim->debugStream) << "SelectAndTransmit()\n";
    });
    
    if (!busy) {

	// First try to find an eligible real-time cell.
	for (i = 0; i < numLevels; i++) {
	    cell = eligible[i].Dequeue();
	    if (cell) {
		level = i;
		break;
	    }
	}

	// If we couldn't find an eligible real-time cell, send a signalling cell
	// if one exists.
	if (!cell) {
	    cell = sigQueue.Dequeue();
	    level = CellQueueRcspLevelSig;
	}

	// If weren't any signalling cells, look for a best-effort cell.
	if (!cell) {
	    cell = bestEffortQueue.Dequeue();
	    level = CellQueueRcspLevelBestEffort;
	}

	// If we're operating in work-conserving mode and we haven't found
	// anything yet, then try to find a real-time cell that's suitable,
	// even if it's not eligible.
	if ((!cell) && (mode == CellQueueRcspModeWorkConserving)) {
	    int s, l;

	    // We really should optimize for the case when there are no
	    // cells queued.  To do this we try to keep track of the part of
	    // the regulator that actually holds cells.  We cache what
	    // information we know about it this range and use it to set new
	    // bounds for scanning the regulator for the next cell.

	    if ((regulatorTop >= 0) && (regulatorBottom >= 0)) {

		IF_DEBUG(5, {
		    DEBUG_PREFIX("CellQueueRcsp");
		    *(sim->debugStream) << "scan " << regulatorTop << " --> " << regulatorBottom << "\n";
		});

		// Scan through the queues looking for the next real-time cell
		// that would get to be transmitted.
		for (s = regulatorTop; s <= regulatorBottom; s++) {
		    for (l = 0; l < numLevels; l++) {
			cell = queues[l][(s + offset) % numSubqueues].Dequeue();
			if (cell) {
			    level = l;
			    regulatorTop = s; // Since we got up through s, we
			    // know that no small subqueues
			    // have cells.

			    IF_DEBUG(3, {
				DEBUG_PREFIX("CellQueueRcsp");
				*(sim->debugStream) << "Found cell at subqueue +" << s << "\n";
			    });

			    break;
			}
		    }
		    if (cell) {
			break;
		    }
		}

		if (!cell) {
		    IF_DEBUG(3, {
			DEBUG_PREFIX("CellQueueRcsp");
			*(sim->debugStream) << "No cells in regulator, setting empty flag\n";
		    });
		    regulatorTop = -1;
		    regulatorBottom = -1;
		}
		else {
		    IF_DEBUG(3, {
			DEBUG_PREFIX("CellQueueRcsp");
			*(sim->debugStream) << "Sending cell in regulator level " << l << " subqueue +" << s << "\n";
		    });
		}
	    }
	    else {
		IF_DEBUG(5, {
		    DEBUG_PREFIX("CellQueueRcsp");
		    *(sim->debugStream) << "No cells in regulator (cached)\n";
		});
	    }
	    
	}

	// If we have a cell to send, do it.
	if (cell) {
	    busy = true;

	    // Send the cell
	    Lookup(sinkObject, sink);
	    Post(time + latency + AtmCellSize / bitrate, sinkObject,
		 EventCellArrival, cell);
	    cellsBuffered[cell->vcid]--;
	    buffersUsed--;
	    cellsOut++;
	    intervalCells++;

	    // Wake us up after some suitable time
	    Post(time + AtmCellSize / bitrate, this,
		 EventCellTransmitComplete, NULL);

	    IF_DEBUG(3, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "sent cell at level " << level << " vcid " << cell->vcid << "\n";
	    });

	}
	else {
	    busy = false;
	}
    }
}

//
// CellQueueRcsp::NewVc
//
// Input:  Current simulation time (time), vcid (vcid), priority level
// (newLevel), traffic parameters (newXmin, newXave, newI), buffer
// requirements (newBuffers).
//
// Output:  None
//
// Set up parameters for a new VC.
//
void
CellQueueRcsp::NewVc(double time, int vcid, int &newLevel, double &newXmin,
		     double &newXave, double &newI, int &newBuffers)
{

    int i;
    
    IF_DEBUG(2, {
	DEBUG_PREFIX("CellQueueRcsp");
	*(sim->debugStream) << "NewVc()\n";
    });
    
    // We can't (yet) modify an existing VC.
    if (eligibleTime[vcid] != NULL) {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("CellQueueRcsp");
	    *(sim->debugStream) << "attempted to NewVc but VC was in use\n";
	});
	return;
    }

    // Store parameters
    level[vcid] = newLevel;
    I[vcid] = newI;
    Xmin[vcid] = newXmin;
    Xave[vcid] = newXave;

    // For real-time channels, set up rate control parameters and state
    if (newLevel >= 0) {

	// Compute size of eligible times array.  This is the number of cells
	// that can be transmitted in I time, as dictated by I and Xave.
	eligibleTimeSize[vcid] = (int) floor(newI / newXave);
	
	// Allocate and initialize eligibleTime array.  We don't want the
	// contents of this array to affect the rate control until at least
	// floor(I / Xave) cells have been received.  We initialize them
	// to an arbitrarily small value so they won't affect the max
	// function used when computing eligibility times.  It turns out that
	// time - I is good enough.
	eligibleTime[vcid] = new double[eligibleTimeSize[vcid]];
	for (i = 0; i < eligibleTimeSize[vcid]; i++) {
	    eligibleTime[vcid][i] = time - newI;
	}

	// Initialize offset (actually, by symmetry, it doesn't really matter
	// what we initialize the offset to).
	eligibleTimeOffset[vcid] = 0;

    }
    
    // Initialize buffering
    cellsMax[vcid] = newBuffers;
    buffersAllocated += cellsMax[vcid];
    cellsBuffered[vcid] = 0;
    discarding[vcid] = false;

}

//
// CellQueueRcsp::DeleteVc
//
// Input:  time (time), VCID to delete (vcid)
//
// Output:  None
//
// Delete a channel through a queue object, releasing all of the
// memory, etc. associated with it.
//
void CellQueueRcsp::DeleteVc(double time, int vcid)
{
    
    IF_DEBUG(2, {
	DEBUG_PREFIX("CellQueueRcsp");
	*(sim->debugStream) << "DeleteVc()\n";
    });
    
    // If we need to delete a real-time channel, there's some memory
    // to deallocate for it.  Do so.
    if (level[vcid] >= 0) {

	// Delete the eligible time array.
	if (eligibleTime[vcid]) {
	    delete [] eligibleTime[vcid];
	    eligibleTime[vcid] = NULL;
	}
	else {
	    IF_DEBUG(0, {
		DEBUG_PREFIX("CellQueueRcsp");
		*(sim->debugStream) << "tried to delete real-time channel " << vcid << " at level " << level[vcid] << " without eligible time array\n";
	    });
	}

    }

    // Note that there may still be cells left in the queue when this
    // procedure gets called.  If we were really pedantic, we would
    // go through the queues at the appropriate level and nuke all
    // of the cells for this VC that just got deleted.  We'll see if
    // we can get away with not doing this for right now...instead
    // we'll demote this channel to best-effort.
    level[vcid] = CellQueueRcspLevelBestEffort;
    buffersAllocated -= cellsMax[vcid];
    cellsMax[vcid] = 0;
    
}

//
// CellQueueRcsp::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
CellQueueRcsp::Output()
{

    String out;
    char scratch[80];

    out = name + " [CellQueueRcsp] ";

    sprintf(scratch, "mode %s ", CellQueueRcspModeName[mode]);
    out += scratch;

    sprintf(scratch, "subqueues %d ", numSubqueues);
    out += scratch;

    sprintf(scratch, "levels %d ", numLevels);
    out += scratch;

    sprintf(scratch, "tick %f ", tick);
    out += scratch;

    sprintf(scratch, "buffersAllocated %d ", buffersAllocated);
    out += scratch;

    out += Output2();

    return out;
    
}

//
// CellQueueRcspCmd
//
// (Tcl command procedure)
//
int
CellQueueRcspCmd(ClientData clientData, Tcl_Interp *interp,
		 int argc, char *argv[]) {

    if (argc != 6) {
	Tcl_SetResult(interp,
		      "usage:  CellQueueRcsp name maxvcid levels subqueues ticktime",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	char *nm;		// name of object
	int mv;			// maximum VCID
	int l;			// number of levels
	int s;			// number of subqueues
	double t;		// tick time
	CellQueueRcsp *cqr;	// new object
	extern Simulation *sim;	// simulation global

	// get arguments
	nm = argv[1];

	rc = Tcl_GetInt(interp, argv[2], &mv); // maxvcid
	if (rc == TCL_OK) {

	    rc = Tcl_GetInt(interp, argv[3], &l); // levels
	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[4], &s); // subqueues
		if (rc == TCL_OK) {

		    rc = Tcl_GetDouble(interp, argv[5], &t); // ticktime
		    if (rc == TCL_OK) {

			// Make new object and make a Tcl command corresponding
			// to the object.
			cqr = new CellQueueRcsp(nm, sim, mv, l, s, t);
			Tcl_CreateCommand(interp, nm,
					  CellQueueRcspObjectCmd,
					  (ClientData) cqr,
					  CellQueueRcspDeleteProc);

		    }
		}
	    }
	}

	return rc;
    }

}

//
// CellQueueRcspObjectCmd
//
// (Tcl object command procedure)
//
int
CellQueueRcspObjectCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    CellQueueRcsp *queue = (CellQueueRcsp *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }
    else {

	// cq mode mode
	if (!strcmp(argv[1], "mode")) {

	    if (argc == 3) {
		if (!strcmp(argv[2], "work-conserving")) {
		    queue->mode = CellQueueRcspModeWorkConserving;
		    return(TCL_OK);
		}
		else if (!strcmp(argv[2], "non-work-conserving")) {
		    queue->mode = CellQueueRcspModeNonWorkConserving;
		    return(TCL_OK);
		}
		else if (!strcmp(argv[2], "no-rate-control")) {
		    queue->mode = CellQueueRcspModeNoRateControl;
		    return(TCL_OK);
		}
		else {
		    Tcl_SetResult(interp, "bad mode", TCL_STATIC);
		    return(TCL_ERROR);
		}
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	//
	// cq statinterval time
	//
	else if (!strcmp(argv[1], "statinterval")) {

	    if (argc == 3) {
		double inter;      // statistics printing interval
		int rc;		   // Tcl return code
		
		rc = Tcl_GetDouble(interp, argv[2], &inter);
		if (rc == TCL_OK) {
		    
		    queue->interval = inter;
		    queue->Post(0.0 + inter, queue, EventTimer2, NULL);

		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// cq netstat
	//
	else if (!strcmp(argv[1], "netstat")) {

	    int i;
	    cout << "vcid level max current discarding\n";
	    for (i = 0; i < queue->maxVcid; i++) {
		if (queue->cellsMax[i] > 0) {
		    cout << i << " " <<
		      queue->level[i] << " " <<
		      queue->cellsMax[i] << " " <<
		      queue->cellsBuffered[i] << " " <<
		      queue->discarding[i] << "\n";
		}
	    }

	    return(TCL_OK);
	}

	else {
	    return CellQueueObjectCmd(clientData, interp, argc, argv);
	}
    }
}

//
// CellQueueRcspDeleteProc
//
// (Tcl delete procedure)
//
void
CellQueueRcspDeleteProc(ClientData clientData) {
    delete (CellQueueRcsp *) clientData;
}

//
// operator<<
//
// Input:  Output stream (s), queueing mode (mode)
//
// Output:  Output stream
//
// Print the name of the mode (mode) on the output stream s.
//
ostream& operator<<(ostream &s, CellQueueRcspMode mode)
{
    char *output;
    
    output = CellQueueRcspModeName[mode];

    return s << output;

}
