// -*- c++ -*-
// $Id: CellQueueRcsp.h,v 1.13 1996-05-29 14:34:48-07 bmah Exp $
//
// CellQueueRcsp.h
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//

#ifndef CELLQUEUERCSP_H
#define CELLQUEUERCSP_H

#include <iostream.h>
#include <String.h>
#include "Calendar.h"
#include "ScheduledObject.h"
#include "AtmCell.h"
#include "AtmCellList.h"
#include "Tcl.h"
#include "CellQueue.h"
#include "SigRcsp.h"
#include "SigHostRcsp.h"

enum CellQueueRcspMode {
    CellQueueRcspModeWorkConserving,
    CellQueueRcspModeNonWorkConserving,
    CellQueueRcspModeNoRateControl,
};

// Special RCSP queue levels.  These values must be outside the normal
// values for queue levels (which are, in this implementation numbered
// starting from 0).
//
// The best-effort queue is the lowest priority.  The next higher queue
// is the signalling queue, which handles all the signalling traffic.
// The real-time queues are prioritized above the signalling queue.
const int CellQueueRcspLevelBestEffort = -1;
const int CellQueueRcspLevelSig = -2;

class SigMessage;

class CellQueueRcsp : public CellQueue {

    friend int
    CellQueueRcspCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[]);
    friend int
    CellQueueRcspObjectCmd(ClientData clientData,
			   Tcl_Interp *interp,
			   int argc, char *argv[]);
			    
    friend SigReturnCode
    SigRcsp::AdmissionControl(double time, int link, int vcid,
			      SigMessage *mesg, void *traffic);

    friend void
    SigRcsp::CommitResources(double time, int link, int vcid, void *traffic);

    friend void
    SigRcsp::ReleaseResources(double time, int link, int vcid);

    friend SigReturnCode
    SigHostRcsp::AdmissionControl(double time, int vcid, void *traffic);

    friend void
    SigHostRcsp::CommitResources(double time, int vcid, void *traffic);

    friend void
    SigHostRcsp::ReleaseResources(double time, int vcid);

    public:

    CellQueueRcsp(const String &name, Simulation *si, int mv,
		  int l, int s, double t);
    virtual ~CellQueueRcsp();

    virtual String Output();
			     
    virtual void Handler(Event *e);

    // Global state
    const int maxVcid;		// maximum number of VCIDs supported
    const int numLevels;	// number of levels
    const int numSubqueues;	// number of subqueues
    const double tick;		// tick interval

    protected:

    // Global state
    CellQueueRcspMode mode;     // operating mode

    int offset;			// subqueue offset
    int buffersAllocated;	// number of buffers allocated in queue module
    int buffersUsed;		// nubmer of buffers used in queue module

    int cellsIn;		// number of input cells
    int cellsSig;		// number of signalling cells input
    int cellsBestEffort;	// number of best effort cells input
    int cellsRealTime;		// number of real time cells input
    int cellsOut;		// number of output cells (transmitted)
    int cellsNoMem;		// cells dropped due to lack of memory
    int cellsScheduler;		// cells dropped due to scheduling constraints
    int cellsEpd;		// cells dropped due to early packet drop
    int cellsNoVc;		// cells received for non-existent VC

    double interval;		// recording interval length
    int intervalCells;		// number of cells in recording interval

    // These members bound the range of the regulator that might actually
    // contain cells (relative to the offset which represents the front
    // of the regulator).  We use this to bound the number of subqueues we
    // need to scan when running in work-conserving mode.
    // If these bounds are -1, the regulator is empty.
    int regulatorTop;		// top of regulator cells, -1 if empty
    int regulatorBottom;	// bottom of regulator cells, -1 if empty

    // Per-scheduler state.  The regulator is implemented as a 2-D array.
    // For each priority level, there is a set of subqueues (implemented
    // as a linked list of ATM cells).  For every tick, one subqueue per
    // level is appended to the eligible queue for that level.  We keep
    // track of (numSubqueues) number of subqueues...for implementation
    // efficiency, we rotate around the array.  The variable (offset)
    // tells us which of the subqueues is "current"...insertions into
    // subqueues are made relative to this offset.
    bool busy;			// scheduler currently busy flag
    double nextTick;		// next time for scheduler tick
    AtmCellList **queues;	// holding queues (implement regulators)
    AtmCellList *eligible;	// cells actively scheduled for transmission
    AtmCellList sigQueue;	// signalling queue
    AtmCellList bestEffortQueue; // best effort queue

    // Per-VCID state
    int *level;			// level to queue cells for each VCID,
				// -1 == best effort
    int *cellsBuffered;		// number of cells buffered each VCID
    int *cellsMax;		// max number of cells buffered each VCID
    double *Xmin, *Xave, *I;	// rate control parameters

    // For each VCID we keep track of the last ceil(I/Xave) eligible
    // times.  We rotate through this array, once per packet transmission;
    // eligibleTimeOffset keeps this state for each virtual circuit.
    double **eligibleTime;	// array of eligible times for each VC
    int *eligibleTimeOffset;	// offset into eligible times array
    int *eligibleTimeSize;	// size of individual eligible times arrays

    bool *discarding;		// array of flags for early packet discard

    void SelectAndTransmit(double time);
    void NewVc(double time, int vcid, int &newLevel,
	       double &newXmin, double &newXave, double &newI,
	       int &newBuffers);
    void DeleteVc(double time, int vcid);
};

extern int CellQueueRcspCmd(ClientData clientData, Tcl_Interp *interp,
			    int argc, char *argv[]);
extern int CellQueueRcspObjectCmd(ClientData clientData, Tcl_Interp *interp,
				  int argc, char *argv[]);
extern void CellQueueRcspDeleteProc(ClientData clientData);

extern ostream& operator<<(ostream &s, CellQueueRcspMode mode);

#endif // CELLQUEUERCSP_H
