// -*- c++ -*-
// $Id: EventList.h,v 1.2 1996-01-31 16:44:38-08 bmah Exp $
//   based on:
// $ID: TcpDataList.h,v 1.1 1995-10-19 14:22:33-07 bmah Exp $
//   based on:
// $ID: AtmCellList.h,v 1.3 1995-09-12 10:15:48-07 bmah Exp $
//
// EventList.h
//
// Copyright (c) 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Maintain a DLL of Events with head and tail pointers
//

#ifndef EVENTLIST_H
#define EVENTLIST_H

#include <stdlib.h>
#include <stdio.h>
#include "Event.h"

class EventList {

    public:

    //
    // Constructor initializes the pointers to the ends of the lists
    //
    EventList() {
	head = NULL;
	tail = NULL;
	length = 0;
    }

    //
    // Destructor deallocates all of the cells that were in the list.
    //
    ~EventList() {
	Event *current, *next;
	for (current = head; current != NULL; current = next) {
	    next = current->ev_next;
	    delete current;
	}
    }

    // Enqueue and dequeue member functions do the expected things
    EnqueueTail(Event *data) {
	data->ev_next = NULL;
	data->ev_prev = tail;
	if (length > 0) {
	    tail->ev_next = data;
	}
	else {
	    head = data;
	}
	tail = data;
	length++;

	Sanity(*this);

    }

    EnqueueHead(Event *data) {
	data->ev_next = head;
	data->ev_prev = NULL;
	if (length > 0) {
	    head->ev_prev = data;
	}
	else {
	    tail = data;
	}
	head = data;
	length++;

	Sanity(*this);

    }

    Event *Dequeue() {
	Event *data;

	data = head;
	if (length > 0) {
	    head = head->ev_next;
	    length--;
	    if (length > 0) {
		head->ev_prev = NULL;
	    }
	    else {
		tail = NULL;
	    }
	    length--;
	}

	Sanity(*this);

	return data;
    }

    // Append this list to another.  Current list is cleared.
    void AppendTo(EventList &other) {
	if (other.length == 0) {
	    other.head = head;
	}
	else {
	    other.tail->ev_next = head;
	    if (length > 0) {
		head->ev_prev = other.tail;
	    }
	}
	if (length > 0) {
	    other.tail = tail;
	}
	other.length += length;

	Sanity(other);

	head = NULL;
	tail = NULL;
	length = 0;

	Sanity(*this);

    }

    // DequeueData:  Search the list for an Event with a particular
    // type field.  If no matching type field can be found, return NULL.
    Event *DequeueType(EventType type) {
	Event *current;

	// Find the first event in the list that matches the data field.
	for (current = head;
	     current != NULL;
	     current = current->ev_next) {
	    if (current->ev_type == type) {
		Remove(current);
		break;
	    }
	}

	Sanity(*this);

	return current;

    }

    // DequeueData:  Search the list for an Event with a particular
    // data field.  If no matching data field can be found, return NULL.
    Event *DequeueData(void *data) {
	Event *current;

	// Find the first event in the list that matches the data field.
	for (current = head;
	     current != NULL;
	     current = current->ev_next) {
	    if (current->ev_data == data) {
		Remove(current);
		break;
	    }
	}

	Sanity(*this);

	return current;

    }

    // DequeueTcpArrival:  Dequeue a particular TCP arrival (necessary)
    // because of the TCP Pcb pointer for a TcpArrival is buried inside
    // the TcpCall call block.
    Event *DequeueTcpArrival(TcpPcb *tp) {
	Event *current;

	// Find the first event in the list that matches the data field.
	for (current = head;
	     current != NULL;
	     current = current->ev_next) {
	    if (current->ev_type == EventTcpArrival) {
		TcpCall *tc = (TcpCall *) current->ev_data;
		if (tc->pcb == tp) {
		    Remove(current);
		    break;
		}
	    }
	}

	Sanity(*this);

	return current;
    }

    unsigned int length;

    protected:
    Event *head, *tail;

    // Remove takes an element out of the middle of the list.  The
    // elements around the one in question have their pointers adjusted.
    Event *Remove(Event *ev) {

	if (ev->ev_prev) {
	    // Patch up previous event
	    ev->ev_prev->ev_next = ev->ev_next;
	}
	else {
	    // No previous event (first element) set head
	    head = ev->ev_next;
	}
	
	if (ev->ev_next) {
	    // Patch up next event
	    ev->ev_next->ev_prev = ev->ev_prev;
	}
	else {
	    // No next event (last element) set tail
	    tail = ev->ev_prev;
	}
	length--;
	
	Sanity(*this);

	return ev;

    }

    // Sanity check on lists
    void Sanity(EventList &list) {
	if (list.length < 0) { 
	    cerr << "EventList:  negative length!\n";
//	    abort();
	}
	else if (list.length == 0) {
	    if ((list.head != NULL) || (list.tail != NULL)) {
		cerr << "EventList:  zero length but non-null pointer!\n";
//		abort();
	    }
	}
	else if (list.length > 0) {
	    if ((list.head == NULL) || (list.tail == NULL)) {
		cerr << "EventList:  non-zero length but null pointer!\n";
//		abort();
	    }
	}
    }

};

#endif // EVENTLIST_H
