static char rcsid[] = "$Id: FtpClient.cc,v 1.13 1996-03-08 09:02:32-08 bmah Exp $";
//
// $Id: FtpClient.cc,v 1.13 1996-03-08 09:02:32-08 bmah Exp $
//
// FtpClient.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// FTP client process class.
//

#include "FtpClient.h"
#include "EmpiricalIntegral.h"
#include "IpPacket.h"
#include "Event.h"
#include "inline.h"

//
// State names
//
char *FtpClientStateName[] = {
    "FtpClientStateStart",
    "FtpClientStateAwaitControl",
    "FtpClientStateAwaitPortResponse",
    "FtpClientStateDataTransfer",
    "FtpClientStateDoneData",
    "FtpClientStateDone"
};

// Global constants
static double gapItem = 0.001;	// 1 ms gap between items
static double epsilon = 0.000001; // 1 us interval between other events

//
// Imported probability distributions
//
extern RNG *rng;		// global random number generator

// Number of FTP items to transfer
extern const DoubleDistElement *FtpConnectionsDistTable;
extern int FtpConnectionsDistLength;

// Length of FTP request (on control connection)
extern const DoubleDistElement *FtpRequestLengthDistTable;
extern int FtpRequestLengthDistLength;

//
// FtpClient::Handler
//
// (Event handler)
//
void FtpClient::Handler(Event *e)
{
    
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalIntegral FtpConnectionsDist(FtpConnectionsDistLength,
						FtpConnectionsDistTable,
						rng);
    static EmpiricalIntegral FtpRequestLengthDist(FtpRequestLengthDistLength,
						  FtpRequestLengthDistTable,
						  rng);

    IF_DEBUG(2, {
	DEBUG_PREFIX("FtpClient");
	*(sim->debugStream) << "received " << type << " in state " << state << "\n";
    });

    // Note:  Because we're using our own private state space,
    // we can never use the base class Handler method for anything.
    switch (state) {

	//
	// FtpClientStateStart:  First (uninitialized state).
	//
	case (FtpClientStateStart):
	{
	    switch (type) {

		// Timer event:  Make control connection
		case (EventTimer):
		{
		    tpControl =
		      tcpModule->UserConnect(time + epsilon, server,
					     port, this);
		    state = FtpClientStateConnectingControl;
		    break;
		}
		break;

		default:
		goto bad;
	    }
	}
	break;

	//
	// FtpClientStateConnectingControl:  Waiting for connect()
	// to complete.  We can either get EventTcpConnect or EventTcpError.
	//
	case (FtpClientStateConnectingControl):
	{

	    switch (type) {
		
		// Establishment successful.  Bind to a data port and send a
		// request on the control connection.
		case (EventTcpConnect):
		{

		    // How many items to transfer?
//		    numItems = ftp_nitems();
		    numItems = (int) FtpConnectionsDist();
		    totalItems = numItems;

		    IF_DEBUG(2, {
			DEBUG_PREFIX("FtpClient");
			*(sim->debugStream) << " requesting " << numItems << " items\n";
		    });

sendport:
		    if (numItems) {

			// Bind a data socket to an ephemeral port.
			tpDataListen = tcpModule->UserListen(time + epsilon,
							     0, this);

			// Make a data tag to send the port number
			TcpData *portTag = new TcpData;
			portTag->data_int = tpDataListen->tcp_lport;

			IF_DEBUG(2, {
			    DEBUG_PREFIX("FtpClient");
			    *(sim->debugStream) << " sending PORT for port " << portTag->data_int << "\n";
			});

			// Send a control message on the control connection.
			// A data tag gives the "payload" which is the
			// information
			// contained in a FTP PORT command (basically, the port
			// number to connect to in order to send data back.
//			int controlLength = ftp_ctlsize();
			int controlLength = (int) FtpRequestLengthDist();
			tcpModule->UserSend(time + (2 * epsilon), tpControl,
					    controlLength, portTag);

			state = FtpClientStateAwaitPortResponse;
			transferStartTime = time;
			transferBytes = 0;
		    }
		    else {
			tcpModule->UserClose(time + epsilon, tpControl);
			endTime = time;
			state = FtpClientStateDone;
			delete this;
		    }
		}
		break;

		// Establishment failed.  Go away.
		case (EventTcpClosed):
		{
		    state = FtpClientStateDone;

		    IF_DEBUG(0, {
			DEBUG_PREFIX("FtpClient");
			*(sim->debugStream) << "connection failed\n";
		    });

		    delete this;
		}
		break;

		default:
		goto bad;
	    }	    
	}
	break;
	
	//
	// FtpClientStateAwaitPortResponse:  Waiting for the remote FTP
	// daemon to send a response to the port command.
	//
	case (FtpClientStateAwaitPortResponse):
	{
	    switch (type) {

		// We could get an arrival on the control connection
		// (this is the reponse to the PORT command.  Right now
		// we just throw it away.
		case (EventTcpArrival):
		{
		    TcpCall *tc = (TcpCall *) data;
		    
		    if (tc->pcb != tpControl) {
			IF_DEBUG(0, {
			    DEBUG_PREFIX("FtpClient");
			    *(sim->debugStream) <<
			      "received data on unknown stream\n";
			});
		    }
		    delete tc;
		};
		break;

		// We could get notification of a new TCP connection
		// (presumably to the data port). 
		case (EventTcpConnect):
		{
		    tpData = (TcpPcb *) data;
		    state = FtpClientStateDataTransfer;
		}
		break;

		default:
		goto bad;

	    }
	}
	break;

	//
	// FtpClientStateDataTransfer:  Receiving data from remote FTP
	// server.
	//
	case (FtpClientStateDataTransfer):
	{
		    
	    switch (type) {

		// Data received.  Should be on the FTP data connection
		// but could potentially be a control packet too.
		case (EventTcpArrival):
		{
		    TcpCall *tc = (TcpCall *) data;
		    if (tc->pcb == tpData) {
			transferBytes += tc->th_ulen;
			totalBytes += tc->th_ulen;
		    }
		    else if (tc->pcb == tpControl) {
			// ignore control packets sent
		    }
		    else {
			IF_DEBUG(0, {
			    DEBUG_PREFIX("FtpClient");
			    *(sim->debugStream) <<
			      "received data on unknown stream\n";
			});
		    };
		    delete tc;
		};
		break;

		// TCP connection closed.
		case (EventTcpClosed):
		{

		    // Complete close.
		    tcpModule->UserClose(time + epsilon, tpData);
		    tcpModule->UserClose(time + (2 * epsilon), tpDataListen);
		    tpData = NULL;
		    tpDataListen = NULL;
		    state = FtpClientStateDoneData;

		    Post(time + gapItem, this, EventTimer, NULL);

		    // log end of transfer
		    *(sim->logStream) << name << " [FtpClient] file time " << time - transferStartTime << " bytes " << transferBytes << "\n";
		};
		break;

		default:
		goto bad;

	    }
	}
	break;

	//
	// FtpClientStateDoneData:  Ready to select next item or quit
	//
	case (FtpClientStateDoneData):
	{

	    switch (type) {
	    
		// Should get a timer event at this point...
		case (EventTimer):
		{
		    // If we still have items left to send, then prepare to
		    // do the next one.
		    if (--numItems) {
			goto sendport;
		    }
		    
		    // All gone!
		    else {
			
			// Close control connection
			tcpModule->UserClose(time + epsilon, tpControl);
			
			endTime = time;
			state = FtpClientStateDone;
			delete this;
		    }
		}
		break;

		default:
		goto bad;
	    }
	}
	break;

	case (FtpClientStateDone):
	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX("FtpClient");
	    *(sim->debugStream) << "bad FtpClientState\n";
	});
	break;
	
    }
    return;

bad:
    IF_DEBUG(0, {
	DEBUG_PREFIX("FtpClient");
	*(sim->debugStream) << "bad EventType " << type << " in FtpClientState " << state << "\n";
    });
    return;

}

//
// FtpClient::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String FtpClient::Output()
{
    char c[30];
    sprintf(c, "%d", port);

    return name + " [FtpClient] tcp " + tcpModuleName + " port " + c +
      " server " + InetNtoA(server);
}

//
// FtpClientCmd
//
// (Tcl command procedure)
//
int FtpClientCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 6) {
	Tcl_SetResult(interp, "usage:  FtpClient name start tcp server port", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	FtpClient *fc;		// new client object
	char *nm;		// name of new client object
	double startTime;	// starting time to run
	String tcpObjectName;	// name of Tcp object
	InAddr server;		// server address to connect to
	int port;		// port number
	extern Simulation *sim;	// simulation object

	nm = argv[1];
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    tcpObjectName = argv[3];
	    server.s_addr = InetAddr(argv[4]);
	    rc = Tcl_GetInt(interp, argv[5], &port);

	    if (rc == TCL_OK) {

		// Get pointer to TCP object (implication:  Tcp
		// object *must* be created before any applications
		// are created.
		fc = new FtpClient(nm, sim, startTime,
				   tcpObjectName,
				   server, port);
		    Tcl_CreateCommand(interp, nm,
				      FtpClientObjectCmd,
				      (ClientData) fc,
				      FtpClientDeleteProc);
	    }
	}
	return rc;
    }
}

//
// FtpClientObjectCmd
//
// (Tcl object command procedure)
//
int FtpClientObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {
    return ScheduledObjectCmd(clientData, interp, argc, argv);
}

//
// FtpClientDeleteProc
//
// (Tcl delete procedure)
//
void FtpClientDeleteProc(ClientData clientData) {
    delete (FtpClient *) clientData;
}

//
// operator<<
//
// Input:  Output stream (s), FTP client state (ftpcs)
//
// Output:  Output stream
//
// Print the name of the state ftpcs on the output stream s.
//
ostream& operator<<(ostream &s, FtpClientState ftpcs)
{
    char *output;
    
    output = FtpClientStateName[ftpcs];

    return s << output;

}

