static char rcsid[] = "$Id: FtpServer.cc,v 1.12 1996-03-11 15:51:04-08 bmah Exp $";
//
// $Id: FtpServer.cc,v 1.12 1996-03-11 15:51:04-08 bmah Exp $
//
// FtpServer.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// FTP daemon class.
//

#include "FtpServer.h"
#include "EmpiricalIntegral.h"
#include "IpPacket.h"
#include "Event.h"
#include "inline.h"

//
// State names
//
char *FtpServerStateName[] = {
    "FtpServerStateStart",
    "FtpServerStateAwaitPort",
    "FtpServerStateConnectData",
    "FtpServerStateData",
    "FtpServerStateDone"
};

//
// Miscellaneous constants
//
static const double gap = 0.001; // 1 ms between sends
static double epsilon = 0.000001; // 1 us interval between other events

//
// Imported probability distributions
//
extern RNG *rng;		// global random number generator

// Length of FTP file (on data connection)
extern const DoubleDistElement *FtpReplyLengthDistTable;
extern int FtpReplyLengthDistLength;

//
// FtpServer::Handler
//
// (Event handler)
//
void FtpServer::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalIntegral FtpReplyLengthDist(FtpReplyLengthDistLength,
						FtpReplyLengthDistTable,
						rng);

    // In the FtpServer object, all of the EventTimer events have a
    // pointer to an FtpServerSession in their data field.  The
    // sole exception is the first EventTimer, which is used to force
    // initialization of the FtpServer object.  In this case, its
    // data field is NULL.
    if ((type == EventTimer) && (data == NULL)) {

	tpControlListen = tcpModule->UserListen(time + epsilon, port, this);

	IF_DEBUG(2, {
	    DEBUG_PREFIX("FtpServer");
	    *(sim->debugStream) << "initialized\n";
	});

    }
    else {

	FtpServerSession *session;
	TcpPcb *tp;
	TcpCall *tc;

	// Figure out what session an event was far.  This determination
	// is dependent on the event type.  After the switch() below,
	// session should be set appropriately.
	switch (type) {

	    case (EventTimer):
	    {
		session = (FtpServerSession *) data;
	    }
	    break;

	    case (EventTcpConnect):
	    {
		tp = (TcpPcb *) data;

		// If we already know about this connection, it's one we
		// initiated, and we should use its session.  Otherwise,
		// the session was initiated by the client and it's a new
		// control connection.  We need a new session for it.
		session = FindSession(tp);
		if (!session) {

		    session = new FtpServerSession;
		    session->state = FtpServerStateStart;
		    session->tpControl = NULL;
		    session->tpData = NULL;
		    session->next = sessions;
		    sessions = session;

		}		    
	    }
	    break;

	    case (EventTcpArrival):
	    {
		tc = (TcpCall *) data;
		tp = tc->pcb;
		session = FindSession(tp);

		if (session == NULL) {
		    IF_DEBUG(0, {
			DEBUG_PREFIX("FtpServer");
			*(sim->debugStream) << type << " arrived but couldn't grok session\n";
		    });
		}
	    }
	    break;

	    case (EventTcpClosed):
	    {
		tp = (TcpPcb *) data;
		session = FindSession(tp);

		if (session == NULL) {
		    IF_DEBUG(0, {
			DEBUG_PREFIX("FtpServer");
			*(sim->debugStream) << type << " arrived but couldn't grok session\n";
		    });
		}
	    }
	    break;

	    default:
	    {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("FtpServer");
		    *(sim->debugStream) << "bad EventType " << type << "\n";
		});
	    }
	    break;

	}

	IF_DEBUG(2, {
	    DEBUG_PREFIX("FtpServer");
	    *(sim->debugStream) << "received " << type << " in state " << session->state << "\n";
	});

	// Now run a standard state machine
	switch (session->state) {

	    //
	    // FtpServerStateStart:  Just got a connect (we think) for
	    // a new control connection.
	    //
	    case (FtpServerStateStart):
	    {
		switch (type) {
		    case (EventTcpConnect):
		    {
			session->tpControl = (TcpPcb *) data;
			session->state = FtpServerStateAwaitPort;
			break;
		    }

		    default:
		    goto bad;
		}
	    }
	    break;
	
	    //
	    // FtpServerStateAwaitPort:  Waiting for PORT command from
	    // client.  If we get a data tag (with a port number in it),
	    // try to connect to the client's data port.
	    //
	    case (FtpServerStateAwaitPort):
	    {
		switch(type) {

		    // EventTcpArrival:  Arrival on control connection.
		    // Throw away all arrivals unless there's a data
		    // tag, in which case it's the port number to connect to
		    // for a data transfer.
		    case (EventTcpArrival):
		    {
			TcpCall *tc = (TcpCall *) data;
			if (tc->pcb == session->tpControl) {

			    TcpData *td = tc->data.Dequeue();
			    if (td) {
				int clientPort = td->data_int;

				IF_DEBUG(2, {
				    DEBUG_PREFIX("FtpServer");
				    *(sim->debugStream) << "received PORT for port " << clientPort << "\n";
				});

				// Connect to client to begin data transfer
				session->tpData =
				  tcpModule->UserConnect(time + epsilon,
							 session->tpControl->tcp_faddr,
							 clientPort,
							 this,
							 dataPort);
				session->state = FtpServerStateConnectData;

				delete td;
			    }
			}
			else {
			    IF_DEBUG(0, {
				DEBUG_PREFIX("FtpServer");
				*(sim->debugStream) << "got " << type <<
				  " but not for control connection\n";
			    });
			}
			delete tc;
		    }
		    break;

		    // EventTcpClosed:  Close control connection.
		    case (EventTcpClosed):
		    {
			TcpPcb *tp = (TcpPcb *) data;
			if (tp == session->tpControl) {
			    tcpModule->UserClose(time + epsilon, tp);
			    session->state = FtpServerStateDone;
			    session->tpControl = NULL;

			    // Delete session
			    FtpServerSession *current, *prev;
			    for (current = sessions, prev = NULL;
				 current != NULL; current = current->next) {
				if (current == session) {
				    if (prev) {
					prev->next = current->next;
				    }
				    else {
					sessions = current->next;
				    }
				    delete current;
				    break;
				}
				prev = current;
			    }
			}
			else {
			    IF_DEBUG(0, {
				DEBUG_PREFIX("FtpServer");
				*(sim->debugStream) << "got " << type <<
				  " but not for control connection\n";
			    });
			}
		    }
		    break;

		    default:
		    goto bad;
		}
	    }
	    break;
	    
	    //
	    // FtpServerStateConnectData:  Waiting for the connection to
	    // the client's data port.
	    //
	    case (FtpServerStateConnectData):
	    {
		switch (type) {

		    // EventTcpConnect:  Hopefully this means we got
		    // connected to the client FTP process, and can
		    // now begin transferring data.
		    case (EventTcpConnect): {
			TcpPcb *tp = (TcpPcb *) data;
			if (tp == session->tpData) {
//			    bytesRemaining = ftp_itemsize();
			    bytesRemaining = (int) FtpReplyLengthDist();

			    IF_DEBUG(2, {
				DEBUG_PREFIX("FtpServer");
				*(sim->debugStream) << "item is " << bytesRemaining << " bytes\n";
			    });

			    session->state = FtpServerStateData;
			    Post(time + gap, this, EventTimer, session);
			}
			else {
			    IF_DEBUG(0, {
				DEBUG_PREFIX("FtpServer");
				*(sim->debugStream) << "got " << type <<
				  " but not for data connection\n";
			    });
			}
		    }
		    break;

		    // EventTcpClosed:  Connection establishment failed
		    case (EventTcpClosed): {

			IF_DEBUG(0, {
			    DEBUG_PREFIX("FtpServer");
			    *(sim->debugStream) << "connection failed\n";
			});

			session->tpData =
			  tcpModule->UserConnect(time + epsilon,
						 session->tpControl->tcp_faddr,
						 session->clientPort,
						 this,
						 dataPort);

		    }
		    break;

		    default:
		    goto bad;
		}
		
	    }
	    break;

	    //
	    // FtpServerStateData:  Transferring data.  When we're done
	    // go to FtpServerStateAwaitPort to await the next item.
	    //
	    case (FtpServerStateData):
	    {
		switch (type) {

		    // EventTimer:  Send the next segment.  If there are
		    // none left, then close the data connection.
		    case (EventTimer):
		    {

			int sendSize = bytesRemaining;
			if (sendSize == 0) {
			    // We're done.  Close.
			    tcpModule->UserClose(time + epsilon,
						 session->tpData);
			    session->tpData = NULL;
			    session->state = FtpServerStateAwaitPort;
			}
			else {
			    // Make up the next segment and send it.
			    if (sendSize > session->tpData->maxseg) {
				sendSize = session->tpData->maxseg;
			    }
			    tcpModule->UserSend(time + epsilon,
						session->tpData,
						sendSize);
			    Post(time + gap, this, EventTimer, session);

			    bytesRemaining -= sendSize;
			}
			    
		    }
		    break;

		    default:
		    goto bad;
		}
	    }
	    break;

	    default:
	    {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("FtpServer");
		    *(sim->debugStream) << "bad EventType " << type << "\n";
		});
	    }
	    break;
	}		
	return;

bad:
	IF_DEBUG(0, {
	    DEBUG_PREFIX("FtpServer");
	    *(sim->debugStream) << "bad EventType " << type << " in FtpServerState " << session->state << "\n";
	});
	return;
    }
    
}

//
// FtpServer::FindSession
//
// Input:  TcpPcb (tp)
//
// Output:  Pointer to an FtpServerSession (in return value).
//
// Locate the session associated with a given TCP connection.  Returns
// NULL if no session can be found.
//
FtpServerSession *FtpServer::FindSession(TcpPcb *tp)
{
    FtpServerSession *session;
    
    for (session = sessions; session != NULL; session = session->next) {
	if ((tp == session->tpControl) || (tp == session->tpData)) {
	    break;
	}
    }

    return session;
}

//
// FtpServer::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String FtpServer::Output()
{

    char c[30];
    sprintf(c, "%d dataport %d", port, dataPort);

    return name + " [FtpServer] tcp " + tcpModuleName + " ctlport " + c;

}

//
// FtpServerCmd
//
// (Tcl command procedure)
//
int FtpServerCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 6) {
	Tcl_SetResult(interp, "usage:  FtpServer name start tcp ctlport dataport", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	FtpServer *fs;		// new server object
	char *nm;		// name of new server object
	double startTime;	// starting time to run
	String tcpObjectName;	// name of Tcp object
	int port;		// port number
	int dp;			// data port number
	extern Simulation *sim;	// simulation object

	nm = argv[1];
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    tcpObjectName = argv[3];
	    rc = Tcl_GetInt(interp, argv[4], &port);

	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[5], &dp);
		
		if (rc == TCL_OK) {

		    // Get pointer to TCP object (implication:  Tcp
		    // object *must* be created before any applications
		    // are created.
		    fs = new FtpServer(nm, sim, startTime, tcpObjectName,
				       port, dp);
		    Tcl_CreateCommand(interp, nm,
				      FtpServerObjectCmd,
				      (ClientData) fs,
				      FtpServerDeleteProc);
		}
	    }
	}
	return rc;
    }
}

//
// FtpServerObjectCmd
//
// (Tcl object command procedure)
//
int FtpServerObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {
    return ScheduledObjectCmd(clientData, interp, argc, argv);
}

//
// FtpServerDeleteProc
//
// (Tcl delete procedure)
//
void FtpServerDeleteProc(ClientData clientData) {
    delete (FtpServer *) clientData;
}

//
// operator<<
//
// Input:  Output stream (s), FTP server state (ftpss)
//
// Output:  Output stream
//
// Print the name of the state ftpss on the output stream s.
//
ostream& operator<<(ostream &s, FtpServerState ftpss)
{
    char *output;
    
    output = FtpServerStateName[ftpss];

    return s << output;

}

