static char rcsid[] = "$Id: HttpClient.cc,v 1.13 1996-03-08 09:20:15-08 bmah Exp $";
//
// $Id: HttpClient.cc,v 1.13 1996-03-08 09:20:15-08 bmah Exp $
//   based on:
// $ID: FtpClient.cc,v 1.3 1995-10-25 14:12:23-07 bmah Exp $
//
// HttpClient.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// HTTP client class.
//

#include "HttpClient.h"
#include "EmpiricalContinuous.h"
#include "EmpiricalIntegral.h"
#include "IpPacket.h"
#include "Event.h"
#include "inline.h"
#include "httpproto.h"

//
// State names
//
char *HttpClientStateName[] = {
    "HttpClientStateStart",
    "HttpClientStateAwaitConnect",
    "HttpClientStateAwaitReply",
    "HttpClientStateNextPage"
};

// Global constants
static double gapItem = 0.001;	// 1 ms gap between items

//
// Imported probability distributions
//
extern RNG *rng;		// global random number generator

// Number of TCP connections per WWW page
extern const DoubleDistElement *HttpConnectionsDistTable;
extern int HttpConnectionsDistLength;

// Number of bytes per request (combined primary and secondary)
extern const DoubleDistElement *HttpRequestLengthDistTable;
extern int HttpRequestLengthDistLength;

// Think time between WWW pages
extern const DoubleDistElement *HttpThinkTimeDistTable;
extern int HttpThinkTimeDistLength;

// Number of connections to same WWW server
extern const DoubleDistElement *HttpServerStayDistTable;
extern int HttpServerStayDistLength;

//
// HttpClient::Handler
//
// (Event handler)
//
void
HttpClient::Handler(Event *e)
{
    
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalIntegral HttpConnectionsDist(HttpConnectionsDistLength,
						 HttpConnectionsDistTable,
						 rng);
    static EmpiricalIntegral HttpRequestLengthDist(HttpRequestLengthDistLength,
						 HttpRequestLengthDistTable,
						 rng);
    static EmpiricalContinuous HttpThinkTimeDist(HttpThinkTimeDistLength,
						 HttpThinkTimeDistTable,
						 rng);

    IF_DEBUG(2, {
	DEBUG_PREFIX("HttpClient");
	*(sim->debugStream) << "received " << type << " in state " << state << "\n";
    });

    // Note:  Because we're using our own private state space,
    // we can never use the base class Handler method for anything.
    switch (state) {

	//
	// HttpClientStateStart:  Uninitialized state.
	//
	case (HttpClientStateStart):
	{

	    switch (type) {
		
		// First timer event...initialize variables and select
		// and connect to first server.
		case (EventTimer):
		{

		    // Select first server
		    server.s_addr = 0;
		    server = SelectServer(server);
		    numItems = (int) HttpConnectionsDist();

		    IF_DEBUG(1, {
			DEBUG_PREFIX("HttpClient");
			*(sim->debugStream) << "new page " <<
			  InetNtoA(server) << " " << numItems <<
			    " items\n";
		    });

		    // Start accounting
		    itemStartTime = time;
		    itemRequestBytes = 0;
		    itemReplyBytes = 0;
		    pageStartTime = time;
		    pageRequestBytes = 0;
		    pageReplyBytes = 0;
		    pageItems = 0;
		    totalRequestBytes = 0;
		    totalReplyBytes = 0;
		    totalItems = 0;
		    totalPages = 0;
		    
		    // Connect to Web server
		    tpData = tcpModule->UserConnect(time, server, port, this);

		    state = HttpClientStateAwaitConnect;

		}
		break;

		default:
		goto bad;
		break;

	    }
	    
	}
	break;

	//
	// HttpClientStateAwaitConnect:  Trying to connect to HTTP
	// server process.
	//
	case (HttpClientStateAwaitConnect):
	{
	    
	    switch(type) {

		// EventTcpConnect:  TCP connection to server completed.
		case (EventTcpConnect):
		{

		    TcpPcb *tp = (TcpPcb *) data;

		    // Sanity checking on the TCP connection.
		    if (tp != tpData) {
			IF_DEBUG(0, {
			    DEBUG_PREFIX("HttpClient");
			    *(sim->debugStream) << "got " << type <<
			      " but not for data connection\n";
			});
		    }
		    else {

			// Send out request to HTTP server.  This request
			// is a standard TCP send with a single data tag,
			// which gives some information as to the type
			// of request.
			int requestLen;
			TcpData *sendData = new TcpData;

			requestLen = (int) HttpRequestLengthDist();
			sendData->data_int = (int) HttpRequestTypePrimary;
			
			tcpModule->UserSend(time, tp, requestLen, sendData);

			IF_DEBUG(3, {
			    DEBUG_PREFIX("HttpClient");
			    *(sim->debugStream) << "sending request of " <<
			      requestLen << " bytes\n";
			});

			state = HttpClientStateAwaitReply;

			itemRequestBytes = requestLen;
			pageRequestBytes += requestLen;
			totalRequestBytes += requestLen;

		    }

		}
		break;

		// EventTcpClosed:  Connection failed
		case (EventTcpClosed):
		{

		    IF_DEBUG(0, {
			DEBUG_PREFIX("HttpClient");
			*(sim->debugStream) << "connection failed\n";
		    });

		    // Try to reconnect
		    tpData = tcpModule->UserConnect(time, server, port, this);

		    state = HttpClientStateAwaitConnect;

		}
		break;

		default:
		goto bad;
		break;
	    }

	}
	break;

	//
	// HttpClientStateAwaitReply:  Waiting for reply from server.
	//
	case (HttpClientStateAwaitReply):
	{
	    switch (type) {

		// EventTcpArrival:  Got TCP segment.
		case (EventTcpArrival):
		{

		    TcpCall *tc = (TcpCall *) data;

		    // Sanity checking on the TCP connection.
		    if (tc->pcb != tpData) {
			IF_DEBUG(0, {
			    DEBUG_PREFIX("HttpClient");
			    *(sim->debugStream) << "got " << type <<
			      " but not for data connection\n";
			});
		    }
		    else {

			// Tally reply bytes received for this
			// this connection/session/whatever.  Shouldn't
			// be any data tags so we don't look for them.
			itemReplyBytes += tc->th_ulen;
			pageReplyBytes += tc->th_ulen;
			totalReplyBytes += tc->th_ulen;

			IF_DEBUG(3, {
			    DEBUG_PREFIX("HttpClient");
			    *(sim->debugStream) << "received " <<
			      tc->th_ulen << " reply bytes\n";
			});

		    }

		    delete tc;
		}
		break;

		// EventTcpClosed:  Server closed TCP connection (end of
		// data)
		case (EventTcpClosed):
		{

		    TcpPcb *tp = (TcpPcb *) data;
		    
		    // Sanity checking on the TCP connection.
		    if (tp != tpData) {
			IF_DEBUG(0, {
			    DEBUG_PREFIX("HttpClient");
			    *(sim->debugStream) << "got " << type <<
			      " but not for data connection\n";
			});
		    }
		    else {

			// First complete the connection close
			tcpModule->UserClose(time, tp);

			pageItems++;
			totalItems++;

			    *(sim->logStream) << name << " [HttpClient] item lifetime " << time - itemStartTime << " requestbytes " << itemRequestBytes << " replybytes " << itemReplyBytes << "\n";

			itemRequestBytes = 0;
			itemReplyBytes = 0;

			// First check to see if there are any more items
			// on the current page.
			if (--numItems >= 1) {

			    itemStartTime = time;
			    itemRequestBytes = 0;
			    itemReplyBytes = 0;

			    // There exist more items to be done...initiate
			    // another connection (to the same server,
			    // on the assumption that all items for a single
			    // page will come from the same server).
			    tpData = tcpModule->UserConnect(time,
							    server,
							    port,
							    this);
			    itemStartTime = time;
			    state = HttpClientStateAwaitConnect;

			    IF_DEBUG(2, {
				DEBUG_PREFIX("HttpClient");
				*(sim->debugStream) <<
				  "next URL on same page\n";
			    });

			}
			else {

			    // No more items on current page, but there
			    // are more pages to be done.
			    totalPages++;

			    // Write out statistics to log file
			    *(sim->logStream) << name << " [HttpClient] page lifetime " << time - pageStartTime << " requestbytes " << pageRequestBytes << " replybytes " << pageReplyBytes << " items " << pageItems << "\n";

			    // Calculate user "think time".  If it's less
			    // than our threshold, then set a timer.
			    // Otherwise, we declare this HttpClient object
			    // to be finished processing and delete it.
			    double thinkTime;
			    thinkTime = HttpThinkTimeDist();

			    if (thinkTime < threshold) {
				Post(time + thinkTime, this, EventTimer, NULL);

				IF_DEBUG(2, {
				    DEBUG_PREFIX("HttpClient");
				    *(sim->debugStream) <<
				      "thinking about next page for " <<
					thinkTime << "\n";
				});

				state = HttpClientStateNextPage;
			    }
			    else {
				delete this;
			    }
			}
		    }
		}
		break;

		default:
		goto bad;
		break;

	    }

	}
	break;

	//
	// HttpClientStateNextPage:  Waiting to select next page
	//
	case (HttpClientStateNextPage):
	{

	    switch (type) {

		// EventTimer:  Ready for next page, so select and
		// connect.
		case (EventTimer):
		{

		    // Select next server.
		    server = SelectServer(server);
		    numItems = (int) HttpConnectionsDist();

		    IF_DEBUG(1, {
			DEBUG_PREFIX("HttpClient");
			*(sim->debugStream) << "new page " <<
			  InetNtoA(server) << " " << numItems <<
			    " items\n";
		    });

		    // Start accounting for next session
		    itemStartTime = time;
		    itemRequestBytes = 0;
		    itemReplyBytes = 0;
		    pageStartTime = time;
		    pageRequestBytes = 0;
		    pageReplyBytes = 0;
		    pageItems = 0;

		    // Connect to (possibly new) server
		    tpData = tcpModule->UserConnect(time, server, port, this);

		    state = HttpClientStateAwaitConnect;

		}
		break;

		default:
		goto bad;
		break;
	    }

	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("HttpClient");
		*(sim->debugStream) << "bad state\n";
	    });
	}
	break;

    }
    
    return;

bad:
    IF_DEBUG(0, {
	DEBUG_PREFIX("HttpClient");
	*(sim->debugStream) << "bad EventType " << type << " in state " << state << "\n";
    });

}

//
// HttpClient::SelectServer
//
// Input:  Current HTTP server the client is connected to.
// An IP address of 0 means there is no current server.
//
// Output:  New HTTP server address (in return value).
//
// Select the HTTP server for the next simulated web page retrieval.
// We assume there is a given probability for locality of reference.
// If a new server is selected, we use Zipf's Law to choose a new
// server from among the list of servers.
//
InAddr
HttpClient::SelectServer(InAddr oldServer) {

    static EmpiricalIntegral HttpServerStayDist(HttpServerStayDistLength,
						HttpServerStayDistTable,
						rng);
    
    if ((oldServer.s_addr == 0) || (--numPagesOnServer < 1)) {
	
	// Set number of pages to retrieve from the new server
	numPagesOnServer = (int) HttpServerStayDist();

	// Select new server.  Remember that a Zipf's Law distribution
	// returns values starting from 1, but C++ arrays are indexed
	// starting from 0.
	InAddr newServer = servers[((int) ((*serverIndex)())) - 1];

	return (newServer);

    }
    else {

	// Use the previous server.
	return(oldServer);

    }
}

//
// HttpClient::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
HttpClient::Output()
{
    char c[30];
    sprintf(c, "%d", port);
    
    return name + " [HttpClient] tcp " + tcpModuleName + " port " + c;
}

//
// HttpClientCmd
//
// (Tcl command procedure)
//
int
HttpClientCmd(ClientData clientData, Tcl_Interp *interp,
	      int argc, char *argv[]) {

    if (argc != 7) {
	Tcl_SetResult(interp, "usage:  HttpClient name start tcp serverlist port", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	HttpClient *fc;		// new client object
	char *nm;		// name of new client object
	double startTime;	// starting time to run
	String tcpObjectName;	// name of Tcp object
	int ns;			// number of HTTP servers to use
	char **hss;		// HTTP server strings
	InAddr *hsa;		// HTTP server addresses
	int port;		// port number
	double th;		// think time threshold
	extern Simulation *sim;	// simulation object
	int i;			// yet another loop counter

	// Get name
	nm = argv[1];

	// Get start time
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    // Get TCP object name
	    tcpObjectName = argv[3];

	    // Get array of IP addresses
	    rc = Tcl_SplitList(interp, argv[4], &ns, &hss);
	    if (rc == TCL_OK) {
		
		// Parse IP addresses
		hsa = new InAddr[ns];
		for (i = 0; i < ns; i++) {
		    hsa[i].s_addr = InetAddr(hss[i]);
		}
		free(hss);

		// Get port number
		rc = Tcl_GetInt(interp, argv[5], &port);
		if (rc == TCL_OK) {
		    
		    // Get timeout threshhold
		    rc = Tcl_GetDouble(interp, argv[6], &th);
		    if (rc == TCL_OK) {

			// Get pointer to TCP object (implication:  Tcp
			// object *must* be created before any applications
			// are created.
			fc = new HttpClient(nm, sim, startTime,
					    tcpObjectName,
					    ns, hsa,
					    port, th);
			Tcl_CreateCommand(interp, nm,
					  HttpClientObjectCmd,
					  (ClientData) fc,
					  HttpClientDeleteProc);
		    }
		}
	    }
	}
	return rc;
    }
}

//
// HttpClientObjectCmd
//
// (Tcl object command procedure)
//
int
HttpClientObjectCmd(ClientData clientData, Tcl_Interp *interp,
		    int argc, char *argv[]) {
    return ScheduledObjectCmd(clientData, interp, argc, argv);
}

//
// HttpClientDeleteProc
//
// (Tcl delete procedure)
//
void
HttpClientDeleteProc(ClientData clientData) {
    delete (HttpClient *) clientData;
}

//
// operator<<
//
// Input:  Output stream (s), HTP client state (httpcs)
//
// Output:  Output stream
//
// Print the name of the state httpcs on the output stream s.
//
ostream&
operator<<(ostream &s, HttpClientState httpcs)
{
    char *output;
    
    output = HttpClientStateName[httpcs];

    return s << output;

}

