static char rcsid[] = "$Id: HttpServer.cc,v 1.8 1996-03-05 15:00:31-08 bmah Exp $";
//
// $Id: HttpServer.cc,v 1.8 1996-03-05 15:00:31-08 bmah Exp $
//   based on:
// $ID: FtpServer.cc,v 1.2 1995-10-25 11:30:23-07 bmah Exp $
//
// HttpServer.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// HTTP server class.  This server is a multi-threaded server with
// no contention for resources.  While this doesn't simulate "real"
// Web servers under load very well, it is a bit easier to work with
// given our event-driven model of the world than the previous iteration
// of this object, which tried to simulate a single-threaded server.
//

#include "HttpServer.h"
#include <ACG.h>
#include "EmpiricalIntegral.h"
#include "IpPacket.h"
#include "Event.h"
#include "inline.h"

//
// State names
//
char *HttpServerStateName[] = {
    "HttpServerStateStart",
    "HttpServerStateConnected"
};

//
// Miscellaneous constants
//
static const double gap = 0.001; // 1 ms between sends

//
// Imported probability distributions
//
extern RNG *rng;		// global random number generator

// Number of bytes per reply (combined primary and secondary)
extern const DoubleDistElement *HttpReplyLengthDistTable;
extern int HttpReplyLengthDistLength;

//
// HttpServer::Handler
//
// (Event handler)
//
void HttpServer::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalIntegral HttpReplyLengthDist(HttpReplyLengthDistLength,
						 HttpReplyLengthDistTable,
						 rng);

    IF_DEBUG(2, {
	DEBUG_PREFIX("HttpServer");
	*(sim->debugStream) << "received " << type << " in state " << state << "\n";
    });

    switch (type) {

	// First timeout event is for initialization.
	// Set up the socket to listen for new connections.
	case (EventTimer):
	{
	    tpListen = tcpModule->UserListen(time, port, this);

	    IF_DEBUG(2, {
		DEBUG_PREFIX("HttpServer");
		*(sim->debugStream) << "initialized\n";
	    });
	}
	break;

	// In the case that we got a new connection, make up a new session
	// structure.
	case (EventTcpConnect):
	{
	    TcpPcb *tp = (TcpPcb *) data; // PCB for new connection

	    HttpServerSession *session = new HttpServerSession;
	    session->state = HttpServerStateConnected;
	    session->tp = tp;
	    session->next = sessions;
	    sessions = session;
	}
	break;

	// We got an arrival for one of the connections we have open
	// (we think!)
	case (EventTcpArrival):
	{

	    HttpServerSession *session;
	    TcpCall *tc = (TcpCall *) e->ev_data;
	    TcpPcb *tp = tc->pcb;;
	    
	    // Search through the session list to find the session.
	    for (session = sessions;
		 session != NULL;
		 session = session->next) {

		if (session->tp == tc->pcb) {
		    break;
		}

	    }


	    // See if we've heard of this connection before
	    if (session == NULL) {

		// Apparently not...
		IF_DEBUG(0, {
		    DEBUG_PREFIX("HttpServer");
		    *(sim->debugStream) <<
		      "received unknown EventTcpArrival\n";
		});

	    }
	    else {

		// Tally request bytes from th_ulen.
		totalRequestBytes += tc->th_ulen;

		// See if we got a data tag.  If we didn't, ignore the
		// data.  If we did, it's the end of the client request.
		TcpData *td = tc->data.Dequeue();
		if (td) {

		    HttpRequestType request = (HttpRequestType) td->data_int;
		    int replyLength;

		    // Select reply length from the appropriate distribution
		    replyLength = (int) HttpReplyLengthDist();

		    totalReplyBytes += replyLength;
		    totalItems++;

		    IF_DEBUG(2, {
			DEBUG_PREFIX("HttpServer");
			*(sim->debugStream) << "sending " <<
			  replyLength << " reply bytes\n";
		    });

		    // Blast the data down the pipe, then close the
		    // connection (this is how most clients determine
		    // the end of the reply content).
		    time += gap;
		    tcpModule->UserSend(time, tp, replyLength);

		    time += gap;
		    tcpModule->UserClose(time, tp);

		    // Get rid of the session record
		    HttpServerSession *current, *prev;
		    for (current = sessions, prev = NULL;
			 current != NULL; current = current->next) {
			if (current == session) {
			    if (prev) {
				prev->next = current->next;
			    }
			    else {
				sessions = current->next;
			    }
			    delete current;
			    break;
			}
			prev = current;
		    }

		    // Delete the data tag
		    delete td;

		}

		// Delete the call block
		delete tc;

	    }

	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("HttpServer");
		*(sim->debugStream) << "received bad event type " << type << " in state " << state << "\n";
	    });
	}
	break;

    }
    

}

//
// HttpServer::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String HttpServer::Output()
{

    char c[30];
    sprintf(c, "%d", port);
    
    return name + " [HttpServer] tcp " + tcpModuleName + " port " + c;

}

//
// HttpServerCmd
//
// (Tcl command procedure)
//
int HttpServerCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 5) {
	Tcl_SetResult(interp, "usage:  HttpServer name start tcp port", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	HttpServer *hs;		// new server object
	char *nm;		// name of new server object
	double startTime;	// starting time to run
	String tcpObjectName;	// name of Tcp object
	int port;		// port number
	extern Simulation *sim;	// simulation object

	nm = argv[1];
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    tcpObjectName = argv[3];
	    rc = Tcl_GetInt(interp, argv[4], &port);

	    if (rc == TCL_OK) {

		// Get pointer to TCP object (implication:  Tcp
		// object *must* be created before any applications
		// are created.
		hs = new HttpServer(nm, sim, startTime, tcpObjectName,
				    port);
		Tcl_CreateCommand(interp, nm,
				  HttpServerObjectCmd,
				  (ClientData) hs,
				  HttpServerDeleteProc);
	    }
	}
	return rc;
    }
}

//
// HttpServerObjectCmd
//
// (Tcl object command procedure)
//
int HttpServerObjectCmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[]) {
    return ScheduledObjectCmd(clientData, interp, argc, argv);
}

//
// HttpServerDeleteProc
//
// (Tcl delete procedure)
//
void HttpServerDeleteProc(ClientData clientData) {
    delete (HttpServer *) clientData;
}

//
// operator<<
//
// Input:  Output stream (s), HTTP server state (httpss)
//
// Output:  Output stream
//
// Print the name of the state ftpss on the output stream s.
//
ostream& operator<<(ostream &s, HttpServerState httpss)
{
    char *output;
    
    output = HttpServerStateName[httpss];

    return s << output;

}

