static char rcsid[] = "$Id: Ip.cc,v 1.16 1996-03-04 14:23:33-08 bmah Exp $";
//
// $Id: Ip.cc,v 1.16 1996-03-04 14:23:33-08 bmah Exp $
//
// Ip.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// IP module class.
//

#include "Ip.h"
#include "IpPacket.h"
#include "Event.h"
#include "inline.h"

//
// Ip::~Ip
//
// Object destructor
//
Ip::~Ip()
{

    // Deallocate all our memory
    IpInterface *i, *inext;
    IpRoute *r, *rnext;
    IpProto *p, *pnext;

    for (i = interfaces; i != NULL; i = inext) {
	inext = i->next;
	delete i;
    }
    for (r = routes; r != NULL; r = rnext) {
	rnext = r->next;
	delete r;
    }
    for (p = protocols; p != NULL; p = pnext) {
	pnext = p->next;
	delete p;
    }

}

//
// Ip::Handler
//
// (Event handler)
//
void
Ip::Handler(Event *e)
{
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    IpPacket *ipPacket = (IpPacket *) data; // IP packet that arrived
    
    switch (type) {

	// All packets arriving at this IP module either showed up from
	// the transport entity above or from the network layer below.
	// In any case, we treat these packets identically.
	case (EventIpSend):
	case (EventPacketArrival):
	{

	    IF_DEBUG(2, {
		DEBUG_PREFIX("Ip");
		*(sim->debugStream) << "packet " << ipPacket << "\n";
	    });

	    // Scan through the IP addresses in the interface table to
	    // see if the destination of this packet is here.  If so,
	    // then pass it up to the appropriate transport layer.
	    bool local = false;
	    IpInterface *currentInterface = interfaces;

	    while (currentInterface) {
		if (ipPacket->ip_dst.s_addr ==
		    currentInterface->ia_addr.s_addr) {
		    local = true;
		    break;
		}
		currentInterface = currentInterface->next;
	    }
	    if (local) {

		if (type == EventIpSend) {

		    // If the packet came from the layer above and not
		    // from a device driver, and its destination is local,
		    // then it was a looped back packet, and so we should
		    // increment input and output counters for this packet.
		    // Note we're short-circuiting the normal update process,
		    // which is normally done by the device driver.  We do
		    // this to allow the device drivers to be ignorant of
		    // their IP addresses (which is necessary above to see
		    // whether a packet should be looped back).
		    Lookup(currentInterface->ia_object,
			   currentInterface->ia_objectName);
		    currentInterface->ia_object->if_ipackets++;
		    currentInterface->ia_object->if_ibytes += ipPacket->ip_len;
		    currentInterface->ia_object->if_opackets++;
		    currentInterface->ia_object->if_obytes += ipPacket->ip_len;
		}
		
		IpProto *currentProtocol;

		// Traverse our SLL of protocols to find the appropriate
		// transport layer entity.
		for (currentProtocol = protocols;
		     currentProtocol != NULL;
		     currentProtocol = currentProtocol->next) {
		    if (currentProtocol->pr_protocol ==
			ipPacket->ip_p) {
			break;
		    }
		}

		// If we found one, then pass the entire IP packet up
		// to it (TCP and UDP, for example, need the header to
		// help them with demuxing)
		if (currentProtocol) {
		    Lookup(currentProtocol->pr_object,
			   currentProtocol->pr_objectName);
		    Post(time + 0, currentProtocol->pr_object,
			 EventIpArrival, data); // XXX 
		}

		// Otherwise, we couldn't find an appropriate transport layer
		// protocol, so we need to drop it.
		else {

		    IF_DEBUG(0, {
			DEBUG_PREFIX("Ip");
			*(sim->debugStream) << "couldn't get transport protocol for " << ipPacket << "\n";
		    });

		    delete ipPacket;
		}

	    }

	    // Packet isn't local.
	    // Compute route using the routing table.  Iterate over
	    // all routes, trying to find the best match.  This isn't
	    // particularly efficient, but then again we don't expect
	    // to be simulating large numbers of routes either, so
	    // a simple solution is probably the best.
	    else {
		IpRoute *bestRoute = RouteLookup(ipPacket->ip_dst);

		// If we couldn't find a route, then drop the packet.  In real
		// life we'd send an IGMP message back to the source.
		if (bestRoute == NULL) {

		    IF_DEBUG(0, {
			DEBUG_PREFIX("Ip");
			*(sim->debugStream) << "couldn't get route for " <<
			  ipPacket << "\n";
		    });

		    delete ipPacket;
		    // XXX update statistics

		}
		else {

		    // If we haven't determined the interface record
		    // for this particular route yet, then do so.
		    if (bestRoute->rt_if == NULL) {
			GetRouteInterface(bestRoute);
		    }

		    // If the source address of this packet is 0,
		    // it originated locally, and we need to set
		    // the source address based on the route
		    // taken.
		    if (ipPacket->ip_src.s_addr == 0) {
			ipPacket->ip_src.s_addr =
			  bestRoute->rt_if->ia_addr.s_addr;
		    }

		    // Set the gateway
		    // address for this packet (sort of analogous
		    // to ARP lookup)
		    if (bestRoute->rt_flags & RouteFlagGateway) {
			ipPacket->ip_next = bestRoute->rt_gateway;
		    }
		    else {
			ipPacket->ip_next = ipPacket->ip_dst;
		    }

		    // XXX ttl--
		    
		    // Send the packet out the right interface
		    Lookup(bestRoute->rt_if->ia_object,
			   bestRoute->rt_if->ia_objectName);
		    Post(time + 0, bestRoute->rt_if->ia_object,
			 EventPacketSend, data);
		}
	    }
	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("Ip");
		*(sim->debugStream) << "bad EventType " << type << "\n";
	    });
	}
	break;

    }
}

//
// Ip::GetRouteInterface
//
// Input:  Route in question (r)
//
// Output;  None
//
// Find the interface associated with a given route.
//
void
Ip::GetRouteInterface(IpRoute *r)
{
    
    // Find the first interface whose netmask and address
    // matches the gateway address for the route
    for (IpInterface *currentInterface = interfaces;
	 currentInterface != NULL;
	 currentInterface = currentInterface->next) {
	if ((r->rt_gateway.s_addr &
	     currentInterface->ia_subnetmask.s_addr) ==
	    (currentInterface->ia_addr.s_addr &
	     currentInterface->ia_subnetmask.s_addr)) {
	    r->rt_if = currentInterface;
		break;
	}
    }

}


//
// Ip::RouteLookup
//
// Input:  Target address (target)
//
// Output:  Pointer to best matching route if any (in return value).
//
// Perform a routing table lookup for the target address.  If no match
// can be found, returns NULL.
//
IpRoute
*Ip::RouteLookup(InAddr target)
{
    IpRoute *bestRoute = NULL;
    IpRoute *currentRoute = routes;

    while (currentRoute) {

	// Basically we want to find a route whose mask and route
	// destination match the target address, and which
	// is also more specific (largest number of 1s in a
	// contiguous netmask) than the best route we've encountered
	// so far.
	if (((target.s_addr & currentRoute->rt_mask.s_addr) ==
	     currentRoute->rt_dst.s_addr) &&
	    ((bestRoute == NULL) ||
	     (bestRoute->rt_mask.s_addr < currentRoute->rt_mask.s_addr))) {
	    bestRoute = currentRoute;
	}
	currentRoute = currentRoute->next;
    }
    return bestRoute;
}



//
// Ip::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
Ip::Output()
{
    return name + " [Ip]";
}

//
// IpCmd
//
// (Tcl command procedure)
//
int
IpCmd(ClientData clientData, Tcl_Interp *interp,
      int argc, char *argv[]) {

    if (argc != 2) {
	Tcl_SetResult(interp, "usage:  Ip name", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of IP module
	Ip *ipModule;		// new Ip module
	extern Simulation *sim;	// global simulation object

	// One argument only...the name of the new Ip module.
	// Kind of hard for anything to go wrong...
	nm = argv[1];

	ipModule = new Ip(nm, sim);
	Tcl_CreateCommand(interp, nm,
			  IpObjectCmd,
			  (ClientData) ipModule,
			  IpDeleteProc);

	return (TCL_OK);

    }

}

//
// IpObjectCmd
//
// (Tcl object command procedure)
//
int
IpObjectCmd(ClientData clientData, Tcl_Interp *interp,
	    int argc, char *argv[]) {

    Ip *ipModule = (Ip *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }
    else {
	
	// ip route ...
	if (!strcmp(argv[1], "route")) {

	    if (argc == 2) {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	    // ip route query addr
	    if (!strcmp(argv[2], "query")) {

		if (argc != 4) {
		    Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		    return(TCL_ERROR);
		}
		else {
		    InAddr queryAddress;
		    queryAddress.s_addr = InetAddr(argv[3]);
		    
		    // First see if the address is local
		    IpInterface *interface = ipModule->interfaces;
		    while (interface != NULL) {
			if (queryAddress.s_addr == interface->ia_addr.s_addr) {
			    strcpy(interp->result, "0.0.0.0");
			    return (TCL_OK);
			}
			interface = interface->next;
		    }

		    // Next query the routing table to find the next-hop
		    // gateway router.  If there's a legitimate address,
		    // then return the IP address of the next hop gateway.
		    // Otherwise, return an empty string.
		    IpRoute *route = ipModule->RouteLookup(queryAddress);
		    if (route) {
			strcpy(interp->result, InetNtoA(route->rt_gateway));
		    }
		    return (TCL_OK);
		}
	    }

	    // ip route add addr mask gateway flags
	    else if (!strcmp(argv[2], "add")) {

		if ((argc != 7) && (argc != 6)) {
		    Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		    return(TCL_ERROR);
		}
		else {
		    InAddr rt_dst, rt_mask, rt_gateway; // addresses
							 // for this route
		    RouteFlags rt_flags = RouteFlagNone; // flags for new route

		    // get IP addresses
		    rt_dst.s_addr = InetAddr(argv[3]);
		    rt_mask.s_addr = InetAddr(argv[4]);
		    rt_gateway.s_addr = InetAddr(argv[5]);
		    
		    if (argc == 7) {
			// get flags
			char *f;
			for (f = argv[6]; (*f) != '\0'; f++) {
			    switch (*f) {

				// G:  gateway route
				case 'g':
				case 'G':
				{
				    rt_flags |= RouteFlagGateway;
				}
				break;

				default:
				{
				    Tcl_SetResult(interp, "bad flag",
						  TCL_STATIC);
				    return(TCL_ERROR);
				}
				break;

			    }
			}
		    }

		    // Construct new IpRoute structure.
		    IpRoute *route = new IpRoute;
		    route->rt_dst = rt_dst;
		    route->rt_mask = rt_mask;
		    route->rt_gateway = rt_gateway;
		    route->rt_flags = rt_flags;
		    route->rt_if = 0; // We set the pointer to the IpInterface
				      // record to zero right now, because it
				      // might not have been ifconfig-ed
				      // at this point...the GetRouteInterface
				      // call below might fail.
		    ipModule->GetRouteInterface(route);

		    if (route->rt_if == NULL) {
			Tcl_SetResult(interp,
				      "warning: no interface found for route",
				      TCL_STATIC);
			cerr << "warning: no interface found for route\n";
		    }

		    route->next = ipModule->routes;
		    ipModule->routes = route;

		}
	    }
	    else {
		Tcl_SetResult(interp, "bad route subcommand", TCL_STATIC);
		return(TCL_ERROR);
	    }

	    return (TCL_OK);

	}

	// ip netstat ...
	else if (!strcmp(argv[1], "netstat")) {
	    
	    if (argc == 2) {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	    // ip netstat interfaces
	    if ((!strcmp(argv[2], "interfaces")) ||
		(!strcmp(argv[2], "-i"))) {

		IpInterface *currentInterface;
		char buf[128];	// temporary printing buffer

		cout << "Name             Address          Netmask          Ipkts Ierrs Opkts Oerrs\n";

		// Loop over all interfaces, printing info
		for (currentInterface = ipModule->interfaces;
		     currentInterface != NULL;
		     currentInterface = currentInterface->next) {

		    sprintf(buf, "%-16s ",
			    (const char *) currentInterface->ia_objectName);
		    cout << buf;
			    
		    sprintf(buf, "%-16s ",
			    InetNtoA(currentInterface->ia_addr));
		    cout << buf;

		    sprintf(buf, "%-16s ",
			    InetNtoA(currentInterface->ia_subnetmask));
		    cout << buf;

		    ipModule->Lookup(currentInterface->ia_object,
				     currentInterface->ia_objectName, false);
		    if (currentInterface->ia_object) {
			sprintf(buf, "%5d %5d %5d %5d",
				currentInterface->ia_object->if_ipackets,
				currentInterface->ia_object->if_ierrors,
				currentInterface->ia_object->if_opackets,
				currentInterface->ia_object->if_oerrors);
			cout << buf;
		    }

		    cout << '\n';
		}

	    }
	    else if ((!strcmp(argv[2], "routes")) ||
		     (!strcmp(argv[2], "-r"))) {

		IpRoute *currentRoute;
		char buf[128];	// temporary printing buffer

		cout << "Destination      Mask             Gateway          Flags\n";

		// Loop over all interfaces, printing info
       		for (currentRoute = ipModule->routes;
		     currentRoute != NULL;
		     currentRoute = currentRoute->next) {

		    sprintf(buf, "%-16s ",
			    InetNtoA(currentRoute->rt_dst));
		    cout << buf;

		    sprintf(buf, "%-16s ",
			    InetNtoA(currentRoute->rt_mask));
		    cout << buf;

		    sprintf(buf, "%-16s ",
			    InetNtoA(currentRoute->rt_gateway));
		    cout << buf;

		    cout << (currentRoute->rt_flags & RouteFlagGateway ?
			     'G' : ' ');

		    cout << '\n';
		}

	    }
	    else if (!strcmp(argv[2], "protocols")) {

		IpProto *currentProto;
		char buf[128];	// temporary printing buffer

		cout << "Proto Handler              \n";

		for (currentProto = ipModule->protocols;
		     currentProto != NULL;
		     currentProto = currentProto->next) {

		    switch (currentProto->pr_protocol) {
			case (IPPROTO_ICMP):
			sprintf(buf, "icmp  ");
			break;

			case (IPPROTO_IGMP):
			sprintf(buf, "igmp  ");
			break;

			case (IPPROTO_TCP):
			sprintf(buf, "tcp   ");
			break;

			case (IPPROTO_UDP):
			sprintf(buf, "udp   ");
			break;

			case (IPPROTO_RAW):
			sprintf(buf, "raw   ");
			break;

			default:
			sprintf(buf, "%-5d ", currentProto->pr_protocol);
			break;
		    }
		    cout << buf;

		    sprintf(buf, "%-20s ", (const char *)
			    currentProto->pr_objectName);
		    cout << buf;

		    cout << '\n';

		}
		
	    }

	    else {
		Tcl_SetResult(interp, "bad netstat subcommand", TCL_STATIC);
		return(TCL_ERROR);
	    }
	    return(TCL_OK);

	}

	// ip ifconfig name addr netmask broadcast
	else if (!strcmp(argv[1], "ifconfig")) {
	    
	    if (argc != 6) {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	    else {

		String nm;	// name of interface to ifconfig
		InAddr ia_addr, ia_subnetmask, ia_broadaddr; // various
							     // addresses
							     // associated
							     // w/interface
		int rc = TCL_OK; // Tcl return code
		
		nm = argv[2];
		ia_addr.s_addr = InetAddr(argv[3]);
		ia_subnetmask.s_addr = InetAddr(argv[4]);
		ia_broadaddr.s_addr = InetAddr(argv[5]);

		IpInterface *newInterface = new IpInterface;
		
		newInterface->ia_addr = ia_addr;
		newInterface->ia_subnetmask = ia_subnetmask;
		newInterface->ia_broadaddr = ia_broadaddr;
		newInterface->mtu = 0;
		newInterface->ia_objectName = nm;
		newInterface->ia_object = NULL;
		newInterface->next = ipModule->interfaces;
		ipModule->interfaces = newInterface;
		
		return (rc);
	    }

	}

	// ip proto ...
	else if (!strcmp(argv[1], "proto")) {

	    if (argc == 2) {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	    // ip proto attach proto handler
	    if (!strcmp(argv[2], "attach")) {

		if (argc == 5) {

		    int pr_protocol; // protocol number
		    String nm;	       // protocol handler name
		    int rc = TCL_OK;   // Tcl return code

		    // get protocol number
		    if (!strcmp(argv[3], "icmp")) {
			pr_protocol = IPPROTO_ICMP;
		    }
		    else if (!strcmp(argv[3], "igmp")) {
			pr_protocol = IPPROTO_IGMP;
		    }
		    else if (!strcmp(argv[3], "tcp")) {
			pr_protocol = IPPROTO_TCP;
		    }
		    else if (!strcmp(argv[3], "udp")) {
			pr_protocol = IPPROTO_UDP;
		    }
		    else if (!strcmp(argv[3], "raw")) {
			pr_protocol = IPPROTO_RAW;
		    }
		    else {
			rc = Tcl_GetInt(interp, argv[3], &pr_protocol);
		    }

		    if (rc == TCL_OK) {

			// get handler name
			nm = argv[4];

			IpProto *proto = new IpProto;
			proto->pr_protocol = pr_protocol;
			proto->pr_objectName = nm;
			proto->pr_object = NULL;
			proto->next = ipModule->protocols;
			ipModule->protocols = proto;

		    }
		    return(rc);

		}

	    }
	    else {
		Tcl_SetResult(interp, "bad proto subcommand", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	else {
	    return (ScheduledObjectCmd(clientData, interp, argc, argv));
	}
    }

}

//
// IpDeleteProc
//
// (Tcl delete procedure)
//
void
IpDeleteProc(ClientData clientData) {
    delete (Ip *) clientData;
}
