static char rcsid[] = "$Id: Lan.cc,v 1.8 1996-02-21 16:13:45-08 bmah Exp $";
//
// $Id: Lan.cc,v 1.8 1996-02-21 16:13:45-08 bmah Exp $
//
// Lan.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Simple shared-medium LAN class.  This LAN has "magic" FIFO scheduling
// and no backoff delays (a la Ethernet).
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//

#include "Lan.h"
#include "LanFrame.h"
#include "Event.h"
#include "inline.h"

//
// Lan::Lan
//
// Input:  various
//
// Output:
//
// Object constructor
//

Lan::Lan(const String &name, Simulation *si, int ma, double l, double bw) :
ScheduledObject(name, si), maxArp(ma), latency(l), bandwidth(bw) {

    expectedTime = 0.0;
    
    // Grab space for an ARP table
    arpMap = new ArpEntry[maxArp];
    for (int i = 0; i < maxArp; i++) {
	arpMap[i].ip.s_addr = 0;
    }
    
    intervalBytes = 0;
    intervalFrames = 0;
    totalBytes = 0;
    totalFrames = 0;
    
}

//
// Lan:~Lan
//
// Input:  None
//
// Output:  None
//
// Object destructor
//
Lan::~Lan() {
    
    *(sim->logStream) << name << " [Lan] frames " << totalFrames << " bytes " << totalBytes << "\n";
    
    delete [] arpMap;
}

//
// Lan::Handler
//
// (Event handler)
//
void Lan::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    switch (type) {

	//
	// FrameSend...outbound frame.
	//
	case (EventFrameSend):
	{

	    LanFrame *frame = (LanFrame *) data;

	    // Make sure we don't try to forward the frame before the
	    // previous frame's gone...compute time to deliver packet.
	    if (time < expectedTime) {
		time = expectedTime;
	    }
	    time += ((frame->length + LanHeaderSize) * 8) / bandwidth +
	      latency;

	    // Deliver the frame to the destination and send a completion
	    // event back to the sender.
	    Post(time, frame->dst, EventFrameArrival, data);
	    Post(time, frame->src, EventFrameTransmitComplete, NULL);

	    intervalFrames++;
	    intervalBytes += frame->length + LanHeaderSize;
	    totalFrames++;
	    totalBytes += frame->length + LanHeaderSize;

	    // Get ready for the next frame!
	    expectedTime = time;

	}
	break;

	//
	// EventTimer2...statistics printing timer.
	//
	case (EventTimer2):
	{

	    // Print out statistics over this interval
	    IF_DEBUG(0, {
		DEBUG_PREFIX("Lan");
		*(sim->debugStream) << "interval " << interval << " frames " << intervalFrames << " bytes " << intervalBytes << " utilization " << intervalBytes / (interval * bandwidth / 8) << "\n";
	    });

	    // Reset counters
	    intervalFrames = 0;
	    intervalBytes = 0;

	    // Wake us up again later
	    Post(time + interval, this, EventTimer2, NULL);
	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("Lan");
		cout << "bad EventType " << type << "\n";
	    });
	}
	break;

    }

}

//
// Lan::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String Lan::Output()
{
    return name + " [Lan]";
}

//
// LanCmd
//
// (Tcl command procedure)
//
int LanCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 5) {
	Tcl_SetResult(interp, "Usage:  Lan name maxarp latency bandwidth",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc;			// Tcl return code
	Lan *lan;		// new Lan object
	char *nm;		// name of new Lan object
	int ma;			// max size of arp table
	double l, bw;		// latency, bandwidth
	extern Simulation *sim;	// simulation object

	// get string arguments
	nm = argv[1];
	// get numeric arguments
	rc = Tcl_GetInt(interp, argv[2], &ma);
	if (rc == TCL_OK) {
	    rc = Tcl_GetDouble(interp, argv[3], &l);
	    if (rc == TCL_OK) {
		rc = Tcl_GetDouble(interp, argv[4], &bw);
		if (rc == TCL_OK) {
		    
		    // Make new object and make a Tcl command corresponding
		    // to the object.
		    lan = new Lan(nm, sim, ma, l, bw);
		    Tcl_CreateCommand(interp, nm,
				      LanObjectCmd,
				      (ClientData) lan,
				      LanDeleteProc);
		}
	    }
	}
	return rc;
    }
}

//
// LanObjectCmd
//
// (Tcl object command procedure)
//
int LanObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {
    Lan *lan = (Lan *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }
    else {

	//
	// lan statinterval time
	//
	if (!strcmp(argv[1], "statinterval")) {

	    if (argc == 3) {
		double inter;      // statistics printing interval
		int rc;		   // Tcl return code
		
		rc = Tcl_GetDouble(interp, argv[2], &inter);
		if (rc == TCL_OK) {
		    
		    lan->interval = inter;
		    lan->Post(0.0 + inter, lan, EventTimer2, NULL);

		}
		return rc;
	    }
	}

	//
	// lan arp ip-addr driver-name
	//
	else if (!strcmp(argv[1], "arp")) {

	    if (argc == 4) {

		struct InAddr ipAddr; // ip address
		String nm;	      // device driver name
		int rc = TCL_OK;      // return code
		
		// Get IP address
		ipAddr.s_addr = InetAddr(argv[2]);
		
		// Get ATM address
		nm = argv[3];

		if (rc == TCL_OK) {

		    int unusedEntry = -1;
		    bool done = false;
		    
		    // Insert entry into address mapping table.
		    // If the driver name is empty, we should delete
		    // the entry.  If it's non-null, replace an existing
		    // entry and reset the IP address.  Keep track
		    // of an empty entry in case this is a new
		    // arp entry.
		    for (int i = 0; i < lan->maxArp; i++) {
			if (lan->arpMap[i].ip.s_addr == ipAddr.s_addr) {
			    if (nm.length == 0) {
				lan->arpMap[i].ip.s_addr = 0;
			    }
			    else {
				lan->arpMap[i].objectName = nm;
				lan->arpMap[i].object = NULL;
			    }
			    done = true;
			    break;
			}
			else if ((lan->arpMap[i].ip.s_addr == 0) &&
				 (unusedEntry < 0)) {
			    unusedEntry = i;
			}
		    }

		    // If we couldn't find a matching entry, but we did
		    // find an unused one, make up an entry.  Otherwise,
		    // return an error.
		    if (!done) {
			if (unusedEntry >= 0) {
			    lan->arpMap[unusedEntry].ip.s_addr =
			      ipAddr.s_addr;
			    lan->arpMap[unusedEntry].objectName = nm;
			    lan->arpMap[unusedEntry].object = NULL;
			}
			else {
			    Tcl_SetResult(interp,
					  "couldn't find arp entry",
					  TCL_STATIC);
			    rc = TCL_ERROR;
			}
		    }
		    
		}
		return(rc);
		
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	else {
	    return(ScheduledObjectCmd(clientData, interp, argc, argv));
	}

    }
}

//
// LanDeleteProc
//
// (Tcl delete procedure)
//
void LanDeleteProc(ClientData clientData) {
    delete (Lan *) clientData;
}
