static char rcsid[] = "$Id: LanDeviceDriver.cc,v 1.10 1996-02-16 07:36:59-08 bmah Exp $";
//
// $Id: LanDeviceDriver.cc,v 1.10 1996-02-16 07:36:59-08 bmah Exp $
//
// LanDeviceDriver.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// LAN device driver class.
//

#include "LanDeviceDriver.h"
#include "IpPacket.h"
#include "Event.h"
#include "inline.h"

//
// LanDeviceDriver::Handler
//
// (Event handler)
//
void LanDeviceDriver::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    switch (type) {

	//
	// EventPacketSend:  Packet outbound to Lan
	//
	case (EventPacketSend):
	{

	    IpPacket *packet = (IpPacket *) data;

	    // If we already
	    // have the maximum number of frames queued up, drop the
	    // packet.
	    if (queued > maxQueued) {

		IF_DEBUG(1, {
		    DEBUG_PREFIX("LanDeviceDriver");
		    *(sim->debugStream) << "output queue overflow for " << packet << "\n";
		});

		delete packet;
		if_oerrors++;

	    }
	    else {
		
		// Need to find the address of the recipient (like consulting
		// the ARP cache get a datalink address).
		Lookup(down, downName);
		int maxArp = ((Lan *) down)->maxArp;
		ArpEntry *arpMap = ((Lan *) down)->arpMap;

		ScheduledObject *dst = NULL;
		for (int i = 0; i < maxArp; i++) {
		    if (packet->ip_next.s_addr ==  arpMap[i].ip.s_addr) {
			Lookup(arpMap[i].object, arpMap[i].objectName);
			dst = arpMap[i].object;
			break;
		    }
		}

		// If we couldn't get an object, that's an error
		if (dst == NULL) {

		    IF_DEBUG(0, {
			DEBUG_PREFIX("LanDeviceDriver");
			*(sim->debugStream) << "couldn't get linklayer destination for " << packet << "\n";
		    });

		    delete packet;
		    if_oerrors++;
		}
		else {
		    
		    // Build up the datalink frame.
		    LanFrame *frame = new LanFrame(packet);
		    frame->src = this;
		    frame->dst = dst;
		    frame->length = packet->ip_len;
		    
		    // Queue the packet on the circular queue.
		    queue[head] = frame;
		    queued++;
		    head = (head + 1) % maxQueued;
		    
		    // If nothing's queued up, we can send immediately
		    if (queued == 1) {
			Lookup(down, downName);
			Post(time + latency, down,
			     EventFrameSend, (void *) frame);
			if_opackets++;
			if_obytes += packet->ip_len;
		    }
		}
	    }
	}
	break;

	//
	// EventFrameTransmitComplete:  Frame just finished sending,
	// we can queue up another one.
	//
	case (EventFrameTransmitComplete):
	{

	    // We're done with this frame.  If there are any more left
	    // in our queue, grab the next one and post it to the net.
	    queued--;
	    tail = (tail + 1) % maxQueued;

	    if (queued) {
		Lookup(down, downName);
		Post(time + latency, down,
		     EventFrameSend, (void *) queue[tail]);

		if_opackets++;
		if_obytes += ((IpPacket *) queue[tail]->payload)->ip_len;
	    }

	}
	break;

	//
	// EventFrameArrival:  Lan frame arrived, pass it up to a
	// higher layer protocol for processing
	//
	case (EventFrameArrival):
	{
	    LanFrame *frame = (LanFrame *) data;
	    IpPacket *ipPacket = (IpPacket *) frame->payload;

	    if (ipPacket) {

		// Decrement the refcount and detatch the frame from the
		// IP packet.
		frame->Detatch();

		// Pass the new packet up to the IP layer
		Lookup(up, upName);
		Post(time + latency, up,
		     EventPacketArrival, (void *) ipPacket);
		
		// Update statistics
		if_ipackets++;
		if_ibytes += ipPacket->ip_len;
	    
	    }
	    else {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("LanDeviceDriver");
		    *(sim->debugStream) << "got null packet\n";
		});
	    }

	    // Throw away what's left of the frame
	    delete frame;
	}
	break;

	default:
	    IF_DEBUG(0, {
		DEBUG_PREFIX("LanDeviceDriver");
		cout << "bad EventType " << type << "\n";
	    });
	break;

    }

}

//
// LanDeviceDriver::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String LanDeviceDriver::Output()
{
    return name + " [LanDeviceDriver]";
}

//
// LanDeviceDriverCmd
//
// (Tcl command procedure)
//
int LanDeviceDriverCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 4) {
	Tcl_SetResult(interp,
		      "usage:  LanDeviceDriver name maxqueue latency",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc;			// Tcl return code
	char *nm;		// name of new LanDeviceDriver
	int mq;			// maximum queue length
	double l;		// driver latency
	LanDeviceDriver *ldd;	// new object
	extern Simulation *sim;	// simulation object

	// get string argument
	nm = argv[1];

	// get numeric arguments
	rc = Tcl_GetInt(interp, argv[2], &mq);
	if (rc == TCL_OK) {

	    rc = Tcl_GetDouble(interp, argv[3], &l);
	    if (rc == TCL_OK) {

		// Make new object and make a Tcl command corresponding
		// to the object.
		ldd = new LanDeviceDriver(nm, sim, mq, l);
		Tcl_CreateCommand(interp, nm,
				  LanDeviceDriverObjectCmd,
				  (ClientData) ldd,
				  LanDeviceDriverDeleteProc);
	    }
	}
	return rc;
    }
}

//
// LanDeviceDriverObjectCmd
//
// (Tcl object command procedure)
//
int LanDeviceDriverObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {
    return DeviceDriverObjectCmd(clientData, interp, argc, argv);
}

//
// LanDeviceDriverDeleteProc
//
// (Tcl delete procedure)
//
void LanDeviceDriverDeleteProc(ClientData clientData) {
    delete (LanDeviceDriver *) clientData;
}
