static char rcsid[] = "$Id: NntpServer.cc,v 1.2 1996-05-29 14:34:52-07 bmah Exp $";
//
// $Id: NntpServer.cc,v 1.2 1996-05-29 14:34:52-07 bmah Exp $
//   based on:
// $ID: SmtpClient.cc,v 1.1.2.1 1996-04-27 14:30:56-07 bmah Exp $
//   based on:
// $ID: SampleTcpClient.cc,v 1.9 1996-01-27 23:14:14-08 bmah Exp $
//
// NntpServer.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Simple NNTP server simulator.  This process connects to the NNTP
// port and sends a number of article-sized files.  Not really realistic
// because NNTP handshaking and bidirectional data transfer aren't
// simulated at all.  Probably good enough for background load though.
//

#include <String.h>
#include "NntpServer.h"
#include "EmpiricalDiscrete.h"
#include "Event.h"
#include "inline.h"

//
// Imported probability distributions
//
extern RNG *rng;		// global random number generator

// News article size (in bytes)
extern const DoubleDistElement *NntpItemSizeDistTable;
extern int NntpItemSizeDistLength;

// News batch size (in articles)
extern const DoubleDistElement *NntpNItemsDistTable;
extern int NntpNItemsDistLength;

//
// NntpServer::Handler
//
// (Event handler)
//
void
NntpServer::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalDiscrete NntpItemSizeDist(NntpItemSizeDistLength,
					      NntpItemSizeDistTable,
					      rng);

    static EmpiricalDiscrete NntpNItemsDist(NntpNItemsDistLength,
					    NntpNItemsDistTable,
					    rng);

    if ((state == AppTcpClientStateEstablishing) &&
	(type == EventTcpConnect)) {
	Post(time + gap, this, EventTimer, NULL);
	state = AppTcpClientStateEstablished;

	dataStartTime = time;
	count = (int) NntpNItemsDist();
	remaining = count;

	IF_DEBUG(2, {
	    DEBUG_PREFIX("NntpServer");
	    *(sim->debugStream) << count << " messages to send this batch\n";
	});

    }

    else if ((state == AppTcpClientStateEstablished) &&
	(type == EventTimer)) {

	if (remaining--) {
	    length = (int) NntpItemSizeDist();

	    IF_DEBUG(2, {
		DEBUG_PREFIX("NntpServer");
		*(sim->debugStream) << "send " <<
		  length << " byte message (" <<
		  remaining << " remaining)\n";
	    });

	    tcpModule->UserSend(time, tp, length);
	    Post(time + gap, this, EventTimer, NULL);
	}
	else {
	    tcpModule->UserClose(time, tp);
	    totalBytesSent = tp->snd_bytes;
	    totalBytesReceived = tp->rcv_bytes;
	    endTime = time;
	    state = AppTcpClientStateDown;
	    delete this;
	}
    }

    else {
	AppTcpClient::Handler(e);
    }

}

//
// NntpServer::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
NntpServer::Output()
{
    return name + " [NntpServer]";
}

//
// NntpServerCmd
//
// (Tcl command procedure)
//
int
NntpServerCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 7) {
	Tcl_SetResult(interp,
		      "Usage:  NntpServer name start tcp server port gap",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	NntpServer *ns;		// new client object
	char *nm;		// name of new client object
	double startTime;	// starting time to run
	String tcpObjectName;	// name of Tcp object
	InAddr server;		// server address to connect to
	int port;		// port number
	double g;		// gap
	extern Simulation *sim;	// simulation object

	nm = argv[1];
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    tcpObjectName = argv[3];
	    server.s_addr = InetAddr(argv[4]);
	    rc = Tcl_GetInt(interp, argv[5], &port);

	    if (rc == TCL_OK) {

		rc = Tcl_GetDouble(interp, argv[6], &g);

		if (rc == TCL_OK) {
			    
		    // Get pointer to TCP object (implication:  Tcp
		    // object *must* be created before any applications
		    // are created.
		    ns = new NntpServer(nm, sim, startTime,
					tcpObjectName,
					server, port, g);
		    Tcl_CreateCommand(interp, nm,
				      NntpServerObjectCmd,
				      (ClientData) ns,
				      NntpServerDeleteProc);
		}
	    }
	}

	return rc;
    }
}

//
// NntpServerObjectCmd
//
// (Tcl object command procedure)
//
int
NntpServerObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {
    return ScheduledObjectCmd(clientData, interp, argc, argv);
}

//
// NntpServerDeleteProc
//
// (Tcl delete procedure)
//
void
NntpServerDeleteProc(ClientData clientData) {
    delete (NntpServer *) clientData;
}


