static char rcsid[] = "$Id: ScheduledObject.cc,v 1.18 1996-02-22 18:53:00-08 bmah Exp $";
//
// $Id: ScheduledObject.cc,v 1.18 1996-02-22 18:53:00-08 bmah Exp $
//
// ScheduledObject.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// ScheduledObject class.  A ScheduledObject is any object that can
// receive events at times specified by the event scheduler.  All
// "real" ScheduledObjects (such as switch queues, switch controllers,
// IP modules, and so on) are subclasses of ScheduledObject.
//
#include "util.h"
#include "Tcl.h"
#include "ScheduledObject.h"
#include "String.ScheduledObjectPtr.CHMap.h"
#include "Calendar.h"
#include "Simulation.h"

//
// Object registry.
//
StringScheduledObjectPtrCHMap *(ScheduledObject::registry);

//
// Flush flag (used by debugging macros)
//
bool ScheduledObject::flushFlag;

//
// ScheduledObject::ScheduledObject
//
// Input:  name of object
//
// Output:  none
//
// Base class constructor.  Enter the new object into the object
// registry (its name must be unique).
//
ScheduledObject::ScheduledObject(const String &newName, Simulation *s) :
name(newName)
{
    //
    // First, check to see if the named object exists already.
    // If so, this is an error (no duplicates allowed).
    //
    if (registry->contains(newName)) {

	//
	// There really should be a more graceful way of dealing
	// with this.  One way would be to throw an exception.
	//
	panic("ScheduledObject duplicate");
    }

    (*registry)[name] = this;

    sim = s;			// save simulation object
    cache = true;		// always cache object references
    debug = 0;			// turn off debugging this object

}

//
// ScheduledObject::~ScheduledObject
//
// Input:  none
//
// Output:  none
//
// Base class constructor.  Delete object from the object
// registry.
//
ScheduledObject::~ScheduledObject()
{
    registry->del(name);
}

//
// ScheduledObject::Handler
//
// (Event handler)
//
// This is the default event handler.  It should never get called
// in normal operation...failure to override this method means that
// someone forgot to define a Handler for some class.  It causes
// a fatal run-time error.
//
void
ScheduledObject::Handler(Event *e)
{
    double time = e->ev_time;

    IF_DEBUG(0, {
	DEBUG_PREFIX("ScheduledObject");
	*(sim->debugStream) << "no Handler defined!";
    });
    exit(1);
}

//
// ScheduledObject::Output
//
// Input:  None
//
// Output:  String with ASCII data about object.
//
// Base class print method.
//
String ScheduledObject::Output()
{
    return ((String) name + " [unknown ScheduledObject] ");
}

//
// ScheduledObject::Post
//
// Input:  time, recipient of event (object), object type (type),
//         event data (data)
//
// Output:  None
//
// Post an object to the simulation object's scheduler.
//
void ScheduledObject::Post(double time, ScheduledObject *object,
				  EventType type, void *data)
{
    Event *e = new Event(time, object, type, data);
    e->ev_sender = this;
    sim->scheduler->Insert(e);
}

//
// ScheduledObject::Lookup
//
// Input:  Object pointer (theObject), object name (theName), fatal flag
//         (fatal, defaults to true)
//
// Output:  None
//
// Look up the name of an object.  If the cache flag for *this* object
// is set and the pointer is valid (non-NULL), then bypass the lookup.
// If the fatal flag is set and the lookup fails, then panic.
//
void ScheduledObject::Lookup(ScheduledObjectPtr &theObject,
			     const String &theName, bool fatal)
{
    if (!cache || !theObject) {
	theObject = (*registry)[theName];
	if (!theObject && fatal) {
	    panic("ScheduledObject::Lookup failure: " + name +
		  " tried to lookup " + theName);
	}
    }
}

// Glue to get around dogmatic type enforcing.
void
ScheduledObject::Lookup(Aal *&theObject, const String &theName,
			bool fatal) {
    ScheduledObject *temp = (ScheduledObject *) theObject;
    Lookup(temp, theName, fatal);
    theObject = (Aal *) temp;
};

void
ScheduledObject::Lookup(CellGoBackN *&theObject, const String &theName,
			bool fatal) {
    ScheduledObject *temp = (ScheduledObject *) theObject;
    Lookup(temp, theName, fatal);
    theObject = (CellGoBackN *) temp;
};

void
ScheduledObject::Lookup(DeviceDriver *&theObject, const String &theName,
			bool fatal) {
    ScheduledObject *temp = (ScheduledObject *) theObject;
    Lookup(temp, theName, fatal);
    theObject = (DeviceDriver *) temp;
};

void
ScheduledObject::Lookup(Ip *&theObject, const String &theName,
			bool fatal) {
    ScheduledObject *temp = (ScheduledObject *) theObject;
    Lookup(temp, theName, fatal);
    theObject = (Ip *) temp;
};

void
ScheduledObject::Lookup(SwitchModule *&theObject, const String &theName,
			bool fatal) {
    ScheduledObject *temp = (ScheduledObject *) theObject;
    Lookup(temp, theName, fatal);
    theObject = (SwitchModule *) temp;
};

void
ScheduledObject::Lookup(Tcp *&theObject, const String &theName,
			bool fatal) {
    ScheduledObject *temp = (ScheduledObject *) theObject;
    Lookup(temp, theName, fatal);
    theObject = (Tcp *) temp;
};

void
ScheduledObject::Lookup(Udp *&theObject, const String &theName,
			bool fatal) {
    ScheduledObject *temp = (ScheduledObject *) theObject;
    Lookup(temp, theName, fatal);
    theObject = (Udp *) temp;
};

//
// operator<<
//
// Input:  stream (s), ScheduledObject (o)
//
// Output:  stream (in return value)
//
// Call object-specific (possibly virtual) print method.
//
ostream& operator<<(ostream &s, ScheduledObject &o)
{
    return s << o.Output();
}

//
// ScheduledObjectCommand
//
// Input:  (as a Tcl comand handler)
//
// Output:
//
// Handle Tcl commands corresponding to (generically) one ScheduledObject.
// Note that this function is similar to a virtual function, except that
// it never gets called directly from a Tcl program (only indirectly
// through the command handlers corresponding to subclasses).
//
int ScheduledObjectCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[])
{
    
    // Calling an object's name (without giving a subcommand) is identical
    // to calling the object with a print subcommand.
    if (argc == 1) {
	String output = ((ScheduledObject *) clientData)->Output();
	Tcl_SetResult(interp, (char *) (const char *) output, TCL_VOLATILE);
	return(TCL_OK);
    }
    else {

	//
	// object print
	//
	// Invoke the object's Output method.
	//
	if (!strcmp(argv[1], "print")) {
	    String output = ((ScheduledObject *) clientData)->Output();
	    Tcl_SetResult(interp, (char *) (const char *) output, TCL_VOLATILE);
	    return(TCL_OK);
	}

	//
	// object debug flag
	//
	// Set debugging flag for this object
	//
	else if (!strcmp(argv[1], "debug")) {

	    if (argc == 3) {
		
		int rc;		// Tcl return value
		rc = Tcl_GetInt(interp, argv[2],
				&((ScheduledObject *) clientData)->debug);
		return rc;
		
	    }
	    else {
		Tcl_SetResult(interp, "bad # args", TCL_STATIC);
		return(TCL_OK);
	    }
	}

	//
	// object address
	//
	// Return the address of the object
	//
	else if (!strcmp(argv[1], "address")) {

	    sprintf(interp->result, "0x%x", clientData);
	    return(TCL_OK);

	}

	else {
	    Tcl_SetResult(interp, "bad subcommand", TCL_STATIC);
	    return(TCL_ERROR);
	}

    }

}
