static char rcsid[] = "$Id: Sig.cc,v 1.24 1996-05-29 14:34:54-07 bmah Exp $";
//
// $Id: Sig.cc,v 1.24 1996-05-29 14:34:54-07 bmah Exp $
//
// Sig.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Base class for ATM signalling entities, which implements the basic
// signalling protocol.  Subclasses of Sig implement scheduler-specific
// admission control tests.
//

#include "SigMessage.h"
#include "Sig.h"
#include "SwitchModule.h"
#include "Simulation.h"
#include "CellGoBackN.h"
#include "inline.h"

//
// Delay constants in seconds.  It's questionable whether or not these
// should be configurable, on a per-subclass or a per-object basis, but
// let's get something working first.
//
// Values chosen here are consistent with observed values of RCAP
// running on mid-range DECstations.  XUNET values will be somewhat
// higher, especially for the reverse pass delay.
//
static const double delayAdmission = 0.005; // 5 ms for forward pass
static const double delayCommit = 0.002;    // 2 ms for reverse pass
static const double delayRelease = 0.002;   // 2 ms for releasing resources
static const double delayOther = 0.001;	    // 1 ms for other messages

// State names
char *SigStateName[] = {
    "SigStateUnused",
    "SigStateEstablishing",
    "SigStateEstablished",
    "SigStateDeleting",
    "SigStateDeleted",
    "SigStateReserved"
};

char *SigModeName[] = {
    "work-conserving",
    "non-work-conserving",
    "no-rate-control",
};

//
// Sig::Sig
//
// Object constructor
//
Sig::Sig(const String &name, Simulation *si, int p, int mv, int mr) :
ScheduledObject(name, si), numPorts(p), maxVcids(mv), maxRoutes(mr) {

    int i, j;				// loop indices

    // Allocate memory for translation tables
    forward = new (SigTranslateEntry *)[numPorts];
    reverse = new (SigTranslateEntry *)[numPorts];
    state = new (SigStateEntry *)[numPorts];
    for (i = 0; i < numPorts; i++) {

	forward[i] = new (SigTranslateEntry)[maxVcids];
	reverse[i] = new (SigTranslateEntry)[maxVcids];
	state[i] = new (SigStateEntry)[maxVcids];

	// By convention, reserve VCIDs 0-15.
	for (j = 0; j < SigProtoReservedChannels; j++) {
	    reverse[i][j].port = 0;
	    forward[i][j].port = 0;
	    forward[i][j].vcid = 0;
	    state[i][j].state = SigStateReserved;
	}

	// Reflect signalling mapping (this isn't really necessary)
	forward[i][0].port = 0;
	forward[i][0].vcid = i;
	
	for (j = SigProtoReservedChannels; j < maxVcids; j++) {
	    reverse[i][j].port = -1;
	    state[i][j].state = SigStateUnused;
	}

    }

    // Allocate memory for output ports
    outputPortName = new (String)[numPorts];
    outputPort = new (ScheduledObject *)[numPorts];

    for (i = 0; i < numPorts; i++) {
	outputPort[i] = NULL;
    }

    // Allocate memory for routes and initialize
    routes = new (int)[maxRoutes];
    for (i = 0; i < maxRoutes; i++) {
	routes[i] = -1;
    }

    intervalCalls = 0;
    mode = SigModeNonWorkConserving;

}

//
// Sig::~Sig
//
// Object destructor
//
Sig::~Sig() {

    // Free up translation tables' memory.
    for (int i = 0; i < numPorts; i++) {
	delete [] forward[i];
	delete [] reverse[i];
	delete [] state[i];
    }
    delete [] forward;
    delete [] reverse;
    delete [] state;
    delete [] routes;
    delete [] outputPortName;
    delete [] outputPort;
}

//
// Sig::Handler
//
// (Event handler)
//
void Sig::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    SigTranslateEntry theVC;	// inbound VCI of VC in question
    SigTranslateEntry scratchVC; // scratch entry
    AtmCell *cell;		// inbound ATM cell, if any
    SigMessage *sigMessage;	// signalling message, if any

    // See what kind of event triggered the handler.  First check
    // for a signalling message.
    if (type == EventCellArrival) {

	cell = (AtmCell *) data;
	sigMessage = (SigMessage *) cell->payload;

	IF_DEBUG(3, {
	    DEBUG_PREFIX("Sig");
	    *(sim->debugStream) << "got message " << sigMessage << "\n";
	});

	// Look at the message type, use the message type to determine the
	// message direction.  We need to do this as a part of figuring out
	// which VC is being referred to, because we internally refer
	// to all connections by their upstream port/VCID.
	switch (sigMessage->type) {

	    // Forward-direction signalling, use incoming port and VCID
	    case (SigTypeEstablish):
	    case (SigTypeDelete):
	    
	    theVC.port = cell->vcid;
	    theVC.vcid = sigMessage->vcid;
	    break;

	    // Reverse-directon signalling message...reverse-translate
	    // the incoming port and VCID.
	    case (SigTypeEstablishAck):
	    case (SigTypeEstablishNak):

	    scratchVC.port = cell->port;
	    scratchVC.vcid = sigMessage->vcid;
	    theVC = reverse[scratchVC.port][scratchVC.vcid];
	    break;

	}

    }

    // Check for a timeout
    else if (type == EventTimer) {
    }
    // Statistics timer
    else if (type == EventTimer2) {

	// Print out statistics over this interval
	IF_DEBUG(0, {
	    DEBUG_PREFIX("Sig");
	    *(sim->debugStream) << "interval " << interval << " calls " << intervalCalls << "\n";
	});

	// Reset counters
	intervalCalls = 0;
	
	// Wake us up again later
	Post(time + interval, this, EventTimer2, NULL);

	// Premature return from this function
	return;

    }

    IF_DEBUG(3, {
	DEBUG_PREFIX("Sig");
	*(sim->debugStream) << "SigMessage carried VCID " << theVC.port <<
	  "." << theVC.vcid << "\n";
	DEBUG_PREFIX("Sig");
	*(sim->debugStream) << "VCID state " << (state[theVC.port][theVC.vcid]).state <<
	  "\n";
    });


    // Switch on the state of the VC, then handle individual message
    // types.  The VC state is keyed by the upstream port number and
    // VCID.
    switch ((state[theVC.port][theVC.vcid]).state) {

	// SigStateUnused:  VC currently not in use.  Note that this
	// state is the only one in which the upstream<->downstream
	// VCID mappings are not valid.
	case (SigStateUnused):

	switch (sigMessage->type) {

	    //
	    // Establish connection (from unused state)
	    //
	    case (SigTypeEstablish):
	    {
		SigReturnCode rc = SigOK;	// return code from admission tests
		SigTranslateEntry downstreamVC; // outgoing link and VCID

		// Compute next hop port number from routing table
		downstreamVC.port = routes[sigMessage->destinationAddress];
		if (downstreamVC.port == -1) {
		    rc = SigBadRoute; // no route to host
		}
		
		// Allocate VCID for next downstream hop.
		if (rc == SigOK) {
		    downstreamVC.vcid = -1;
		    for (int i = 0; i < maxVcids; i++) {
			if (reverse[downstreamVC.port][i].port < 0) {
			    downstreamVC.vcid = i;
			    break;
			}
		    }
		    if (downstreamVC.vcid < 0) {
			rc = SigNoVcids; // couldn't get any VCIDs, return an error
		    }
		}

		// Run admission control tests (we needed to do routing
		// first, in order that we knew what the outgoing link
		// was).  Note that admission control tests are run on
		// the downstream hop, but we keep track of connections
		// by their upstream hop.
		if (rc == SigOK) {
		    rc = AdmissionControl(time,
					  downstreamVC.port,
					  downstreamVC.vcid,
					  sigMessage,
					  sigMessage->traffic);
		}
		if (rc == SigOK) {
		    
		    // Set up association between the upstream and downstream sides
		    // of the channel, for signalling messages.  We don't
		    // do the switch setup for data delivery until the reverse
		    // pass.  Update channel state.
		    forward[theVC.port][theVC.vcid] = downstreamVC;
		    reverse[downstreamVC.port][downstreamVC.vcid] = theVC;

		    state[theVC.port][theVC.vcid].state = SigStateEstablishing;

		    // Send message on its way...
		    cell->vcid = downstreamVC.port;
		    sigMessage->vcid = downstreamVC.vcid;
		    sigMessage->hopcount++;

		    Lookup(theGoBackN, theGoBackNName);
		    Post(time + delayAdmission, theGoBackN, EventCellSend, cell);
		    
		    IF_DEBUG(3, {
			DEBUG_PREFIX("Sig");
			*(sim->debugStream) << "new channel picked VCID " <<
			  downstreamVC.port << "." << downstreamVC.vcid <<
			    "\n";
		    });

		}
		else {

		    // Send back a NAK.
		    sigMessage->type = SigTypeEstablishNak;
		    sigMessage->code = rc;
		    Lookup(theGoBackN, theGoBackNName);
		    Post(time + delayAdmission, theGoBackN, EventCellSend, cell);

		    IF_DEBUG(3, {
			DEBUG_PREFIX("Sig");
			*(sim->debugStream) << "return error " << rc << "\n";
		    });

		}

		// Increment statistics
		intervalCalls++;

		break;
	    }
	    
	    //
	    // By definition we shouldn't get any other kind of
	    // signalling message from this state.
	    //
	    default:
	    {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("Sig");
		    *(sim->debugStream) << 
		      "received illegal message type from SigStateUnused\n";
		});
	    }
	    break;

	}

	break;
		
	//
	// SigStateEstablishing:  VC being established, awaiting ACK
	// or NAK from downstream neighbor.
	//
	case (SigStateEstablishing):

	switch (sigMessage->type) {

	    //
	    // Establishment in progress for this channel was successful!
	    //
	    case (SigTypeEstablishAck):
	    {
		
		// Commit resources and set up forwarding in the switch
		CommitResources(time,
				cell->vcid, sigMessage->vcid,
				sigMessage->traffic);

		state[theVC.port][theVC.vcid].state = SigStateEstablished;
		Lookup(theSwitch, theSwitchName);
		theSwitch->Set(theVC.port, theVC.vcid,
			       cell->vcid, sigMessage->vcid);

		// Pass on the ACK for the establishment.
		cell->vcid = theVC.port;
		sigMessage->vcid = theVC.vcid;
		// Lookup(theGoBackN, theGoBackNName);
		Post(time + delayCommit, theGoBackN, EventCellSend, cell);

	    }
	    break;

	    //
	    // Establishment in progress met with an error of some
	    // kind downstream.
	    //
	    case (SigTypeEstablishNak):
	    {
		SigTranslateEntry downstreamVC;

		// Release tentatively-commited resources for the channel.
		ReleaseResources(time, cell->vcid, sigMessage->vcid);

		reverse[cell->vcid][sigMessage->vcid].port = -1;
		reverse[cell->vcid][sigMessage->vcid].vcid = -1;
		state[theVC.port][theVC.vcid].state = SigStateUnused;

		// Pass on the NAK.
		cell->vcid = theVC.port;
		sigMessage->vcid = theVC.vcid;
		
		Lookup(theGoBackN, theGoBackNName);
		Post(time + delayOther, theGoBackN, EventCellSend, cell);
	    }
	    break;

	    //
	    // Ignore (but flag) any other types of messages.
	    //
	    default:
	    {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("Sig");
		    *(sim->debugStream) << 
		      "received illegal message type from SigStateEstablishing\n";
		});
	    }
	    break;

	}


	break;
	
	//
	// SigStateEstablished:  VC established and active.
	//
	case (SigStateEstablished):

	switch (sigMessage->type) {

	    //
	    // Request to delete channel
	    //
	    case (SigTypeDelete):
	    {
		// Release all committed resources; note that because
		// we're an output queued switch we need to release
		// resources on the downstream hop (this means we need
		// to translate the upstream identifier)
		ReleaseResources(time,
				 forward[theVC.port][theVC.vcid].port,
				 forward[theVC.port][theVC.vcid].vcid);

		theSwitch->Unset(theVC.port, theVC.vcid);

		// Forward the delete message downstream.
		cell->vcid = forward[theVC.port][theVC.vcid].port;
		sigMessage->vcid = forward[theVC.port][theVC.vcid].vcid;

		state[theVC.port][theVC.vcid].state = SigStateUnused;
		reverse[forward[theVC.port][theVC.vcid].port][forward[theVC.port][theVC.vcid].vcid].port = -1;

		Lookup(theGoBackN, theGoBackNName);
		Post(time + delayRelease, theGoBackN, EventCellSend, cell);

	    }
	    break;

	    //
	    // Other message types constitute an error.
	    //
	    default:
	    {
		IF_DEBUG(0, {
		    DEBUG_PREFIX("Sig");
		    *(sim->debugStream) << 
		      "received illegal message type from SigStateEstablished\n";
		});
	    }
	    break;
	}

	break;

	//
	// Other (non-valid) state
	//
	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX("Sig");
	    *(sim->debugStream) << "in illegal state\n";
	});
	abort();
	break;
	
    }

}

//
// Sig::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String Sig::Output()
{

    String out;
    char buf[80];

    sprintf(buf, "%d", numPorts);
    out += name + " [Sig] mode " + SigModeName[mode] + " numPorts " + buf;

    sprintf(buf, "%d", maxVcids);
    out += (String) " maxVcis " + buf;

    sprintf(buf, "%d", maxRoutes);
    out += (String) " maxRoutes " + buf;

    return out;
    
}

//
// Sig::AdmissionControl
//
// Input:  traffic characteristics (traffic), outgoing link number (link)
//
// Output:  Return code
//
// Perform admission control tests for the appropriate outgoing link
// and given traffic descriptor.  If tests success (return code ==
// SigOK), make tentative resource reservations.
//
// This method should be overridden for use with different scheduling
// disciplines.
//
SigReturnCode Sig::AdmissionControl(double time,
				    int link, int vcid,
				    SigMessage *mesg, void *traffic)
{
    return SigOK;
}

//
// Sig::CommitResources
//
// Input:  traffic characteristics (traffic), outgoing link number (link)
// virtual circuit (vcid)
//
// Output:  Return code
//
// Commit reserved resources to a channel.
//
// This method should be overridden for use with different scheduling
// disciplines.
//
void Sig::CommitResources(double time,
			  int link, int vcid, void *traffic)
{
    return;
}

//
// Sig::ReleaseResources
//
// Input:  outgoing link number (link), virtual circuit ID (vcid)
//
// Output:  Return code
//
// Release resources allocated to a channel, both with respect to the
// signalling entity's accounting and the actual data transport entity's
// records.
//
// This method should be overridden for use with different scheduling
// disciplines.
//
void Sig::ReleaseResources(double time,
			   int link, int vcid)
{
    return;
}

//
// SigCmd
//
// (Tcl command handler procedure)
//
// Sig name numports maxvci maxroutes
//
int SigCmd(ClientData clientData, Tcl_Interp *interp,
		    int argc, char *argv[])
{

    if (argc != 5) {
	Tcl_SetResult(interp,
		      "Usage:  Sig name numports maxvci maxroute",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of new Sig
	int np;			// num ports
	int maxvci;		// maximum number of VCIs
	int maxroute;		// maximum number of routes supported
	int rc;			// return code from Tcl library
	Sig *sig;	// new switch module object
	extern Simulation *sim;	// global simulation object

	// get name
	nm = argv[1];

	// get numeric arguments
	rc = Tcl_GetInt(interp, argv[2], &np);
	if (rc == TCL_OK) {

	    rc = Tcl_GetInt(interp, argv[3], &maxvci);
	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[4], &maxroute);
		if (rc == TCL_OK) {

		    sig = new Sig(nm, sim, np, maxvci, maxroute);
		    Tcl_CreateCommand(interp, nm,
				      SigObjectCmd,
				      (ClientData) sig,
				      SigDeleteProc);
		    
		}

	    }

	}

	return rc;
	
    }

}

//
// SigObjectCmd
//
// (Tcl object handler procedure)
//
int SigObjectCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[])
{
    
    Sig *sig = (Sig *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }
    else {

	//
	// sig set input-port input-vcid output-port output-vcid
	//
	// Manipulate translation table
	//
	if (!strcmp(argv[1], "set")) {

	    if (argc == 6) {

		int inPort, inVcid, outPort, outVcid; // input parameters
		int rc;				      // Tcl return code
		
		rc = Tcl_GetInt(interp, argv[2], &inPort);
		if (rc == TCL_OK) {
		    
		    if ((inPort < 0) || (inPort >= sig->numPorts)) {
			Tcl_SetResult(interp, "input port out of bounds",
				      TCL_STATIC);
			return(TCL_ERROR);
		    }
		    
		    rc = Tcl_GetInt(interp, argv[3], &inVcid);
		    if (rc == TCL_OK) {

			if ((inVcid < 0) || (inVcid >= sig->maxVcids)) {
			    Tcl_SetResult(interp, "input VCID out of bounds",
					  TCL_STATIC);
			    return(TCL_ERROR);
			}
		    
			rc = Tcl_GetInt(interp, argv[4], &outPort);
			if (rc == TCL_OK) {

			    if ((outPort < 0) || (outPort >= sig->numPorts)) {
				Tcl_SetResult(interp,
					      "output port out of bounds",
					      TCL_STATIC);
				return(TCL_ERROR);
			    }
		    
			    rc = Tcl_GetInt(interp, argv[5], &outVcid);
			    if (rc == TCL_OK) {

				if ((outVcid < 0) || (outVcid >= sig->maxVcids)) {
				    Tcl_SetResult(interp,
						  "output VCID out of bounds",
						  TCL_STATIC);
				    return(TCL_ERROR);
				}
		    
				// got args, set translation entry
				sig->forward[inPort][inVcid].port = outPort;
				sig->forward[inPort][inVcid].vcid = outVcid;
				sig->reverse[outPort][outVcid].port = inPort;
				sig->reverse[outPort][outVcid].vcid = inVcid;

			    }
			}
		    }
		}
		return (rc);

	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
		
	}

	//
	// sig route dest port
	//
	// Set routing table for signalling software
	else if (!strcmp(argv[1], "route")) {

	    if (argc == 4) {

		int dest, port;                       // input parameters
		int rc;				      // Tcl return code
		
		rc = Tcl_GetInt(interp, argv[2], &dest);
		if (rc == TCL_OK) {
		    
		    if ((dest < 0) || (dest >= sig->maxRoutes)) {
			Tcl_SetResult(interp, "dest out of bounds",
				      TCL_STATIC);
			return(TCL_ERROR);
		    }
		    
		    rc = Tcl_GetInt(interp, argv[3], &port);
		    if (rc == TCL_OK) {

			if ((port < 0) || (port >= sig->numPorts)) {
			    Tcl_SetResult(interp, "port out of bounds",
					  TCL_STATIC);
			    return(TCL_ERROR);
			}
			
			// got args, set translation entry
			sig->routes[dest] = port;

		    }

		}
		return (rc);

	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
		
	}

	//
	// sig switch object
	//
	// Associate signalling entity with its switch module
	//
	else if (!strcmp(argv[1], "switch")) {

	    if (argc == 3) {

		String outName;	// object name

		outName = argv[2];
		
		// set new object...zero out cached object pointer
		sig->theSwitchName = outName;
		sig->theSwitch = NULL;

		return(TCL_OK);
		
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// sig gobackn object
	//
	// Associate signalling entity with its reliability transport layer
	//
	else if (!strcmp(argv[1], "gobackn")) {

	    if (argc == 3) {

		String outName;	// object name

		outName = argv[2];
		
		// set new object...zero out cached object pointer
		sig->theGoBackNName = outName;
		sig->theGoBackN = NULL;

		return(TCL_OK);
		
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// sig output port object
	//
	// Set pointer to output port object
	//
	else if (!strcmp(argv[1], "output")) {

	    if (argc == 4) {

		String portName;                      // output port name
		int port;	                      // port number
		int rc;				      // Tcl return code
		
		rc = Tcl_GetInt(interp, argv[2], &port);
		if (rc == TCL_OK) {
		    
		    if ((port < 0) || (port >= sig->numPorts)) {
			Tcl_SetResult(interp, "port out of bounds",
				      TCL_STATIC);
			return(TCL_ERROR);
		    }
		    
		    portName = argv[3];
			
		    // got args, set output port
		    sig->outputPortName[port] = portName;

		}
		return (rc);

	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
		
	}

	//
	// sighost netstat
	//
	else if (!strcmp(argv[1], "netstat")) {

	    if (argc > 2) {

		if (!strcmp(argv[2], "-v")) {
		    int i, j;
		    for (i = 0; i < sig->numPorts; i++) {
			for (j = 0; j < sig->maxVcids; j++) {
			    if (sig->state[i][j].state != SigStateUnused) {
				cout << i << "." << j << " --> " <<
				  sig->forward[i][j].port << "." <<
				    sig->forward[i][j].vcid << "\t" << 
				      sig->state[i][j].state << "\n";
			    }
			}
		    }
		    return (TCL_OK);
		}
		else {
		    Tcl_SetResult(interp, "bad netstat subcommand",
				  TCL_STATIC);
		    return(TCL_ERROR);
		}

	    }
	    else {

		int i, j;
		for (i = 0; i < sig->numPorts; i++) {
		    int count = 0;
		    for (j = 0; j < sig->maxVcids; j++) {
			if (sig->state[i][j].state != SigStateUnused) {
			    count++;
			}
		    }
		    cout << "port " << i << ": " << count << "\n";
		}
		return (TCL_OK);

	    }
	}

	//
	// sig statinterval time
	//
	else if (!strcmp(argv[1], "statinterval")) {

	    if (argc == 3) {
		double inter;      // statistics printing interval
		int rc;		   // Tcl return code
		
		rc = Tcl_GetDouble(interp, argv[2], &inter);
		if (rc == TCL_OK) {
		    
		    sig->interval = inter;
		    sig->Post(0.0 + inter, sig, EventTimer2, NULL);

		}
		return rc;
	    }
	}

	// sig mode mode
	else if (!strcmp(argv[1], "mode")) {

	    if (argc == 3) {
		if (!strcmp(argv[2], "work-conserving")) {
		    sig->mode = SigModeWorkConserving;
		    return(TCL_OK);
		}
		else if (!strcmp(argv[2], "non-work-conserving")) {
		    sig->mode = SigModeNonWorkConserving;
		    return(TCL_OK);
		}
		else if (!strcmp(argv[2], "no-rate-control")) {
		    sig->mode = SigModeNoRateControl;
		    return(TCL_OK);
		}
		else {
		    Tcl_SetResult(interp, "bad mode", TCL_STATIC);
		    return(TCL_ERROR);
		}
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	else {
	    return(ScheduledObjectCmd(clientData, interp, argc, argv));
	}
	
    }
}

//
// SigDeleteProc
//
// (Tcl DeleteProc)
//
void SigDeleteProc(ClientData clientData)
{
    delete ((Sig *) clientData);
}

//
// operator<<
//
// Input:  Output stream (s), state (ss)
//
// Output:  Output stream
//
// Print the name of the SigHost state shs on the output stream s.
//
ostream& operator<<(ostream &s, SigState ss)
{
    char *output;
    
    output = SigStateName[ss];

    return s << output;

}

//
// operator<<
//
// Input:  Output stream (s), queueing mode (mode)
//
// Output:  Output stream
//
// Print the name of the mode (mode) on the output stream s.
//
ostream& operator<<(ostream &s, SigMode mode)
{
    char *output;
    
    output = SigModeName[mode];

    return s << output;

}
