static char rcsid[] = "$Id: SigHost.cc,v 1.27 1996-05-29 14:34:57-07 bmah Exp $";
//
// $Id: SigHost.cc,v 1.27 1996-05-29 14:34:57-07 bmah Exp $
//
// SigHost.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Signalling entity for ATM-attached host.
//

#include "SigMessage.h"
#include "apiproto.h"
#include "SigHost.h"
#include "SwitchModule.h"
#include "Simulation.h"
#include "AtmCell.h"
#include "CellGoBackN.h"
#include "inline.h"

//
// Delay constants in seconds.  Ideally these would be configurable
// on a per-controller basis, rather than hard-coded as done here.
//
// Values chosen here are consistent with observed values of RCAP
// running on mid-range DECstations.  XUNET values will be somewhat
// higher, especially for the reverse pass delay.
//
static const double delayAdmission = 0.005; // 5 ms for forward pass
static const double delayCommit = 0.002;    // 2 ms for reverse pass
static const double delayRelease = 0.002;   // 2 ms for releasing resources
static const double delayOther = 0.001;	    // 1 ms for other messages

// Names of states
char *SigHostStateName[] = {
    "SigHostStateUnused",
    "SigHostStateEstablishing",
    "SigHostStateEstablished",
    "SigHostStateDeleting",
    "SigHostStateDeleted",
    "SigHostStateReserved"
};

char *SigHostModeName[] = {
    "work-conserving",
    "non-work-conserving",
    "no-rate-control",
};

//
// SigHost::SigHost
//
// Input:  Name of new SigHost object (name), simulation object pointer (si),
// maximum number of VCIDs supported (mv), maximum number of downcalls
// supported (down), ATM address (addr)
//
// Output:  None
//
// Object constructor.
//
SigHost::SigHost(const String &name, Simulation *si, int mv, int down,
		 int addr) :
ScheduledObject(name, si), maxVcids(mv),
maxDownCalls(down), address(addr) {

    int i, j;			// loop indices

    // Allocate memory for channel state and downcall pointers
    state = new (SigHostStateEntry *)[sighostPorts];
    for (i = 0; i < sighostPorts; i++) {
	state[i] = new (SigHostStateEntry[maxVcids]);
	
	for (j = 0; j < SigProtoReservedChannels; j++) {
	    state[i][j].state = SigHostStateReserved;
	}
	for (j = SigProtoReservedChannels; j < maxVcids; j++) {
	    state[i][j].state = SigHostStateUnused;
	}
    }

    downCalls = new(ApiMessage *)[maxDownCalls];
    for (i = 0; i < maxDownCalls; i++) {
	downCalls[i] = NULL;
    }

    // Zero out object pointers (we'll look them up later)
    networkPort = NULL;
    aal = NULL;
    theSwitch = NULL;

    intervalCalls = 0;
    mode = SigHostModeNonWorkConserving;

}

//
// SigHost::~SigHost
//
// Input:  None
//
// Output:  None
//
// Object destructor.
//
SigHost::~SigHost() {

    int i;
    
    delete [] downCalls;
    for (i = 0; i < sighostPorts; i++) {
	delete [] state[i];
    }
    delete [] state;

}

//
// SigHost::Handler
//
// (Event handler)
//
void SigHost::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;
    
    AtmCell *cell;
    SigMessage *sigMessage;
    ApiMessage *apiMessage;

    int port;
    int vcid;

    // Figure out which virtual circuit is being referred to, so we
    // can check its state.  To do this, we look both at the event
    // type and the data in the message.
    if (type == EventCall) {

	// Arrival of a call from a host application.
	apiMessage = (ApiMessage *) data;

	// All calls refer to a channel going in its forward direction.
	port = sighostPortHost;

	// If the call was to establish a new virtual circuit,
	// we need to grab a new VCID.  Otherwise, we will be
	// working with the one passed in the message.
	if (apiMessage->type == ApiTypeEstablish) {

	    // Try to find an unused VCID for the new virtual
	    // circuit.  If we can't find one, then return an
	    // error to the application (mutate the existing
	    // message, kick it back to the app, and return).
	    vcid = -1;
	    for (int i = 0; i < maxVcids; i++) {
		if ((state[port][i]).state == SigHostStateUnused) {
		    vcid = i;
		    break;
		}
	    }
	    if (vcid == -1) {
		apiMessage->type = ApiTypeReturn;
		apiMessage->code = ApiNoVcids;
		Post(time, apiMessage->client, EventCall, data);
		return;
	    }
	    
	    IF_DEBUG(3, {
		DEBUG_PREFIX("SigHost");
		*(sim->debugStream) << 	
		  "ApiTypeEstablish picked VCID " << vcid << "\n";
	    });

	}
	else {
	    vcid = apiMessage->vcid;

	    IF_DEBUG(3, {
		DEBUG_PREFIX("SigHost");
		*(sim->debugStream) << 	
		  "ApiMessage type " <<
		  ApiMessageTypeString(apiMessage->type) <<
		  " carried VCID " << vcid << "\n";
	    });
	}

    }
    else if (type == EventCellArrival) {

	// Arrival of a signalling cell from the network.  Use
	// the VCID to figure out what VC we're referring to.
	cell = (AtmCell *) data;
	sigMessage = (SigMessage *) cell->payload;
	vcid = sigMessage->vcid;

	IF_DEBUG(3, {
	    DEBUG_PREFIX("SigHost");
	    *(sim->debugStream) <<
	      "got signalling message " << sigMessage << "\n";
	});

	switch (sigMessage->type) {

	    case SigTypeEstablish:
	    case SigTypeDelete:
	    port = cell->vcid;
	    break;

	    case SigTypeEstablishAck:
	    case SigTypeEstablishNak:
	    port = sighostPorts - cell->vcid;
	    break;
	    
	    default:
	    IF_DEBUG(0, {
		DEBUG_PREFIX("SigHost");
		*(sim->debugStream) << "got unrecognized message type\n";
	    });
	    break;

	}

    }

    // Statistics timer
    else if (type == EventTimer2) {

	// Print out statistics over this interval
	IF_DEBUG(0, {
	    DEBUG_PREFIX("SigHost");
	    *(sim->debugStream) << "interval " << interval << " calls " << intervalCalls << "\n";
	});

	// Reset counters
	intervalCalls = 0;
	
	// Wake us up again later
	Post(time + interval, this, EventTimer2, NULL);

	// Premature return from this function
	return;

    }


    IF_DEBUG(3, {
	DEBUG_PREFIX("SigHost");
	*(sim->debugStream) << 
	  "VCID[" << port << "][" << vcid <<
	  "] state " << state[port][vcid].state << " event type " <<
	  type <<"\n";
    });

    // Now look at the state of the VC in question, and process the
    // incoming messages appropriately.
    switch ((state[port][vcid]).state) {

	//
	// Establish connections (from unused state).  This happens
	// either by a call from the host (outgoing call) or an
	// establishment message from the network (call with
	// destination on this host).
	//
	case (SigHostStateUnused):

	// See if the attached host wanted to initiate a connection
	// (downcall with type of Establish).
	if ((type == EventCall) &&
	    (apiMessage->type == ApiTypeEstablish)) {

	    SigReturnCode rc = SigOK;
	    
	    // Save downcall's ID...we'll need it later to generate
	    // a reply (note that this operation can fail due to a lack
	    // of downcall descriptors)
	    int dc; 			// saved downcall slot number
	    apiMessage->vcid = vcid;
	    if (rc == SigOK) {
		rc = ApiNoDownCalls;
		for (int i = 0; i < maxDownCalls; i++) {
		    if (downCalls[i] == NULL) {
			rc = SigOK;
			dc = i;
			break;
		    }
		}
	    }
	    
	    // Run admission tets
	    if (rc == SigOK) {
		rc = AdmissionControl(time, vcid, apiMessage->traffic);
	    }
	    
	    IF_DEBUG(3, {
		DEBUG_PREFIX("SigHost");
		*(sim->debugStream) << " admission tests returned " <<
		  rc << "\n";
	    });

	    if (rc == SigOK) {

		// Kick off a channel establishment message.
		SigMessage *sigMessage = new SigMessage;

		sigMessage->type = SigTypeEstablish;
		sigMessage->vcid = vcid;
		sigMessage->code = SigOK;
		sigMessage->sourceAddress = apiMessage->sourceAddress;
		sigMessage->sourceService = apiMessage->sourceService;
		sigMessage->destinationAddress =
		  apiMessage->destinationAddress;
		sigMessage->destinationService =
		  apiMessage->destinationService;
		sigMessage->hopcount = 1;
		sigMessage->traffic = new char[apiMessage->trafficLength];
		bcopy((char *) apiMessage->traffic,
		      (char *) sigMessage->traffic,
		      apiMessage->trafficLength);
		sigMessage->trafficLength = apiMessage->trafficLength;

		// A little cleanup before the message goes out
		if (apiMessage->sourceAddress == 0) {
		    apiMessage->sourceAddress = address;
		}
		
		AtmCell *cell = new AtmCell(AtmCellTypeData,
					    sigMessage,
					    0, 0);
		cell->vcid = sighostPortNet; // set VCID for this cell to
					     // route it towards the net
		cell->port = sighostPortSigHost; // came from signalling port

		Lookup(theSwitch, theSwitchName);
		Post(time + delayAdmission, (ScheduledObject *) theSwitch,
		     EventCellSend, (void *) cell);

		state[port][vcid].state = SigHostStateEstablishing;
		downCalls[dc] = apiMessage;

		intervalCalls++;

		IF_DEBUG(4, {
		    DEBUG_PREFIX("SigHost");
		    *(sim->debugStream) << "registered downcall " <<
		      dc << " from " <<
		      downCalls[dc]->client->name << "\n";
		});

	    }
	    else {

		// Return from downcall with an error message.
		apiMessage->type = ApiTypeReturn;
		apiMessage->code = (SigReturnCode) rc;
		Post(time + delayAdmission, apiMessage->client,
		     EventCall, (void *) apiMessage);

		IF_DEBUG(4, {
		    DEBUG_PREFIX("SigHost");
		    *(sim->debugStream) << "channel reject " <<
		      rc << " notify " <<
			apiMessage->client->name << "\n";
		});
	    }
	}

	// See if an establishment message came from the network.  This
	// means that the destination of the connection should be this
	// host.
	else if ((type == EventCellArrival) &&
		 (sigMessage->type == SigTypeEstablish)) {

	    SigReturnCode rc = SigOK;	// return code

	    // First make sure the message reached the proper
	    // destination.  If not, send a rejection message
	    // back towards the source.
	    if (sigMessage->destinationAddress != address) {
		rc = SigBadDest;
	    }

	    // Check the service type.  At present, only an IP
	    // service is supported.
	    if (rc == SigOK) {
		if (sigMessage->destinationService != SigServiceIP) {
		    rc = SigBadService;
		}
	    }

	    // Set state of higher-layer protocols
	    Lookup(aal, aalName);
	    aal->ResetIn(vcid, 0);

	    // For valid connections (as defined by the above tests)
	    // set the connection state and send back an acknowledgement.
	    if (rc == SigOK) {
		state[port][vcid].state = SigHostStateEstablished;
		sigMessage->type = SigTypeEstablishAck;
		
		Lookup(theSwitch, theSwitchName);
		Post(time + delayCommit, (ScheduledObject *) theSwitch,
		     EventCellSend, (void *) cell);
	    }

	    // Otherwise, if there was an error of some kind, plug that
	    // back into the message, and return a NAK.
	    else {
		sigMessage->type = SigTypeEstablishNak;
		sigMessage->code = (SigReturnCode) rc;

		Lookup(theSwitch, theSwitchName);
		Post(time + delayOther, (ScheduledObject *) theSwitch,
		     EventCellSend, (void *) cell);
	    }
	    
	}
	
	// Shouldn't get down here.
	else {
	    IF_DEBUG(0, {
		DEBUG_PREFIX("SigHost");
		*(sim->debugStream) << 	
		  "received illegal message type from SigHostStateUnused\n";
	    });
	}

	break;
	
	//
	// Currently establishing this connection from the host.
	// We're expecting some result (either positive or negative)
	// from the network.
	//
	case (SigHostStateEstablishing):

	// See if the network accepted the connection.  If so,
	// go to the connected state.
	if ((type == EventCellArrival) &&
	    (sigMessage->type == SigTypeEstablishAck)) {

	    // Commit resources.  They should already have been reserved
	    // by the forward pass by an AdmissionControl() call.
	    CommitResources(time, vcid, sigMessage->traffic);

	    // Notify of new outgoing channel
	    Lookup(aal, aalName);
	    aal->ResetOut(vcid, 0);

	    state[port][vcid].state = SigHostStateEstablished;

	    // Inform client (retrieve the downcall ApiMessage and
	    // post the return back to the client)
	    for (int i = 0; i < maxDownCalls; i++) {
		if ((downCalls[i]) && (downCalls[i]->vcid == vcid)) {
		    downCalls[i]->type = ApiTypeReturn;
		    downCalls[i]->code = SigOK;

		    Post(time + delayCommit, downCalls[i]->client,
			 EventCall, (void *) downCalls[i]);

		    IF_DEBUG(4, {
			DEBUG_PREFIX("SigHost");
			*(sim->debugStream) << "channel accepted notify " <<
			  downCalls[i]->client->name << " via downcall " <<
			  i << "\n";
		    });

		    downCalls[i] = NULL;
		    break;
		}
	    }

	    // Deallocate signalling message
	    delete cell;

	}

	// Otherwise, the network rejected the connection.
	else if ((type == EventCellArrival) &&
		 (sigMessage->type == SigTypeEstablishNak)) {

	    // Deallocate resources
	    ReleaseResources(time, vcid);

	    state[port][vcid].state = SigHostStateUnused;

	    // Inform client (retrieve downcall ApiMessage,
	    // as above)
	    for (int i = 0; i < maxDownCalls; i++) {
		if ((downCalls[i]) && (downCalls[i]->vcid == vcid)) {
		    downCalls[i]->type = ApiTypeReturn;
		    downCalls[i]->code = (SigReturnCode) sigMessage->code;

		    Post(time + delayRelease, downCalls[i]->client,
			 EventCall, (void *) downCalls[i]);

		    IF_DEBUG(4, {
			DEBUG_PREFIX("SigHost");
			*(sim->debugStream) << "channel reject " <<
			  sigMessage->code << " notify " <<
			  downCalls[i]->client->name << " via downcall " <<
			  i << "\n";
		    });
		    downCalls[i] = NULL;
		    break;
		}
	    }

	    // Deallocate signalling message
	    delete cell;

	}

	// Shouldn't get here...
	else {
	    IF_DEBUG(0, {
		DEBUG_PREFIX("SigHost");
		*(sim->debugStream) <<
		  "received illegal message type from SigHostStateEstablishing\n";
	    });
	}
	
	break;
	
	//
	// Channel currently established.  Right now, the only thing
	// that can happen is that someone wants to tear down the
	// channel.
	//
        // A potential problem here is that a Delete message should
	// only come from the network towards the destination of a
	// connection and a Delete calldown should only be done on the
	// source endpoint of a connection.  Currently we don't check for
	// this, but it should be pretty easy to do.
	//
	case (SigHostStateEstablished):

	// Check for a Delete downcall
	if ((type == EventCall) &&
	    (apiMessage->type == ApiTypeDelete)) {

	    // Release outgoing resources
	    ReleaseResources(time, vcid);

	    // Compose a new delete message to send out to the network
	    SigMessage *sigMessage = new SigMessage;
	    sigMessage->type = SigTypeDelete;
	    sigMessage->vcid = vcid;
	    sigMessage->code = SigOK;
	    sigMessage->sourceAddress = 0;
	    sigMessage->sourceService = SigServiceNone;
	    sigMessage->destinationAddress = 0;
	    sigMessage->destinationService = SigServiceNone;
	    sigMessage->trafficLength = 0;
	    sigMessage->traffic = NULL;

	    AtmCell *cell = new AtmCell(AtmCellTypeData,
					sigMessage,
					0, 0);
	    cell->vcid = sighostPortNet; // set VCID for this cell to route it
					 // out the network port
	    cell->port = sighostPortSigHost; // came from signalling port

	    Lookup(theSwitch, theSwitchName);
	    Post(time + delayRelease, (ScheduledObject *) theSwitch,
		 EventCellSend, (void *) cell);

	    // With our current network model, it's impossible to fail,
	    // so return the downcall.

	    apiMessage->type = ApiTypeReturn;
	    apiMessage->code = SigOK;
	    Post(time + delayRelease + delayOther, apiMessage->client,
		 EventCall, (void *) apiMessage);
	    
	    state[port][vcid].state = SigHostStateUnused;

	}

	// Now check for a Delete message from the network
	else if ((type == EventCellArrival) &&
		 (sigMessage->type == SigTypeDelete)) {

	    // Set state of channel, do notification to client?
	    state[port][vcid].state = SigHostStateUnused;

	    if (sigMessage->traffic) {
		delete sigMessage->traffic;
	    }
//	    delete sigMessage;
	    delete cell;

	}

	// Shouldn't get down here...
	else {
	    IF_DEBUG(0, {
		DEBUG_PREFIX("SigHost");
		*(sim->debugStream) << 	
		  "received illegal message type from SigHostStateEstablished\n";
	    });
	}
	
	break;
	
	//
	// Shouldn't get here.
	//
	default:
	IF_DEBUG(0, {
	    DEBUG_PREFIX("SigHost");
	    *(sim->debugStream) <<
	      "in illegal state\n";
	});
	abort();
	break;
    }

}

//
// SigHost::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String SigHost::Output()
{

    String out;
    char buf[80];

    sprintf(buf, "%d", maxVcids);
    out += name + " [SigHost] mode " + SigHostModeName[mode] +
      " maxVcids " + buf;

    sprintf(buf, "%d", maxDownCalls);
    out += (String) " maxDownCalls " + buf;

    sprintf(buf, "%d", address);
    out += (String) " address " + buf;

    out += (String) " aal " + aalName;
    
    return out;

}

//
// SigHost::AdmissionControl
//
// Input:  VCID (vcid), traffic characteristics (traffic)
//
// Output:  status (in return value)
//
// Run admission control tests for the outgoing link and the given
// traffic descriptor.  If tests are successful, make tentative
// resource reservations.
//
// This method will be overridden for use with various scheduling
// disciplines.
//
SigReturnCode SigHost::AdmissionControl(double time,
					int vcid, void *traffic)
{
    return SigOK;
}

//
// SigHost::CommitResources
//
// Input:  VCID (vcid), traffic characteristics (traffic)
//
// Output:  status (in return value)
//
// Commit resources for the specified VCID.
//
// This method will be overridden for use with various scheduling
// disciplines.
//
void SigHost::CommitResources(double time,
			      int vcid, void *traffic)
{
    return;
}

//
// SigHost::ReleaseResources
//
// Input:  VCID (vcid)
//
// Output:  status (in return value)
//
// Release resources for the specified outgoing VC.
//
// This method will be overridden for use with various scheduling
// disciplines.
//
void SigHost::ReleaseResources(double time,
			       int vcid)
{
    return;
}

//
// SigHostCmd
//
// (Tcl command handler procedure)
//
// SigHost name maxVcids downCalls address
//
int SigHostCmd(ClientData clientData, Tcl_Interp *interp,
	       int argc, char *argv[])
{

    if (argc != 5) {
	Tcl_SetResult(interp,
		      "usage:  SigHost name maxvcid downcalls address",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of new SigHost object
	int mv;			// maximum number of VCIs
	int dc;			// maximum number of concurrent
				// signalling downcalls from clients
	int addr;		// ATM address
	int rc;			// return code from Tcl library
	SigHost *sigHost;	// pointer to new SigHost object
	extern Simulation *sim;	// global simualtion object

	// get name
	nm = argv[1];

	// get three numeric argments into mv, dc, addr
	rc = Tcl_GetInt(interp, argv[2], &mv);
	if (rc == TCL_OK) {

	    rc = Tcl_GetInt(interp, argv[3], &dc);
	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[4], &addr);
		if ((rc == TCL_OK) && (addr < 1)) {
		    Tcl_SetResult(interp, "address out of range", TCL_STATIC);
		    rc = TCL_ERROR;
		}
		if (rc == TCL_OK) {

		    sigHost = new SigHost(nm, sim, mv, dc, addr);
		    Tcl_CreateCommand(interp, nm,
				      SigHostObjectCmd,
				      (ClientData) sigHost,
				      SigHostDeleteProc);

		}
	    }
	}

	return rc;

    }

}

//
// SigHostObjectCmd
//
// (Tcl object handler procedure)
//
int SigHostObjectCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{

    SigHost *sigHost = (SigHost *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }
    else {

	//
	// sighost net net-name
	//
	if (!strcmp(argv[1], "net")) {

	    if (argc == 3) {
		String on;	// object name

		on = argv[2];
		sigHost->networkPortName = on;
		sigHost->networkPort = NULL;
		return TCL_OK;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	//
	// sighost aal aal-name
	//
	else if (!strcmp(argv[1], "aal")) {
	    if (argc == 3) {
		String on;	// object name

		on = argv[2];
		sigHost->aalName = on;
		sigHost->aal = NULL;
		return TCL_OK;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	//
	// sighost switch switchmod-name
	//
	else if (!strcmp(argv[1], "switch")) {
	    if (argc == 3) {
		String on;	// object name

		on = argv[2];
		sigHost->theSwitchName = on;
		sigHost->theSwitch = NULL;
		return TCL_OK;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	//
	// sighost netstat
	//
	else if (!strcmp(argv[1], "netstat")) {

	    if (argc > 2) {

		if (!strcmp(argv[2], "-v")) {

		    int i, j;
		    for (i = 0; i < sighostPorts; i++) {
			for (j = 0; j < sigHost->maxVcids; j++) {
			    if (sigHost->state[i][j].state != SigHostStateUnused) {
				cout << i << "." << j << "\t" <<
				  sigHost->state[i][j].state << "\n";
			    }
			}
		    }
		    return (TCL_OK);
		}
		else {
		    Tcl_SetResult(interp, "bad netstat subcommand",
				  TCL_STATIC);
		    return(TCL_ERROR);
		}
	    }
	    else {
		int i, j;
		for (i = 0; i < sighostPorts; i++) {
		    int count = 0;
		    for (j = 0; j < sigHost->maxVcids; j++) {
			if (sigHost->state[i][j].state != SigHostStateUnused) {
			    count++;
			}
		    }
		    cout << "port " << i << ": " << count << "\n";
		}
		return (TCL_OK);
	    }
	}

	//
	// sighost statinterval time
	//
	else if (!strcmp(argv[1], "statinterval")) {

	    if (argc == 3) {
		double inter;      // statistics printing interval
		int rc;		   // Tcl return code
		
		rc = Tcl_GetDouble(interp, argv[2], &inter);
		if (rc == TCL_OK) {
		    
		    sigHost->interval = inter;
		    sigHost->Post(0.0 + inter, sigHost, EventTimer2, NULL);

		}
		return rc;
	    }
	}

	// sighost mode mode
	else if (!strcmp(argv[1], "mode")) {

	    if (argc == 3) {
		if (!strcmp(argv[2], "work-conserving")) {
		    sigHost->mode = SigHostModeWorkConserving;
		    return(TCL_OK);
		}
		else if (!strcmp(argv[2], "non-work-conserving")) {
		    sigHost->mode = SigHostModeNonWorkConserving;
		    return(TCL_OK);
		}
		else if (!strcmp(argv[2], "no-rate-control")) {
		    sigHost->mode = SigHostModeNoRateControl;
		    return(TCL_OK);
		}
		else {
		    Tcl_SetResult(interp, "bad mode", TCL_STATIC);
		    return(TCL_ERROR);
		}
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	else {
	    return(ScheduledObjectCmd(clientData, interp, argc, argv));
	}
    }

}

//
// SigHostDeleteProc
//
// (Tcl DeleteProc)
//
void SigHostDeleteProc(ClientData clientData)
{
    delete ((SigHost *) clientData);
}

//
// operator<<
//
// Input:  Output stream (s), state (shs)
//
// Output:  Output stream
//
// Print the name of the SigHost state shs on the output stream s.
//
ostream& operator<<(ostream &s, SigHostState shs)
{
    char *output;
    
    output = SigHostStateName[shs];

    return s << output;

}

//
// operator<<
//
// Input:  Output stream (s), queueing mode (mode)
//
// Output:  Output stream
//
// Print the name of the mode (mode) on the output stream s.
//
ostream& operator<<(ostream &s, SigHostMode mode)
{
    char *output;
    
    output = SigHostModeName[mode];

    return s << output;

}
