static char rcsid[] = "$Id: SigHostRcsp.cc,v 1.10 1996-05-29 14:34:59-07 bmah Exp $";
//
// $Id: SigHostRcsp.cc,v 1.10 1996-05-29 14:34:59-07 bmah Exp $
//
// SigHostRcsp.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// RCSP host signalling entity class.  Mostly, this class overrides
// certain methods of the parent SigHost class related to admission
// control tests.
//
// For the most part, the contents of this file were derived from
// SigRcsp.cc, optimized for the special case of a single output link.
//

#include "SigMessage.h"
#include "SigHostRcsp.h"
#include "SwitchModule.h"
#include "Simulation.h"
#include "CellGoBackN.h"
#include "CellQueueRcsp.h"
#include "inline.h"

//
// SigHostRcsp::SigHostRcsp
//
// Object constructor
//
SigHostRcsp::SigHostRcsp(const String &name, Simulation *si, int mv,
			 int down, int addr) :
SigHost(name, si, mv, down, addr) {
    int i;

    channelState = new ChannelState[maxVcids];
    for (i = 0; i < maxVcids; i++) {
	channelState[i].level = -1;
    }

    buffersExcess = SigHostRcspBuffersExcessDefault;

}

//
// SigHostRcsp::~SigHostRcsp
//
// Object destructor
//
SigHostRcsp::~SigHostRcsp() {
    delete [] channelState;
}

//
// SigHostRcsp::Output
//
// Input:  None
//
// Output:  String with ASCII representation of this object
//
String SigHostRcsp::Output() {

    String out;
    char buf[80];
    extern char *SigHostModeName[];

    sprintf(buf, "%d ", maxVcids);
    out = name + " [SigHostRcsp] mode " + SigHostModeName[mode] +
      " maxVcids " + buf;

    sprintf(buf, "excess %d ", buffersExcess);
    out += buf;

    sprintf(buf, "maxDownCalls %d ", maxDownCalls);
    out += buf;

    sprintf(buf, "address %d ", address);
    out += buf;

    out += (String) "aal " + aalName + " ";
    
    return out;

}

//
// SigHostRcsp::AdmissionControl
//
// Input:  VCID (vcid), traffic characteristics (traffic)
//
// Output:  status (in return value)
//
// Run admission control tests for the outgoing link and the given
// traffic descriptor.  If tests are successful, make tentative
// resource reservations.  This method is a near-copy of the corresponding
// method in the SigRcsp class.
//
SigReturnCode SigHostRcsp::AdmissionControl(double time,
					    int vcid, void *traffic)
{
    QosRcsp *parms = (QosRcsp *) traffic; // got parameters
    int i;				  // Loop Variables 'R' Us
    int k;				  // priority level

    IF_DEBUG(2, {
	DEBUG_PREFIX("SigHostRcsp");
	*(sim->debugStream) << "in SigHostRcsp::AdmissionControl\n";
    });

    if (parms == NULL) {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("SigHostRcsp");
	    *(sim->debugStream) << "missing traffic parameters\n";
	});
	return (SigBadParms);
    }

    if (parms->type != QosTypeRcsp) {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("SigHostRcsp");
	    *(sim->debugStream) << "bad type (" << parms->type << ")\n";
	});
	return (SigBadParms);
    }

    // Compute priority level k
    if (parms->d < 0) {

	// best effort channel
	k = -1;

    }
    else {
	
	// If the local delay requested is smaller than our minimum,
	// we need to reject the channel.
	if (parms->d < linkState.d[0]) {
	    return (SigFailDelay);
	}

	// If local delay requested is larger than we are prepared to
	// support, then change the local delay to the maximum.
	if (parms->d > linkState.d[linkState.n - 1]) {
	    parms->d = linkState.d[linkState.n - 1];
	}

	// Determine channel level...basically find the first level
	// that meets the requirements for this channel.  From the
	// tests above, we know that at least one channel will work.
	for (k = (linkState.n) - 1; k >= 0; k--) {
	    if (parms->d >= linkState.d[k]) {
		break;
	    }
	}

//////////////////////
	
	// If we're not doing rate control, then bypass admission control
	// tests.
	if (mode == SigHostModeNoRateControl) {
	    IF_DEBUG(3, {
		DEBUG_PREFIX("SigHostRcsp");
		*(sim->debugStream) << "Bypassing admission control\n";
	    });
	}
	else {
	    
	    //
	    // Run admission control tests
	    //
	    {
		// Stability/bandwidth test
		double cells; 		// number of cells per second
		int v;			// loop counter

		// Initialize with average rate of current connection
		cells = 1.0 / (parms->Xmin);
		
		// Add in average rate of all established real-time connections
		for (v = 0; v < maxVcids; v++) {
		    if (channelState[v].level >= 0) {
			cells += 1.0 / (channelState[v].Xave);
		    }
		}
		
		IF_DEBUG(3, {
		    DEBUG_PREFIX("SigHostRcsp");
		    *(sim->debugStream) << "Bandwidth test:  " << cells << " vs. " << (linkState.l / AtmCellSize) << "\n";
		});

		// Bandwidth test fails if this is greater than the cell rate
		// on this link.
		if (cells > (linkState.l / AtmCellSize)) {
		    return SigFailBandwidth;
		}

	    }

	    {
		// Delay test.  Objective is to calculate the backlog (in
		// cells)
		// that can accumulate during the time of a cell's delay bound.
		// We need to check the bounds for all channels at the current
		// level or lower.  Note that every level has a delay
		// bound associated with it, which simplifies our testing a bit.
		int level;
		for (level = k; level < linkState.n; level++) {

		    IF_DEBUG(3, {
			DEBUG_PREFIX("SigHostRcsp");
			*(sim->debugStream) << "Delay test at level " << level << "\n";
		    });
		    
		    const int maxKnees = 50;
		    double kneePoints[maxKnees];
		    int knees = 0;

		    // Clear out the list of already-checked knee points.
		    for (i = 0; i < maxKnees; i++) {
			kneePoints[i] = -1.0;
		    }

		    // For each level, check the curves at I*Xmin/Xave for
		    // every channel in the current level or lower
		    // (higher priority).  The kneePoints[] array is a cache
		    // of points we've already checked, to avoid us needing
		    // to check anything twice.
		    int v;
		    for (v = 0; v < maxVcids; v++)
		      {

			  // Don't need to check the knee for a channel not at
			  // this level or lower (it won't contribute to the
			  // delay of a cell at the current level).  Also don't
			  // check best-effort channels.
			  if ((channelState[v].level > level) ||
			      (channelState[v].level < 0)) {
			      continue;
			  }

			  double t = channelState[v].I *
			    channelState[v].Xmin / channelState[v].Xave;
			  // t is the current knee

			  // See if we've already checked this kneepoint.
			  // If it's in our cache, then we've already run our
			  // tests at this point and we don't need to do
			  // it again.
			  int cachedKnee = -1;
			  for (i = 0; i < knees; i++) {
			      if (kneePoints[i] == t) {
				  cachedKnee = i;
				  break;
			      }
			  }
			  if (cachedKnee >= 0) {

			      IF_DEBUG(3, {
				  DEBUG_PREFIX("SigHostRcsp");
				  *(sim->debugStream) <<
				    "Skip cached knee point " << t << "\n";
			      });
			      continue;

			  }

			  // We haven't tested this kneepoint, so let's do it.
			  // XXX We should do this test at the kneepoint for the
			  // channel being established too! XXX
			  IF_DEBUG(3, {
			      DEBUG_PREFIX("SigHostRcsp");
			      *(sim->debugStream) << "Test knee point " << t << "\n";
			  });

			  double cells = RcspB(t + parms->d,
					       parms->Xmin,
					       parms->Xave,
					       parms->I);
			  // number of cells in current level
			  IF_DEBUG(4, {
			      DEBUG_PREFIX("SigHostRcsp");
			      *(sim->debugStream) << "Establishing channel contributes " << cells << "\n";
			  });

			  // Iterate over all channels at level k or lower.
			  // Figure out how much they'll contribute to the
			  // total delay, at the kneepoint we've selected.
			  int v2;
			  double b;
			  for (v2 = 0; v2 < maxVcids; v2++) {
			      if ((channelState[v2].level <= k) &&
				  (channelState[v2].level > 0)) {


				  b = RcspB(t + parms->d,
					    channelState[v2].Xmin,
					    channelState[v2].Xave,
					    channelState[v2].I);

				  IF_DEBUG(4, {
				      DEBUG_PREFIX("SigHostRcsp");
				      *(sim->debugStream) << "VCID " << v2 <<
					" (level = " <<
					  channelState[v2].level <<
					", xmin = " << channelState[v2].Xmin <<
					", xave = " << channelState[v2].Xave <<
					", I = " << channelState[v2].I <<
					") contributes " << b << "\n";
				  });

				  cells += b;

			      }
			  }

			  IF_DEBUG(3, {
			      DEBUG_PREFIX("SigHostRcsp");
			      *(sim->debugStream) << "Gross backlog: " << cells << " cells\n";
			  });
			  
			  // Now subtract the amount that can get transmitted
			  // during that time.  What's left is the backlog
			  // that can cause delay for the current cell.
			  cells -= ((linkState.l) / AtmCellSize) *
			    (t);
			  
			  IF_DEBUG(3, {
			      DEBUG_PREFIX("SigHostRcsp");
			      *(sim->debugStream) << "Net backlog: " << cells << " cells\n";
			      DEBUG_PREFIX("SigHostRcsp");
			      *(sim->debugStream) << "Delay: " << cells*AtmCellSize/linkState.l << " vs. " << linkState.d[level] << "\n";
			  });

			  // If the remaining delay is greater than the delay
			  // bound for this level, then we've failed the delay
			  // test.
			  if (cells * AtmCellSize / linkState.l >
			      linkState.d[level]) {
			      return SigFailDelay;
			  }

			  // We've successfully tested this kneepoint, so add it
			  // to our cache.
			  if (knees < maxKnees) {
			      kneePoints[knees++] = t;
			  }
			  else {
			      IF_DEBUG(0, {
				  DEBUG_PREFIX("SigHostRcsp");
				  *(sim->debugStream) <<
				    "Ran out of kneepoint cache entries\n";
			      });
			  }
		      }
		}
	    }
	}
    }
    
    // Compute and store parameters
    if (k >= 0) {

	// Save parameters of new real-time channel
	channelState[vcid].Xmin = parms->Xmin;
	channelState[vcid].Xave = parms->Xave;
	channelState[vcid].I = parms->I;
	channelState[vcid].level = k;

	// Real-time buffer allocation
	channelState[vcid].buffers =
	  (int) (ceil((linkState.d[k] + linkState.tick) / parms->Xmin) +
		 ceil(linkState.d[k] / parms->Xmin));
	// Note:  No compensation needed for work-conserving scheduler:
	// we only get called here at the start of a channel, when there
	// are no previous upstream nodes.

    }
    else {

	// Parameter for new best-effort channel
	channelState[vcid].level = k;

	// Best-effort buffer allocation...
	channelState[vcid].buffers = SigHostRcspBuffersBestEffort;

    }

    // Add on extra margin for host interface buffering
    channelState[vcid].buffers += buffersExcess;

    IF_DEBUG(1, {
	DEBUG_PREFIX("SigHostRcsp");
	*(sim->debugStream) << "accepted channel level " <<
	  channelState[vcid].level << " buffers " <<
	    channelState[vcid].buffers << "\n";
    });

    return (SigOK);

}

//
// SigHostRcsp::CommitResources
//
// Input:  VCID (vcid), traffic characteristics (traffic)
//
// Output:  status (in return value)
//
// Commit resources for the specified VCID.
//
void SigHostRcsp::CommitResources(double time,
				  int vcid, void *traffic)
{

    CellQueueRcsp *port;	// properly-cast version of port object

    IF_DEBUG(2, {
	DEBUG_PREFIX("SigHostRcsp");
	*(sim->debugStream) << "in SigHostRcsp::CommitResources\n";
    });

    // First do a lookup on the output port
    Lookup(networkPort, networkPortName);
    port = (CellQueueRcsp *) networkPort;

    // Setup up variables in rate-control module and scheduler.
    port->NewVc(time, vcid,
		channelState[vcid].level,
		channelState[vcid].Xmin,
		channelState[vcid].Xave,
		channelState[vcid].I,
		channelState[vcid].buffers);

    return;
}

//
// SigHostRcsp::ReleaseResources
//
// Input:  VCID (vcid)
//
// Output:  status (in return value)
//
// Release resources for the specified outgoing VC.
//
void SigHostRcsp::ReleaseResources(double time,
				   int vcid)
{
    CellQueueRcsp *port;	// properly-cast version of port object
    int i;

    IF_DEBUG(2, {
	DEBUG_PREFIX("SigHostRcsp");
	*(sim->debugStream) << "in SigHostRcsp::ReleaseResources\n";
    });

    // First do a lookup on the output port.  We're being a little extra
    // paranoid here...someone had to do the lookup to set up the channel
    // in the first place, right?
    Lookup(networkPort, networkPortName);
    port = (CellQueueRcsp *) networkPort;

    // Delete resources in rate-control module and scheduler.
    port->DeleteVc(time, vcid);

    // Do accounting...if necessary...
    if (channelState[vcid].level >= 0) {
	channelState[vcid].level = -1;
    }
}

//
// SigHostRcspCmd
//
// (Tcl command handler procedure)
//
// SigHostRcsp name maxVcids downCalls address
//
int SigHostRcspCmd(ClientData clientData, Tcl_Interp *interp,
	       int argc, char *argv[])
{

    if (argc != 5) {
	Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of new SigHostRcsp object
	int mv;			// maximum number of VCIs
	int dc;			// maximum number of concurrent
				// signalling downcalls from clients
	int addr;		// ATM address
	int rc;			// return code from Tcl library
	SigHostRcsp *sigHost;	// pointer to new SigHostRcsp object
	extern Simulation *sim;	// global simualtion object

	// get name
	nm = argv[1];

	// get three numeric argments into mv, dc, addr
	rc = Tcl_GetInt(interp, argv[2], &mv);
	if (rc == TCL_OK) {

	    rc = Tcl_GetInt(interp, argv[3], &dc);
	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[4], &addr);
		if ((rc == TCL_OK) && (addr < 1)) {
		    Tcl_SetResult(interp, "address out of range", TCL_STATIC);
		    rc = TCL_ERROR;
		}
		if (rc == TCL_OK) {

		    sigHost = new SigHostRcsp(nm, sim, mv, dc, addr);
		    Tcl_CreateCommand(interp, nm,
				      SigHostRcspObjectCmd,
				      (ClientData) sigHost,
				      SigHostRcspDeleteProc);

		}
	    }
	}

	return rc;

    }

}

//
// SigHostRcspObjectCmd
//
// (Tcl object handler procedure)
//
int SigHostRcspObjectCmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{

    SigHostRcsp *sigHostRcsp = (SigHostRcsp *) clientData;

    if (argc == 1) {
	return(SigHostObjectCmd(clientData, interp, argc, argv));
    }
    else {

	//
	// sighost net net-name
	//
	if (!strcmp(argv[1], "net")) {

	    if (argc == 6) {
		String on;	// object name
		double l;	// link speed
		double tick;	// tick time
		int n;		// number of levels
		char **ds;	// delay strings
		double *dd;	// delay doubles
		int rc = TCL_OK; // Tcl return value
		int i;

		on = argv[2];

		// Get link speed
		rc = Tcl_GetDouble(interp, argv[3], &l);
		if (rc == TCL_OK) {
		    
		    // Get tick time
		    rc = Tcl_GetDouble(interp, argv[4], &tick);
		    if (rc == TCL_OK) {

			// Get array o fdelays
			rc = Tcl_SplitList(interp, argv[5], &n, &ds);
			if (rc == TCL_OK) {

			    // Parse array into doubles
			    dd = new double[n];
			    for (i = 0; i < n; i++) {

				if ((rc =
				     Tcl_GetDouble(interp,
						   ds[i],
						   &dd[i])) != TCL_OK) {
				    delete [] dd;
				    break;
				}
			    }
			    
			    free(ds);

			    if (rc == TCL_OK) {
				sigHostRcsp->networkPortName = on;
				sigHostRcsp->networkPort = NULL;
				sigHostRcsp->linkState.d = dd;
				sigHostRcsp->linkState.tick = tick;
				sigHostRcsp->linkState.l = l;
				sigHostRcsp->linkState.n = n;
			    }
			}
		    }
		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// sighost excess buffersexcess
	//
	else if (!strcmp(argv[1], "excess")) {

	    if (argc == 3) {
		int e;         // excess buffers (per VC)
		int rc;	       // Tcl return code
		
		rc = Tcl_GetInt(interp, argv[2], &e);
		if (rc == TCL_OK) {
		    
		    sigHostRcsp->buffersExcess = e;
		    
		}
		return rc;
	    }
	}

	else {

	    // Pass-through all others.
	    return (SigHostObjectCmd(clientData, interp, argc, argv));

	}
    }
}


//
// SigHostRcspDeleteProc
//
// (Tcl DeleteProc)
//
void SigHostRcspDeleteProc(ClientData clientData)
{
    delete ((SigHostRcsp *) clientData);
}

