static char rcsid[] = "$Id: SigRcsp.cc,v 1.13 1996-05-29 14:35:02-07 bmah Exp $";
//
// $Id: SigRcsp.cc,v 1.13 1996-05-29 14:35:02-07 bmah Exp $
//
// SigRcsp.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// RCSP signalling entity class.  This class merely overrides certain
// methods of the parent Sig class.
//

#include "SigMessage.h"
#include "SigRcsp.h"
#include "SwitchModule.h"
#include "Simulation.h"
#include "CellGoBackN.h"
#include "CellQueueRcsp.h"
#include "inline.h"

//
// RcspB
//
// Input:  Time (u), xmin, xave, I
//
// Output:  Number of cells
//
// Compute the bound on the total number of cells that can be transmitted
// during any interval of length u, given channel characteristics
// xmin, xave, i.
//
double RcspB(double u, double xmin, double xave, double i)
{
    double lastPeriod, lastPeriod1, lastPeriod2;
    double otherPeriods;

    lastPeriod1 = ceil(fmod(u, i) / xmin);
    lastPeriod2 = ceil(i / xave);
    otherPeriods = floor(u / i) * (i / xave);

    if (lastPeriod1 < lastPeriod2) {
	lastPeriod = lastPeriod1;
    }
    else {
	lastPeriod = lastPeriod2;
    }

    return lastPeriod + otherPeriods;
}

//
// SigRcsp::SigRcsp
//
// Object constructor
//
SigRcsp::SigRcsp(const String &name, Simulation *si, int p, int mv, int mr) :
Sig(name, si, p, mv, mr) {

    int i, j;
    
    // Allocate space to hold RCSP-specific state
    linkState = new LinkState[numPorts];
    for (i = 0; i < numPorts; i++) {
	linkState[i].d = NULL;
    }

    channelState = new (ChannelState *)[numPorts];
    for (i = 0; i < numPorts; i++) {
	channelState[i] = new ChannelState[maxVcids];
	for (j = 0; j < maxVcids; j++) {
	    channelState[i][j].level = -1;
	}
    }

    buffersExcess = SigRcspBuffersExcessDefault;

}

//
// SigRcsp::~SigRcsp
//
// Object destructor
//
SigRcsp::~SigRcsp() {

    int i;

    for (i = 0; i < numPorts; i++) {
	delete [] channelState[i];
	if (linkState[i].d) {
	    delete [] linkState[i].d;
	}
    }
    delete [] linkState;
    delete [] channelState;

}


//
// SigRcsp::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String SigRcsp::Output()
{

    String out;
    char buf[30];
    extern char *SigModeName[];

    sprintf(buf, "%d ", numPorts);
    out = name + " [SigRcsp] mode " + SigModeName[mode] + " numPorts " + buf;

    sprintf(buf, "maxVcids %d ", maxVcids);
    out += buf;

    sprintf(buf, "excess %d ", buffersExcess);
    out += buf;

    sprintf(buf, "maxRoutes %d", maxRoutes);
    out += buf;

    return out;
    
}

//
// SigRcsp::AdmissionControl
//
// Input:  traffic characteristics (traffic), outgoing link number (link)
//
// Output:  Return code
//
// Perform admission control tests for the appropriate outgoing link
// and given traffic descriptor.  If tests success (return code ==
// SigOK), make tentative resource reservations.
//
SigReturnCode SigRcsp::AdmissionControl(double time,
					int link, int vcid,
					SigMessage *mesg, void *traffic)
{

    QosRcsp *parms = (QosRcsp *) traffic; // got parameters
    int i;				  // Loop Variables 'R' Us
    int k;				  // priority level

    IF_DEBUG(2, {
	DEBUG_PREFIX("SigRcsp");
	*(sim->debugStream) << "in SigRcsp::AdmissionControl\n";
    });

    if (parms == NULL) {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("SigRcsp");
	    *(sim->debugStream) << "missing traffic parameters\n";
	});
	return (SigBadParms);
    }

    if (parms->type != QosTypeRcsp) {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("SigRcsp");
	    *(sim->debugStream) << "bad type (" << parms->type << ")\n";
	});
	return (SigBadParms);
    }

    IF_DEBUG(3, {
	DEBUG_PREFIX("SigRcsp");
	*(sim->debugStream) << "d = " << parms->d <<
	  ", xmin = " << parms->Xmin <<
	  ", xave = " << parms->Xave <<
	  ", I = " << parms->I <<
	  "\n";
    });

    // Compute priority level k
    if (parms->d < 0) {

	// best effort channel
	k = -1;

    }
    else {
	
	// If the local delay requested is smaller than our minimum,
	// we need to reject the channel.
	if (parms->d < linkState[link].d[0]) {
	    return (SigFailDelay);
	}

	// If local delay requested is larger than we are prepared to
	// support, then change the local delay to the maximum.
	if (parms->d > linkState[link].d[linkState[link].n - 1]) {
	    parms->d = linkState[link].d[linkState[link].n - 1];
	}

	// Determine channel level...basically find the first level
	// that meets the requirements for this channel.  From the
	// tests above, we know that at least one channel will work.
	for (k = (linkState[link].n) - 1; k >= 0; k--) {
	    if (parms->d >= linkState[link].d[k]) {
		break;
	    }
	}

	IF_DEBUG(3, {
	    DEBUG_PREFIX("SigRcsp");
	    *(sim->debugStream) << "Channel level " << k << "\n";
	});

	// If we're not doing rate control, then bypass admission control
	// tests.
	if (mode == SigModeNoRateControl) {
	    IF_DEBUG(3, {
		DEBUG_PREFIX("SigRcsp");
		*(sim->debugStream) << "Bypassing admission control\n";
	    });
	}
	else {

	    //
	    // Run admission control tests
	    //
	    {
		// Stability/bandwidth test
		double cells; 		// number of cells per second
		int v;			// loop counter

		// Initialize with average rate of current connection
		cells = 1.0 / (parms->Xmin);
		
		// Add in average rate of all established real-time connections
		for (v = 0; v < maxVcids; v++) {
		    if (channelState[link][v].level >= 0) {
			cells += 1.0 / (channelState[link][v].Xave);
		    }
		}
		
		IF_DEBUG(3, {
		    DEBUG_PREFIX("SigRcsp");
		    *(sim->debugStream) << "Bandwidth test:  " << cells << " vs. " << (linkState[link].l / AtmCellSize) << "\n";
		});

		// Bandwidth test fails if this is greater than the cell rate
		// on this link.
		if (cells > (linkState[link].l / AtmCellSize)) {

		    IF_DEBUG(2, {
			DEBUG_PREFIX("SigRcsp");
			*(sim->debugStream) << "Fail bandwidth test\n";
		    });

		    return SigFailBandwidth;
		}

	    }

	    {
		// Delay test.  Objective is to calculate the backlog (in cells)
		// that can accumulate during the time of a cell's delay bound.
		// We need to check the bounds for all channels at the current
		// level or lower.  Note that every level has a delay
		// bound associated with it, which simplifies our testing a bit.
		int level;
		for (level = k; level < linkState[link].n; level++) {

		    IF_DEBUG(3, {
			DEBUG_PREFIX("SigRcsp");
			*(sim->debugStream) << "Delay test at level " << level << "\n";
		    });
		    
		    const int maxKnees = 50;
		    double kneePoints[maxKnees];
		    int knees = 0;

		    // Clear out the list of already-checked knee points.
		    for (i = 0; i < maxKnees; i++) {
			kneePoints[i] = -1.0;
		    }

		    // For each level, check the curves at I*Xmin/Xave for
		    // every channel in the current level or lower
		    // (higher priority).  The kneePoints[] array is a cache
		    // of points we've already checked, to avoid us needing
		    // to check anything twice.
		    int v;
		    for (v = 0; v < maxVcids; v++)
		      {

			  // Don't need to check the knee for	a channel not at
			  // this level or lower (it won't contribute to the
			  // delay of a cell at the current level).  Also don't
			  // check best-effort channels.
			  if ((channelState[link][v].level > level) ||
			      (channelState[link][v].level < 0)) {
			      continue;
			  }

			  double t = channelState[link][v].I *
			    channelState[link][v].Xmin / channelState[link][v].Xave;
			  // t is the current knee

			  // See if we've already checked this kneepoint.
			  // If it's in our cache, then we've already run our
			  // tests at this point and we don't need to do it again.
			  int cachedKnee = -1;
			  for (i = 0; i < knees; i++) {
			      if (kneePoints[i] == t) {
				  cachedKnee = i;
				  break;
			      }
			  }
			  if (cachedKnee >= 0) {

			      IF_DEBUG(3, {
				  DEBUG_PREFIX("SigRcsp");
				  *(sim->debugStream) <<
				    "Skip cached knee point " << t << "\n";
			      });
			      continue;

			  }

			  // We haven't tested this kneepoint, so let's do it.
			  // XXX We should do this test at the kneepoint for the
			  // channel being established too! XXX
			  IF_DEBUG(3, {
			      DEBUG_PREFIX("SigRcsp");
			      *(sim->debugStream) << "Test knee point " << t << "\n";
			  });

			  double cells = RcspB(t + parms->d,
					       parms->Xmin,
					       parms->Xave,
					       parms->I);
			  // number of cells in current level
			  IF_DEBUG(4, {
			      DEBUG_PREFIX("SigRcsp");
			      *(sim->debugStream) << "Establishing channel contributes " << cells << "\n";
			  });

			  // Iterate over all channels at level k or lower.
			  // Figure out how much they'll contribute to the
			  // total delay, at the kneepoint we've selected.
			  int v2;
			  double b;
			  for (v2 = 0; v2 < maxVcids; v2++) {
			      if ((channelState[link][v2].level <= k) &&
				  (channelState[link][v2].level > 0)) {


				  b = RcspB(t + parms->d,
					    channelState[link][v2].Xmin,
					    channelState[link][v2].Xave,
					    channelState[link][v2].I);

				  IF_DEBUG(4, {
				      DEBUG_PREFIX("SigRcsp");
				      *(sim->debugStream) << "VCID " << v2 <<
					" (level = " << channelState[link][v2].level <<
					", xmin = " << channelState[link][v2].Xmin <<
					", xave = " << channelState[link][v2].Xave <<
					", I = " << channelState[link][v2].I <<
					") contributes " << b << "\n";
				  });

				  cells += b;

			      }
			  }

			  IF_DEBUG(3, {
			      DEBUG_PREFIX("SigRcsp");
			      *(sim->debugStream) << "Gross backlog: " << cells << " cells\n";
			  });
			  
			  // Now subtract the amount that can get transmitted
			  // during that time.  What's left is the backlog
			  // that can cause delay for the current cell.
			  cells -= ((linkState[link].l) / AtmCellSize) *
			    (t);
			  
			  IF_DEBUG(3, {
			      DEBUG_PREFIX("SigRcsp");
			      *(sim->debugStream) << "Net backlog: " << cells << " cells\n";
			      DEBUG_PREFIX("SigRcsp");
			      *(sim->debugStream) << "Delay: " << cells*AtmCellSize/linkState[link].l << " vs. " << linkState[link].d[level] << "\n";
			  });

			  // If the remaining delay is greater than the delay
			  // bound for this level, then we've failed the delay
			  // test.
			  if (cells * AtmCellSize / linkState[link].l >
			      linkState[link].d[level]) {

			      IF_DEBUG(2, {
				  DEBUG_PREFIX("SigRcsp");
				  *(sim->debugStream) << "Fail delay test\n";
			      });

			      return SigFailDelay;
			  }

			  // We've successfully tested this kneepoint, so add it
			  // to our cache.
			  if (knees < maxKnees) {
			      kneePoints[knees++] = t;
			  }
			  else {
			      IF_DEBUG(0, {
				  DEBUG_PREFIX("SigRcsp");
				  *(sim->debugStream) <<
				    "Ran out of kneepoint cache entries\n";
			      });
			  }
		      }
		}
	    }
	}
    }
    
    // Compute and store parameters
    if (k >= 0) {

	// Save parameters of new real-time channel
	channelState[link][vcid].Xmin = parms->Xmin;
	channelState[link][vcid].Xave = parms->Xave;
	channelState[link][vcid].I = parms->I;
	channelState[link][vcid].level = k;

	// Real-time buffer allocation
	if (mode == SigModeWorkConserving) {

	    // If we're work-conserving, we need to be able to buffer up
	    // a delay equal to the total delay incurred by all the network
	    // switches.  (More accurately, if the other switches upstream
	    // on the path of a channel do work-conserving scheduling,
	    // we need to be able to buffer up cells equal to the sum
	    // of the delay bounds on those switches.)
	    channelState[link][vcid].buffers =
	      (int) (ceil((linkState[link].d[k] + linkState[link].tick) /
			  parms->Xmin) * mesg->hopcount +
		     ceil(linkState[link].d[k] / parms->Xmin));

	}
	else {
	    channelState[link][vcid].buffers =
	      (int) (ceil((linkState[link].d[k] + linkState[link].tick) /
			  parms->Xmin) +
		     ceil(linkState[link].d[k] / parms->Xmin));
	}
	
	if (channelState[link][vcid].buffers < SigRcspMinBuffers) {
	    channelState[link][vcid].buffers = SigRcspMinBuffers; 
	}

    }
    else {

	// Parameter for new best-effort channel
	channelState[link][vcid].level = k;

	// Best-effort buffer allocation...
	channelState[link][vcid].buffers = SigRcspBuffersBestEffort;

    }
    channelState[link][vcid].buffers += buffersExcess;

    IF_DEBUG(1, {
	DEBUG_PREFIX("SigRcsp");
	*(sim->debugStream) << "accepted channel level " <<
	  channelState[link][vcid].level << " buffers " <<
	    channelState[link][vcid].buffers << "\n";
    });

    return (SigOK);

}

//
// SigRcsp::CommitResources
//
// Input:  traffic characteristics (traffic), outgoing link number (link)
// virtual circuit (vcid)
//
// Output:  Return code
//
// Commit reserved resources to a channel.
//
void SigRcsp::CommitResources(double time,
			      int link, int vcid, void *traffic)
{

    CellQueueRcsp *port;	// properly-cast version of port object

    IF_DEBUG(2, {
	DEBUG_PREFIX("SigRcsp");
	*(sim->debugStream) << "in SigRcsp::CommitResources\n";
    });

    // First do a lookup on the output port
    Lookup(outputPort[link], outputPortName[link]);

    port = (CellQueueRcsp *) outputPort[link];

    // Setup up variables in rate-control module and scheduler.
    port->NewVc(time, vcid,
		channelState[link][vcid].level,
		channelState[link][vcid].Xmin,
		channelState[link][vcid].Xave,
		channelState[link][vcid].I,
		channelState[link][vcid].buffers);

}

//
// SigRcsp::ReleaseResources
//
// Input:  outgoing link number (link), virtual circuit ID (vcid)
//
// Output:  Return code
//
// Release resources allocated to a channel, both with respect to the
// signalling entity's accounting and the actual data transport entity's
// records.
//
void SigRcsp::ReleaseResources(double time,
			       int link, int vcid)
{
    CellQueueRcsp *port;	// properly-cast version of port object
    int i;

    IF_DEBUG(2, {
	DEBUG_PREFIX("SigRcsp");
	*(sim->debugStream) << "in SigRcsp::ReleaseResources\n";
    });

    // First do a lookup on the output port.  We're being a little extra
    // paranoid here...someone had to do the lookup to set up the channel
    // in the first place, right?
    Lookup(outputPort[link], outputPortName[link]);
    port = (CellQueueRcsp *) outputPort[link];

    // Delete resources in rate-control module and scheduler.
    port->DeleteVc(time, vcid);

    // Do accounting...
    channelState[link][vcid].level = -1;

}

//
// SigRcspCmd
//
// (Tcl command procedure)
//
// This constructor command is essentially the same as SigCmd.
//
int SigRcspCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 5) {
	Tcl_SetResult(interp,
		      "SigRcsp name maxports maxvcid maxroutes",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of new Sig
	int np;			// num ports
	int maxvci;		// maximum number of VCIs
	int maxroute;		// maximum number of routes supported
	int rc;			// return code from Tcl library
	SigRcsp *sigRcsp;	// new signalling object
	extern Simulation *sim;	// global simulation object

	// get name
	nm = argv[1];

	// get numeric arguments
	rc = Tcl_GetInt(interp, argv[2], &np);
	if (rc == TCL_OK) {

	    rc = Tcl_GetInt(interp, argv[3], &maxvci);
	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[4], &maxroute);
		if (rc == TCL_OK) {

		    sigRcsp = new SigRcsp(nm, sim, np, maxvci, maxroute);
		    Tcl_CreateCommand(interp, nm,
				      SigRcspObjectCmd,
				      (ClientData) sigRcsp,
				      SigRcspDeleteProc);
		    
		}

	    }

	}

	return rc;
	
    }
}

//
// SigRcspObjectCmd
//
// (Tcl object command procedure)
//
int SigRcspObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {

    SigRcsp *sigRcsp = (SigRcsp *) clientData;

    if (argc == 1) {
	return (SigObjectCmd(clientData, interp, argc, argv));
    }
    else {

	if (!strcmp(argv[1], "output")) {

	    if (argc == 7) {

		String portName;                      // output port name
		int port;	                      // port number
		double l;			      // link speed
		double tick;			      // tick
		int n;				      // number of levels
		char **ds;			      // delay (strings)
		double *dd;			      // delay (doubles)
		int rc = TCL_OK;		      // Tcl return code
		int i;
		
		rc = Tcl_GetInt(interp, argv[2], &port);
		if (rc == TCL_OK) {
		    
		    if ((port < 0) || (port >= sigRcsp->numPorts)) {
			Tcl_SetResult(interp, "port out of bounds",
				      TCL_STATIC);
			return(TCL_ERROR);
		    }
		    
		    portName = argv[3];
			
		    rc = Tcl_GetDouble(interp, argv[4], &l);
		    if (rc == TCL_OK) {

			rc = Tcl_GetDouble(interp, argv[5], &tick);
			if (rc == TCL_OK) {
			    
			    // Get array of delays
			    rc = Tcl_SplitList(interp, argv[6], &n, &ds);
			    if (rc == TCL_OK) {
				
				// Parse array into doubles
				dd = new double[n];
				for (i = 0; i < n; i++) {
				    if ((rc =
					 Tcl_GetDouble(interp,
						       ds[i],
						       &dd[i])) != TCL_OK) {
					delete [] dd;
					break;
				    }
				}
				
				// Yes, we use free(3) here, because libtcl
				// did a malloc(3) to get this space in the
				// first place.  Ugly, ugly!  We hates it,
				// my precious!  We hates it!
				free (ds);
				
				if (rc == TCL_OK) {
				    sigRcsp->outputPortName[port] = portName;
				    sigRcsp->linkState[port].d = dd;
				    sigRcsp->linkState[port].tick = tick;
				    sigRcsp->linkState[port].l = l;
				    sigRcsp->linkState[port].n = n;
				}
			    }
			}
		    }
		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// sig excess buffersexcess
	//
	else if (!strcmp(argv[1], "excess")) {

	    if (argc == 3) {
		int e;         // excess buffers (per VC)
		int rc;	       // Tcl return code
		
		rc = Tcl_GetInt(interp, argv[2], &e);
		if (rc == TCL_OK) {
		    
		    sigRcsp->buffersExcess = e;
		    
		}
		return rc;
	    }
	}

	else {
	    return (SigObjectCmd(clientData, interp, argc, argv));
	}

    }

}

//
// SigRcspDeleteProc
//
// (Tcl delete procedure)
//
void SigRcspDeleteProc(ClientData clientData) {
    delete (SigRcsp *) clientData;
}
