static char rcsid[] = "$Id: Simulation.cc,v 1.12 1996-02-09 01:43:54-08 bmah Exp $";
// $Id: Simulation.cc,v 1.12 1996-02-09 01:43:54-08 bmah Exp $
//
// Simulation.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// A Simulation object is sort of a meta-data object for a simulation.
// There will probably only be one of these per run.  Its purpose is
// to serve as a receiver of breakpoints and other kinds of events
// which affect the simulation as a whole.  For example, the simulation
// stop time can be expressed by simply posting a SimulationEventStop
// event into the simulator scheduler, at the appropriate time.
//

#include <Pix.h>
#include "Simulation.h"
#include "String.ScheduledObjectPtr.CHMap.h"
#include "Tcl.h"
#include "inline.h"

#ifdef HAVE_CSRI_MALLOC
extern "C" {
#include "malloc.h"
}
#endif // HAVE_CSRI_MALLOC

//
// Simulation::Simulation
//
// Input:  name of simulation object (name)
// 
// Output:  none
//
// Constructor method for Simulation object.
//
Simulation::Simulation(const String &name) : ScheduledObject(name, this) {

}

//
// Simulation::Handler
//
// Input:  Normal handler arguments
//
// Output:  None
//
// Handle events for a simulation object.
//
void
Simulation::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    switch (type) {

	//
	// EventTimer
	//
	// Used by the "sim at" facility.  Data member holds the
	// Tcl script to execute, which is deallocated after execution.
	//
	case EventTimer:
	{
	    char *script = (char *) data;
	    int rc;		// Tcl return value
	    extern Tcl *tcl;	// Global Tcl interpreter

	    // Set up the script environment:
	    //
	    // o Global variable time is set to current simulation time.
	    tcl->SetVar("time", time, 0);

	    // Run the script
	    rc = tcl->Eval(script);
	    if (rc != TCL_OK) {
		tcl->AddErrorInfo("\n    caused by\na TimerEvent sent to the Simulation " + name);

		cout << tcl->GetVar("errorInfo", TCL_GLOBAL_ONLY) << "\n";
	    }
	    else {
		if (!(tcl->result.empty())) {
		    cout << tcl->result << '\n';
		}
	    }

	    delete script;
	}
	break;

	//
	// EventHalt
	//
	// End of simulation.  Clean up and halt.
	//
	case EventHalt:
	{
	    // cleanup stuff
	    IF_DEBUG(0, {
		DEBUG_PREFIX("Sim");
		*(sim->debugStream) << "Halt\n";
	    });

	    // stop simulation next time in event loop
	    run = false;
	}
	break;

	//
	// EventDebug
	//
	// Cause simulation to drop into CLI/debug interface.
	//
	case EventDebug:
	interactive = true;
	break;

	default:
	break;

    }
    
}

//
// Simulation::Output
//
// Input:  None
//
// Output:  String with ASCII data about object.
//
// Simulation class print method.
//
String Simulation::Output()
{
    String out;
    char buf[30];

    out = name + " [Simulation] events ";

    sprintf(buf, "%d", scheduler->cq_eventNum);
    out += (String) buf;
    
    return (out);
}

//
// SimulationCmd
//
// Input: 
//
// Output: 
//
// Simulation object Tcl command handler.
//
int SimulationCmd(ClientData clientData, Tcl_Interp *interp,
		  int argc, char *argv[])
{

    if (argc == 1) {
	return ScheduledObjectCmd(clientData, interp, argc, argv);
    }
    else {

	//
	// sim go
	//
	// Only meaningful in CLI...this causes the CLI to exit
	// and causes the simulation to continue.
	//
	if (!strcmp(argv[1], "go")) {

	    if (argc == 2) {

		((Simulation *) clientData)->stayInCli = false;
		return(TCL_OK);

	    }
	    else if (argc == 3) {

		double runUntil; // stopping time
		int rc; // Tcl return code

		rc = Tcl_GetDouble(interp, argv[2], &runUntil);
		if (rc == TCL_OK) {

		    if (runUntil > ((Simulation *) clientData)->scheduler->clock) {
		    
			((Simulation *) clientData)->Post(runUntil,
							  (Simulation *) clientData,
							  EventDebug,
							  NULL);
			((Simulation *) clientData)->stayInCli = false;
		    }
		    else {
			Tcl_SetResult(interp, "can't set breakpoint in past", TCL_STATIC);
			return(TCL_ERROR);
		    }
		}
		return rc;

	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }

	}

	//
	// sim list
	//
	// Return a list of the objects in the object registry.
	//
	else if (!strcmp(argv[1], "list")) {
	    
	    Pix ind;
	    String out;
	    StringScheduledObjectPtrCHMap *r =
	      ((Simulation *) clientData)->registry;

	    ind = r->first();
	    while (ind) {
		
		Tcl_AppendElement(interp, (const char *) r->key(ind));
		r->next(ind);
	    }
	    
	    return(TCL_OK);

	}

	//
	// sim when
	//
	// Print current simulation time
	//
	else if (!strcmp(argv[1], "when")) {
	    sprintf(interp->result, "%f",
		    ((Simulation *) clientData)->scheduler->clock);
	    return(TCL_OK);
	}

	//
	// sim at
	//
	// Cause a Tcl script to be executed at a certain time.
	//
	else if (!strcmp(argv[1], "at")) {
	    if (argc == 4) {

		int rc;
		double triggertime;

		rc = Tcl_GetDouble(interp, argv[2], &triggertime);
		if (rc == TCL_OK) {

		    char *script = new char[strlen(argv[3]) + 1];
		    strcpy(script, argv[3]);

		    if (triggertime > ((Simulation *) clientData)->scheduler->clock) {
			((Simulation *) clientData)->Post(triggertime,
							  (Simulation *) clientData,
							  EventTimer,
							  script);
		    }
		    else {
			Tcl_SetResult(interp, "can't run command in past", TCL_STATIC);
			return(TCL_ERROR);
		    }
		}
		return rc;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// sim trace ...
	//
	// Turn on or off event tracing.
	//
	else if (!strcmp(argv[1], "trace")) {
	    if (argc == 3) {

		int rc;
		int trace;
		
		rc = Tcl_GetInt(interp, argv[2], &trace);
		if (rc == TCL_OK) {
		    if (trace) {
			((Simulation *) clientData)->trace = true;
		    }
		    else {
			((Simulation *) clientData)->trace = false;
		    }
		}
		
		return(rc);

	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}
#ifdef HAVE_CSRI_MALLOC
	//
	// sim mdebug ...
	//
	// Set CSRI malloc debug level
	//
	else if (!strcmp(argv[1], "mdebug")) {
	    if (argc == 3) {

		int rc;
		int mdlevel;
		
		rc = Tcl_GetInt(interp, argv[2], &mdlevel);
		if (rc == TCL_OK) {
		    mal_debug(mdlevel);
		}
		
		return(rc);

	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}
#endif // HAVE_CSRI_MALLOC
	
	//
	// sim halt time
	//
	// Post a SimulationEventHalt event to the calendar queue.
	//
	else if (!strcmp(argv[1], "halt")) {
	    if (argc == 3) {

		int rc;
		double time;
		
		rc = Tcl_GetDouble(interp, argv[2], &time);
		if (rc == TCL_OK) {
		    if (time > ((Simulation *) clientData)->scheduler->clock) {
			((Simulation *) clientData)->Post(time,
							  (Simulation *) clientData,
							  EventHalt,
							  NULL);
		    }
		    else {
			Tcl_SetResult(interp, "can't set halt point in past", TCL_STATIC);
			return(TCL_ERROR);
		    }
		}
		
		return(rc);

	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// sim debug time
	//
	// Post a SimulationEventDebug event to the calendar queue.
	//
	else if (!strcmp(argv[1], "break")) {
	    if (argc == 3) {

		int rc;
		double time;
		
		rc =  Tcl_GetDouble(interp, argv[2], &time);
		if (rc == TCL_OK) {
		    if (time > ((Simulation *) clientData)->scheduler->clock) {
			((Simulation *) clientData)->Post(time,
							  (Simulation *) clientData,
							  EventDebug,
							  NULL);
		    }
		    else {
			Tcl_SetResult(interp, "can't set breakpoint in past", TCL_STATIC);
			return(TCL_ERROR);
		    }
		}
		
		return(rc);

	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	else {
	    return ScheduledObjectCmd(clientData, interp, argc, argv);
	}
    }

    return TCL_OK;
}

