static char rcsid[] = "$Id: Tcp.cc,v 1.41 1996-06-30 20:57:28-07 bmah Exp $";
//
// $Id: Tcp.cc,v 1.41 1996-06-30 20:57:28-07 bmah Exp $
//
// Tcp.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// Created by Bruce A. Mah
// All Rights Reserved
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Tcp protocol entity class class.  This code is based heavily on the
// TCP code from BSD 4.4Lite.
//
// Features *omitted* for simplicity:
//   o out-of-band data
//   o per-route statistics
//   o window scaling
//   o TCP header options
//
//
// This file contains material which is:

/*
 * Copyright (c) 1982, 1986, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "Tcp.h"
#include "IpPacket.h"
#include "TcpSegment.h"
#include "Event.h"
#include "util.h"
#include "inline.h"

// XXX Useful utility routines.  Maybe these should be moved to util.h?
inline int min(int a, int b) { return ((a < b) ? a : b); }
inline int max(int a, int b) { return ((a > b) ? a : b); }
#define exchange(a, b, type) { type t; t = a; a = b; b = t; }

#define TcpsHaveRcvdSyn(s) ((s) >= TcpsSynReceived)
#define TcpsHaveRcvdFin(s) ((s) >= TcpsTimeWait)

//
// Globals and stuff
//
const int Tcp::Backoff[] = {
    1, 2, 4, 8, 16, 32, 64, 64, 64, 64, 64, 64, 64
};

const int Tcp::Outflags[] = {
    TH_RST | TH_ACK, 0, TH_SYN, TH_SYN | TH_ACK,
    TH_ACK, TH_ACK, TH_FIN | TH_ACK, TH_FIN | TH_ACK,
    TH_FIN | TH_ACK, TH_ACK, TH_ACK
};

char *TcpsName[] = {
    "CLOSED",
    "LISTEN",
    "SYNSENT",
    "SYNRECEIVED",
    "ESTABLISHED",
    "CLOSEWAIT",
    "FINWAIT1",
    "CLOSING",
    "LASTACK",
    "FINWAIT2",
    "TIMEWAIT"
};

Uniform *Tcp::skewGenerator2; // Random offset to skew TCP timers
Uniform *Tcp::skewGenerator5; // to prevent phased effects

//
// Tcp::Tcp
//
// Input:  name, simulation object
//
// Output:  None
//
// Tcp object constructor
//
Tcp::Tcp(const String &name, Simulation *si) :
ScheduledObject(name, si) {

    // Check skew random number generators, initialize if not already
    // done.  There are two, one for each of the timeouts
    if (skewGenerator2 == NULL) {
	skewGenerator2 = new Uniform(0, 1.0 / PR_FASTHZ,
				     si->rng);
    }
    if (skewGenerator5 == NULL) {
	skewGenerator5 = new Uniform(0, 1.0 / PR_SLOWHZ,
				     si->rng);
    }

    // Initialize first TcpPcb record.  It contains the first
    // (next) "ephemeral" port number.
    tcb = new TcpPcb;
    tcb->tcp_lport = TcpMinEphemeralPort;
    tcb->next = tcb;
    tcb->prev = tcb;

    // Initialize 200 ms and 500 ms timers, with random phase offsets
    // to avoid synchronization effects.
    Post((*skewGenerator2)(), this, EventTimer, NULL);
    Post((*skewGenerator5)(), this, EventTimer2, NULL);

    // Initialize globals for TCP entity
    rexmtthresh = 3;
    mssdflt = 512;		        // default MSS 512 bytes
    rttdflt = 3;			// 3 seconds
    keepidle = 120 * 60 * PR_SLOWHZ;	// 2 hours
    keepintvl = 75 * PR_SLOWHZ;		// 75 seconds
    keepcnt = 8;			// 8 keepalive probes
    maxidle = keepcnt * keepintvl;
    tcp_iss = 0;

    // Initialize timing parameters
    latencyToIp = 0.000001;              // 1 us to network
    latencyToApp = 0.000005;		 // 5 us to application

}

//
// Tcp::~Tcp
//
// Object destructor
//
Tcp::~Tcp()
{
    delete tcb;
}

//
// Tcp::Handler
//
// (Event handler)
//
void
Tcp::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    // Store event time...a bunch of routines depend on this and it's
    // too cumbersome to pass this around to everybody.
    this->time = time;

    // Handle different types of events
    switch (type) {

	// Timer event (fast 200 ms timer).
	case (EventTimer):
	Fast();
	Post(time + (1.0 / PR_FASTHZ), this, EventTimer, NULL);
	break;

	// Timer2 event (slow 500 ms timer).
	case (EventTimer2):
	Slow();
	Post(time + (1.0 / PR_SLOWHZ), this, EventTimer2, NULL);
	break;

	// Arrival of an IP packet
	case (EventIpArrival):
	In((IpPacket *) data);
	break;

	// TCP segment to send from an application.  data contains a pointer
	// to a TcpCall structure, 
	case (EventTcpSend):
	{
	    TcpCall *callblock = (TcpCall *) data;
	    UserSend(time, callblock->pcb, callblock->th_ulen,
		     callblock->data);
	}
	break;

	// Unknown event type
	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("Tcp");
		*(sim->debugStream) << "bad EventType " << type << "\n";
	    });
	}
	break;

    }
}

//
// Tcp::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
Tcp::Output()
{
    return name + " [Tcp]";
}

//
// Tcp::Fast
//
// Input:  none
//
// Output:  none
//
// Handle fast (200 ms) timeouts.  The only timer using this timeout
// is the delayed ACK timer.  [tcp_fasttimo]
//
void
Tcp::Fast(void)
{
    TcpPcb *currentPcb;		// which connection we're currently
				// working with

    IF_DEBUG(4, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Fast()\n";
    });

    // Iterate over all PCBs.  Because this is a circular DLL, we
    // eventually hit the header record again.
    for (currentPcb = tcb->next; currentPcb != tcb;
	 currentPcb = currentPcb->next) {

	// If we're suppose to do a delayed acknowledgement for this
	// connection, reset the flag, set to do an acknowlegement, and
	// send a segment.
	if (currentPcb->flags & TF_DELACK) {
	    currentPcb->flags &= ~TF_DELACK;
	    currentPcb->flags |= TF_ACKNOW;

	    IF_DEBUG(3, {
		DEBUG_PREFIX("Tcp");
		*(sim->debugStream) << "ready to send delayed ACK\n";
	    });

	    Out(currentPcb, 0);
	}
    }
}

//
// Tcp::Slow
//
// Input:  none
//
// Output:  none
//
// Handle slow (500 ms) timeouts.  [tcp_slowtimo]
//
void
Tcp::Slow(void)
{

    TcpPcb *currentPcb;		// which connection we're looking at
				// now
    TcpPcb *nextPcb;		// next PCB to look at
    int i;			// generic loop counter
    
    IF_DEBUG(4, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Slow()\n";
    });

    // Iterate over all the PCBs.  Remember that this is a circular
    // DLL, so we'll eventually come back to the beginning of the list
    // (the header record).
    for (currentPcb = tcb->next; currentPcb != tcb; currentPcb = nextPcb) {

	// Keep track of the next PCB in the DLL.  This is necessary
	// in case the current PCB goes away due to a side-effect of
	// one of the timers going off.
	nextPcb = currentPcb->next;
	
	// Iterate over the timers
	for (i = 0; i < TCPT_NTIMERS; i++) {

	    // If timer was supposed to go off, call Tcp::Timers to
	    // handle it.
	    if (currentPcb->timer[i] && --currentPcb->timer[i] == 0) {
		Timers(currentPcb, i);

		// It's possible the connection (and its PCB) went away,
		// so in this case, skip all remaining processing for this
		// connection.
		if (nextPcb->prev != currentPcb) {
		    goto tpgone;
		}
	    }
	}

	// Increment various counters for this connection.
	currentPcb->idle++;
	if (currentPcb->rtt) {
	    currentPcb->rtt++;
	}

	tpgone:
	;

    }

    // Increment iss for next connection
    tcp_iss += TCP_ISSINCR / PR_SLOWHZ;

}

// Tcp::CancelTimers
//
// Input:  TCP PCB for a TCP connection (tp)
//
// Output:  None
//
// Turn off all timers [tcp_canceltimers]
//
void
Tcp::CancelTimers(TcpPcb *tp)
{
    int i;

    for (i = 0; i < TCPT_NTIMERS; i++) {
	tp->timer[i] = 0;
    }
}

//
// Tcp::Timers
//
// Input;  TCP PCB for a TCP connection (tp), timer number (timer)
//
// Output:  (in return value)
//
// Called to handle various slow timeouts.  [tcp_timers]
//
TcpPcb
*Tcp::Timers(TcpPcb *tp, int timer)
{

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Timers\n";
    });

    // Walk though the various types of timers
    switch (timer) {

	// The 2MSL timer handles timeouts from both the FIN_WAIT_2 and
	// TIME_WAIT states.  Since a connection will never be in both
	// states at once, one timer suffices.
	case (TCPT_2MSL):
	{

	    // If we're not in the TIME_WAIT state, we must be in FIN_WAIT_2.
	    // If this is true and we haven't been idle for more than 10
	    // minutes, run the timer again in another 75 seconds.
	    if ((tp->state != TcpsTimeWait) && (tp->idle <= maxidle)) {
		tp->timer[TCPT_2MSL] = keepintvl;
	    }

	    // Otherwise we were in TIME_WAIT, or in FIN_WAIT_2 and idle
	    // longer than 10 minutes.  Under these circumstances, we close
	    // the connection.
	    else {
		tp = Close(tp);
	    }

	}
	break;

	// The persist timer forces a probe of a zero-sized window.
	case (TCPT_PERSIST):
	{
	    // Calculate new value for persist timer and force a segment
	    // to be output.
	    SetPersist(tp);
	    Out(tp, 1);
	}
	break;

	// The keepalive timer has two functions.  First, it ensures that a
	// new connection enters the established state within 75 seconds.
	// Second, it forces a probe of the other side of any
	// connection that hasn't been heard from for 2 hours.
	// Again, we can reuse this particular timer because a connection
	// either hasn't yet entered the ESTABLISHED state (first case)
	// or it has (second case).
	case (TCPT_KEEP):
	{

	    // If we aren't yet established, then the 75-second timer went
	    // off and we lose.
	    if (tp->state < TcpsEstablished) {
		goto dropit;
	    }

	    // Otherwise, we're in a position to send keepalive packets.
	    if (tp->state < TcpsCloseWait) {

		// If the connection has been idle too long (connection idle
		// plus nine keepalive packets have not been acknowledged),
		// then we lose and give up.
		if (tp->idle >= keepidle + maxidle) {
		    goto dropit;
		}

		// Otherwise, compose a keepalive probe
		Respond(tp, NULL, tp->rcv_nxt, tp->snd_una - 1, 0);
		tp->timer[TCPT_KEEP] = keepintvl;

	    }
	    else {
		tp->timer[TCPT_KEEP] = keepidle;
	    }
	    break;

	    dropit:
	    tp = Drop(tp);

	}
	break;
	
	// Retransmission timer.
	case (TCPT_REXMT):
	{

	    int rexmt;		// used in retransmission timer calculations

	    // Increment the shift count for the retransmit.  If we
	    // had to increment too much, give up on the connection.
	    if (++tp->rxtshift > TCP_MAXRXTSHIFT) {
		tp->rxtshift = TCP_MAXRXTSHIFT;
		tp = Drop(tp);
		break;
	    }

	    rexmt = TCP_REXMTVAL(tp) * Backoff[tp->rxtshift];
	    rexmt = min(rexmt, TCPTV_REXMTMAX);
	    rexmt = max(rexmt, tp->rttmin);
	    tp->rxtcur = rexmt;

	    // If the segment has been retransmitted four or more
	    // times (more than a quarter of the maximum, which is 12)
	    // then clobber the RTT extimate.
	    if (tp->rxtshift > (TCP_MAXRXTSHIFT / 4)) {
		tp->rttvar += (tp->srtt >> TCP_RTT_SHIFT);
		tp->srtt = 0;
	    }
	    
	    tp->snd_nxt = tp->snd_una;

	    tp->rtt = 0;

	    // Close the congestion window to one segment.  See the
	    // Reno source code for more details.
	    {
		unsigned int win = min(tp->snd_wnd, tp->snd_cwnd) /
		  2 / tp->maxseg;
		if (win < 2) {
		    win = 2;
		}
		tp->snd_cwnd = tp->maxseg;
		tp->snd_ssthresh = win * tp->maxseg;
		tp->dupacks = 0;

		IF_DEBUG(3, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "cwnd one segment " <<
		      tp->snd_cwnd << "\n";
		});
	    }

	    // Output another TCP segment
	    Out(tp, 0);

	}
	break;

    }

    return tp;

}

//
// Tcp::SetPersist
//
// Input:  PCB for TCP connection (tp)
//
// Output:  (in return value)
//
// Tcp::SetPersist calculates new value for the persist timer.
// [tcp_setpersist]
//
void
Tcp::SetPersist(TcpPcb *tp)
{
    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::SetPersist\n";
    });

    short t = ((tp->srtt >> 2) + tp->rttvar) >> 1;

    tp->timer[TCPT_PERSIST] = t * Backoff[tp->rxtshift];
    if (tp->timer[TCPT_PERSIST] < TCPTV_PERSMIN) {
	tp->timer[TCPT_PERSIST] = TCPTV_PERSMIN;
    }
    else if (tp->timer[TCPT_PERSIST] > TCPTV_PERSMAX) {
	tp->timer[TCPT_PERSIST] = TCPTV_PERSMAX;
    }
     
    if (tp->rxtshift < TCP_MAXRXTSHIFT) {
	tp->rxtshift++;
    }

}

//
// Tcp::XmitTimer
//
// Input:  TCP PCB (tp), latest round trip figure expressed in slow
// timeout clock ticks (rtt).
//
// Output:  None
//
// Recompute figures relating to RTT estimate for a TCP connection.
// [tcp_xmit_timer]
//
void
Tcp::XmitTimer(TcpPcb *tp, short rtt)
{

    short delta;		// difference between estimate and
				// actual, in slow clock ticks.

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::XmitTimer\n";
    });

    // If srtt > 0, then we have a round-trip estimate figure.
    if (tp->srtt != 0) {
	
	// SRTT <- RTT/8 + (7/8)*SRTT
	delta = rtt - 1 - (tp->srtt >> TCP_RTT_SHIFT);
	if ((tp->srtt += delta) < 0) {
	    tp->srtt = 1;
	}
	
	// Compute smoothed mean difference
	if (delta < 0) {
	    delta = -delta;
	}
	delta -= (tp->rttvar >> TCP_RTTVAR_SHIFT);
	if ((tp->rttvar += delta) <= 0) {
	    tp->rttvar = 1;
	}

    }

    // No smoothed RTT estimates available yet, set SRTT = RTT,
    // RTTVAR = RTT/2.
    else {
	tp->srtt = rtt << TCP_RTT_SHIFT;
	tp->rttvar = rtt << (TCP_RTTVAR_SHIFT - 1);
    }
    
    IF_DEBUG(3, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "Tcp::XmitTimer rtt " << rtt <<
	  " --> srtt " << tp->srtt << ">>" << TCP_RTT_SHIFT <<
	    " rttvar " << tp->rttvar << ">>" << TCP_RTTVAR_SHIFT << "\n";
    });
    
    // Prepare for next RTT estimate.
    tp->rtt = 0;
    tp->rxtshift = 0;

    tp->rxtcur = TCP_REXMTVAL(tp);
    if (tp->rxtcur < tp->rttmin) {
	tp->rxtcur = tp->rttmin;
    }
    if (tp->rxtcur > TCPTV_REXMTMAX) {
	tp->rxtcur = TCPTV_REXMTMAX;
    }
}

//
// Tcp::Drop
//
// Input:  TCP PCB (tp)
//
// Output;  (return value)
//
// Abort a connection by sending an RST.  [tcp_drop]
//
TcpPcb
*Tcp::Drop(TcpPcb *tp)
{

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Drop\n";
    });

    // If we've already exchanged SYN segments with the other side,
    // we need to send an RST.  We do this by setting our state to close
    // and calling the output function.
    if (tp->state >= TcpsSynReceived) {
	tp->state = TcpsClosed;
	Out(tp, 0);
    }
    
    // Release resources
    return Close(tp);

}

//
// Tcp::Close
//
// Input:  TCP PCB (tp)
//
// Output:  (return value)
//
// Finish the close of a TCP connection and release resources.  Note this
// procedure is considerably shorter than its BSD 4.4Lite counterpart,
// because our TCP implementation doesn't support per-route metrics.
// [tcp_close]
//
TcpPcb
*Tcp::Close(TcpPcb *tp)
{

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Close\n";
    });

    // Release segments on reassembly queue.  This is just the standard
    // deallocation of a linked list
    Reassem *next, *current;
    for (current = tp->seg_next; current != NULL; current = next)
    {
	next = current->next;
	delete current;
    }
    
    // We do an asynchronous notification to the process, telling it
    // that the connection's gone.  We need to be careful here, in
    // order to avoid posting an EventTcpClosed for a connection
    // that was closed by order of the local process.
    if ((tp->state != TcpsClosed) &&
	(tp->state <= TcpsEstablished) &&
	(tp->state != TcpsListen)) {

	Post(time + latencyToApp, tp->tcp_up, EventTcpClosed, tp);

	IF_DEBUG(3, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "Tcp::Close posting EventTcpClosed\n";
	});

    }
    tp->prev->next = tp->next;
    tp->next->prev = tp->prev;
    delete(tp);

    return NULL;

}

//
// Tcp::Out
//
// Input:  PCB for TCP connection (tp), force segment output flag (force)
//
// Output:  (in return value)
//
// Tcp::Out is a generalized output routine for all TCP segments
// on a connection.  [tcp_output]
//
int
Tcp::Out(TcpPcb *tp, int force)
{

    bool idleFlag;		// connection idle?
    bool sendalotFlag = false;	// more than one segment to output?

    int len;			// number of bytes possible to send
    int win;			// current applicable window size
    int rwin;			// receive window size for window adverts.
    int off;			// number of bytes in the send buffer
				// awaiting ACK
    int flags;			// TCP flags for TCP header
    double delayOut = 0.0;	// accumulated delay for packets going to IP

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Out\n";
    });

    // First see if we should drop
    // back to slowstart, in case we haven't heard from the other side
    // for awhile and no ACKs are expected.
    idleFlag = (tp->snd_max == tp->snd_una);
    if (idleFlag && tp->idle >= tp->rxtcur) {
	tp->snd_cwnd = tp->maxseg;

	IF_DEBUG(3, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "cwnd idle now " << tp->snd_cwnd << "\n";
	});
    }

    // In a quest to attempt to improve the readability of TCP (at the
    // expense of some performance), we've rewritten the sending loop to
    // eliminate a gotos.
    do {

	off = tp->snd_nxt - tp->snd_una;
	win = min(tp->snd_wnd, tp->snd_cwnd);
	flags = Outflags[tp->state];

	//
	// Test to see if we should send anything.  These routines
	// will set sendFlag true.
	//
	// Handle the case of a persist timeout
	if (force) {

	    // If the window size is zero, then we need to make up a single
	    // byte of data to send.  If we have a byte of data to send,
	    // then send it, otherwise send a FIN (?).
	    if (win == 0) {
		if (off < tp->so_snd_sb_cc) {
		    flags &= ~TH_FIN;
		}
		win = 1;
	    }
	    else {
		tp->timer[TCPT_PERSIST] = 0;
		tp->rxtshift = 0;
	    }

	}

	// Figure out how much data to send (if any)
	len = min(tp->so_snd_sb_cc, win) - off;

	// If FIN sent but not ACKed and not retransmitted, len will be -1.
	// Otherwise we sent data into a window that later shrunk on us.
	// In either case, cancel pending retransmissions and take note of
	// the shrunken window.
	if (len < 0) {
	    len = 0;
	    if (win == 0) {
		tp->timer[TCPT_REXMT] = 0;
		tp->snd_nxt = tp->snd_una;

		IF_DEBUG(3, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "REXMT timer reset\n";
		});

	    }
	}
	
	// If send length greater than maximum segment length,
	// we'll need to do this in multiple segments.
	if (len > tp->maxseg) {
	    len = tp->maxseg;
	    sendalotFlag = true;
	}
	else {
	    sendalotFlag = false;
	}
	    
	// If we can't send out all of the buffered-up data, we
	// sure can't send a FIN (if we were supposed to do so).
	if (SEQ_LT(tp->snd_nxt + len, tp->snd_una + tp->so_snd_sb_cc)) {
	    flags &= ~TH_FIN;
	}

	// Calculate receive window size.  Note this differs from the
	// BSD 4.4Lite code, which re-uses the variable win for this
	// purpose.
	rwin = tp->so_rcv_sb_hiwat - // inline expansion of sbspace()
	  tp->so_rcv_sb_cc;

	// Silly window avoidance (sender side)
	if (len) {
	    if (len == tp->maxseg) {
		goto send;
	    }
	    if ((idleFlag ||
		 (tp->flags & TF_NODELAY)) &&
		(len + off >= tp->so_snd_sb_cc)) {
		goto send;
	    }
	    if (force) {
		goto send;
	    }
	    if (len >= tp->max_sndwnd / 2) {
		goto send;
	    }
	    if (SEQ_LT(tp->snd_nxt, tp->snd_max)) {
		goto send;
	    }
	}

	// See if we should send a window update
	if (rwin > 0) {

	    int adv;		// amount by which the window has increased

	    // If window has advanced by more than two segments, or
	    // it has advanced by more than half of the receive buffer.
	    adv = min(rwin, TCP_MAXWIN) - (tp->rcv_adv - tp->rcv_nxt);
	    if (adv >= (2 * tp->maxseg)) {
		goto send;
	    }
	    if (2 * adv >= tp->so_rcv_sb_hiwat) {
		goto send;
	    }
	}

	// Various other flags may require our TCP entity to send a
	// segment, for example, needing to send an ACK.
	if (tp->flags & TF_ACKNOW) {
	    goto send;
	}
	if (flags & (TH_SYN | TH_RST)) {
	    goto send;
	}
	if (SEQ_GT(tp->snd_up, tp->snd_una)) {
	    goto send;
	}
	if (flags & TH_FIN &&
	    (((tp->flags & TF_SENTFIN) == 0) ||
	     (tp->snd_nxt == tp->snd_una))) {
	    goto send;
	}

	//    
	// Send something only if we're supposed to...otherwise jump out
	// of the sending loop...here!
	//
	// We might need to set the persist timer if we still have data to
	// send.
	//
	if (tp->so_snd_sb_cc &&
	    (tp->timer[TCPT_REXMT] == 0) &&
	    (tp->timer[TCPT_PERSIST] == 0)) {
	    tp->rxtshift = 0;
	    SetPersist(tp);
	}
	break;

	//
	// Send a segment!
	//
	// Compose an output segment and Post it to the Ip module.
	// Originally we wanted to get rid of the label send: below
	// eliminate about a dozen gotos (in hopes of making the code
	// more "structured").  This was a lose...too much work.
	//
send:	

	// If we're doing a SYN, then set the MSS-related parameters
	if (flags & TH_SYN) {
	    Mss(tp);
	}

	// Grab IP and TCP message structures.
	TcpSegment *tcpSegment = new TcpSegment;
	IpPacket *ipPacket = new IpPacket(tcpSegment);
	
	// Initialize selected header fields (this is done in BSD 4.4Lite
	// by copying the t_template buffer for the connection)
	ipPacket->ip_tos = 0;
	ipPacket->ip_len = len + TcpHeaderSize + IpHeaderSize;
	ipPacket->ip_id = 0;
	ipPacket->ip_off = 0;
	ipPacket->ip_ttl = 0;
	ipPacket->ip_p = IPPROTO_TCP;
	ipPacket->ip_src = tp->tcp_laddr;
	ipPacket->ip_dst = tp->tcp_faddr;
	ipPacket->ip_next.s_addr = 0;

	tcpSegment->th_sport = tp->tcp_lport;
	tcpSegment->th_dport = tp->tcp_fport;
	
	// Set push flag if we're sending all of our data
	if (len) {
	    if (off + len == tp->so_snd_sb_cc) {
		flags |= TH_PUSH;
	    }
	}
    
	// If we're resending a FIN flag, be sure not to use a new
	// sequence number.
	if ((flags & TH_FIN) &&
	    (tp->flags & TF_SENTFIN) &&
	    (tp->snd_nxt == tp->snd_max)) {
		tp->snd_nxt--;
	}

	// Set th_seq, th_ack, and th_flags
	if (len || (flags & (TH_SYN | TH_FIN)) || tp->timer[TCPT_PERSIST]) {
	    tcpSegment->th_seq = tp->snd_nxt;
	}
	else {
	    tcpSegment->th_seq = tp->snd_max;
	}

	tcpSegment->th_ack = tp->rcv_nxt;
	tcpSegment->th_flags = flags;
	tp->so_snd.CopyData(tcpSegment->payload, tcpSegment->th_seq, len);
	    
	// Calculate receive window and store it
	if ((rwin < (tp->so_rcv_sb_hiwat / 4)) &&
	    (rwin < tp->maxseg)) {
	    rwin = 0;
	}
	if (rwin > TCP_MAXWIN) {
	    rwin = TCP_MAXWIN;
	}
	if (rwin < (tp->rcv_adv - tp->rcv_nxt)) {
	    rwin = tp->rcv_adv - tp->rcv_nxt;
	}
	tcpSegment->th_win = rwin;

	// If we're sending normally...
	if ((force == 0) || (tp->timer[TCPT_PERSIST] == 0)) {

	    TcpSeq startseq;	// start sequence number this segment,
				// used for timing

	    startseq = tp->snd_nxt;

	    // Advance snd_nxt over this segment
	    if (flags & (TH_SYN | TH_FIN)) {
		if (flags & TH_SYN) {
		    tp->snd_nxt++;
		}
		if (flags & TH_FIN) {
		    tp->snd_nxt++;
		    tp->flags |= TF_SENTFIN;
		}
	    }
	    tp->snd_nxt += len;
	    
	    // Update snd_max if necessary
	    if (SEQ_GT(tp->snd_nxt, tp->snd_max)) {
		tp->snd_max = tp->snd_nxt;

		// If this segment isn't a retransmission and we're not
		// timing anything right now, then start RTT counter
		// and remember the starting sequence number of this
		// segment.
		if (tp->rtt == 0) {
		    tp->rtt = 1;
		    tp->rtseq = startseq;
		}
	    }

	    // If retransmit timer isn't set and we're not doing an
	    // ACK or a keepalive probe, then set retransmit timer.
	    if ((tp->timer[TCPT_REXMT] == 0) &&
		(tp->snd_nxt != tp->snd_una)) {

		tp->timer[TCPT_REXMT] = tp->rxtcur;
		if (tp->timer[TCPT_PERSIST]) {
		    tp->timer[TCPT_PERSIST] = 0;
		    tp->rxtshift = 0;
		}

		IF_DEBUG(3, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "REXMT timer set to " <<
		      tp->timer[TCPT_REXMT] << "\n";
		});

	    }
	}

	// We're persisting...
	else {
	    if (SEQ_GT(tp->snd_nxt + len, tp->snd_max)) {
		tp->snd_max = tp->snd_nxt + len;
	    }
	}

	// Pass off the packet to the IP routing module
	delayOut += latencyToIp;
	Lookup(down, downName);
	Post(time + delayOut, down, EventIpSend, ipPacket);

	IF_DEBUG(2, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "Tcp::Out sent segment " <<
	      ipPacket << "\n";
	});

	// Post-segment processing...remember highest window
	// advertisement.  Reset connection flags.
	if ((rwin > 0) &&
	    SEQ_GT(tp->rcv_nxt + rwin, tp->rcv_adv)) {
	    tp->rcv_adv = tp->rcv_nxt + rwin;
	}
//	tp->last_ack_sent = tp->rcv_nxt;
	tp->flags &= ~(TF_ACKNOW | TF_DELACK);

    } while (sendalotFlag);

    return(0);

}

//
// Tcp::Respond
//
// Input:  PCB for TCP connection (tp), stimulus IP packet (m
// [optional]), acknowldgement and segment
// sequence numbers (ack, seg), TCP flags (flags)
//
// Output:  none
//
// Special-purpose TCP segment output routine.  [tcp_respond]
//
void
Tcp::Respond(TcpPcb *tp, IpPacket *m, TcpSeq ack, TcpSeq seq, int flags)
{

    int win = 0;		// window size for advertisement
    TcpSegment *th;		// TCP header for packet

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Respond\n";
    });

    if (tp) {
	win = tp->so_rcv_sb_hiwat - tp->so_rcv_sb_cc; // expansion of sbspace()
    }
    
    // If m wasn't passed in, we're generating a keepalive probe.  Use
    // the addressing info stored in the TcpPcb to construct a new
    // TCP segment.
    if (m == 0) {

	// Get new packet headers for keepalive
	th = new TcpSegment;
	m = new IpPacket(th);

	// Copy stuff into headers
	m->ip_tos = 0;
	m->ip_id = 0;
	m->ip_off = 0;
	m->ip_ttl = 0;
	m->ip_p = IPPROTO_TCP;
	m->ip_src = tp->tcp_laddr;
	m->ip_dst = tp->tcp_faddr;
	m->ip_next.s_addr = 0;
	m->payload = th;
	
	flags = TH_ACK;

    }

    // If m *was* passed in, we're generating a RST segment in response
    // to incoming segment m.  We reuse the incoming segment and send it
    // right back out.
    else {
	th = (TcpSegment *) m->payload;	// get TCP header for inbound packet

	m->ip_next.s_addr = 0;
	exchange(m->ip_src.s_addr, m->ip_dst.s_addr, u_int);
	exchange(th->th_dport, th->th_sport, u_short);
    }

    // At this point, m points to a (somewhat) valid IP header and
    // th points to a (mostly) valid TCP header.  Complete initialization
    // of all the fields that weren't set above.
    m->ip_len = IpHeaderSize + TcpHeaderSize; // no TCP data
    th->th_seq = seq;
    th->th_ack = ack;
    th->th_flags = flags;
    th->th_win = win;		// no window scaling

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "sent packet " << m << "\n";
    });

    // Send it!!!
    Lookup(down, downName);
    Post(time + latencyToIp, down, EventIpSend, m);

}

//
// Tcp::In
//
// Input:  IP packet from the IP layer (ipPacket)
//
// Output:  none
//
// TCP input processing.  [tcp_input]
//
void
Tcp::In(IpPacket *ipPacket)
{

    bool dropsocket = false;	// drop socket on error (used at connection
				// setup time)
    bool needoutput = false;	// flag to restart retransmit timer
    bool ourfinisacked;		// needed for FIN processing
    TcpPcb *tp;			// PCB for this connection
    TcpPcb *wcPcb;		// wildcard PCB for PCB matching algorithm
    TcpSegment *th = (TcpSegment *) ipPacket->payload; // TCP header/message
    int tiflags;		// working copy of flags
    int seglen;			// segment length
    int acked;			// number of unacked bytes acknowledged
    int todrop;			// amount of data to drop from segment
    int iss = 0;		// initial sequence number
    
    // We can skip a lot of code at the start of this function because
    // we don't need to deal with mbuf manipulation, checksums, or
    // TCP options.
    seglen = ipPacket->ip_len - IpHeaderSize - TcpHeaderSize;
    tiflags = th->th_flags;

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "Tcp::In got segment " << ipPacket << "\n";
    });

findpcb:

    // Locate PCB for this segment.  For right now we don't support the
    // TCP PCB cache, but it should be easy to add in...
    //
    // Caution;  This algorithm isn't too much like the real BSD 4.4Lite
    // PCB finding algorithm...we've kind of simplified things here.
    // A few assumptions:  Either both components of the remote side are
    // specified or they're both wildcards.  PCB endpoints need to be
    // unique.
    tp = NULL;
    wcPcb = NULL;
    for (TcpPcb *tryPcb = tcb->next; tryPcb != tcb; tryPcb = tryPcb->next) {

	// First see if local side matches up (or fits wildcard).
	if ((tryPcb->tcp_lport == th->th_dport) &&
	    ((tryPcb->tcp_laddr.s_addr == 0) ||
	     (tryPcb->tcp_laddr.s_addr == ipPacket->ip_dst.s_addr))) {

	    // Now check to see if remote side matches.  If so, then we're
	    // done.
	    if ((tryPcb->tcp_fport == th->th_sport) &&
		(tryPcb->tcp_faddr.s_addr == ipPacket->ip_src.s_addr)) {
		tp = tryPcb;
		break;
	    }

	    // Otherwise, try to find a wildcard match.  Save this in
	    // case we encounter an exact match later.
	    if ((tryPcb->tcp_faddr.s_addr == 0) &&
		(tryPcb->tcp_fport == 0)) {
		wcPcb = tryPcb;
	    }
	}
    }

    // If we found a wildcard but not an exact match, then use the
    // wildcard (presumably a connection in LISTEN state, but we don't
    // check for that here).
    if (!tp && wcPcb) {
	tp = wcPcb;
    }

    // If the state is closed (i.e. PCB not in Pacific Bell Smart Yellow
    // Pages, er, the DLL of PCBs, and therefore doesn't exist), then
    // drop the segment and send a reset.
    if (tp == NULL) {

	IF_DEBUG(1, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "demux failure on packet from " << ipPacket->ip_src << " remote port " << th->th_sport << " local port " << th->th_dport << "\n";
	});

	goto dropwithreset;
    }
    if (tp->state == TcpsClosed) {
	goto drop;
    }

    // Handle listening sockets here.  We need to clone a new TcpPcb
    // if a SYN came in (set to LISTEN) state.  Set a flag to drop
    // the socket if an error was encountered.
    if (tp->state == TcpsListen) {

	// Clone the PCB and fill in a few more fields to get it linked
	// into the DLL and properly associate it with the connection
	// endpoints
	TcpPcb *newPcb = new TcpPcb(*tp);

	tcb->next->prev = newPcb;
	newPcb->next = tcb->next;
	tcb->next = newPcb;
	newPcb->prev = tcb;

	newPcb->tcp_faddr.s_addr = ipPacket->ip_src.s_addr;
	newPcb->tcp_fport = th->th_sport;
	newPcb->tcp_laddr.s_addr = ipPacket->ip_dst.s_addr;
	newPcb->tcp_lport = th->th_dport;
	newPcb->snd_bytes = 0;
	newPcb->rcv_bytes = 0;
	
	// Work with the new PCB from now on...leave the old one alone.
	tp = newPcb;

	dropsocket = true;
    }

    // Reset idle time and keepalive, since we got what appears to be
    // a valid segment (so far)
    tp->idle = 0;
    tp->timer[TCPT_KEEP] = keepidle;

    // Try to do header prediction.  First see if we can do header
    // prediction on this segment.  Then test for the two cases of
    // the sender and receiver of a bulk data transfer.
    if ((tp->state == TcpsEstablished) &&
	((tiflags & (TH_SYN | TH_FIN | TH_RST | TH_ACK)) == TH_ACK) &&
	(th->th_seq == tp->rcv_nxt) &&
	(th->th_win) &&
	(th->th_win == tp->snd_wnd) &&
	(tp->snd_nxt == tp->snd_max)) {

	// If the segment length is zero, it might be a pure ACK segment.
	// We also need to check to make sure the ack is for outstanding
	// data, the ack isn't past the maximum range of acks, and that
	// we're not in slowstart or congestion avoidance.
	if (seglen == 0) {
	    if (SEQ_GT(th->th_ack, tp->snd_una) &&
		SEQ_LEQ(th->th_ack, tp->snd_max) &&
		tp->snd_cwnd >= tp->snd_wnd) {

		// If we were doing roundtrip timing and the ACK was
		// for some data after the starting sequence number
		// we were timing, then go update the RTT estimators.
		if (tp->rtt &&
		    SEQ_GT(th->th_ack, tp->rtseq)) {
		    XmitTimer(tp, tp->rtt);
		}
		
		// We can get rid of data from the send buffer.  Also
		// we can get rid of the inbound segment (0 bytes of data,
		// just the headers)
		acked = th->th_ack - tp->snd_una;
		tp->snd_una = th->th_ack;
		tp->so_snd_sb_cc -= acked;
//		delete (th);
		delete (ipPacket);

		// If all outstanding data has been ACKed, then stop the
		// retransmit timer, otherwise restart it.
		if (tp->snd_una == tp->snd_max) {
		    tp->timer[TCPT_REXMT] = 0;
		}
		else if (tp->timer[TCPT_PERSIST] == 0) {
		    tp->timer[TCPT_REXMT] = tp->rxtcur;
		}

		IF_DEBUG(3, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "REXMT timer set to " <<
		      tp->timer[TCPT_REXMT] << "\n";
		});

		// If more data left to send, do it
		if (tp->so_snd_sb_cc) {
		    Out(tp, 0);
		}
		return;
	    }
	}

	// It might be a pure data segment.  Make sure it's in-sequence,
	// we don't have any out-of-order segments, and there's space
	// for it in the buffer.
	else if ((th->th_ack == tp->snd_una) &&
		 (tp->seg_next == NULL) && // no out of order segments
		 (seglen <= tp->so_rcv_sb_hiwat -
		  tp->so_rcv_sb_cc)) { // sbspace

	    // Give the data to the application
	    TcpCall *call = new TcpCall;
	    call->pcb = tp;
	    call->th_ulen = seglen;
	    th->payload.AppendTo(call->data);
	    
	    Post(time + latencyToApp,
		 tp->tcp_up,
		 EventTcpArrival,
		 call);

	    time += latencyToApp;

//	    delete th;
	    delete ipPacket;

	    tp->rcv_nxt += seglen;
	    tp->flags |= TF_DELACK;
	    tp->rcv_bytes += seglen;

	    // Sometime after this (on a "real" system) the application
	    // would read the data from its socket buffer.  This causes
	    // tcp_usrreq() to be called with a PRU_RCVD request, which
	    // calls tcp_output(), which may potentially generate a window
	    // update.  Since our assumption is that the application is
	    // fast enough to take all data being sent to it, we call
	    // our Tcp::Out() routine from here.
	    Out(tp, 0);

	    return;

	}
    }

    // If we got down this far, we lose and need to do slow-path
    // protocol processing.  Doh!

    // Calculate amount of space in receive window.
    {
	int win;		// new window size

	win = tp->so_rcv_sb_hiwat - tp->so_rcv_sb_cc; // sbspace
	if (win < 0) {
	    win = 0;
	}
	tp->rcv_wnd = max(win, tp->rcv_adv - tp->rcv_nxt);
    }

    // Input processing.  Figure out what to do with the segment.
    // Basically we do the connection management stuff here.
    switch (tp->state) {

	// LISTEN state...this segment should start a passive
	// open into SYN_RECEIVED.
	case TcpsListen:
	{
	    
	    // Filter out bad segments (bad flags)
	    if (tiflags & TH_RST) {
		goto drop;
	    }
	    if (tiflags & TH_ACK) {
		goto dropwithreset;
	    }
	    if (!(tiflags & TH_SYN)) {
		goto drop;
	    }

	    // Fill in TcpPcb fields that weren't set originally. XXX
	    if (iss) {
		tp->iss = iss;
	    }
	    else {
		tp->iss = tcp_iss;
	    }
	    tp->iss += TCP_ISSINCR / 2;
	    tp->irs = th->th_seq;

#define tcp_rcvseqinit(tp) \
  (tp)->rcv_adv = (tp)->rcv_nxt = (tp)->irs + 1
	    tcp_rcvseqinit(tp);

#define tcp_sendseqinit(tp) \
  (tp)->snd_una = (tp)->snd_nxt = (tp)->snd_max = (tp)->snd_up = (tp)->iss
	    tcp_sendseqinit(tp);

	    tp->flags |= TF_ACKNOW;
	    tp->state = TcpsSynReceived;
	    tp->timer[TCPT_KEEP] = TCPTV_KEEP_INIT;
	    dropsocket = false;

	    IF_DEBUG(2, {
		DEBUG_PREFIX("Tcp");
		*(sim->debugStream) << "cloned new LISTEN connection into SYN_RCVD\n";
	    });

	    goto trimthenstep6;

	}

	// SYN_SENT state.  We're expecting to get back a SYN...make
	// sure we got it.
	case TcpsSynSent:
	{

	    // Drop the packet if it was bad
	    if ((tiflags & TH_ACK) &&
		(SEQ_LEQ(th->th_ack, tp->iss) ||
		 SEQ_GT(th->th_ack, tp->snd_max))) {
		goto dropwithreset;
	    }
	    if (tiflags & TH_RST) {
		if (tiflags &TH_ACK) {
		    tp = Drop(tp);
		    goto drop;
		}
	    }
	    if (!(tiflags & TH_SYN)) {
		goto drop;
	    }

	    // If we got an ACK, then set the boundary of the unacknowledged
	    // data to the ACK value.
	    if (tiflags & TH_ACK) {
		tp->snd_una = th->th_ack;
		if (SEQ_LT(tp->snd_nxt, tp->snd_una)) {
		    tp->snd_nxt = tp->snd_una;
		}
	    }

	    // Set a bunch of other variables...
	    tp->timer[TCPT_REXMT] = 0;
	    tp->irs = th->th_seq;
	    tcp_rcvseqinit(tp);
	    tp->flags |=  TF_ACKNOW;

	    IF_DEBUG(3, {
		DEBUG_PREFIX("Tcp");
		*(sim->debugStream) << "REXMT timer reset\n";
	    });

	    // If we got an ACK and a SYN in this packet, we've already
	    // sent a SYN, so we're established!
	    if ((tiflags & TH_ACK) &&
		SEQ_GT(tp->snd_una, tp->iss)) {
		
		tp->state = TcpsEstablished;

		// Deal with any fragments that might have arrived
		{
		    TcpDataList nullList;
		    Reass(tp, 0, 0, 0, nullList);
		}

		// Use the RTT of the SYN as initial SRTT and RTTvar
		// if we didn't need to retransmit it.
		if (tp->rtt) {
		    XmitTimer(tp, tp->rtt);
		}

		Post(time + latencyToApp, tp->tcp_up, EventTcpConnect, tp);

		IF_DEBUG(2, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "connection ESTABLISHED\n";
		});

	    }

	    // Otherwise, we got a SYN but we need an ACK.  This is a
	    // simultaneous open.
	    else {
		tp->state = TcpsSynReceived;

trimthenstep6:		

		// Handle any data received with the SYN
		th->th_seq++;

		// If the data we just got was more than the window could
		// handle, then truncate it and remove any FIN flag
		if (seglen > tp->rcv_wnd) {
		    seglen = tp->rcv_wnd;
		    tiflags &= ~TH_FIN;
		}
		
		tp->snd_wl1 = th->th_seq - 1;
		goto step6;

	    }

	}

	default:
	break;

    }

    // Trim the segment...first look for duplicate data at the start
    // of the segment.
    //
    // Note:  When we complete an active open (on the initiator side),
    // from the SYN_SENT state going to ESTABLISHED, the following
    // line will make it look like the SYN is a duplicate (so it gets
    // trimmed by the following block).  That's OK...we've already
    // seen the SYN and we'll need to get it out of the window anyways
    // for some error-checking farther down.
    todrop = tp->rcv_nxt - th->th_seq;

    // If we trimmed anything at the start, we need to adjust for
    // not getting a SYN too.
    if (todrop > 0) {
	if (tiflags & TH_SYN) {

	    tiflags &= ~TH_SYN;
	    th->th_seq++;
	    todrop--;

	}

	// If we had to drop more than the total segment length, then we got
	// us here a duplicate packet.
	if (todrop >= seglen) {
	    
	    // If this segment is immediately adjacent to the left edge of
	    // the window, then send an ACK to resynchronize but keep
	    // doing RST or ACK processing.
	    if ((tiflags & TH_FIN) && (todrop == seglen + 1)) {
		
		todrop = seglen;
		tiflags &= ~TH_FIN;
		tp->flags |= TF_ACKNOW;
		
		IF_DEBUG(3, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "immediate ACK\n";
		});

	    }
	    else {
		if ((todrop) || (tiflags & TH_ACK) == 0) {
		    goto dropafterack;
		}
	    }
	}

	// Adjust length and sequence number to handle data we
	// dropped off the left edge of the window.
	th->th_seq += todrop;
	seglen -= todrop;
	th->payload.Trim(th->th_seq, seglen);

    }

    // If new data received on a connection after the user
    // processes are gone, then close and RST the other end.
    if ((tp->state > TcpsCloseWait) && seglen) {

	IF_DEBUG(2, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "received data after Close now resetting\n";
	});

	tp = Close(tp);
	goto dropwithreset;
    }
    
    // If segment ends after the right edge of the window, drop
    // trailing data.
    todrop = (th->th_seq + seglen) - (tp->rcv_nxt + tp->rcv_wnd);
    if (todrop > 0) {
	if (todrop >= seglen) {

	    // If a new connection request is received in TIME_WAIT,
	    // drop the old connection and start over if the sequence
	    // numbers are above the previous ones.
	    if ((tiflags & TH_SYN) &&
		(tp->state == TcpsTimeWait) &&
		(SEQ_GT(th->th_seq, tp->rcv_nxt))) {
		iss = tp->rcv_nxt + TCP_ISSINCR;
		tp = Close(tp);
		goto findpcb;
	    }

	    // If the window is closed, we can only take segments
	    // at the edge of the window...continue processing but
	    // remember to ACK.  Otherwise drop the segment and
	    // send an ACK.
	    if ((tp->rcv_wnd == 0) && (th->th_seq == tp->rcv_nxt)) {
		tp->flags |= TF_ACKNOW;
	    }
	    else {
		goto dropafterack;
	    }

	}
	seglen -= todrop;
	tiflags &= ~(TH_PUSH | TH_FIN);
	th->payload.Trim(th->th_seq, seglen);
    }

    // Handle processing of the RST flag
    if (tiflags & TH_RST) {

	switch (tp->state) {

	    case TcpsSynReceived:
	    // XXX connection refused
	    goto close;
	    break;

	    case TcpsEstablished:
	    case TcpsFinWait1:
	    case TcpsFinWait2:
	    case TcpsCloseWait:
	    // XXX connection reset by peer
close:
	    tp->state = TcpsClosed;
	    tp = Close(tp);
	    goto drop;
	    break;

	    case TcpsClosing:
	    case TcpsLastAck:
	    case TcpsTimeWait:
	    tp = Close(tp);
	    goto drop;
	    break;

	    default:
	    break;

	}
    }

    // Shouldn't be a SYN in the window at this point...
    if (tiflags & TH_SYN) {
	tp = Drop(tp);
	goto dropwithreset;
    }

    // Should have an ACK bit set at this point
    if (!(tiflags)) {
	goto drop;
    }

    // Now do ACK processing!  Isn't this fun?
    switch (tp->state) {

	case TcpsSynReceived:

	// Passive or simultaneous open or self-connect
	if (SEQ_GT(tp->snd_una, th->th_ack) ||
	    SEQ_GT(th->th_ack, tp->snd_max)) {
	    goto dropwithreset;
	}

	// Wake up process on server side waiting for passive open
	// to complete
	Post(time + latencyToApp, tp->tcp_up, EventTcpConnect, tp);

	tp->state = TcpsEstablished;
	{
	    TcpDataList nullList;
	    Reass(tp, 0, 0, 0, nullList);
	}
	tp->snd_wl1 = th->th_seq - 1;

	time += latencyToApp;

	IF_DEBUG(2, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "passive open SYN_RCVD now ESTABLISHED\n";
	});

	// FALL THROUGH...

	// Bunch o' other cases...handle duplicate ACKs, fast retransmit
	// and fast recovery algorithms
	case TcpsEstablished:
	case TcpsFinWait1:
	case TcpsFinWait2:
	case TcpsCloseWait:
	case TcpsClosing:
	case TcpsLastAck:
	case TcpsTimeWait:

	if (SEQ_LEQ(th->th_ack, tp->snd_una)) {

	    if ((seglen == 0) && (th->th_win == tp->snd_wnd)) {

		// Not a consecutive dupack.
		if ((tp->timer[TCPT_REXMT] == 0) ||
		    (th->th_ack != tp->snd_una)) {
		    tp->dupacks = 0;

		    IF_DEBUG(3, {
			DEBUG_PREFIX("Tcp");
			*(sim->debugStream) << "reset dup ACKs\n";
		    });

		}

		// Received a duplicate ack
		else {

		    tp->dupacks++;

		    IF_DEBUG(3, {
			DEBUG_PREFIX("Tcp");
			*(sim->debugStream) << "dup ACK " <<
			  tp->dupacks << "\n";
		    });
		    
		    // If we're at threshhold, then do a fast retransmit.
		    if (tp->dupacks == rexmtthresh) {
			TcpSeq onxt = tp->snd_nxt;

			u_int win = min(tp->snd_wnd, tp->snd_cwnd) / 2 /
			  tp->maxseg;

			// Fast retransmit.  Set variables so we can call
			// Output to retransmit the lost segment, then
			// restore things to continue transmitting segments
			// in sequence.
			if (win < 2) {
			    win = 2;
			}

			tp->snd_ssthresh = win * tp->maxseg;
			tp->timer[TCPT_REXMT] = 0;
			tp->rtt = 0;
			tp->snd_nxt = th->th_ack;
			tp->snd_cwnd = tp->maxseg;

			IF_DEBUG(3, {
			    DEBUG_PREFIX("Tcp");
			    *(sim->debugStream) << "snd_ssthresh now " << tp->snd_ssthresh << "\n";
			    DEBUG_PREFIX("Tcp");
			    *(sim->debugStream) << "fast retransmit and REXMT timer set cwnd to " << tp->snd_cwnd << "\n";
			});

			Out(tp, 0);

			tp->snd_cwnd = tp->snd_ssthresh +
			  tp->maxseg * tp->dupacks;

			IF_DEBUG(3, {
			    DEBUG_PREFIX("Tcp");
			    *(sim->debugStream) << "post-fast retransmit, set cwnd to " << tp->snd_cwnd << "\n";
			});

			if (SEQ_GT(onxt, tp->snd_nxt)) {
			    tp->snd_nxt = onxt;
			}
			goto drop;

		    }
		    
		    // If we've received more packets than the threshhold,
		    // then open up the congestion window (packets continue
		    // to arrive at the destination) and continue normal
		    // transmission.  Drop the duplicate ACK.
		    else if (tp->dupacks > rexmtthresh) {

			tp->snd_cwnd += tp->maxseg;

			IF_DEBUG(3, {
			    DEBUG_PREFIX("Tcp");
			    *(sim->debugStream) << "cwnd fast retransmit now " <<
			      tp->snd_cwnd << "\n";
			});

			Out(tp, 0);
			goto drop;

		    }
		}
	    }
	    else {
		tp->dupacks = 0;

		IF_DEBUG(3, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "reset dup ACKs\n";
		});
	    }
	    break;
	}
	
	// If the congestion window was inflated to account for the
	// other side's cached packets during fast retransmit, retract it.
	// In any case, if we get here, we haven't seen any consecutive
	// duplicate ACKs (for awhile).
	if ((tp->dupacks > rexmtthresh) &&
	    (tp->snd_cwnd > tp->snd_ssthresh)) {
	    tp->snd_cwnd = tp->snd_ssthresh;

	    IF_DEBUG(3, {
		DEBUG_PREFIX("Tcp");
		*(sim->debugStream) << "cwnd end of fast retransmit now " <<
		  tp->snd_cwnd << "\n";
	    });

	}
	tp->dupacks = 0;

	// Throw away ACKs for data we haven't sent yet (!)
	if (SEQ_GT(th->th_ack, tp->snd_max)) {
	    goto dropafterack;
	}
	acked = th->th_ack - tp->snd_una;
	
	// Update RTT estimate
	if (tp->rtt && SEQ_GT(th->th_ack, tp->rtseq)) {
	    XmitTimer(tp, tp->rtt);
	}

	// If all outstanding data is ACKed, stop the retransmit timer
	// (we'll restart it if we get more output or a persist).
	// If more data remains to be ACKed, restart retransmit timer.
	if (th->th_ack == tp->snd_max) {
	    tp->timer[TCPT_REXMT] = 0;
	    needoutput = true;
	}
	else if (tp->timer[TCPT_PERSIST] == 0) {
	    tp->timer[TCPT_REXMT] = tp->rxtcur;
	}
	IF_DEBUG(3, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "REXMT timer set to " <<
	      tp->timer[TCPT_REXMT] << "\n";
	});

	// Open the congestion window...
	{

	    u_int cw = tp->snd_cwnd;
	    u_int incr = tp->maxseg;

	    if (cw > tp->snd_ssthresh) {
		incr = incr * incr / cw;
	    }
	    tp->snd_cwnd = min(cw + incr, TCP_MAXWIN);

	    IF_DEBUG(3, {
		DEBUG_PREFIX("Tcp");
		*(sim->debugStream) << "cwnd open to " << tp->snd_cwnd << "\n";
	    });
	    
	}

	// Remove acknowledged data from the send buffer
	if (acked > tp->so_snd_sb_cc) {
	    tp->snd_wnd -= tp->so_snd_sb_cc;
	    tp->so_snd_sb_cc = 0;
	    ourfinisacked = true;
	}
	else {
	    tp->so_snd_sb_cc -= acked;
	    tp->snd_wnd -= acked;
	    ourfinisacked = false;
	}
	tp->snd_una = th->th_ack;
	tp->so_snd.Trim(tp->snd_una, tp->so_snd_sb_cc);

	if (SEQ_LT(tp->snd_nxt, tp->snd_una)) {
	    tp->snd_nxt = tp->snd_una;
	}


	// DONE WITH GENERAL ACK PROCESSING

	switch (tp->state) {

	    // If we were in FIN_WAIT_1 and we got a FIN acknowledged,
	    // enter FIN_WAIT_2
	    case TcpsFinWait1:
	    if (ourfinisacked) {

		// Assume here that processes (such that they are)
		// only close(2) connections and never do a shutdown(2)
		// to do a half-close.
		tp->timer[TCPT_2MSL] = maxidle;
		// soisdisconnected(so);
		// Post();
		tp->state = TcpsFinWait2;

		IF_DEBUG(2, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "ACK of FIN caused transition from FIN_WAIT_1 to FIN_WAIT_2\n";
		});

	    }
	    break;

	    // If we're CLOSING, and we got an ACK for a FIN, go to
	    // TIME_WAIT.
	    case TcpsClosing:
	    if (ourfinisacked) {
		if (ourfinisacked) {
		    tp->state = TcpsTimeWait;
		    CancelTimers(tp);
		    tp->timer[TCPT_2MSL] = 2 * TCPTV_MSL;

		    IF_DEBUG(2, {
			DEBUG_PREFIX("Tcp");
			*(sim->debugStream) << "ACK of FIN caused transition from CLOSING to TIME_WAIT\n";
		    });

		}
		// soisdisconnected(so);
		// Post()
	    }
	    break;

	    // If we are in LAST_ACK, and we got an ACK for a FIN,
	    // we're done!
	    case TcpsLastAck:
	    if (ourfinisacked) {
		tp = Close(tp);
		goto drop;
	    }
	    break;

	    // If we're in TIME_WAIT, the only thing we should get is
	    // a retransmission of a remote FIN...make sure we
	    // ACK it.
	    case TcpsTimeWait:
	    tp->timer[TCPT_2MSL] = 2 * TCPTV_MSL;
	    goto dropafterack;
	    break;

	    default:
	    break;
	} // switch (tp->state)
	break;

	default:
	break;
    } // switch (tp->state)

step6:

    // Update window information

    if ((tiflags & TH_ACK) &&
	(SEQ_LT(tp->snd_wl1, th->th_seq) ||
	 (tp->snd_wl1 == th->th_seq) &&
	 (SEQ_LT(tp->snd_wl2, th->th_ack) ||
	  (tp->snd_wl2 == th->th_ack) && (th->th_win > tp->snd_wnd)))) {

	tp->snd_wnd = th->th_win;
	tp->snd_wl1 = th->th_seq;
	tp->snd_wl2 = th->th_ack;
	if (tp->snd_wnd > tp->max_sndwnd) {
	    tp->max_sndwnd = tp->snd_wnd;
	}
	needoutput = true;
    }

    // Don't do any urgent mode stuff...we don't support it.

dodata:

    // If we got some real data or a FIN flag in the new segment, and
    // we haven't gotten a FIN on this connection yet, then process
    // the incoming data.
    if ((seglen || (tiflags & TH_FIN)) &&
	(!TcpsHaveRcvdFin(tp->state))) {

	// Code below is a simple inline
	// expansion of TCP_REASS()...to handle the case that we got in-order
	// data, there is no waiting out-of-order data, and we're in
	// ESTABLISHED.  In this case we do some simple accounting and pass
	// the data up to the process.
	if ((th->th_seq == tp->rcv_nxt) &&
	    (tp->seg_next == NULL) &&
	    (tp->state == TcpsEstablished)) {

	    tp->flags |= TF_DELACK;
	    tp->rcv_nxt += seglen;
	    tiflags &= TH_FIN;

	    // If there was data in the segment, give it to the application
	    // (if there was *no* data, the segment only had a FIN flag
	    // set).
	    if (seglen) {

		TcpCall *call = new TcpCall;
		call->pcb = tp;
		call->th_ulen = seglen;
		th->payload.AppendTo(call->data);
		
		Post(time + latencyToApp,
		     tp->tcp_up,
		     EventTcpArrival,
		     call);
		
		time += latencyToApp;
		tp->rcv_bytes += seglen;
		
		// Since the segment has been successfully delivered,
		// call Tcp::Out() in the same way that tcp_output()
		// would be called by tcp_usrreq().
		Out(tp, 0);

	    }
	    
//	    delete th;
	    delete ipPacket;

	}

	// Otherwise, we got some data that requires us to run the TCP
	// reassembly routine.
	else {

	    tiflags = Reass(tp, th->th_seq, seglen, tiflags, th->payload);
	    tp->flags |= TF_ACKNOW;

	    time += latencyToApp;

	    IF_DEBUG(3, {
		DEBUG_PREFIX("Tcp");
		*(sim->debugStream) << "immediate (possibly dup) ACK\n";
	    });

//	    delete th;
	    delete ipPacket;

	}

    }

    // If we get here, we either didn't have any data or we've already
    // seen a FIN on this connection.  In either case, drop the data on
    // the floor.
    else {
//	delete th;
	delete ipPacket;

	tiflags &= ~TH_FIN;
    }

    //
    // FIN processing starts here
    //
    if (tiflags & TH_FIN) {

	// If we haven't seen a FIN yet, then deal with this
	if (!(TcpsHaveRcvdFin(tp->state))) {
	    tp->flags |= TF_ACKNOW;
	    tp->rcv_nxt++;
	}

	// State-specific processing
	switch (tp->state) {

	    // SYN_RECEIVED and ESTABLISHED states go to CLOSE_WAIT
	    case TcpsSynReceived:
	    case TcpsEstablished:
	    {
		tp->state = TcpsCloseWait;
		
		IF_DEBUG(2, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "FIN caused transition from SYN_RECEIVED/ESTABLISHED to CLOSE_WAIT\n";
		});

		Post(time + latencyToApp, tp->tcp_up, EventTcpClosed, tp);

		IF_DEBUG(3, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "Tcp::In posting EventTcpClosed\n";
		});

	    }

	    break;

	    // If we're in FIN_WAIT_1, the FIN hasn't been acked,
	    // so go to CLOSING.
	    case TcpsFinWait1:
	    {
		tp->state = TcpsClosing;
		
		IF_DEBUG(2, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "FIN caused transition from FIN_WAIT_1 to CLOSING\n";
		});
		
	    }
	    break;

	    // If we're in FIN_WAIT_2, then enter TIME_WAIT.
	    case TcpsFinWait2:
	    {
		tp->state = TcpsTimeWait;
		CancelTimers(tp);
		tp->timer[TCPT_2MSL] = 2 * TCPTV_MSL;
		// soisdisconnected(so);
		// Post(); XXX

		IF_DEBUG(2, {
		    DEBUG_PREFIX("Tcp");
		    *(sim->debugStream) << "FIN caused transition from FIN_WAIT_2 to TIME_WAIT\n";
		});

	    }
	    break;

	    // If we're in TIME_WAIT, restart the 2 MSL timer
	    case TcpsTimeWait:
	    tp->timer[TCPT_2MSL] = 2 * TCPTV_MSL;
	    break;

	    default:
	    break;
	}
    }

    // Return any desired output
    if (needoutput || (tp->flags & TF_ACKNOW)) {
	Out(tp, 0);
    }

    return;

    // Generate an ACK and drop the incoming segment
dropafterack:
    
    if (tiflags & TH_RST) {
	goto drop;
    }
//    delete th;
    delete ipPacket;
    tp->flags |= TF_ACKNOW;
    Out(tp, 0);
    return;

    // Generate an RST, dropping the incoming segment.  Make up an ACK...
dropwithreset:

    if (tiflags & TH_ACK) {
	Respond(tp, ipPacket, 0, th->th_ack, TH_RST);
    }
    else {
	if (tiflags & TH_SYN) {
	    seglen++;
	}
	Respond(tp, ipPacket, th->th_seq + seglen, 0, TH_RST | TH_ACK);
    }
	
    if (dropsocket) {
	// soabort(so);
	tp->prev->next = tp->next;
	tp->next->prev = tp->prev;
	delete(tp);
    }

    return;
    
    // Drop space held by incoming segment and return
drop:    

//    delete th;
    delete ipPacket;

    if (dropsocket) {
	// soabort(so);
	tp->prev->next = tp->next;
	tp->next->prev = tp->prev;
	delete(tp);
    }

    return;
}

//
// Tcp::NewTcpPcb
//
// Input:  None
//
// Output:  Pointer to new TcpPcb structure (in return value)
//
// Allocate a new TcpPcb, link it into the DLL of TcpPcbs, and
// do partial initialization of its fields.
//
TcpPcb
*Tcp::NewTcpPcb() {

    TcpPcb *tp;

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::NewTcpPcb\n";
    });

    // Allocate and zero out the TcpPcb.  Zeroing out things is
    // pretty barbaric, but we do it because the BSD 4.4Lite code
    // does this to initialize a bunch of fields to be zero.
    tp = (TcpPcb *) new TcpPcb;
    bzero((char *) tp, sizeof (TcpPcb));
    
    // Link into the DLL
    tp->next = tcb->next;
    tp->prev = tcb;
    tcb->next->prev = tp;
    tcb->next = tp;

    // Find a bunch of things to initialize
    tp->so_snd_sb_cc = 0;
    tp->so_rcv_sb_cc = 0;
    tp->so_rcv_sb_hiwat = 8192;	// XXX
    tp->seg_next = NULL;
    
    tp->maxseg = mssdflt;	// XXX

    tp->srtt = TCPTV_SRTTBASE;
    tp->rttvar = rttdflt * PR_SLOWHZ << 2;
    tp->rttmin = TCPTV_MIN;

    tp->rxtcur = ((TCPTV_SRTTBASE >> 2) + (TCPTV_SRTTDFLT << 2)) >> 1;
    tp->rxtcur = max(tp->rxtcur, TCPTV_MIN);
    tp->rxtcur = min(tp->rxtcur, TCPTV_REXMTMAX);

    tp->snd_cwnd = TCP_MAXWIN; // mss???
    tp->snd_ssthresh = TCP_MAXWIN;
      
    tp->snd_bytes = 0;
    tp->rcv_bytes = 0;

    return tp;

}


//
// Tcp::EphemeralPort
//
// Input:  None
//
// Output:  port number (in return value)
//
// Allocate an ephemeral port for use by a new TCP connection when no
// local endpoint has been specified.
//
unsigned short Tcp::EphemeralPort()
{
    unsigned short tryPort;		// try this port number
    bool OK;				// is this port number OK?

    do {

	// Pick a port number to try
	tryPort = tcb->tcp_lport++;
	if (tcb->tcp_lport >= TcpMaxEphemeralPort) {
	    tcb->tcp_lport = TcpMinEphemeralPort;
	}

	// Scan through all the PCBs to see if this port number is already
	// in use (by any connection)
	TcpPcb *current;
	OK = true;
	for (current = tcb->next; current != tcb; current = current->next) {
	    if (current->tcp_lport == tryPort) {
		OK = false;
		break;
	    }
	}
	
    } while (!OK);

    return tryPort;

}


//
// Tcp::Mss
//
// Input:  TCP PCB pointer (tp)
//
// Output:  None
//
// Compute the MSS for a connection and compute the initial congestion
// window.
//
void
Tcp::Mss(TcpPcb *tp) {

    int mss;			// maximum segment size

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Mss\n";
    });

    // Ordinarily we'd compute the MSS here.  For now we'll just
    // set it to the default.
    mss = mssdflt;

    tp->maxseg = mss;
    tp->snd_cwnd = mss;

    IF_DEBUG(3, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "cwnd initialize to " << tp->snd_cwnd << "\n";
    });
	    
}


//
// Tcp::Reass
//
// Input:  TCP PCB (tp), segment starting sequence number (seq),
//         length (len), TCP flags (flags), payload list (data)
// header flags (flags)
//
// Output;  New th_flags (in return value).
//
int
Tcp::Reass(TcpPcb *tp, TcpSeq seq, int len, int flags, TcpDataList &data)
{

    Reassem *curReassem, *prevReassem; // elements on reassembly queue
    Reassem *re;		       // potential new reassembly element
    int returnLen = 0;		       // how much data to pass to process
    
    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::Reass\n";
    });

    // If length is zero and we don't have a FIN flag, we just need to
    // force the pre-ESTABLISHED data
    // up to the user process
    if ((len == 0) && (!(flags & TH_FIN))) {

	goto present;
	
    }

    // Make up a new struct Reassem element.
    re = new Reassem;
    re->seq = seq;
    re->len = len;
    re->flags = flags;
    data.AppendTo(re->data);

    // The basic algorithm is this:  First find a segment that begins
    // after new segment.  Trim data from the start of new segment if there's
    // an overlap.  Then for succeeding segments, trim overlaps from *them*.
    // Finally, enqueue the new segment in place.
    //
    // Do the first part now.
    for (curReassem = tp->seg_next, prevReassem = NULL;
	 curReassem != NULL;
	 prevReassem = curReassem, curReassem = curReassem->next) {

	if (SEQ_GT(curReassem->seq, re->seq)) {
	    break;
	}

    }

    // See if there was a preceding segment, trim data from new
    // segment if necessary.
    if (prevReassem) {

	// Check for total overlap.  If so, dequeue current segment.
	if (SEQ_GEQ(prevReassem->seq + prevReassem->len,
		   re->seq + re->len)) {
	    delete re;
	    goto present;
	}

	// Check for partial overlap, in which case we trim part of
	// the new segment.
	if (SEQ_GT(prevReassem->seq + prevReassem->len,
		   seq)) {
	    int overlap = prevReassem->seq + prevReassem->len - re->seq;
	    re->seq += overlap;
	    re->len -= overlap;
	    re->data.Trim(re->seq, re->len);
	}

    }

    // Now trim any data from succeeding segments.
    while (curReassem) {

	// Figure out how much the current segment on the queue
	// overlaps the one we're processing
	int overlap = re->seq + re->len - curReassem->seq;

	// Total overlap => dequeue succeeding segment
	if (overlap >= curReassem->len) {

	    Reassem *dieReassem = curReassem;
	    curReassem = dieReassem->next;
	    tp->so_rcv_sb_cc -= dieReassem->len;
	    delete dieReassem;

	}
	// Partial overlap => trim succeeding segment (we also know
	// that it's the last one)
	else if (overlap > 0) {

	    tp->so_rcv_sb_cc -= overlap;
	    curReassem->seq += overlap;
	    curReassem->len -= overlap;
	    curReassem->data.Trim(curReassem->seq, curReassem->len);

	}
	
	// No overlap...we're done!
	else {
	    break;
	}

    }

    // Enqueue segment
    if (prevReassem) {
	prevReassem->next = re;
    }
    else {
	tp->seg_next = re;
    }
    re->next = curReassem;
    tp->so_rcv_sb_cc += re->len;

    // Pass any available data up to the application

present:

    // If we're not in a position to receive any data, don't do anything.
    if (tp->state < TcpsSynReceived) {
	return(0);
    }

    // If we don't have any inorder data, don't do anything either.
    curReassem = tp->seg_next;
    if ((curReassem == NULL) ||
	(curReassem->seq != tp->rcv_nxt)) {
	return(0);
    }
    
    // If we're in SYN_RECEIVED, don't do anything (wait for ESTABLISHED)
    if ((tp->state == TcpsSynReceived) &&
	(curReassem->len)) {
	return(0);
    }

    // Accumulate data
    TcpDataList returnData;
    do {

	Reassem *dieReassem;
	
	tp->rcv_nxt += curReassem->len;
	returnLen += curReassem->len;
	tp->so_rcv_sb_cc -= curReassem->len;
	flags = curReassem->flags & TH_FIN;
	curReassem->data.AppendTo(returnData);

	dieReassem = curReassem;
	curReassem = curReassem->next;
	delete dieReassem;

    } while ((curReassem) && (curReassem->seq == tp->rcv_nxt));
    tp->seg_next = curReassem;

    if (returnLen) {

	// Give the data to the application
	TcpCall *call = new TcpCall;
	call->pcb = tp;
	call->th_ulen = returnLen;
	returnData.AppendTo(call->data);
	
	Post(time + latencyToApp,
	     tp->tcp_up,
	     EventTcpArrival,
	     call);

	tp->rcv_bytes += returnLen;
	
	IF_DEBUG(3, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "posted " << returnLen << " reassembled bytes to application\n";
	});

	// At this point it seems logical that we would (potentially)
	// generate a window update, to simulate
	// tcp_usrreq calling tcp_output().  We don't do this, however.
	// The reason is that immediately after we return to Tcp::In(),
	// it will set TF_ACKNOW, which will cause an acknowledgement
	// segment to be emitted anyways.  This would cause two
	// identical acknowledgements to go out, which would be bad.

    }

    IF_DEBUG(3, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "return flags " << flags << "\n";
    });
    
    return (flags);

}

//
// Tcp::PcbCheck
//
// Input:  PCB pointer (tp)
//
// Output:  In return value...returns tp if valid, otherwise NULL.
//
// PcbCheck scans the list of PCBs to find a given TcpPcb.  If it
// exists, it returns the value...otherwise it returns NULL.
//
TcpPcb *Tcp::PcbCheck(TcpPcb *tp)
{
    TcpPcb *tryPcb;		// current TcpPcb we're looking at

    for (tryPcb = tcb->next; tryPcb != tcb; tryPcb = tryPcb->next) {
	if (tryPcb == tp) {
	    return tp;
	}
    }
    return NULL;
}

//
// Tcp::UserListen
//
// Input:  Current time (time),
// TCP port to listen on (port), application to receive packets (app)
//
// Output:  Pointer to TcpPcb (in return value).
//
// Listen on a TCP port for incoming connections.  Object app will receive
// an asynchronous EventTcpConnect event when connection is completed.
// Analogous to a socket(2)/bind(2)/listen(2) sequence.
//
TcpPcb
*Tcp::UserListen(double time, unsigned short port, ScheduledObject *app)
{

    TcpPcb *tp;			// new Pcb for Listening connection

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::UserListen for port " << port << "\n";
    });

    // Save current time for other procedures that need it.
    this->time = time;

    tp = NewTcpPcb();

    // Bind everything we know about.  For a listen(), this is just
    // the local port number.  Note that NewTcpPcb() initialized
    // everything else to zero.
    tp->tcp_lport = port;

    // Check to see if we should use an ephemeral port on local side...
    if (tp->tcp_lport == 0) {
	tp->tcp_lport = EphemeralPort();
    }
    

    // Set the state and any other fields of this structure we can
    // do now.
    tp->state = TcpsListen;
    tp->tcp_up = app;

    // Application gets a EventTcpConnect event when a new client
    // connects.
    return tp;

}

//
// Tcp::UserConnect
//
// Input:  Current time (time), IP address to connect to (addr),
// port number (port), and application to receive packets (app).
//
// Output:  Pointer to TcpPcb (in return value).
//
// Initiate a TCP connection to another process.  Object app will receive
// an EventTcpConnect event when connection is complete.
// Analogous to a socket(2)/bind(2)/connect(2) sequence.
//
TcpPcb
*Tcp::UserConnect(double time, InAddr addr, unsigned short port,
		  ScheduledObject *app, unsigned short lport)
{

    TcpPcb *tp;			// new Pcb for this connection
    IpRoute *bestRoute;		// best route to host from routing table
				// (need this to fill in source address)

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::UserConnect\n";
    });

    // Save current time for other procedures that need it.
    this->time = time;

    tp = NewTcpPcb();

    // Bind everything we can find.
    tp->tcp_faddr.s_addr = addr.s_addr;
    tp->tcp_fport = port;

    Lookup(down, downName);
    bestRoute = down->RouteLookup(addr);
    if (bestRoute == NULL) {
	IF_DEBUG(1, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "no route to host\n";
	});
	return (NULL);
    }

    tp->tcp_laddr.s_addr = bestRoute->rt_if->ia_addr.s_addr;
    tp->tcp_up = app;

    // Set port on local side...either user-specified or ephemeral
    if (lport == 0) {
	tp->tcp_lport = EphemeralPort();
    }
    else {
	tp->tcp_lport = lport;
    }
    
    // Set other state variables
    tp->state = TcpsSynSent;
    tp->timer[TCPT_KEEP] = TCPTV_KEEP_INIT;
    tp->iss = tcp_iss;
    tcp_iss += TCP_ISSINCR / PR_SLOWHZ;
    tp->snd_una = tp->snd_nxt = tp->snd_max = tp->iss; // tcp_sndseqinit

    // Send initial segment to get the ball rolling.  The SYN flag for
    // the first header comes from Tcp::Outflags[TcpsSynSent].
    Out(tp, 0);
    
    // Client application gets an EventTcpConnect message when the
    // connection is successfully established.
    return (tp);

}

//
// Tcp::UserClose
//
// Input:  Current time (time), TcpPcb for connection to close (tp)
//
// Output:  None.
//
// Close a TCP connection.  Analogous to a close(2).  Embodies part of
// [tcp_disconnect] and [tcp_usrclosed].
//
void
Tcp::UserClose(double time, TcpPcb *tp)
{

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::UserClose entry\n";
    });

    // Save current time for other procedures that need it.
    this->time = time;

    if (PcbCheck(tp)) {
	// If the connection is "embryonic", we can just do a Close().  This
	// ought to get rid of the TcpPcb and everything.  Otherwise, the
	// initiation of a close is a little more complex...
	if (tp->state > TcpsListen) {

	    // If we're not established yet, then use the Close() method.
	    if (tp->state < TcpsEstablished) {
		tp = Close(tp);
	    }
	    else {

		// Otherwise, set new state
		switch (tp->state) {

		    case (TcpsClosed):
		      case (TcpsListen):
			case (TcpsSynSent):
			  tp->state = TcpsClosed;
			  tp = Close(tp);

			  IF_DEBUG(2, {
			      DEBUG_PREFIX("Tcp");
			      *(sim->debugStream) << "UserClose caused transition from CLOSE/LISTEN/SYNSENT to CLOSED\n";
			  });
			  break;
			  
			  case (TcpsSynReceived):
			    case (TcpsEstablished):
			      tp->state = TcpsFinWait1;

			  IF_DEBUG(2, {
			      DEBUG_PREFIX("Tcp");
			      *(sim->debugStream) << "UserClose caused transition from SYN_RCVD/ESTABLISHED to FIN_WAIT_1\n";
			  });
			  break;

			  case (TcpsCloseWait):
			    tp->state = TcpsLastAck;

			  IF_DEBUG(2, {
			      DEBUG_PREFIX("Tcp");
			      *(sim->debugStream) << "UserClose caused transition from CLOSE_WAIT to LAST_ACK\n";
			  });
			  break;

			  default:
			    break;

		      }

		// If the connection is still around, then send out a FIN
		// segment...
		if (tp) {
		    Out(tp, 0);
		}
	    }
	}
	else {
	    Close(tp);
	}
    }
    else {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) <<
	      "UserClose called for non-existent connection";
	});
	abort();
    }
}

//
// Tcp::UserSend
//
// Input:  Current time (time), TcpPcb for connection, length of data to send
//
// Output:  None.
//
// Send data on a TCP connection.  Three versions of this function are
// provided; two support the sending of data segments/entities/tags
// and the other doesn't.
//
void
Tcp::UserSend(double time, TcpPcb *tp, short len, TcpDataList &data)
{

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::UserSend\n";
    });

    // Save current time for other procedures that need it.
    this->time = time;

    if (PcbCheck(tp)) {
	tp->so_snd_sb_cc += len;
	tp->snd_bytes += len;
//        data.AppendTo(tp->so_snd);	// XXX put sequence numbers in data
	Out(tp, 0);
    }
    else {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) <<
	      "UserSend called for non-existent connection";
	});
	abort();
    }
    
}

void
Tcp::UserSend(double time, TcpPcb *tp, short len, TcpData *data)
{

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::UserSend\n";
    });

    // Save current time for other procedures that need it.
    this->time = time;

    if (PcbCheck(tp)) {
	// If we really had a data tag for this send, then process it,
	// otherwise, use the UserSend that doesn't tag data.
	if (data) {
	    data->seq = tp->snd_una + tp->so_snd_sb_cc + len - 1;
	    tp->so_snd_sb_cc += len;
	    tp->snd_bytes += len;
	    tp->so_snd.EnqueueTail(data);
	    Out(tp, 0);
	}
	else {
	    UserSend(time, tp, len);
	}
    }
    else {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) <<
	      "UserSend called for non-existent connection";
	});
	abort();
    }
}

void
Tcp::UserSend(double time, TcpPcb *tp, short len)
{

    IF_DEBUG(2, {
	DEBUG_PREFIX("Tcp");
	*(sim->debugStream) << "in Tcp::UserSend\n";
    });

    // Save current time for other procedures that need it.
    this->time = time;

    if (PcbCheck(tp)) {
	tp->so_snd_sb_cc += len;
	tp->snd_bytes += len;
	Out(tp, 0);
    }
    else {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) <<
	      "UserSend called for non-existent connection";
	});
	abort();
    }
}

//
// Tcp::UserSet
//
// Input:  Current time (time), TcpPcb for connection, parameter flag to set
//
// Output:  None.
//
// Set a flag in the TCP PCB (i.e. TF_NODELAY).
//
void
Tcp::UserSet(double time, TcpPcb *tp, int parm)
{
    if (PcbCheck(tp)) {
	tp->flags |= parm;

	IF_DEBUG(2, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "Tcp::UserSet\n";
	});
    }
    else {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) <<
	      "UserSet called for non-existent connection";
	});
	abort();
    }
}

//
// Tcp::UserReset
//
// Input:  Current time (time), TcpPcb for connection, parameter flag to unset
//
// Output:  None.
//
// Reset a flag in the TCP PCB (i.e. TF_NODELAY).
//
void
Tcp::UserReset(double time, TcpPcb *tp, int parm)
{
    if (PcbCheck(tp)) {
	tp->flags &= ~parm;

	IF_DEBUG(2, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) << "Tcp::UserReset\n";
	});
    }
    else {
	IF_DEBUG(0, {
	    DEBUG_PREFIX("Tcp");
	    *(sim->debugStream) <<
	      "UserReset called for non-existent connection";
	});
	abort();
    }

}


//
// TcpCmd
//
// (Tcl command procedure)
//
int
TcpCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 2) {
	Tcl_SetResult(interp, "usage:  Tcp name", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of TCP module
	Tcp *tcpModule;		// new Tcp module
	extern Simulation *sim;	// global simulation object

	// One argument only...the name of the new Tcp module.
	nm = argv[1];

	tcpModule = new Tcp(nm, sim);
	Tcl_CreateCommand(interp, nm,
			  TcpObjectCmd,
			  (ClientData) tcpModule,
			  TcpDeleteProc);

	return (TCL_OK);

    }
}

//
// TcpObjectCmd
//
// (Tcl object command procedure)
//
int
TcpObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {

    Tcp *tcpModule = (Tcp *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }

    else {

	//
	// tcp down object
	//
	if (!strcmp(argv[1], "down")) {
	    if (argc == 3) {
		String on;	// object name

		on = argv[2];
		tcpModule->downName = on;
		tcpModule->down = NULL;
		return TCL_OK;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// tcp netstat
	//
	else if (!strcmp(argv[1], "netstat")) {
	    if (argc == 2) {
		TcpPcb *tp = tcpModule->tcb->next;
		char buf[128];	// temporary printing buffer

		while (tp != tcpModule->tcb) {
		    
		    sprintf(buf, "%s.%d ",
			    InetNtoA(tp->tcp_laddr),
			    tp->tcp_lport);
		    cout << buf;

		    sprintf(buf, "%s.%d ",
			    InetNtoA(tp->tcp_faddr),
			    tp->tcp_fport);
		    cout << buf;

		    cout << tp->state;

		    cout << "\n";

		    tp = tp->next;
		}
		return (TCL_OK);
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	else {
	    return(ScheduledObjectCmd(clientData, interp, argc, argv));
	}

    }


}

//
// TcpDeleteProc
//
// (Tcl delete procedure)
//
void
TcpDeleteProc(ClientData clientData) {
    delete ((Tcp *) clientData);
}

//
// operator<<
//
// Input:  Output stream (s), TCP state (state)
//
// Output:  Output stream
//
// Print the name of the state on the output stream s.
//
ostream& operator<<(ostream &s, Tcps state)
{
    char *output;
    
    output = TcpsName[state];

    return s << output;

}

