static char rcsid[] = "$Id: TelnetClient.cc,v 1.21 1996-06-07 09:36:28-07 bmah Exp $";
//
// $Id: TelnetClient.cc,v 1.21 1996-06-07 09:36:28-07 bmah Exp $
//
// TelnetClient.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// telnet (remote login) client program.
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//

#include <String.h>
#include "TelnetClient.h"
#include "EmpiricalContinuous.h"
#include "Event.h"
#include "inline.h"
#include "telnetproto.h"

//
// Imported probability distributions
//
extern RNG *rng;		// global random number generator

// telnet connection duration
extern const DoubleDistElement *TelnetDurationDistTable;
extern int TelnetDurationDistLength;

// telnet interarrivals
extern const DoubleDistElement *TelnetInterarrivalDistTable;
extern int TelnetInterarrivalDistLength;

//
// TelnetClient::TelnetClient
//
// (object constructor)
//
TelnetClient::TelnetClient(const String &name, Simulation *si, double initTime,
	     String &t, InAddr s, short p, int nt) :
    AppTcpClient(name, si, initTime, t, s, p), numTimestamps(nt) {
	startTime = initTime;
	endTime = initTime;
	tagNumber = 0;
	lastReply = 0;
	overruns = 0;
	totalBytesSent = 0;
	totalBytesReceived = 0;
	timestamps = new double[numTimestamps];
	dormant = false;
	rttStat = new SampleHistogram(0.0, 0.5, 0.05);
}

//
// TelnetClient::~TelnetClient
//
// (object destructor)
//
TelnetClient::~TelnetClient() {
    int i;
    
    *(sim->logStream) << name << " [TelnetClient] lifetime " <<
	lifeTime << " roundtrips " <<
	lastReply << " overruns " << overruns << 
	" sent " << totalBytesSent << " received " << totalBytesReceived <<
	" mean " << rttStat->mean() << " var " << rttStat->var() <<
	" min " << rttStat->min() << " max " << rttStat->max() << "\n";
    *(sim->logStream) << name << " [TelnetClient] rtts ";
    
    for (i = 0; i < rttStat->buckets(); i++) {
	*(sim->logStream) << rttStat->bucketThreshold(i) << " " << rttStat->inBucket(i) << " ";
    }
    *(sim->logStream) << "\n";
    
    delete [] timestamps;
    delete rttStat;
}


//
// TelnetClient::Handler
//
// (Event handler)
//
void
TelnetClient::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalContinuous TelnetDurationDist(TelnetDurationDistLength,
						  TelnetDurationDistTable,
						  rng);
    static EmpiricalContinuous TelnetInterarrivalDist(TelnetInterarrivalDistLength,
						 TelnetInterarrivalDistTable,
						 rng);

    switch (state) {

	case (AppTcpClientStateDown):
	{
	    switch (type) {

		case (EventTimer):
		{
		    // We got a Timer event...kick off a connection
		    // establishment.
		    tp = tcpModule->UserConnect(time, server, port, this);
		    state = AppTcpClientStateEstablishing;
		    break;
		}
		break;

		default:
		{
		    IF_DEBUG(0, {
			DEBUG_PREFIX("TelnetClient");
			*(sim->debugStream) << "bad event type " << type << " in state " << state << "\n";
		    });
		}
		break;

	    }
	}
	break;

	case (AppTcpClientStateEstablishing):
	{

	    switch (type) {

		case (EventTcpConnect):
		{
		    // Update our state.
		    // Find out how long we're going to be active.
		    dataStartTime = time;
		    lifeTime = TelnetDurationDist();
		    state = AppTcpClientStateEstablished;

		    // Set TF_NODELAY mode for this connection
		    tcpModule->UserSet(time, tp, TF_NODELAY);

		    IF_DEBUG(2, {
			DEBUG_PREFIX("TelnetClient");
			*(sim->debugStream) << "conversation lifetime " << lifeTime << "\n";
		    });

		    *(sim->logStream) << name <<
		      " [TelnetClient] connect-took " <<
		      (time - startTime) << "\n";

		    // Set timer for first telnet packet
		    double interarrival = TelnetInterarrivalDist();

		    Post(time + interarrival, this, EventTimer, NULL);
	
		    IF_DEBUG(2, {
			DEBUG_PREFIX("TelnetClient");
			*(sim->debugStream) << "first client send at " <<
			  time + interarrival << " (+" << interarrival <<
			  ")\n";
		    });

		}
		break;

		case (EventTcpClosed):
		{
		    IF_DEBUG(0, {
			DEBUG_PREFIX("TelnetClient");
			*(sim->debugStream) << "connection failed\n";
		    });

		    lifeTime = 0;
		    delete this;
		}
		break;

		default:
		{
		    IF_DEBUG(0, {
			DEBUG_PREFIX("TelnetClient");
			*(sim->debugStream) << "bad event type " << type << " in state " << state << "\n";
		    });
		}
		break;

	    }
	}
	break;

	case (AppTcpClientStateEstablished):
	{

	    switch (type) {

		case (EventTcpArrival):
		{
		    TcpCall *tc = (TcpCall *) data;
		    
		    IF_DEBUG(2, {
			DEBUG_PREFIX("TelnetClient");
			*(sim->debugStream) << "received " <<
			  tc->th_ulen << " bytes\n";
		    });

		    TcpData *receivedData;
		    while (receivedData = tc->data.Dequeue()) {
			
			// Record round-trip time for stats
			double rtt = time -
			  timestamps[receivedData->data_int % numTimestamps];
			lastReply = receivedData->data_int;
			
			(*rttStat) += rtt;
			
			*(sim->logStream) << name <<
			  " [TelnetClient] rtt " <<
			  rtt << "\n";

			IF_DEBUG(2, {
			    DEBUG_PREFIX("TelnetClient");
			    *(sim->debugStream) << "received reply tag " <<
			      receivedData->data_int << " rtt " << rtt << "\n";
			});
			
			// If we were dormant, we can wake up now (because
			// we have a timestamp slot to use) so wake up and
			// prepare to send another packet.
			if (dormant) {
		
			    dormant = false;
		
			    // Set a timeout for the time to send the
			    // next packet.
			    double interarrival = TelnetInterarrivalDist();
		
			    Post(time + interarrival, this, EventTimer, NULL);
			    
			    IF_DEBUG(0, {
				DEBUG_PREFIX("TelnetClient");
				*(sim->debugStream) <<
				  "waking up next client send at " <<
				  time + interarrival << " (+" <<
				  interarrival << ")\n";
			    });

			}

			// Get rid of data payload
			delete receivedData;

		    }

		    delete tc;

		}
		break;

		case (EventTimer):
		{
		    // If there's still time left, then send another packet.
		    // Otherwise, close down.
		    if (time < dataStartTime + lifeTime) {

			// First make sure that we haven't
			// used up all of outstanding timestamp slots.
			if (tagNumber - lastReply >= numTimestamps - 1) {
			    
			    dormant = true;
			    overruns++;

			    IF_DEBUG(0, {
				DEBUG_PREFIX("TelnetClient");
				*(sim->debugStream) <<
				  "ran out of timestamp slots...going dormant\n";
			    });
			}
			else {
			    
			    // Make up a new packet and send it.
			    TcpData *sendData = new TcpData;
			    sendData->data_int = tagNumber++;
			    
			    timestamps[sendData->data_int % numTimestamps] =
			      time;
			    
			    IF_DEBUG(2, {
				DEBUG_PREFIX("TelnetClient");
				*(sim->debugStream) <<
				  "sending client data tag " <<
				  sendData->data_int << "\n";
			    });
		
			    tcpModule->UserSend(time, tp,
						TelnetClientSendLength,
						sendData);

			    // Set a timeout for the time to send the next
			    // packet.
			    double interarrival = TelnetInterarrivalDist();
		
			    Post(time + interarrival, this, EventTimer, NULL);
		
			    IF_DEBUG(2, {
				DEBUG_PREFIX("TelnetClient");
				*(sim->debugStream) <<
				  "next client send at " <<
				  time + interarrival << " (+" <<
				  interarrival << ")\n";
			    });
			    
			}
			
		    }
		    else {

			// Close the "socket", but we can't go away yet.
			// In some circumstances, we can execute this
			// close between the time that an event is posted
			// to us and the time that it is delivered (for
			// example a TcpArrival event).  So we need to
			// sit in a Zombie state for a little while to
			// absorb any events posted to us before we can
			// go away.  The length of the Zombie state is,
			// unfortunately, a heuristic.
			//
			// Yet another artifact of doing a pure event-driven
			// simulation with dynamic objects.
			tcpModule->UserClose(time, tp);
			endTime = time + TelnetClientZombieTime;
			state = AppTcpClientStateZombie;
			Post(endTime, this, EventTimer, NULL);
			break;
		    }

		}
		break;

		case (EventTcpClosed):
		{
		    IF_DEBUG(0, {
			DEBUG_PREFIX("TelnetClient");
			*(sim->debugStream) << "connection reset? event type " << type << " in state " << state << "\n";
		    });
		}
		break;

		default:
		{
		    IF_DEBUG(0, {
			DEBUG_PREFIX("TelnetClient");
			*(sim->debugStream) << "bad event type " << type << " in state " << state << "\n";
		    });
		}
		break;

	    }

	}
	break;

	case (AppTcpClientStateZombie):
	{
	    switch (type) {

		case (EventTimer):
		{

		    // We can finally go away (we think)
		    totalBytesSent = tp->snd_bytes;
		    totalBytesReceived = tp->rcv_bytes;
		    state = AppTcpClientStateDown;
		    delete this;
		}
		break;

		// We throw away all events posted to us in Zombie state
		// (well those that aren't Timers).  Should we try to
		// do something more intelligent with events?
		default:
		{
		}
		break;
	    }
	}
	break;
	      

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("TelnetClient");
		*(sim->debugStream) << "bad state " << state << "\n";
	    });
	}
	break;
    }
}

//
// TelnetClient::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
TelnetClient::Output()
{
    return name + " [TelnetClient]";
}

//
// TelnetClientCmd
//
// (Tcl command procedure)
//
int
TelnetClientCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 7) {
	Tcl_SetResult(interp,
		      "Usage:  TelnetClient name start tcp server port timestamps",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	TelnetClient *tc;	// new client object
	char *nm;		// name of new client object
	double startTime;	// starting time to run
	String tcpObjectName;	// name of Tcp object
	InAddr server;		// server address to connect to
	int port;		// port number
	int nt;			// number of outstanding timestamps
	extern Simulation *sim;	// simulation object

	nm = argv[1];
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    tcpObjectName = argv[3];
	    server.s_addr = InetAddr(argv[4]);
	    rc = Tcl_GetInt(interp, argv[5], &port);

	    if (rc == TCL_OK) {

		rc = Tcl_GetInt(interp, argv[6], &nt);

		if (rc == TCL_OK) {
		    
		    // Get pointer to TCP object (implication:  Tcp
		    // object *must* be created before any applications
		    // are created.
		    tc = new TelnetClient(nm, sim, startTime,
					  tcpObjectName,
					  server, port, nt);
		    Tcl_CreateCommand(interp, nm,
				      TelnetClientObjectCmd,
				      (ClientData) tc,
				      TelnetClientDeleteProc);
		}

	    }

	}

	return rc;
    }
}

//
// TelnetClientObjectCmd
//
// (Tcl object command procedure)
//
int
TelnetClientObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {
    return ScheduledObjectCmd(clientData, interp, argc, argv);
}

//
// TelnetClientDeleteProc
//
// (Tcl delete procedure)
//
void
TelnetClientDeleteProc(ClientData clientData) {
    delete (TelnetClient *) clientData;
}
