static char rcsid[] = "$Id: TelnetServer.cc,v 1.11 1996-02-15 19:31:20-08 bmah Exp $";
//
// $Id: TelnetServer.cc,v 1.11 1996-02-15 19:31:20-08 bmah Exp $
//   based on:
// $ID: SampleTcpServer.cc,v 1.2 1995-10-09 10:37:42-07 bmah Exp $
//
// TelnetServer.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Sample TCP client application, used to test TCP.
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//

#include "TelnetServer.h"
#include "EmpiricalIntegral.h"
#include "telnetproto.h"
#include "Event.h"
#include "inline.h"

//
// Imported probability distributions
//
extern RNG *rng;		// global random number generator

// Length of telnet reply packet
extern const DoubleDistElement *TelnetReplyLengthDistTable;
extern int TelnetReplyLengthDistLength;

//
// TelnetServer::Handler
//
// (Event handler)
//
void
TelnetServer::Handler(Event *e)
{
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalIntegral TelnetReplyLengthDist(TelnetReplyLengthDistLength,
						   TelnetReplyLengthDistTable,
						   rng);

    if (!initFlag) {
	AppTcpServer::Handler(e);
    }
    else {
	switch (type) {
	    case (EventTcpConnect):
	    {

		// Connection bookkeeping
		TelnetServerConnection *newConnection =
		  new TelnetServerConnection;
		newConnection->tcp = (TcpPcb *) data;
		newConnection->startTime = time;
		newConnection->totalBytesSent = 0;
		newConnection->totalBytesReceived = 0;
		newConnection->next = active;
		active = newConnection;

		// Set nodelay for this connection
		tcpModule->UserSet(time, newConnection->tcp, TF_NODELAY);

		AppTcpServer::Handler(e);
	    }
	    break;

	    case (EventTcpArrival):
	    {
		TelnetServerConnection *current;
		TelnetServerConnection *previous = NULL;
		TelnetServerConnection *conn = NULL;
		TcpPcb *tp;

		TcpCall *tc = (TcpCall *) data;

		// Locate connection
		for (current = active;
		     current != NULL;
		     current = current->next) {

		    if (current->tcp == tc->pcb) {
			conn = current;
			break;
		    }
		}

		if (conn) {
		    conn->totalBytesReceived += tc->th_ulen;

		    IF_DEBUG(2, {
			DEBUG_PREFIX("TelnetServer");
			*(sim->debugStream) << "received " << tc->th_ulen << " bytes total " << conn->totalBytesReceived << "\n";
		    });

		}

		TcpData *receivedData;
		while (receivedData = tc->data.Dequeue()) {

		    // Grab a telnet server reply length from the distribution
		    // and send it.
//		    int replyLength = telnet_pktsize();
		    int replyLength = (int) TelnetReplyLengthDist();

		    IF_DEBUG(2, {
			DEBUG_PREFIX("TelnetServer");
			*(sim->debugStream) << "sending reply tag " <<
			  receivedData->data_int << " with " <<
			    replyLength << " bytes\n";
		    });

		    tcpModule->UserSend(time + TelnetServerLatency, tc->pcb,
					replyLength, receivedData);
		}

		delete tc;

	    }
	    break;

	    case (EventTcpClosed):
	    {
		TelnetServerConnection *current;
		TelnetServerConnection *previous = NULL;
		TelnetServerConnection *die = NULL;
		TcpPcb *tp = (TcpPcb *) data;

		// Delete this connection from the connection list
		for (current = active;
		     current != NULL;
		     previous = current, current = current->next) {

		    if (current->tcp == tp) {
			if (previous) {
			    previous->next = current->next;
			}
			else {
			    active = current->next;
			}
			die = current;
			break;
		    }
		}

		tcpModule->UserClose(time, tp);

		//
		if (die) {
		    *(sim->logStream) << name <<
		      " [TelnetServer] end send " <<
		      die->totalBytesSent << " received " <<
			die->totalBytesReceived <<
			" time " << (time - die->startTime) << "\n";

		    delete die;
		}

	    }
	    break;

	    default:
	    AppTcpServer::Handler(e);
	    break;
	}
    }
}

//
// TelnetServer::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
TelnetServer::Output()
{
    return name + " [TelnetServer]";
}

//
// TelnetServerCmd
//
// (Tcl command procedure)
//
int TelnetServerCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 5) {
	Tcl_SetResult(interp,
		      "Usage:  TelnetServerCmd name start tcp port",
		      TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	TelnetServer *ts;	// new server object
	char *nm;		// name of new server object
	double startTime;	// starting time to run
	String tcpObjectName;	// name of Tcp object
	int port;		// port number
	extern Simulation *sim;	// simulation object

	nm = argv[1];
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    tcpObjectName = argv[3];
	    rc = Tcl_GetInt(interp, argv[4], &port);

	    if (rc == TCL_OK) {

		// Get pointer to TCP object (implication:  Tcp
		// object *must* be created before any applications
		// are created.
		ts = new TelnetServer(nm, sim, startTime, tcpObjectName,
					  port);
		Tcl_CreateCommand(interp, nm,
				  TelnetServerObjectCmd,
				  (ClientData) ts,
				  TelnetServerDeleteProc);
	    }

	}

	return rc;
    }
}

//
// TelnetServerObjectCmd
//
// (Tcl object command procedure)
//
int TelnetServerObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {
    return ScheduledObjectCmd(clientData, interp, argc, argv);
}

//
// TelnetServerDeleteProc
//
// (Tcl delete procedure)
//
void TelnetServerDeleteProc(ClientData clientData) {
    delete (TelnetServer *) clientData;
}


