static char rcsid[] = "$Id: Udp.cc,v 1.13 1996-02-07 16:38:24-08 bmah Exp $";
//
// $Id: Udp.cc,v 1.13 1996-02-07 16:38:24-08 bmah Exp $
//
// Udp.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// UDP protocol module class
//

#include "Udp.h"
#include "IpPacket.h"
#include "UdpPacket.h"
#include "Event.h"
#include "inline.h"

//
// Udp::Handler
//
// (Event handler)
//
void Udp::Handler(Event *e)
{

    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    switch (type) {

	//
	// IpArrival:  IP packet showed up that was demuxed at the IP
	// layer to UDP.  UDP now needs to demux to a process.
	//
	case (EventIpArrival):
	{

	    IpPacket *packet = (IpPacket *) data;
	    bool found = false;
	    
	    IF_DEBUG(2, {
		DEBUG_PREFIX("Udp");
		*(sim->debugStream) << "got packet " << packet << "\n";
	    });

	    if (packet->ip_p != IPPROTO_UDP) {

		IF_DEBUG(0, {
		    DEBUG_PREFIX("Udp");
		    *(sim->debugStream) << "bad UDP demux\n";
		});

		// drop packet?
		delete packet;

	    }
	    else {

		UdpPacket *udpHeader = (UdpPacket *) packet->payload;

		// Walk the DLL of PCBs, try to find something that matches
		// the incoming packet.
		UdpPcb *current = udb->next;
		while (current != udb) {
		    
		    // Matching criteria for PCB:  foreign address,
		    // local address, and foreign port must each match
		    // the packet header or be set to 0 (wildcard).
		    // Local port must match the packet.
		    if (((current->udp_faddr.s_addr == 0) ||
			 (current->udp_faddr.s_addr == packet->ip_src.s_addr)) &&
			((current->udp_laddr.s_addr == 0) ||
			 (current->udp_laddr.s_addr == packet->ip_dst.s_addr)) &&
			((current->udp_fport == 0) ||
			 (current->udp_fport == udpHeader->uh_sport)) &&
			((current->udp_lport == udpHeader->uh_dport))) {
			found = true;
			break;
		    }

		    current = current->next;
		    
		}
		    
		if (found) {
		    
		    // Post the packet payload + PCB pointer
		    UdpCall *call = new UdpCall;
		    call->pcb = current;
		    call->udp_faddr.s_addr = packet->ip_src.s_addr;
		    call->udp_fport = udpHeader->uh_sport;
		    call->uh_ulen = udpHeader->uh_ulen;
		    udpHeader->payload.AppendTo(call->data);
		    
		    Post(time + 0, current->udp_up, EventUdpArrival,
			 (void *) call);
		    
		    // Release resources
		    delete packet;
		    
		}
		else {

		    // Drop packet.  In real life we'd send an ICMP
		    // message back to the source.
		    delete packet;

		    IF_DEBUG(0, {
			DEBUG_PREFIX("Udp");
			*(sim->debugStream) <<
			  "dropped packet inbound port " <<
			  udpHeader->uh_dport << "\n";
		    });

		}
	    }
	}
	break;

	//
	// UdpSend:  Application wanted to send a UDP packet.
	//
	case (EventUdpSend):
	{
	    UdpCall *call = (UdpCall *) data;

	    UserSend(time, call->pcb, call->uh_ulen, NULL);
	    delete call;
	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("Udp");
		*(sim->debugStream) << "bad EventType " << type << "\n";
	    });
	}
	break;

    }

}

//
// Udp::UserBind
//
// Input:  Current simulation time (time, not used),
//         Foreign address and port (faddr, fport), local address and
//         port (laddr, lport).  Owning object (recipient of packets).
//
// Output:  Pointer to a UdpPcb to be used for outgoing packets.
//
// Bind a UdpPcb to some addresses and port numbers (this is roughly
// equivalent to creating a socket and binding that socket to addresses).
// 
UdpPcb *Udp::UserBind(double time,
		      InAddr faddr, unsigned short fport,
		      InAddr laddr, unsigned short lport,
		      ScheduledObject *owner)
{

    // Create a new Pcb and fill in its fields
    UdpPcb *pcb = new UdpPcb;
    pcb->udp_faddr = faddr;
    pcb->udp_fport = fport;
    pcb->udp_laddr = laddr;
    if (lport) {
	pcb->udp_lport = lport;	// fixed local port number
    }
    else {
	pcb->udp_lport = udb->udp_lport++; // ephemeral port number
    }
    pcb->udp_up = owner;

    // Drop it at the head of the DLL
    pcb->next = udb->next;
    pcb->prev = udb;
    pcb->next->prev = pcb;
    udb->next = pcb;
    
    return pcb;

}

//
// Udp::UserSend
//
// Input:  sending time (time), pointer to UDP pcb (up), packet length
//         (len), data field (data), packet type (type)
//
// Output:  None
//
// Send a UDP datagram.
//
void Udp::UserSend(double time, UdpPcb *up, short len, UdpData *data,
		   UdpPayloadType type)
{

    // Make up the UDP and IP headers
    UdpPacket *udpHeader = new UdpPacket;
    IpPacket *ipHeader = new IpPacket(udpHeader);
    
    ipHeader->ip_tos = 0;
    ipHeader->ip_len = len + UdpHeaderSize + IpHeaderSize;
    ipHeader->ip_id = 0; // Ip fills this in 
    ipHeader->ip_off = 0; // No fragments
    ipHeader->ip_ttl = 255;
    ipHeader->ip_p = IPPROTO_UDP;
    ipHeader->ip_src.s_addr = 0; // Ip fills this in
    ipHeader->ip_dst = up->udp_faddr;
    ipHeader->ip_next.s_addr = 0; // Ip fills this in
    ipHeader->payload = udpHeader;
    
    udpHeader->uh_sport = up->udp_lport;
    udpHeader->uh_dport = up->udp_fport;
    udpHeader->uh_ulen = len;
    udpHeader->type = type;
    if (data) {
	udpHeader->payload.EnqueueTail(data);
    }
    
    // Post the packet
    Lookup(down, downName);
    Post(time, down, EventIpSend, (void *) ipHeader);

    IF_DEBUG(2, {
	DEBUG_PREFIX("Udp");
	*(sim->debugStream) << "packet outbound " << ipHeader << "\n";
    });

}

//
// Udp::UserSend
//
// Input:  sending time (time), pointer to UDP pcb (up), packet length
//         (len), data tag (data)
//
// Output:  None
//
// Send a UDP datagram, with a UDP data tag but without a type.
//
void Udp::UserSend(double time, UdpPcb *up, short len, UdpData *data)
{
    UserSend(time, up, len, data, UdpPayloadTypeNull);
}

//
// Udp::UserSend
//
// Input:  sending time (time), pointer to UDP pcb (up), packet length
//         (len)
//
// Output:  None
//
// Send a UDP datagram, without a UDP data tag or type.
//
void Udp::UserSend(double time, UdpPcb *up, short len)
{
    UserSend(time, up, len, (UdpData *) NULL, UdpPayloadTypeNull);
}

//
// Udp::UserClose
//
// Input:  current simulation time (time, not used), PCB to close/deallocate
//
// Output;  none
//
// Release a UDP PCB.
//
void Udp::UserClose(double time, UdpPcb *pcb)
{
    pcb->next->prev = pcb->prev;
    pcb->prev->next = pcb->next;
    delete pcb;
}

//
// Udp::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String Udp::Output()
{
    return name + " [Udp]";
}

//
// UdpCmd
//
// (Tcl command procedure)
//
int UdpCmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) {

    if (argc != 2) {
	Tcl_SetResult(interp, "Usage:  Udp name", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	char *nm;		// name of UDP module
	Udp *udpModule;		// new UDP module
	extern Simulation *sim;	// global simulation object

	// One argument only...the name of the new UDP module.
	nm = argv[1];

	udpModule = new Udp(nm, sim);
	Tcl_CreateCommand(interp, nm,
			  UdpObjectCmd,
			  (ClientData) udpModule,
			  UdpDeleteProc);

	return (TCL_OK);

    }
}

//
// UdpObjectCmd
//
// (Tcl object command procedure)
//
int UdpObjectCmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]) {

    Udp *udpModule = (Udp *) clientData;

    if (argc == 1) {
	return(ScheduledObjectCmd(clientData, interp, argc, argv));
    }

    else {

	//
	// udp down object
	//
	if (!strcmp(argv[1], "down")) {
	    if (argc == 3) {
		String on;	// object name

		on = argv[2];
		udpModule->downName = on;
		udpModule->down = NULL;
		return TCL_OK;
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}

	//
	// udp netstat
	//
	else if (!strcmp(argv[1], "netstat")) {
	    if (argc == 2) {
		UdpPcb *up = udpModule->udb->next;
		char buf[128];	// temporary printing buffer

		while (up != udpModule->udb) {
		    
		    sprintf(buf, "%s.%d ",
			    InetNtoA(up->udp_laddr),
			    up->udp_lport);
		    cout << buf;

		    sprintf(buf, "%s.%d ",
			    InetNtoA(up->udp_faddr),
			    up->udp_fport);
		    cout << buf;

		    cout << "\n";

		    up = up->next;
		}
		return (TCL_OK);
	    }
	    else {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return(TCL_ERROR);
	    }
	}
	else {
	    return(ScheduledObjectCmd(clientData, interp, argc, argv));
	}

    }
}

//
// UdpDeleteProc
//
// (Tcl delete procedure)
//
void UdpDeleteProc(ClientData clientData) {
    delete (Udp *) clientData;
}



