static char rcsid[] = "$Id: VideoSource.cc,v 1.3 1996-04-16 11:03:18-07 bmah Exp $";
//
// $Id: VideoSource.cc,v 1.3 1996-04-16 11:03:18-07 bmah Exp $
//   based on:
// $ID: AudioSource.cc,v 1.3 1996-01-22 13:33:27-08 bmah Exp $
//
// VideoSource.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Video client (sender) application.  This simple source sends a
// CBR stream of constant-sized UDP packets.
//

#include "VideoSource.h"
#include "Event.h"
#include "inline.h"

//
// VideoSource::VideoSource
//
// Input:  name, simulation object, start time, UDP object, server
//         IP address, server port
//
// Output:  None
//
// Object constructor
//
VideoSource::VideoSource(const String &name, Simulation *si, double initTime,
			 String &u, InAddr s, short p, int pay,
			 double g) :
			 AppUdpClient(name, si, initTime, u, s, p),
			 payloadBytes(pay), gap(g)
{

    InAddr local;

    lifeTime = 100.0;		// XXX
    startTime = initTime;
    endTime = lifeTime + initTime;

    local.s_addr = 0;		// XXX INADDR_ANY
    up = udpModule->UserBind(initTime, s, p, local, 0, this);

    packetsSent = 0;

}

//
// VideoSource::Handler
//
// (Event handler)
//
void
VideoSource::Handler(Event *e)
{
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    IF_DEBUG(3, {
	DEBUG_PREFIX("VideoSource");
	*(sim->debugStream) << "got event " << type << "\n";
    });

    // See if we're done.
    if (time >= endTime) {

	// Delete object
	udpModule->UserClose(time, up);
	delete this;
	
    }

    else {

	UdpData *ud = new UdpData;
	ud->data_time = time;
	ud->seq = packetsSent;
	
	udpModule->UserSend(time, up, payloadBytes, ud,
			    UdpPayloadTypeVideo);
	Post(time + gap, this, EventTimer, NULL);
		
	IF_DEBUG(2, {
	    DEBUG_PREFIX("VideoSource");
	    *(sim->debugStream) << "sent packet " << packetsSent << " " << payloadBytes << " bytes\n";
	});
	
	packetsSent++;

    }

}

//
// VideoSource::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
VideoSource::Output()
{
    return name + " [VideoSource]";
}

//
// VideoSourceCmd
//
// (Tcl command procedure)
//
int
VideoSourceCmd(ClientData clientData, Tcl_Interp *interp,
	       int argc, char *argv[])
{
    if (argc != 8) {
	Tcl_SetResult(interp, "usage:  VideoSource name start udp server port payload gap", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	VideoSource *vs;	// new client object
	char *nm;		// name of new client object
	double startTime;	// starting time to run
	String udpObjectName;	// name of Udp object
	InAddr destaddr;	// destination address
	int port;		// port number
	int pay;		// payload size
	double g;		// gap
	extern Simulation *sim;	// simulation object
	int i;			// yet another loop counter

	// Get name
	nm = argv[1];

	// Get start time
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    // Get UDP object name
	    udpObjectName = argv[3];

	    // Get server address
	    destaddr.s_addr = InetAddr(argv[4]);

	    // Get port number
	    rc = Tcl_GetInt(interp, argv[5], &port);
	    if (rc == TCL_OK) {
		
		// Get payload
		rc = Tcl_GetInt(interp, argv[6], &pay);
		if (rc == TCL_OK) {

		    // Get gap
		    rc = Tcl_GetDouble(interp, argv[7], &g);
		    if (rc == TCL_OK) {

			vs = new VideoSource(nm, sim, startTime,
					     udpObjectName,
					     destaddr, port,
					     pay, g);
			Tcl_CreateCommand(interp, nm,
					  VideoSourceObjectCmd,
					  (ClientData) vs,
					  VideoSourceDeleteProc);

		    }
		}
	    }
	}
    return rc;
    }
}


//
// VideoSourceObjectCmd
//
// (Tcl object command procedure)
//
int
VideoSourceObjectCmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
    
    VideoSource *vs = (VideoSource *) clientData;

    if (argc == 1) {
	return ScheduledObjectCmd(clientData, interp, argc, argv);
    }
    else {

	//
	// vs lifetime ...
	//
	if (!strcmp(argv[1], "lifetime")) {

	    //
	    // vs lifetime
	    //
	    if (argc == 2) {
		char ch[80];
		sprintf(ch, "%f", vs->lifeTime);
		Tcl_SetResult(interp, ch, TCL_VOLATILE);
		return TCL_OK;
	    }
	    
	    //
	    // vs lifetime lt
	    //
	    else if (argc == 3) {
		int rc = TCL_OK;
		double lt;

		rc = Tcl_GetDouble(interp, argv[2], &lt);
		if (rc == TCL_OK) {
		    vs->lifeTime = lt;
		    vs->endTime = vs->lifeTime + vs->startTime;
		}
		return rc;
	    }		

	}
	else {
	    return ScheduledObjectCmd(clientData, interp, argc, argv);
	}
    }
}

//
// VideoSourceDeleteProc
//
// (Tcl delete procedure)
//
void
VideoSourceDeleteProc(ClientData clientData)
{
    delete (VideoSource *) clientData;
}
