static char rcsid[] = "$Id: VideoSource2.cc,v 1.2 1996-04-16 11:03:19-07 bmah Exp $";
//
// $Id
//   based on:
// $ID: VideoSource.cc,v 1.2 1996-01-27 23:14:20-08 bmah Exp $
//   based on:
// $ID: AudioSource.cc,v 1.3 1996-01-22 13:33:27-08 bmah Exp $
//
// VideoSource2.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//
// Video client (sender) application.  This simple source sends a
// VBR video stream, based on a two-state model of sending packets
// for conditional replenishment and for background updates.  The
// traffic model is taken from network traces.
//

#include "VideoSource2.h"
#include "Event.h"
#include "EmpiricalContinuous.h"
#include "EmpiricalIntegral.h"
#include "inline.h"

//
// Imported probability distributions and ancillary objects
//
extern RNG *rng; 		// global random number generator

// Conditional replenishment mode consecutive packets
extern const DoubleDistElement *VideoCrStateDistTable;
extern int VideoCrStateDistLength;

// Background update mode consecutive packets
extern const DoubleDistElement *VideoBgStateDistTable;
extern int VideoBgStateDistLength;

// Conditional replenishment mode packet size
extern const DoubleDistElement *VideoCrPacketSizeDistTable;
extern int VideoCrPacketSizeDistLength;

// Background update mode packet size
extern const DoubleDistElement *VideoBgPacketSizeDistTable;
extern int VideoBgPacketSizeDistLength;

//
// VideoSource2::VideoSource2
//
// Input:  name, simulation object, start time, UDP object, server
//         IP address, server port, bitrate, framerate
//
// Output:  None
//
// Object constructor
//
VideoSource2::VideoSource2(const String &name, Simulation *si, double initTime,
			   String &u, InAddr s, short p, double br,
			   double f) :
			   AppUdpClient(name, si, initTime, u, s, p),
			   bitrate(br), fps(f)
{

    InAddr local;

    lifeTime = 100.0;		// XXX
    startTime = initTime;
    endTime = lifeTime + initTime;

    local.s_addr = 0;		// XXX INADDR_ANY
    up = udpModule->UserBind(initTime, s, p, local, 0, this);

    packetsSent = 0;

    // Set Cr mode to start (arbitrarily)
    state = VideoSource2StateCr;
    packetsThisState = 1;	// By setting this to 1 we eliminate some
				// problems associated with making the
				// constructor access the distributions.
				// The effect on the results should be
				// small.

}

//
// VideoSource2::Handler
//
// (Event handler)
//
void
VideoSource2::Handler(Event *e)
{
    double time = e->ev_time;
    EventType type = e->ev_type;
    void *data = e->ev_data;

    static EmpiricalContinuous
      VideoCrStateDist(VideoCrStateDistLength,
		       VideoCrStateDistTable,
		       rng);
    static EmpiricalContinuous
      VideoBgStateDist(VideoBgStateDistLength,
		       VideoBgStateDistTable,
		       rng);
    static EmpiricalIntegral
      VideoCrPacketSizeDist(VideoCrPacketSizeDistLength,
			    VideoCrPacketSizeDistTable,
			    rng);
    static EmpiricalIntegral
      VideoBgPacketSizeDist(VideoBgPacketSizeDistLength,
			    VideoBgPacketSizeDistTable,
			    rng);
    
    // Should only get timeout events...check for it...
    switch (type) {

	case (EventTimer):
	{

	    // See if we're done.
	    if (time >= endTime) {
		
		// Delete object
		udpModule->UserClose(time, up);
		delete this;
		
	    }
	    
	    else {
		
		int payloadBytes;	// how many bytes to send?
		double gap;		// how long til next packet?
		
		// We're not done yet...sending a packet but let's find out
		// what it is and how long we should wait before going to
		// the next packet.
		
		switch (state) {
		    
		    // VideoSource2StateCr:  Conditional replenishment
		    case (VideoSource2StateCr):
		    {
			payloadBytes = (int) VideoCrPacketSizeDist();
			gap = (double) payloadBytes * 8.0 / bitrate;
		    }
		    break;
	
		    // VideoSource2StateBg:  Background updates
		    case (VideoSource2StateBg):
		    {
			payloadBytes = (int) VideoBgPacketSizeDist();
			gap = 1.0 / fps;
		    }
		    break;
		    
		    default:
		    {
		        IF_DEBUG(0, {
			    DEBUG_PREFIX("VideoSource2");
			    *(sim->debugStream) << "bad event type " <<
			      type << "\n";
			});
		    }
		    break;

		}
		
		UdpData *ud = new UdpData;
		ud->data_time = time;
		ud->seq = packetsSent;
		
		udpModule->UserSend(time, up, payloadBytes, ud,
				    UdpPayloadTypeVideo);
		Post(time + gap, this, EventTimer, NULL);
		
		IF_DEBUG(2, {
		    DEBUG_PREFIX("VideoSource");
		    *(sim->debugStream) << "sent packet " << packetsSent << " " << payloadBytes << " bytes gap " << gap << "\n";
		});
		
		packetsSent++;

		// See if it's time to change states?
		packetsThisState--;
		if (packetsThisState < 1) {
		    if (state == VideoSource2StateBg) {
			state = VideoSource2StateCr;
			packetsThisState = (int) VideoCrStateDist();
		    }
		    else if (state == VideoSource2StateCr) {
			state = VideoSource2StateBg;
			packetsThisState = (int) VideoBgStateDist();
		    }
		}
	    }
	}
	break;

	default:
	{
	    IF_DEBUG(0, {
		DEBUG_PREFIX("VideoSource2");
		*(sim->debugStream) << "bad event type " << type << "\n";
	    });
	}
	break;
    }
}

//
// VideoSource2::Output
//
// Input:  None
//
// Output:  String with an ASCII representation of this object.
//
String
VideoSource2::Output()
{
    return name + " [VideoSource2]";
}

//
// VideoSource2Cmd
//
// (Tcl command procedure)
//
int
VideoSource2Cmd(ClientData clientData, Tcl_Interp *interp,
		int argc, char *argv[])
{
    if (argc != 8) {
	Tcl_SetResult(interp, "usage:  VideoSource2 name start udp server port bitrate framerate", TCL_STATIC);
	return(TCL_ERROR);
    }
    else {
	int rc = TCL_OK;	// Tcl return code
	VideoSource2 *vs;	// new client object
	char *nm;		// name of new client object
	double startTime;	// starting time to run
	String udpObjectName;	// name of Udp object
	InAddr destaddr;	// destination address
	int port;		// port number
	double br;		// bitrate
	double fr;		// framerate
	extern Simulation *sim;	// simulation object
	int i;			// yet another loop counter

	// Get name
	nm = argv[1];

	// Get start time
	rc = Tcl_GetDouble(interp, argv[2], &startTime);
	if (rc == TCL_OK) {

	    // Get UDP object name
	    udpObjectName = argv[3];

	    // Get server address
	    destaddr.s_addr = InetAddr(argv[4]);

	    // Get port number
	    rc = Tcl_GetInt(interp, argv[5], &port);
	    if (rc == TCL_OK) {
		
		// Get bitrate
		rc = Tcl_GetDouble(interp, argv[6], &br);
		if (rc == TCL_OK) {

		    // Get framerate
		    rc = Tcl_GetDouble(interp, argv[7], &fr);
		    if (rc == TCL_OK) {

			vs = new VideoSource2(nm, sim, startTime,
					      udpObjectName,
					      destaddr, port,
					      br, fr);
			Tcl_CreateCommand(interp, nm,
					  VideoSource2ObjectCmd,
					  (ClientData) vs,
					  VideoSource2DeleteProc);

		    }
		}
	    }
	}
    return rc;
    }
}


//
// VideoSource2ObjectCmd
//
// (Tcl object command procedure)
//
int
VideoSource2ObjectCmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
    
    VideoSource2 *vs = (VideoSource2 *) clientData;

    if (argc == 1) {
	return ScheduledObjectCmd(clientData, interp, argc, argv);
    }
    else {

	//
	// vs lifetime ...
	//
	if (!strcmp(argv[1], "lifetime")) {

	    //
	    // vs lifetime
	    //
	    if (argc == 2) {
		char ch[80];
		sprintf(ch, "%f", vs->lifeTime);
		Tcl_SetResult(interp, ch, TCL_VOLATILE);
		return TCL_OK;
	    }
	    
	    //
	    // vs lifetime lt
	    //
	    else if (argc == 3) {
		int rc = TCL_OK;
		double lt;

		rc = Tcl_GetDouble(interp, argv[2], &lt);
		if (rc == TCL_OK) {
		    vs->lifeTime = lt;
		    vs->endTime = vs->lifeTime + vs->startTime;
		}
		return rc;
	    }		

	}
	else {
	    return ScheduledObjectCmd(clientData, interp, argc, argv);
	}
    }
}

//
// VideoSource2DeleteProc
//
// (Tcl delete procedure)
//
void
VideoSource2DeleteProc(ClientData clientData)
{
    delete (VideoSource2 *) clientData;
}
