static char rcsid[] = "$Id: main.cc,v 1.31 1996-06-12 20:20:31-07 bmah Exp $";
//
// $Id: main.cc,v 1.31 1996-06-12 20:20:31-07 bmah Exp $
//
// main.cc
// Bruce A. Mah (bmah@CS.Berkeley.EDU)
//
// Copyright (c) 1995, 1996 The Regents of the University of California
// All Rights Reserved
// Created by Bruce A. Mah
// Department of Electrical Engineering and Computer Sciences,
//   Computer Science Division
//

#include <time.h>
#include <stdio.h>
#include <fstream.h>
#include <iostream.h>
#include <GetOpt.h>
#include <ACG.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>

#include "util.h"
#include "Calendar.h"
#include "ScheduledObject.h"
#include "String.ScheduledObjectPtr.CHMap.h"
#include "Simulation.h"
#include "user.h"

#ifdef HAVE_CSRI_MALLOC
extern "C" {
#include "malloc.h"
}
#endif //HAVE_CSRI_MALLOC

//
// Global objects
//
Calendar *scheduler;              // simulation scheduler
Simulation *sim;		  // global simulation object
ACG *rng;			  // global random number generator

char StatBuffer[1024];		  // status buffer
char HostName[MAXHOSTNAMELEN];	  // hostname
int Pid;			  // process ID

ofstream *DebugFileStream;	  // debug file stream (if needed)
ofstream *LogFileStream;	  // log file stream (if needed)

int events = 0;			  // total number of events processed

String SimName = "sim";

//
// Deal with configuration macros that might not have been defined
// by the Makefile or the command line.
//
#ifndef INSANE_CONF
#define INSANE_CONF "insane.conf"
#endif // INSANE_CONF

#ifndef INSANE_CONFPATH
#define INSANE_CONFPATH "."
#endif // INSANE_CONFPATH

#ifndef INSANE_LIBDIR
#define INSANE_LIBDIR "../../lib/tcl"
#endif // INSANE_LIBDIR

//
// Configuration variables
//
String ConfigFile(INSANE_CONF);   // configuration file name
String DebugFile("-");		  // debug file
bool EFlag = false;		  // event info
bool FFlag = false;		  // flush debugging output
bool IFlag = false;		  // interactive flag
String LibDir(INSANE_LIBDIR);	  // Tcl autoload directory
String LogFile("-");		  // log file
#ifdef HAVE_CSRI_MALLOC
int MDebug = 0;			  // malloc debugging level
#endif
bool RFlag = false;		  // -R debugging turned on
int TimeStamps = 0;		  // timestamps every...
int Zeed = 0;			  // random number generator seed

//
// Initialize
//
// Input:  None
//
// Output:  None
//
// Initialize simulation meta-data structures (calendar queue, object
// registry, etc.).  Basically, set up all the things we need to do before
// we read the configuration file.
//
void Initialize(void) {
    
    scheduler = new Calendar;
    ScheduledObject::registry = new StringScheduledObjectPtrCHMap(NULL);
    ScheduledObject::flushFlag = FFlag;

    // Initialize random number generators
    rng = new ACG(Zeed);

    sim = new Simulation(SimName);
    sim->trace = EFlag;
    sim->run = true;
    sim->interactive = IFlag;
    sim->scheduler = scheduler;
    sim->rng = rng;
    
    // Open debugging stream 
    if (DebugFile == "-") {
 	(sim->debugStream) = &cout;
    }
    else {
 	DebugFileStream = new ofstream(DebugFile);
 	if (!(*DebugFileStream)) {
	    
 	    perror("Couldn't open debug file");
 	    exit(1);
 	    
 	}
	sim->debugStream = DebugFileStream;
    }

    // Open logging stream
    if (LogFile == "-") {
 	(sim->logStream) = &cout;
    }
    else {
 	LogFileStream = new ofstream(LogFile);
 	if (!(*LogFileStream)) {
	    
 	    perror("Couldn't open log file");
 	    exit(1);
 	    
 	}
	sim->logStream = LogFileStream;
    }

}

//
// PrintVersion
//
// Input:  None.
//
// Output:  None.
//
void PrintVersion()
{
    extern char *Version, *Copyright, *Build;
    cerr << Version << "\n";
    cerr << Copyright << "\n";
    cerr << Build << "\n";
}


//
// PrintUsage
//
// Input:  Command vector (argc, argv)
//
// Output:  None.
//
void PrintUsage(int argc, char **argv)
{
    PrintVersion();
    cerr << "Usage: " << argv[0] << " [-d file] [-D var=val] [-f file] [-h]\n";
    cerr << "       [-i] [-l file] [-L dir] [-t time] [-T expr] [-v]\n";
    cerr << "       [-z seed]\n";
    cerr << "-d file\t\tSet debug file (default is " << DebugFile <<
      ", - for stdout)\n";
    cerr << "-D var=val\tDefine Tcl variable\n";
    cerr << "-E\t\tTurn on event tracing\n";
    cerr << "-f file\t\tSet configuration file (default is " <<
      ConfigFile << ")\n";
    cerr << "-F\t\tFlush debugging output after every line\n";
    cerr << "-h\t\tPrint this message\n";
    cerr << "-i\t\tSet interactive mode (default is off)\n";
    cerr << "-l file\t\tSet log file (default is " << LogFile <<
      ", - for stdout)\n";
    cerr << "-L dir\t\tSet Tcl autoload directory (default is " <<
      LibDir << ")\n";
#ifdef HAVE_CSRI_MALLOC
    cerr << "-M debug\tSet malloc debugging level (default is " <<
      MDebug << ")\n";
#endif //HAVE_CSRI_MALLOC
    cerr << "-R host/port\tLog status reports to host/port\n";
    cerr << "-t time\t\tPrint simulation timestamps to log file every\n";
    cerr << "\t\ttime events (default is " << TimeStamps <<
      ", 0 to disable)\n";
    cerr << "-T expr\t\tEvaluate Tcl expression\n";
    cerr << "-v\t\tPrint version information\n";
    cerr << "-z seed\t\tSet random number generator seed (default is " <<
      Zeed << ")\n";
}

//
// main
//
// Input:  Command vector (argc, argv)
//
// Output:  None.
//
main(int argc, char **argv)
{

    int i;
    GetOpt getopt(argc, argv, "d:D:Ef:Fhil:L:M:R:t:T:vz:");
    int option;
    Event *e;
    String insaneInit2;		// Second (command-line) INSANE init string

    int sock;			// socket to send status updates on
    struct sockaddr_in addr;	// sending address
    short port;			// port number
    String cmd;			// concatenated command-line options

    cmd = argv[0];
    for (i = 1; i < argc; i++) {
	cmd += " ";
	cmd += argv[i];
    }

    // Parse command-line options
    while ((option = getopt()) != EOF) {

	switch (option) {

	    case 'd':
	    DebugFile = getopt.optarg;
	    break;

	    case 'D':
	    {
		String optarg = getopt.optarg;
		String tclVar;
		String tclVal;
		
		tclVar = optarg.before("=");
		tclVal = optarg.after("=");

		insaneInit2 += ("set " + tclVar + " " + tclVal + "\n");
	    }
	    break;

	    case 'E':
	    {
		EFlag = true;
	    }
	    break;
	      
	    case 'f':
	    ConfigFile = getopt.optarg;
	    break;

	    case 'F':
	    FFlag = true;
	    break;

	    case 'h':
	    case '?':
	    PrintUsage(argc, argv);
	    exit(0);
	    break;

	    case 'i':
	    IFlag = true;
	    break;

	    case 'l':
	    LogFile = getopt.optarg;
	    break;
	    
	    case 'L':
	    LibDir = getopt.optarg;
	    break;

	    case 'M':
#ifdef HAVE_CSRI_MALLOC
	    MDebug = atoi(getopt.optarg);
#else
	    cerr << "-M:  Option not supported, CSRI malloc not installed\n";
	    exit(1);
#endif
	    break;

	    case 'R':
	    {
		String arg, hostArg, portArg; // parsed arguments from cmdline
		struct hostent *host;	      // host

		// Parse host/port arg
		arg = (String) getopt.optarg;
		hostArg = arg.before("/");
		portArg = arg.after("/");

		// Do a nameserver lookup to get the monitoring host
		port = atoi((const char *) portArg);
		host = gethostbyname((const char *) hostArg);
		if (host == NULL) {
		    endhostent();
		    perror("gethostbyname");
		    exit(1);
		}

		// Fill in the addr structure so we can send data
		bzero((char *) &addr, sizeof(addr));
		addr.sin_family = AF_INET;
		bcopy((const char *) host->h_addr,
		      (char *) &addr.sin_addr,
		      host->h_length);
		addr.sin_port = htons(port);

		RFlag = true;
		
	    }
	    break;

	    case 't':
	    TimeStamps = atoi(getopt.optarg);
	    break;

	    case 'T':
	    insaneInit2 += (String) getopt.optarg + "\n";
	    break;

	    case 'v':
	    PrintVersion();
	    exit(0);
	    break;
	    
	    case 'z':
	    Zeed = atoi(getopt.optarg);
	    break;

	}

    }

    // Initialize global objects
    Initialize();

    // User interface initialization
    if (InitUser(insaneInit2, ConfigFile, sim) < 0) {
	perror("Couldn't read configuration file");
	exit(1);
    }

    //
    // Get some identifying data.  Mostly this is used for status
    // update packets, but we also write it to the log file.
    //
    // Hostname...
    if (gethostname(HostName, MAXHOSTNAMELEN) < -1) {
	perror("gethostname");
	exit(1);
    }
    
    // ...and process ID
    Pid = getpid();

    // If we're writing via UDP, then do the setup stuff we need to
    // do.
    if (RFlag) {
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
	    perror("socket");
	    exit(1);
	}

    }

    // Write build information to the log
    extern char *Build, *Version;
    time_t start, now;
    
    *(sim->logStream) << "insane [simulation] " << Version << "\n";
    *(sim->logStream) << "insane [simulation] " << Build << "\n";

    time(&start);
    *(sim->logStream) << "insane [simulation] start time " << (int) start << " " << ctime(&start);

    *(sim->logStream) << "insane [simulation] invocation " << cmd << "\n";

    *(sim->logStream) << "insane [simulation] running on " << HostName <<
      " as PID " << Pid << "\n";
    
#ifdef HAVE_CSRI_MALLOC
    mal_debug(MDebug);
#endif

    // Event loop
    while (sim->run) {

	// Drop into command line interface if needed
	if (sim->interactive) {

	    // reset interactive mode
	    sim->interactive = false;
	    Cli(sim);

	}

	// get next event
	e = scheduler->Get();

	// no event => calendar queue empty
	if (e == NULL) {
	    *(sim->logStream) << "scheduler [scheduler] no events left\n";
	    break;
	}
	
	// Event debugging, if requested.
	if (sim->trace) {
	    *(sim->debugStream) << e->ev_time << ": scheduler deliver " << e->ev_type << " to " << e->ev_object->name;
	    if (e->ev_sender) {
		*(sim->debugStream) << " from " << e->ev_sender->name;
	    }
	    *(sim->debugStream) << "\n";
	    if (FFlag) {
		*(sim->debugStream) << flush;
	    }
	}

	// dispatch
	e->ev_object->Handler(e);
	delete(e);
	
	events++;

	// Write timestamps:  If Rflag is on, then we send a UDP packet
	// with a status report.  Otherwise, we write to the logstream.
	if ((TimeStamps != 0) && (events % TimeStamps == 0)) {

	    if (RFlag) {

		// Compute dynamic stuff
		time(&now);

		// Compose a status report and send it.
		sprintf(StatBuffer,
			"system insane host %s pid %d hosttype %s args {%s} status %s start %d time %d events %d simtime %f",
			HostName,
			Pid,
			HOST_TYPE,
			(const char *) cmd,
			"running",
			(int) start,
			(int) now,
			events,
			sim->scheduler->clock);
		sendto(sock, StatBuffer, strlen(StatBuffer) + 1, 0,
		       (sockaddr *) &addr, sizeof(addr));

	    }
	    else {
		*(sim->logStream) << sim->scheduler->clock << " scheduler [scheduler] events " << events << "\n";
	    }
	    
#ifdef HAVE_CSRI_MALLOC
	    mal_verify(1);
#endif
	    sim->scheduler->Sanity();
	}

    }

    //
    // End-of-simulation processing
    //

    // Write global end statistics to the log
    time(&now);
    *(sim->logStream) << "insane [simulation] end time " << (int) now << " " << ctime(&now);
    *(sim->logStream) << "insane [simulation] " << events << " events processed\n";

    // Iterate over all simulation objects and deallocate them.  This
    // in turn calls their Shutdown methods, which writes statistics to
    // our output files.
    Pix p; 			// Loop index analog
    String n;			// name
    ScheduledObject *s;		// object

    p = sim->registry->first();
    while (p != 0) {
	n = sim->registry->key(p);
	s = sim->registry->contents(p);

	// Need to advance over the item we're looking at, because when
	// we delete it, the entry in the registry will go away.  If
	// we delete, then advance, we'll find that the corresponding
	// registry entry disappeared out from under us.
	sim->registry->next(p);

	// Don't delete the simulation object yet!
	if (n != SimName) {
//	    *(sim->debugStream) << s->name << "\n";	    
	    delete s;
	}

    }

    // Close additional streams if they exist
    if (LogFileStream) {
	delete LogFileStream;
    }
    if (DebugFileStream) {
	delete DebugFileStream;
    }

    // We're done!
    exit(0);

}
